{*****************************************************************************}
{                                                                             }
{  llBool.pas                                                                 }
{                                                                             }
{  Misc routines dealing primarily with boolean arguments or results.         }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llBool;

interface

uses
  SysUtils;

type
  TBooleanOperator = (boNOOP, boNOT, boAND, boOR, boXOR);
  TComparisonOperator = (coNOOP, coEQ, coNE, coLT, coGT, coLE, coGE);

const
  BoolOpWords: array[TBooleanOperator] of string =
    ('', 'NOT', 'AND', 'OR', 'XOR');
  BoolOpSymbols: array[TBooleanOperator] of string =
    ('', '!', '&&', '||', '^');
  CompOpLetters: array[TComparisonOperator] of string =
    ('', 'EQ', 'NE', 'LT', 'GT', 'LE', 'GE');
  CompOpWords: array[TComparisonOperator] of string =
    ('', 'Equal', 'Not equal', 'Less than', 'Greater than',
     'Less than or equal', 'Greater than or equal');
  CompOpSymbols: array[TComparisonOperator] of string =
    ('', '=', '<>', '<', '>', '<=', '>=');
  CompOpAltSymbols: array[TComparisonOperator] of string =
    ('', '==', '!=', '<', '>', '<=', '>=');

  function BtoS(const B: Boolean): string;
  function StoB(const S: string; const BlankDefault: Boolean = False): Boolean;

  function StrToBoolOp(const S: string): TBooleanOperator;
  function StrToCompOp(const S: string): TComparisonOperator;

  function EvalBoolExpr(const ABoolOp: TBooleanOperator;
                        const A, B: Boolean): Boolean;

  {+ /trim=2}
  function EvalComparison(const ACompOp: TComparisonOperator;
                          const A, B: string): Boolean; overload;
  function EvalComparison(const ACompOp: TComparisonOperator;
                          const A, B: Integer): Boolean; overload;
  function EvalComparison(const ACompOp: TComparisonOperator;
                          const A, B: Int64): Boolean; overload;
  function EvalComparison(const ACompOp: TComparisonOperator;
                          const A, B: Extended): Boolean; overload;
  function EvalComparison(const ACompOp: TComparisonOperator;
                          const A, B: Currency): Boolean; overload;
  {
    The overloaded EvalComparison functions let you compare two values using a
    comparison operator and returns the boolean result of the expression
    (A operator B). If the operator is coNOOP, the function always returns True.

    TComparisonOperator = (coNOOP, coEQ, coNE, coLT, coGT, coLE, coGE);

    The Extended version works for Single, Double, Extended and TDateTime.
    Normal cautions apply when comparing floating point values for [in]equality.
    String comparisons are case sensitive.

    Examples: EvalComparison(coEQ, 0, 1) -> (0 = 1 evaluates to False)
              EvalComparison(coLT, 0, 1) -> (0 < 1 evaluates to True)

    See also: StrToCompOp
  -}

  {+ /trim=2}
  function IIf(const ABool: Boolean; const ATrue, AFalse: string): string; overload;
  function IIf(const ABool: Boolean; const ATrue, AFalse: Integer): Integer; overload;
  function IIf(const ABool: Boolean; const ATrue, AFalse: Int64): Int64; overload;
  function IIf(const ABool: Boolean; const ATrue, AFalse: Extended): Extended; overload;
  function IIf(const ABool: Boolean; const ATrue, AFalse: Currency): Currency; overload;
  {
    Inline-If functions overloaded for common data types.
    The Extended version works for Single, Double, Extended and TDateTime.
  -}

implementation

{+}
function BtoS(const B: Boolean): string;
{
  A simpler alternative to BoolToStr.  Returns '1' for True, '0' for False -
  the values typical to ini file booleans.
-}
begin
  if B then
    Result := '1'
  else
    Result := '0';
end;

{+}
function StoB(const S: string; const BlankDefault: Boolean = False): Boolean;
{
  A simpler version of StrToBool that never raises an exception. It uses my
  predefined list of True values and considers anything else as False.
  If the input is blank (''), the function returns the BlankDefault value.

  The case-insensitive True values are anything that starts with Y, T, +, *, or
  a non-zero numeric.
-}
var
  x: Extended;
begin
  // Try the most likely values before converting to a number
  if S = '' then
    Result := BlankDefault
  else if (S = '0') or IsDelimiter('FfNn', S, 1) then
    Result := False
  else if IsDelimiter('123456789TtYy+*', S, 1) then
    Result := True
  else if TryStrToFloat(S, x) then
  	Result := x <> 0
  else
    Result := False;
end;

{+}
function StrToBoolOp(const S: string): TBooleanOperator;
{
  Takes a string and returns the TBooleanOperator equivalent. The comparison is
  case-insensitive. If a blank or other invalid value is supplied, the function
  returns boNOOP.

  TBooleanOperator = (boNOOP, boNOT, boAND, boOR, boXOR);

  The following constant arrays define the valid input for this function, and
  can also be used to convert a TBooleanOperator to a string:

  BoolOpWords: array[TBooleanOperator] of string =
    ('', 'NOT', 'AND', 'OR', 'XOR');
  BoolOpSymbols: array[TBooleanOperator] of string =
    ('', '!', '&&', '||', '^');

  See also: EvalBoolExpr
-}
begin
  Result := High(TBooleanOperator);
  while Result > boNOOP do begin
    if SameText(S, BoolOpWords[Result]) or (S = BoolOpSymbols[Result]) then
      break;
    Dec(Result);
  end;
end;

{+}
function StrToCompOp(const S: string): TComparisonOperator;
{
  Takes a string and returns the equivalent TComparisonOperator. The comparison
  is case-insensitive. If a blank or other invalid value is supplied, the
  function returns coNOOP.

  TComparisonOperator = (coNOOP, coEQ, coNE, coLT, coGT, coLE, coGE);

  The following constant arrays define the valid input for this function, and
  can also be used to convert a TComparisonOperator to a string:

  CompOpLetters: array[TComparisonOperator] of string =
    ('', 'EQ', 'NE', 'LT', 'GT', 'LE', 'GE');
  CompOpWords: array[TComparisonOperator] of string =
    ('', 'Equal', 'Not equal', 'Less than', 'Greater than',
     'Less than or equal', 'Greater than or equal');
  CompOpSymbols: array[TComparisonOperator] of string =
    ('', '=', '<>', '<', '>', '<=', '>=');
  CompOpAltSymbols: array[TComparisonOperator] of string =
    ('', '==', '!=', '<', '>', '<=', '>=');

  See also: EvalComparison
-}
begin
  Result := High(TComparisonOperator);
  while Result > coNOOP do begin
    if SameText(S, CompOpLetters[Result]) or
       SameText(S, CompOpWords[Result]) or
       (S = CompOpSymbols[Result]) or
       (S = CompOpAltSymbols[Result]) then
      break;
    Dec(Result);
  end;
end;

{+}
function EvalBoolExpr(const ABoolOp: TBooleanOperator;
                      const A, B: Boolean): Boolean;
{
  The function returns the boolean result of (A operator B) except in the case
  of boNOT, when it returns (not A). If the operator is boNOOP, the function
  just returns A.

  TBooleanOperator = (boNOOP, boNOT, boAND, boOR, boXOR);

  See also: StrToBoolOp
-}
begin
  case ABoolOp of
    boNOT:  Result := not A;
    boAND:  Result := A and B;
    boXOR:  Result := A xor B;
    boOR:   Result := A or B;
    else    Result := A; // boNOOP
  end;
end;

{----- EvalComparison ---------------------------------------------------------}

function EvalComparison(const ACompOp: TComparisonOperator;
                        const A, B: string): Boolean;
begin
  case ACompOp of
    coEQ: Result := A = B;
    coNE: Result := A <> B;
    coLT: Result := A < B;
    coGT: Result := A > B;
    coLE: Result := A <= B;
    coGE: Result := A >= B;
    else  Result := True; // coNOOP
  end;
end;

function EvalComparison(const ACompOp: TComparisonOperator; const A, B: Integer): Boolean;
begin
  case ACompOp of
    coEQ: Result := A = B;
    coNE: Result := A <> B;
    coLT: Result := A < B;
    coGT: Result := A > B;
    coLE: Result := A <= B;
    coGE: Result := A >= B;
    else  Result := True; // coNOOP
  end;
end;

function EvalComparison(const ACompOp: TComparisonOperator; const A, B: Int64): Boolean;
begin
  case ACompOp of
    coEQ: Result := A = B;
    coNE: Result := A <> B;
    coLT: Result := A < B;
    coGT: Result := A > B;
    coLE: Result := A <= B;
    coGE: Result := A >= B;
    else  Result := True; // coNOOP
  end;
end;

function EvalComparison(const ACompOp: TComparisonOperator; const A, B: Extended): Boolean;
begin
  case ACompOp of
    coEQ: Result := A = B;
    coNE: Result := A <> B;
    coLT: Result := A < B;
    coGT: Result := A > B;
    coLE: Result := A <= B;
    coGE: Result := A >= B;
    else  Result := True; // coNOOP
  end;
end;

function EvalComparison(const ACompOp: TComparisonOperator; const A, B: Currency): Boolean;
begin
  case ACompOp of
    coEQ: Result := A = B;
    coNE: Result := A <> B;
    coLT: Result := A < B;
    coGT: Result := A > B;
    coLE: Result := A <= B;
    coGE: Result := A >= B;
    else  Result := True; // coNOOP
  end;
end;

{----- IIf --------------------------------------------------------------------}

function IIf(const ABool: Boolean; const ATrue, AFalse: string): string;
begin
  if ABool then
    Result := ATrue
  else
    Result := AFalse;
end;

function IIf(const ABool: Boolean; const ATrue, AFalse: Integer): Integer;
begin
  if ABool then
    Result := ATrue
  else
    Result := AFalse;
end;

function IIf(const ABool: Boolean; const ATrue, AFalse: Int64): Int64;
begin
  if ABool then
    Result := ATrue
  else
    Result := AFalse;
end;

function IIf(const ABool: Boolean; const ATrue, AFalse: Extended): Extended;
begin
  if ABool then
    Result := ATrue
  else
    Result := AFalse;
end;

function IIf(const ABool: Boolean; const ATrue, AFalse: Currency): Currency;
begin
  if ABool then
    Result := ATrue
  else
    Result := AFalse;
end;

end.
