{*****************************************************************************}
{                                                                             }
{  llBytes.pas                                                                }
{                                                                             }
{  Encoding and simple encryption to/from byte arrays.                        }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llBytes;

interface

uses SysUtils, Character, Math;

const
  HEX_CHARS = '0123456789ABCDEF';
  BASE64_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=';

  function HtoI(const AHexChar: Char): Integer;
  function IsHex(const AHex: string): Boolean;
  function BytesToHex(const ABytes: TBytes): string;
  function HexToBytes(const AHex: string): TBytes;

  function IsBase64(const ABase64: string): Boolean;
  function BytesToBase64(const ABytes: TBytes): string;
  function Base64ToBytes(const ABase64: string): TBytes;

  function BytesToASCII(const ABytes: TBytes): string;
  function ASCIIToBytes(const ASCII: string): TBytes;

  function EncodeURLParam(const S: string): string;

implementation

uses llStrings; {TokenList for ASCIItoBytes}

{+}
function HtoI(const AHexChar: Char): Integer;
{
  Returns a value 0..15 for a valid hex character, or -1 otherwise.
  Case insensitive.

  See also: IsHex, BytesToHex, HexToBytes
-}
const
  HexVals: array['0'..'f'] of Integer =
    ( 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,-1,-1,-1,-1,-1,-1, {0..9}
     -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1, {A..F}
     -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
     -1,10,11,12,13,14,15);                           {a..f}
begin
  if (AHexChar >= '0') and (AHexChar <= 'f') then
    Result := HexVals[AHexChar]
  else
    Result := -1;
end;

{+}
function IsHex(const AHex: string): Boolean;
{
  Returns True if the input string contains only valid hex characters. An empty
  string returns True. Case insensitive.

  See also: HtoI, BytesToHex, HexToBytes
-}
var
  i: Integer;
begin
  Result := False;
  for i := 1 to Length(AHex) do
    if HtoI(AHex[i]) < 0 then
      Exit;
  Result := True;
end;

{+}
function BytesToHex(const ABytes: TBytes): string;
{
  Returns a string of hex characters representing the array of bytes passed in.

  See also: HexToBytes, HtoI, IsHex
-}
var
  iLen, b, r: Integer;
begin
  iLen := Length(ABytes);
  SetLength(Result, iLen * 2);
  r := 0;
  for b := 0 to iLen - 1 do begin
    Inc(r);
    Result[r] := HEX_CHARS[(ABytes[b] shr 4) + 1];
    Inc(r);
    Result[r] := HEX_CHARS[(ABytes[b] and $F) + 1];
  end;
end;

{+}
function HexToBytes(const AHex: string): TBytes;
{
  Converts a string of hex characters into an array of bytes. The input MUST be
  an even number of hex characters. Raises an EConvertError on invalid input.

  See also: BytesToHex, HtoI, IsHex
-}
var
  iLen: Integer;
  i, r, h: Integer;
  function GetVal: Integer;
  begin
    Result := HtoI(AHex[i]);
    if Result < 0 then
      raise EConvertError.CreateFmt(
        'HexToBytes invalid character "%s" at position %d', [AHex[i], i]);
  end;
begin
  // Input must be an even-length string of valid hex characters
  iLen := Length(AHex);
  if iLen mod 2 = 1 then
    raise EConvertError.Create('HexToBytes input length must be even.');
  SetLength(Result, iLen div 2);
  i := 1;
  r := 0;
  while i < iLen do begin
    h := GetVal shl 4;
    Inc(i);
    h := h + GetVal;
    Inc(i);
    Result[r] := h;
    Inc(r);
  end;
end;

{+}
function IsBase64(const ABase64: string): Boolean;
{
  Returns True if the input string contains only valid Base64 characters with a
  legal length.

  See also: BytesToBase64, Base64ToBytes
-}
var
  iLen, i: Integer;
begin
  Result := False;
  // Valid length?
  iLen := Pos(BASE64_CHARS[65], ABase64);
  if iLen > 0 then
    Dec(iLen)
  else
    iLen := Length(ABase64);
  if (iLen mod 4) = 1 then
    Exit;
  // Valid characters?
  for i := 1 to iLen do
    if Pos(ABase64[i], BASE64_CHARS) < 1 then
      Exit;
  Result := True;
end;

{+}
function BytesToBase64(const ABytes: TBytes): string;
{
  Encodes an array of bytes as a Base64 string. Empty input results in empty
  output.

  See also: Base64ToBytes, IsBase64
-}
var
  iLen, b, r, x, sh: Integer;
begin
  { The output is always a multiple of four characters, with 0-2 padding bytes. }
  iLen := Length(ABytes);
  r := Ceil(iLen / 3) * 4;
  SetLength(Result, r);
  if r = 0 then
    Exit;
  { Turn groups of three 8-bit bytes into four 6-bit characters: }
  b := 0;
  r := 1;
  while b < iLen do begin
    x := 0;
    sh := 16;
    while (b < iLen) and (sh >= 0) do begin
      x := x + ABytes[b] shl sh;
      Inc(b);
      Dec(sh, 8);
    end;
    sh := 18;
    while sh >= 0 do begin
      Result[r] := BASE64_CHARS[((x shr sh) and 63) + 1];
      Inc(r);
      Dec(sh, 6);
    end;
  end;
  { Pad with equal signs if the input wasn't a multiple of 24 bits: }
  b := iLen mod 3;
  if b = 1 then
    Result[r - 2] := BASE64_CHARS[65];
  if b > 0 then
    Result[r - 1] := BASE64_CHARS[65];
end;

{+}
function Base64ToBytes(const ABase64: string): TBytes;
{
  Decodes a Base64 string into an array of bytes. Invalid input will raise an
  EConvertError exception.

  See also: BytesToBase64, IsBase64
-}
var
  iLen, iRes, i, r, x, y, sh: Integer;
begin
  { Ignore any equal sign padding: }
  iLen := Pos(BASE64_CHARS[65], ABase64);
  if iLen > 0 then
    Dec(iLen)
  else
    iLen := Length(ABase64);
  { Pre-calculate the length of the result:
    After stripping any padding, the input length must be an even multiple of 4
    characters, or have 2 or 3 characters remaining. }
  i := iLen mod 4;
  if i = 1 then
    raise EConvertError.Create('Illegal Base64 data length');
  iRes := ((iLen div 4) * 3) + // full 24-bit runs +
          ((i * 6) div 8);     // 8-bit chars from remaining 6-bit chars
  SetLength(Result, iRes);
  // Turn groups of four 6-bit characters into three 8-bit characters:
  i := 1;
  r := 0;
  while i < iLen do begin // Use < as there should never be only 1 char left!
    x := 0;
    sh := 18;
    while (i <= iLen) and (sh >= 0) do begin
      y := Pos(ABase64[i], BASE64_CHARS) - 1;
      if y < 0 then
        raise EConvertError.CreateFmt(
          'Illegal Base64 character "%s" at position %d', [ABase64[i], i]);
      x := x + (y shl sh);
      Inc(i);
      Dec(sh, 6);
    end;
    sh := 16;
    while (r <= iRes) and (sh >= 0) do begin
      Result[r] := (x shr sh) and 255;
      Inc(r);
      Dec(sh, 8);
    end;
  end;
end;

{+}
function BytesToASCII(const ABytes: TBytes): string;
{
  Converts an array of bytes into a Delphi-style ASCII string (i.e. '#13#10').
  Empty input results in empty output.

  See also: ASCIIToBytes
-}
var
  i: Integer;
begin
  Result := '';
  for i := 0 to Length(ABytes) - 1 do
    Result := Result + '#' + IntToStr(ABytes[i]);
end;

{+}
function ASCIIToBytes(const ASCII: string): TBytes;
{
  Converts a string in Delphi-style ASCII notation (e.g. '#13#10') into an array
  of bytes. Invalid input will raise an EConvertError exception.

  See also: BytesToASCII
-}
var
  iLen, i, j, r, b: Integer;
  sVal: string;
begin
  iLen := Length(ASCII);
  SetLength(Result, iLen); // Will always be <= the input length
  if iLen = 0 then
    Exit;
  if ASCII[1] <> '#' then
    raise EConvertError.Create('ASCIIToBytes: First character must be #');
  r := 0;
  j := 2;
  repeat
    i := Pos('#', ASCII, j);
    if i = 0 then
      i := iLen + 1;
    sVal := Copy(ASCII, j, i - j);
    b := StrToInt(sVal);
    if (b < 0) or (b > 255) then
      raise EConvertError.CreateFmt(
        'ASCIIToBytes invalid value #%s at position %d', [sVal, j - 1]);
    Result[r] := b;
    Inc(r);
    j := i + 1;
  until j > iLen;
  SetLength(Result, r);
end;

{+}
function EncodeURLParam(const S: string): string;
{
  Encodes a URL parameter with percent/hex escaping after first converting to
  UTF-8. Meant for single parameters only, not entire URLs because it encodes
  things like equal signs and ampersands, which are used to separate parameters
  in a URL.
-}
var
  Bytes: TBytes;
  r, i: Integer;
  c: Char;
begin
  Bytes := TEncoding.UTF8.GetBytes(S);
  SetLength(Result, Length(Bytes) * 3); // Output may be up to 3x input
  r := 0;
  for i := 0 to Length(Bytes) - 1 do begin
    Inc(r);
    c := Char(Bytes[i]);
    // 48..122 is 0..z, the rest is the stuff in the middle (underscore allowed)
    if (Bytes[i] < 48) or (Bytes[i] > 122) or (Pos(c, ':;<=>?@[\]^`') > 0) then begin
      Result[r] := '%';
      Inc(r);
      Result[r] := HEX_CHARS[(Bytes[i] shr 4) + 1];
      Inc(r);
      Result[r] := HEX_CHARS[(Bytes[i] and $F) + 1];
    end else
      Result[r] := c;
  end;
  SetLength(Result, r);
end;

end.
