{*****************************************************************************}
{                                                                             }
{  llCSV.pas                                                                  }
{                                                                             }
{  CSV utility routines and classes.                                          }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llCSV;

interface

uses
  Classes, SysUtils;

type
  {+ /trim=2 TCSVFile}
  TCSVQuotingStyle = (qsNormal, qsSpaces, qsEverything);

  TCSVFile = class(TObject)
  {-}
  private
    FData: TStringList;
    FFields: TStringList;
    FHeader: TStringList;
    FLastRow: Integer;
    FEatWS: Boolean;
    FMultiLine: Boolean;
    FHasHeader: Boolean;
    FHasMultiLine: Boolean;
    FDataError: string;
    FDiffRow: Integer;
    FFilename: string;
    FNewLine: string;
    FQStyle: TCSVQuotingStyle;
    procedure SetNewLine(const Value: string);
    function GetDelim: Char;
    function GetQuote: Char;
    function GetRowCount: Integer;
    procedure GetStartFinish(const IncludeHeader: Boolean; const ARow: Integer;
                             out s, f: Integer);
    procedure ParseHeader;
    procedure ParseRow(const ARow: Integer);
    procedure SetDelim(const Value: Char);
    procedure SetHasHeader(const Value: Boolean);
    procedure SetQuote(const Value: Char);
    procedure SetQStyle(const Value: TCSVQuotingStyle);
    procedure StoreData(const ARow: Integer);
    function Validate: Boolean;
  {+ /trim=2 /addline=OFF TCSVFile}
  public
    constructor Create; overload;
    destructor Destroy; override;               // Property defaults:
    property DataError: string read FDataError; // ''
    property Delimiter: Char read GetDelim write SetDelim; // comma (,)
    property EatWhitespace: Boolean read FEatWS write FEatWS; // False
    property Filename: string read FFilename write FFilename; // ''
    property HasHeader: Boolean read FHasHeader write SetHasHeader; // False
    property HasMultiLine: Boolean read FHasMultiLine;
    property NewLine: string read FNewLine write SetNewLine; // '\r\n'
    property QuoteChar: Char read GetQuote write SetQuote; // double quote (")
    property QuotingStyle: TCSVQuotingStyle read FQStyle write SetQStyle; // qsNormal
    property RowCount: Integer read GetRowCount;
    property CountsDifferAtRow: Integer read FDiffRow write FDiffRow;
    procedure AddField(const ARow: Integer; const Value: string);
    procedure AddRow;
    procedure ChangeDelimiters(const ANewQuote, ANewDelim: Char);
    procedure ClearRow(const ARow: Integer);
    procedure DeleteField(const ARow, ACol: Integer);
    procedure DeleteRow(const ARow: Integer);
    function FieldCount(const ARow: Integer): Integer;
    function GetColByName(const AName: string): Integer;
    function GetFieldName(const ACol: Integer): string;
    function GetValue(const ARow, ACol: Integer): string;
    procedure InsertField(const ARow, ACol: Integer; const Value: string);
    procedure InsertRow(const ARow: Integer);
    procedure LoadFromFile(const AFilename: string; const AHasHeader: Boolean);
    procedure MoveField(const ARow, AFrom, ATo: Integer);
    procedure MoveRow(const AFrom, ATo: Integer);
    function RowText(const ARow: Integer): string;
    procedure Save(const ASplitMultiLine: Boolean = False);
    procedure SaveToFile(const AFilename: string;
                         const ASplitMultiLine: Boolean = False);
    procedure SetFieldName(const ACol: Integer; const AName: string);
    procedure SetValue(const ARow, ACol: Integer; const Value: string);
  end;
  {+ /trim=OFF /addline=OFF

  TCSVFile is a utility class for working with CSV files. It uses the other CSV
  utility routines to parse and store the data.

  It loads the whole file and keeps it in memory using a TStringList, so be
  sensible about the size of files you load. It's not designed for giant files!

  The file is parsed entirely when loaded to check for errors. If DataError is
  non-blank after calling LoadFromFile, you have bad data. Parsing stops as soon
  as the first error is encountered, so the data should not be used.

  If all of the rows in the file don't have the same field count, the
  CountsDifferAtRow property will contain the row index of the first row that
  has a different count than previous rows. This is only checked at load time.

  During the load, any multi-line fields are put back together using the NewLine
  property. When saving, set the ASplitMultiLine argument to True to have them
  broken back out into multiple lines at each instance of NewLine. The
  HasMultiLine property will be set to True after loading if any of these were
  found.

  Only one row is kept parsed into fields at a time. Trying to group all field
  operations for a given row together will minimize the need to re-parse the
  data.

  The methods related to Field Names only work when HasHeader is True. Calling
  GetColByName without a header will return -1.

  A number of the methods that operate on fields can operate over the whole file
  if the ARow argument is set to -1. The following methods allow this:

  AddField
  DeleteField
  InsertField
  MoveField
  SetValue (Doesn't affect row zero if HasHeader is True)

  The quote and delimiter characters can be changed by setting the respective
  properties. If changing both at the same time, the ChangeDelimiters method is
  less expensive.

  The QuotingStyle only affects how the data is stored and written. Any quoting
  style can be read regardless of what the current setting is. qsNormal quotes
  only values that contain the delimiter or other quotes (standard DelimitedText
  style). qsSpaces will also quote any fields that contain spaces or tabs (uses
  BuildCSV), since these confuse TStringList.CommaText without StrictDelimiter.
  qsEverything will quote every field (using BuildCSV with ForceQuote=True).

  The ClearRow method deletes the data from the fields in a row, but leaves the
  fields there (1,2,3 -> ,,) while DeleteRow removes the row entirely. The Add/
  InsertRow methods create new rows with the correct number of empty fields.

  The RowText function returns an entire row as it is currently stored, with
  quoting and delimiters.

  See also: BuildCSV, ParseCSVLine
  -}

{+}
function BuildCSV(AFields: TStrings; const QS: TCSVQuotingStyle): string; overload;
function BuildCSV(AFields: TStrings; const AForceQuote: Boolean = False): string; overload; deprecated;
{
  Works like TStrings.DelimitedText, but sometimes you need to force quoting of
  certain values or all fields.

  The quoting style argument controls what fields get quoted:
  qsNormal:     Only fields that contain QuoteChar or Delimiter.
  qsSpaces:     As above, but also includes spaces and tabs.
  qsEverything: All fields are quoted, including empty ones.

  The QuoteChar and Delimiter are determined by the AFields properties of the
  same names.

  The deprecated boolean overload version works like qsSpaces unless ForceQuote
  is True, when it works like qsEverything.
-}

{+}
function ParseCSVLine(const ALine: string;
                      Fields: TStrings;
                      var MultiLine: Boolean;
                      const EatWhitespace: Boolean = False;
                      const AQuote: Char = '"';
                      const ADelim: Char = ',';
                      const ANewline: string = #10;
                      const APreserveQuoting: Boolean = False): Boolean;
{
  Parses a string of CSV-formatted text into individual fields and places them
  into a TStrings list. Unlike TStringList.CommmaText/DelimitedText*, it only
  recognizes a single delimiter character and will not break unquoted fields
  that contain spaces or tabs. It can also handle fields that span multiple
  lines.

  *As of BDS2006, TStrings has a StrictDelimiter property to ignore all but the
  specified delimiter character.

  --- Returns:

  A boolean indicating if ALine contained properly formatted CSV data.
  Currently, only improperly quoted fields will trigger a False result.

  --- Arguments:

  ALine - A line of text from a CSV file.

  Fields - A TStrings/TStringlist object to hold the parsed fields.

  MultiLine - A var parameter that gets set to True when the procedure needs
    the next line from the CSV file to continue the current field/record.

  EatWhitespace - A boolean to determine behavior with regards to unquoted
    whitespace. If False (the default), unquoted whitespace is preserved. If
    True, this whitespace is discarded. Quoted whitespace is always preserved,
    while whitespace between quoted fields and delimiters/EOL is always
    discarded.

    Examples: Given the following input: ' aaa, bbb , "ccc" , " d " '

     EatWhitespace | Input is parsed equivalent to...
    ---------------+----------------------------------
         False     | '" aaa"," bbb ","ccc"," d "'
         True      | '"aaa","bbb","ccc"," d "'

  AQuote - The text quote character. Default=double quote (")

  ADelim - The field delimiter character. Default=comma (,)

  ANewline - The character(s) to be used in place of the actual end-of-line
    characters (i.e. #13#10/CRLF) when storing multiline fields in the Fields
    stringlist. Default=#10/LF. Other possible values: '\r\n', '<br/>', etc.

  APreserveQuoting - If set to True, quoting will not be stripped out of the
    fields parsed from the line of text. The Fields data will contain the raw
    data, with any surrounding quotes and doubled literals still present.

  --- Usage:

  Initially, call the procedure with MultiLine set to False. The Fields list 
  will be cleared and the line parsed according to the other options.

  If the function encounters a quoted field that is continued on the next line
  it will set MultiLine to True and return. The caller should then call the
  procedure again, passing the next line from the CSV file with MultiLine still
  set to True. The parsing of the record will continue where it left off, adding
  to the Fields list. MultiLine will be set to False when it finds the end of
  the record.

  When the field list contains multiline fields, the ANewline character(s) will
  be inserted in place of the actual end-of-line characters (usually #13#10 aka
  CRLF). Select a character or characters that won't get confused with the rest
  of the data and won't cause Fields.Count to be erroneously increased. It can
  be as simple as #10, something fairly common like "\r\n", or as wacky as a
  dozen alternating tildes & underscores. The caller can then replace this value
  with the desired line terminator using something like this:

  Memo.Text := StringReplace(Fields[x], ANewLine, #13#10, [rfReplaceAll]);

  The CSV RFC: http://www.ietf.org/rfc/rfc4180.txt

  See Also: SimpleCSV, TokenList

-}

{+}
function SimpleCSV(const ALine: string; Fields: TStrings;
                   const EatWhitespace: Boolean = True): Boolean;
{
  This is a simplified wrapper to ParseCSVLine. You need not provide a boolean
  MultiLine variable and EatWhitespace is set to True by default. The quote and
  delimiter characters are taken directly from Fields.QuoteChar and
  Fields.Delimiter.

  The function just returns the value returned by ParseCSVLine, but will also
  return False if MultiLine got set to True, indicating that you're not working
  with simple, single-line CSV data.

  See also: ParseCSVLine, TokenList
-}

implementation

uses llStrings;

function BuildCSV(AFields: TStrings; const QS: TCSVQuotingStyle): string; overload;
var
  i, j, iLen: Integer;
  sChars: string;
  bQuote: Boolean;
begin
  Result := '';
  if AFields.Count = 0 then
    Exit;
  if QS = qsEverything then begin
    Result := Enquote(AFields[0], AFields.QuoteChar);
    for i := 1 to AFields.Count - 1 do
      Result := Result + AFields.Delimiter + Enquote(AFields[i], AFields.QuoteChar);
  end else begin
    sChars := AFields.QuoteChar + AFields.Delimiter;
    if QS = qsSpaces then
      sChars := sChars + ' ' + #9;
    for i := 0 to AFields.Count - 1 do begin
      // Does this field need to be quoted?
      bQuote := False;
      iLen := Length(AFields[i]);
      j := 1;
      while (j <= iLen) and not bQuote do begin
        bQuote := IsDelimiter(sChars, AFields[i], j);
        Inc(j);
      end;
      if i > 0 then
        Result := Result + AFields.Delimiter;
      if bQuote then
        Result := Result + Enquote(AFields[i], AFields.QuoteChar)
      else
        Result := Result + AFields[i];
    end;
  end;
end;

function BuildCSV(AFields: TStrings; const AForceQuote: Boolean = False): string; overload; deprecated;
begin
  if AForceQuote then
    Result := BuildCSV(AFields, qsEverything)
  else
    Result := BuildCSV(AFields, qsSpaces);
end;

function ParseCSVLine(const ALine: string;
                      Fields: TStrings;
                      var MultiLine: Boolean;
                      const EatWhitespace: Boolean;
                      const AQuote, ADelim: Char;
                      const ANewline: string;
                      const APreserveQuoting: Boolean): Boolean;
var
  iLen, iPos, iAdj, iFld: Integer;
  cSpace: Char;
  sField: string;
  bDone: Boolean;
  procedure AdjustPos;
  begin
    // Adjusts iPos to skip past any unwanted leading whitespace.
    if EatWhitespace then begin
      // It's an unconditional trim:
      while (iPos <= iLen) and
            (ALine[iPos] <= cSpace) and
            (ALine[iPos] <> ADelim) and
            (ALine[iPos] <> AQuote) do
        Inc(iPos);
    end else begin
      // Only trim if the first non-whitespace character is a quote:
      iAdj := iPos;
      while (iAdj < iLen) and
            (ALine[iAdj] <= cSpace) and
            (ALine[iAdj] <> ADelim) and
            (ALine[iAdj] <> AQuote) do begin
        if ALine[iAdj + 1] = AQuote then begin
          iPos := iAdj + 1;
          Break;
        end;
        Inc(iAdj);
      end;
    end;
  end; // AdjustPos
  procedure AddPreservedQuote;
  begin
    if APreserveQuoting then begin
      Inc(iFld);
      sField[iFld] := AQuote;
    end;
  end;
begin
  Result := True;
  cSpace := ' ';
  iPos := 1;
  iLen := Length(ALine);
  SetLength(sField, iLen); // Preallocate. No single field can be > the whole!

  // If MultiLine is True upon entry, we're continuing the previous record.
  if MultiLine then begin
    Assert(Fields.Count > 0,
      'ParseCSVLine: Fields.Count must be > 0 if continuing a MultiLine record.');
    // A completely blank line in the middle of a multi-line field?
    if iLen = 0 then begin
      Fields[Fields.Count - 1] := Fields[Fields.Count - 1] + ANewline;
      Exit;
    end;
  end else begin
    Fields.Clear;
    AdjustPos;
  end;

  while iPos <= iLen do begin

    iFld := 0; // Reset the field character count

    // Is this field surrounded by quotes or the continuation of a quoted field
    // from the previous line?
    if MultiLine or (ALine[iPos] = AQuote) then begin

      if not MultiLine then begin // Skip the opening quote
        Inc(iPos);
        AddPreservedQuote;
      end;

      bDone := False;
      while iPos <= iLen do begin
        if ALine[iPos] <> AQuote then begin
          // Not a quote - add this character to our field and continue.
          Inc(iFld);
          sField[iFld] := ALine[iPos];
        end else if iPos = iLen then begin
          // Last character is a quote - we're done!
          Inc(iPos);
          bDone := True;
          break;
        end else if ALine[iPos + 1] = ADelim then begin
          // Quote followed by a delimiter - we're done!
          Inc(iPos, 2);
          bDone := True;
          break;
        end else if ALine[iPos + 1] = AQuote then begin
          // It's a doubled, literal quote.  Keep one and skip the other.
          Inc(iFld);
          sField[iFld] := AQuote;
          Inc(iPos);
          AddPreservedQuote;
        end else begin
          { If only whitespace appears before reaching the next delimiter or
            EOL, we're done and can discard the post-quote whitespace.
            If not, then this data is badly formatted and the next step is
            undefined. I have chosen to continue with the current field, as if
            it were an "undoubled literal". }
          iAdj := iPos + 1;
          while iAdj <= iLen do begin
            if ALine[iAdj] = ADelim then begin
              iPos := iAdj + 1;
              bDone := True;
              break;
            end;
            if (ALine[iAdj] > cSpace) or (ALine[iAdj] = AQuote) then begin
              Result := False; // Your data smells of elderberries!
              Inc(iFld);
              sField[iFld] := ALine[iPos]; // Grab this char and keep going
              break;
            end;
            Inc(iAdj);
          end;
          if iAdj > iLen then begin
            iPos := iAdj; // We reached EOL encountering only whitespace
            bDone := True;
          end;
          if bDone then
            break;
        end;
        Inc(iPos);
      end; // while iPos <= iLen (in a quoted field)

      if bDone then
        AddPreservedQuote; // Add the closing quote

      // Add the field to the list:
      if MultiLine then
        Fields[Fields.Count - 1] :=
          Fields[Fields.Count - 1] + Copy(sField, 1, iFld)
      else
        Fields.Add(Copy(sField, 1, iFld));

      // If we reached EOL without finding a closing quote, it means that this
      // field is continued on the next line:
      if not bDone then begin
        Fields[Fields.Count - 1] := Fields[Fields.Count - 1] + ANewline;
        MultiLine := True;
        Exit;
      end;

    end else begin // Not a quoted field - just find the next delimiter or EOL:

      while iPos <= iLen do begin
        if ALine[iPos] = ADelim then begin
          Inc(iPos);
          break;
        end else begin
          Inc(iFld);
          sField[iFld] := ALine[iPos];
          if ALine[iPos] = AQuote then
            Result := False; // Stray/mismatched quotes!
        end;
        Inc(iPos);
      end;
      if EatWhitespace then // Perform a "TrimRight" on our field:
        while (iFld > 0) and (sField[iFld] <= cSpace) do
          Dec(iFld);
      Fields.Add(Copy(sField, 1, iFld));

    end; // if (not) quoted

    AdjustPos;
    MultiLine := False;

  end; // while iPos <= iLen

  { If the last non-whitespace character is a delimiter, we need to add a blank
    field to the list. If EatWhitespace is False, there is no need to check past
	  the very last character since any spaces would have already been added, as
	  opposed to having been skipped over by AdjustPos. }
  while iLen > 0 do begin
    if ALine[iLen] = ADelim then begin
      Fields.Add('');
      break;
    end;
    if (ALine[iLen] > cSpace) or
       (ALine[iLen] = AQuote) or
       not EatWhitespace then
      break;
    Dec(iLen);
  end;

end;

function SimpleCSV(const ALine: string; Fields: TStrings;
                   const EatWhitespace: Boolean = True): Boolean;
var
  bMultiLine: Boolean;
begin
  bMultiLine := False;
  Result := ParseCSVLine(ALine, Fields, bMultiLine, EatWhitespace,
                         Fields.QuoteChar, Fields.Delimiter);
  if bMultiLine then
    Result := False; // You should have used the full-featured version!
end;

{ TCSVFile }

procedure TCSVFile.AddField(const ARow: Integer; const Value: string);
var
  i, s, f: Integer;
begin
  GetStartFinish(True, ARow, s, f);
  for i := s to f do begin
    ParseRow(i);
    FFields.Add(Value);
    StoreData(i);
  end;
  if FHasHeader and (s = 0) then
    ParseHeader;
end;

procedure TCSVFile.AddRow;
begin
  InsertRow(FData.Count);
end;

procedure TCSVFile.ChangeDelimiters(const ANewQuote, ANewDelim: Char);
var
  slTemp: TStringList;
  i: Integer;
begin
  // Replace the quotes/delimiters in each line:
  FMultiLine := False;
  slTemp := TStringList.Create;
  try
    slTemp.StrictDelimiter := True;
    slTemp.QuoteChar := ANewQuote;
    slTemp.Delimiter := ANewDelim;
    for i := 0 to FData.Count - 1 do begin
      ParseRow(i);
      slTemp.Clear;
      slTemp.AddStrings(FFields);
      FData[i] := BuildCSV(slTemp, FQStyle);
    end;
    FFields.QuoteChar := ANewQuote;
    FFields.Delimiter := ANewDelim;
    FFields.Clear;
    FLastRow := -1;
  finally
    slTemp.Free;
  end;
end;

procedure TCSVFile.ClearRow(const ARow: Integer);
begin
  FData.Delete(ARow);
  InsertRow(ARow);
  if FLastRow = ARow then
    FLastRow := -1;
end;

constructor TCSVFile.Create;
begin
  FData := TStringList.Create;
  FHeader := TStringList.Create;
  FHeader.StrictDelimiter := True;
  FFields := TStringList.Create;
  FFields.StrictDelimiter := True;
  FLastRow := -1;
  FEatWS := False;
  FHasHeader := False;
  FHasMultiLine := False;
  FDiffRow := -1;
  FDataError := '';
  FFilename := '';
  FNewLine := '\r\n';
  FQStyle := qsNormal;
end;

procedure TCSVFile.DeleteField(const ARow, ACol: Integer);
var
  i, s, f: Integer;
begin
  GetStartFinish(True, ARow, s, f);
  for i := s to f do begin
    ParseRow(i);
    FFields.Delete(ACol);
    StoreData(i);
  end;
  if FHasHeader and (s = 0) then
    ParseHeader;
end;

procedure TCSVFile.DeleteRow(const ARow: Integer);
begin
  FData.Delete(ARow);
  if FLastRow >= ARow then
    FLastRow := -1;
end;

destructor TCSVFile.Destroy;
begin
  FData.Free;
  FFields.Free;
  FHeader.Free;
  inherited;
end;

function TCSVFile.FieldCount(const ARow: Integer): Integer;
begin
  Result := 0;
  if ARow < FData.Count then begin
    ParseRow(ARow);
    Result := FFields.Count;
  end else if FHasHeader then
    Result := FHeader.Count;
end;

function TCSVFile.GetColByName(const AName: string): Integer;
begin
  if FHasHeader then
    Result := FHeader.IndexOf(AName)
  else
    Result := -1;
end;

function TCSVFile.GetDelim: Char;
begin
  Result := FFields.Delimiter;
end;

function TCSVFile.GetFieldName(const ACol: Integer): string;
begin
  if FHasHeader then
    Result := FHeader[ACol]
  else
    Result := '';
end;

function TCSVFile.GetQuote: Char;
begin
  Result := FFields.QuoteChar;
end;

function TCSVFile.GetRowCount: Integer;
begin
  Result := FData.Count;
end;

procedure TCSVFile.GetStartFinish(const IncludeHeader: Boolean;
                                  const ARow: Integer;
                                  out s, f: Integer);
begin
  if ARow >= 0 then begin
    s := ARow;
    f := ARow;
  end else begin
    if FHasHeader and not IncludeHeader then
      s := 1
    else
      s := 0;
    f := FData.Count - 1;
  end;
end;

function TCSVFile.GetValue(const ARow, ACol: Integer): string;
begin
  // Not checking for index out of range - it will just raise EListError anyway
  ParseRow(ARow);
  if ACol <= FFields.Count then
    Result := FFields[ACol]
  else
    Result := '';
end;

procedure TCSVFile.InsertField(const ARow, ACol: Integer; const Value: string);
var
  i, s, f: Integer;
begin
  GetStartFinish(True, ARow, s, f);
  for i := s to f do begin
    ParseRow(i);
    FFields.Insert(ACol, Value);
    StoreData(i);
  end;
  if FHasHeader and (s = 0) then
    ParseHeader;
end;

procedure TCSVFile.InsertRow(const ARow: Integer);
var
  iCount: Integer;
  slTemp: TStringList;
begin
  if FHasHeader then
    iCount := FHeader.Count
  else if FLastRow >= 0 then
    iCount := FFields.Count
  else if FData.Count > 0 then begin
    ParseRow(0);
    iCount := FFields.Count;
  end else
    iCount := 0;
  slTemp := TStringList.Create;
  try
    while slTemp.Count < iCount do
      slTemp.Add('');
    FData.Insert(ARow, BuildCSV(slTemp, FQStyle));
  finally
    slTemp.Free;
  end;
  if ARow <= FLastRow then
    FLastRow := -1;
end;

procedure TCSVFile.LoadFromFile(const AFilename: string;
                                const AHasHeader: Boolean);
begin
  FDataError := '';
  if FFilename = '' then
    FFilename := AFilename;
  FHasHeader := AHasHeader;
  FHeader.Clear;
  FFields.Clear;
  FLastRow := -1;
  FData.LoadFromFile(AFilename);
  if FHasHeader then
    ParseHeader;
  if FDataError = '' then
    Validate;
end;

procedure TCSVFile.MoveField(const ARow, AFrom, ATo: Integer);
var
  i, s, f: Integer;
begin
  GetStartFinish(True, ARow, s, f);
  for i := s to f do begin
    ParseRow(i);
    FFields.Move(AFrom, ATo);
    StoreData(i);
  end;
  if FHasHeader and (s = 0) then
    ParseHeader;
end;

procedure TCSVFile.MoveRow(const AFrom, ATo: Integer);
begin
  FData.Move(AFrom, ATo);
  if (FLastRow >= AFrom) or (FLastRow >= ATo) then
    FLastRow := -1;
end;

procedure TCSVFile.ParseHeader;
begin
  FHeader.Clear;
  if FData.Count = 0 then
    Exit;
  FMultiLine := False;
  if not ParseCSVLine(FData[0], FHeader, FMultiLine, FEatWS,
                      FFields.QuoteChar, FFields.Delimiter, FNewLine) then
    FDataError := 'Invalid data in header'
  else if FMultiLine then
    FDataError := 'MultiLine fields not supported in header';
end;

procedure TCSVFile.ParseRow(const ARow: Integer);
begin
  if (ARow <> FLastRow) or (FFields.Count = 0) then begin
    FMultiLine := False;
    if ParseCSVLine(FData[ARow], FFields, FMultiLine, FEatWS,
                    FFields.QuoteChar, FFields.Delimiter, FNewLine) then begin
      if FMultiLine then
        FDataError := Format('MultiLine error on row %d', [ARow]);
      FLastRow := ARow;
    end else
      FDataError := Format('Data error on row %d', [ARow]);
  end;
end;

function TCSVFile.RowText(const ARow: Integer): string;
begin
  Result := FData[ARow];
end;

procedure TCSVFile.Save(const ASplitMultiLine: Boolean);
begin
  SaveToFile(FFilename, ASplitMultiLine);
end;

procedure TCSVFile.SaveToFile(const AFilename: string;
                              const ASplitMultiLine: Boolean);
var
  i, j: Integer;
  slTemp, slOutput: TStringList;
begin
  Assert(AFilename <> '', 'TCSVFile.SaveToFile - Filename required.');
  if not ASplitMultiLine then
    FData.SaveToFile(AFilename)
  else begin
    slTemp := TStringList.Create;
    slOutput := TStringList.Create;
    try
      for i := 0 to FData.Count - 1 do begin
        if Pos(FNewLine, FData[i]) = 0 then
          slOutput.Add(FData[i])
        else begin
          slTemp.Clear;
          TokenList(FData[i], FNewLine, slTemp);
          for j := 0 to slTemp.Count - 1 do
            slOutput.Add(slTemp[j]);
        end;
      end;
      slOutput.SaveToFile(AFilename, FData.Encoding);
    finally
      slTemp.Free;
      slOutput.Free;
    end;
  end;
end;

procedure TCSVFile.SetDelim(const Value: Char);
begin
  if Value <> FFields.Delimiter then
    ChangeDelimiters(FFields.QuoteChar, Value);
end;

procedure TCSVFile.SetFieldName(const ACol: Integer; const AName: string);
begin
  if FHasHeader then begin
    FHeader[ACol] := AName;
    ParseRow(0);
    FFields[ACol] := AName;
    StoreData(0);
  end;
end;

procedure TCSVFile.SetHasHeader(const Value: Boolean);
begin
  if FHasHeader <> Value then begin
    if (not FHasHeader) and (FData.Count > 0) then
      ParseHeader;
    FHasHeader := Value;
    if not FHasHeader then
      FHeader.Clear;
  end;
end;

procedure TCSVFile.SetNewLine(const Value: string);
var
  i: Integer;
begin
  if Value <> FNewLine then begin
    for i := 0 to FData.Count - 1 do
      FData[i] := StringReplace(FData[i], FNewLine, Value, [rfReplaceAll]);
    FNewLine := Value;
  end;
end;

procedure TCSVFile.SetQuote(const Value: Char);
begin
  if Value <> FFields.QuoteChar then
    ChangeDelimiters(Value, FFields.Delimiter);
end;

procedure TCSVFile.SetQStyle(const Value: TCSVQuotingStyle);
var
  i: Integer;
begin
  if Value <> FQStyle then begin
    FQStyle := Value;
    for i := 0 to FData.Count - 1 do begin
      ParseRow(i);
      StoreData(i);
    end;
  end;
end;

procedure TCSVFile.SetValue(const ARow, ACol: Integer; const Value: string);
var
  i, s, f: Integer;
begin
  GetStartFinish(False, ARow, s, f); // Don't include the header if present
  for i := s to f do begin
    ParseRow(i);
    while ACol >= FFields.Count do
      FFields.Add('');
    FFields[ACol] := Value;
    StoreData(i);
  end;
  if FHasHeader and (s = 0) then
    ParseHeader;
end;

procedure TCSVFile.StoreData(const ARow: Integer);
begin
  // Save the current contents of FFields back into the given row
  FData[ARow] := BuildCSV(FFields, FQStyle);
  if FHasHeader and (ARow = 0) then
    ParseHeader;
end;

function TCSVFile.Validate: Boolean;
var
  i, iActualRow, iLeast, iMost: Integer;
begin
  // Parse every line of data, combining multi-line fields/rows into single rows.
  Result := False;
  FDataError := '';
  FHasMultiLine := False;
  FLastRow := -1;
  iLeast := MaxInt;
  iMost := -1;
  i := 0;
  iActualRow := -1;
  FDiffRow := -1;
  FMultiLine := False;
  while i < FData.Count do begin
    repeat
      if FMultiLine then
        FHasMultiLine := True;
      if not ParseCSVLine(FData[i], FFields, FMultiLine, FEatWS,
                          FFields.QuoteChar, FFields.Delimiter, FNewLine) then begin
        FDataError := Format('Error parsing line %d', [i]);
        Exit;
      end;
      Inc(i);
    until (i >= FData.Count) or not FMultiLine;
    if FMultiLine then begin
      FDataError := 'Incomplete MultiLine field on last line';
      Exit;
    end;
    Inc(iActualRow);
    StoreData(iActualRow);
    if FFields.Count < iLeast then
      iLeast := FFields.Count;
    if FFields.Count > iMost then
      iMost := FFields.Count;
    if (FDiffRow < 0) and (iLeast <> iMost) then
      FDiffRow := iActualRow;
  end;
  Inc(iActualRow);
  while FData.Count > iActualRow do
    FData.Delete(FData.Count - 1);
  Result := True;
end;

end.


