{*****************************************************************************}
{                                                                             }
{  llControls.pas                                                             }
{                                                                             }
{  VCL control utility routines                                               }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llControls;

interface

  uses
    SysUtils, Classes, Windows, Messages, StdCtrls, Checklst;

  function AnyChecked(Lbx: TCheckListBox): Boolean;
  procedure CheckListChange(Lbx: TCheckListBox; const AChange: string);
  procedure CheckListCheckObject(const Lbx: TCheckListBox;
                                 const AValue: Integer;
                                 const ACheck: Boolean = True);

  procedure ScrollToEnd(AMemo: TMemo); // TODO: Overload for other control types?
  procedure ScrollToTop(AMemo: TMemo);

  {+ /trim=2}
  function GetItemObjectInt(Ctl: TComboBox): Integer; overload;
  function GetItemObjectInt(Ctl: TListBox): Integer; overload;
  function GetItemObjectInt(Ctl: TCheckListBox): Integer; overload;
  {
    Gets an integer object value from the currently selected item in a
    combobox, listbox, or checklistbox.
    If no item is currently selected the function will return -1.
    Unassigned objects (nil) become zero when cast as integers.
  -}

  {+ /trim=2}
  procedure SetItemIndex(Ctl: TComboBox; const AValue: string;
                         const ADefault: Integer = -1); overload;
  procedure SetItemIndex(Ctl: TListBox; const AValue: string;
                         const ADefault: Integer = -1); overload;
  procedure SetItemIndex(Ctl: TCheckListBox; const AValue: string;
                         const ADefault: Integer = -1); overload;
  {
    Finds a string value in a combobox/listbox/checklistbox Items property and
    sets the ItemIndex to that value, or to ADefault if not found.
    Uses llStrings.IndexOfPartial to find partial matches.
  -}

  procedure SetTabStops(const AMemo: TMemo; const ATabChars: Integer);

implementation

uses
  llStrings; // IndexOfPartial

{+}
function AnyChecked(Lbx: TCheckListBox): Boolean;
{
  Returns True if any of the items in the TCheckListBox are checked.
-}
var
  i: Integer;
begin
  Result := True;
  for i := 0 to Lbx.Items.Count - 1 do
    if Lbx.Checked[i] then
      Exit;
  Result := False;
end;

{+}
procedure CheckListChange(Lbx: TCheckListBox; const AChange: string);
{
  Updates the check marks in a TCheckListBox depending on AChange:
  'A' = All
  'N' = None (default when AChange <> A or R)
  'R' = Reverse
-}
var
  i: Integer;
  bCheck: Boolean;
begin
  if SameText(AChange, 'R') then begin
    for i := 0 to Lbx.Items.Count - 1 do
      Lbx.Checked[i] := not Lbx.Checked[i];
  end else begin
    bCheck := SameText(AChange, 'A');
    for i := 0 to Lbx.Items.Count - 1 do
      Lbx.Checked[i] := bCheck;
  end;
end;

{+}
procedure CheckListCheckObject(const Lbx: TCheckListBox;
                               const AValue: Integer;
                               const ACheck: Boolean = True);
{
  Checks an item in a TCheckListBox whose integer Object value matches AValue.
  If ACheck is False, the item is un-checked. If there is more than one matching
  object in the list, they will all be [un-]checked.
-}
var
  i: Integer;
begin
  for i := 0 to Lbx.Items.Count - 1 do
    if Integer(Lbx.Items.Objects[i]) = AValue then
      Lbx.Checked[i] := ACheck;
end;

{+}
procedure ScrollToEnd(AMemo: TMemo);
{
  Moves the cursor position and scrolls to the end of the text in a TMemo.
-}
begin
  AMemo.SelStart := AMemo.GetTextLen;
  AMemo.SelLength := 0;
  AMemo.Perform(EM_SCROLLCARET, 0, 0);
end;

{+}
procedure ScrollToTop(AMemo: TMemo);
{
  Moves the cursor position and scrolls to the start of the text in a TMemo.
-}
begin
  AMemo.SelStart := 0;
  AMemo.SelLength := 0;
  AMemo.Perform(EM_SCROLLCARET, 0, 0);
end;

function GetItemObjectInt(Ctl: TComboBox): Integer; overload;
begin
  if Ctl.ItemIndex >= 0 then
    Result := Integer(Ctl.Items.Objects[Ctl.ItemIndex])
  else
    Result := -1;
end;

function GetItemObjectInt(Ctl: TListBox): Integer; overload;
begin
  if Ctl.ItemIndex >= 0 then
    Result := Integer(Ctl.Items.Objects[Ctl.ItemIndex])
  else
    Result := -1;
end;

function GetItemObjectInt(Ctl: TCheckListBox): Integer; overload;
begin
  if Ctl.ItemIndex >= 0 then
    Result := Integer(Ctl.Items.Objects[Ctl.ItemIndex])
  else
    Result := -1;
end;

procedure SetItemIndex(Ctl: TComboBox; const AValue: string;
                       const ADefault: Integer);
begin
  Ctl.ItemIndex := IndexOfPartial(Ctl.Items, AValue, ADefault);
end;

procedure SetItemIndex(Ctl: TListBox; const AValue: string;
                       const ADefault: Integer);
begin
  Ctl.ItemIndex := IndexOfPartial(Ctl.Items, AValue, ADefault);
end;

procedure SetItemIndex(Ctl: TCheckListBox; const AValue: string;
                       const ADefault: Integer);
begin
  Ctl.ItemIndex := IndexOfPartial(Ctl.Items, AValue, ADefault);
end;

{+}
procedure SetTabStops(const AMemo: TMemo; const ATabChars: Integer);
{
  Sets the approximate size of a tab stop in a TMemo.
-}
var
  iChars: Integer;
begin
  iChars := ATabChars * 4; // Average of 4 dialog units per character
  AMemo.Perform(EM_SETTABSTOPS, 1, NativeInt(@iChars));
end;

end.
