{*****************************************************************************}
{                                                                             }
{  llDates.pas                                                                }
{                                                                             }
{  Date manipulation and formatting routines                                  }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llDates;

interface

uses
  SysUtils, Classes, DateUtils, Math, Character;

const
  ANSI_DATE = 'yyyy-mm-dd';
  ANSI_DATETIME = ANSI_DATE + ' hh:nn:ss';
  ANSI_DATETIMEZ = ANSI_DATETIME + '.zzz';
  OneMillisecond = 1.0 / MSecsPerDay;
  MSecsPerHour = MSecsPerSec * SecsPerHour;
  MSecsPerMin  = MSecsPerSec * SecsPerMin;

type
  {+ /trim=2 TDateTimePrecision}
  TDateTimePrecision = (dtpMillennium, dtpCentury, dtpDecade,
                        dtpYear, dtpMonth, dtpDay,
                        dtpHour, dtpMinute, dtpSecond, dtpMillisecond,
                        dtpE = 0, dtpC = 1, dtpX = 2,
                        dtpY = 3, dtpM = 4, dtpD = 5,
                        dtpH = 6, dtpN = 7, dtpS = 8, dtpZ = 9, dtpMSec = 9);
{+ /trim=OFF}
const
  dtpWords: array[TDateTimePrecision] of string =
    ('Millennium', 'Century', 'Decade', 'Year', 'Month', 'Day',
     'Hour', 'Minute', 'Second', 'Millisecond');
  dtpLetters: array[TDateTimePrecision] of Char =
    ('E', 'C', 'X', 'Y', 'M', 'D', 'H', 'N', 'S', 'Z'); // E=Epoch, X=Roman ten
{
  The TDateTimePrecision enumeration is used by various functions in the
  llDates unit. The dtpMillisecond and dtpMSec values are synonyms, as are the
  shortened variations that match the format string letters (dtpY, dtpM, etc.).

  See also: StoDTP
-}
  lopTrimZeros = dtpMonth;   // Synonyms for IntervalStr
  lopDayOrHour = dtpYear;
  lopSmallSeps = dtpDecade;
  lopSpaceSeps = dtpCentury;

type
  {+ /trim=2 TDayOfWeek}
  TDayOfWeek = (dowAnyDay, dowSunday, dowMonday, dowTuesday,
                dowWednesday, dowThursday, dowFriday, dowSaturday,
                dowAny = 0, dowSun = 1, dowMon = 2, dowTue = 3,
                dowWed = 4, dowThu = 5, dowFri = 6, dowSat = 7);
  {+ /trim=OFF
  The ordinal values of the TDayOfWeek enumeration equate to the return values
  of the DayOfWeek function for dowSunday..dowSaturday. The abbreviated names
  are equivalent to their longer counterparts.
  dowAny[Day] is a custom value (0) used by the WeekdayOfMonth function.

  See also: WeekdayOfMonth
  -}

function AdjustDate(const AAging: string; var VDate: TDateTime): Boolean;
function AgeOnDate(const ADOB, ADate: TDateTime): Integer;
function AnsiDate(const DT: TDateTime): string;
function AnsiDateTime(const DT: TDateTime): string;
function AssumeCentury(const y: Integer): Integer;
function DateDiff(const AUnit: TDateTimePrecision;
                  const AStart, AEnd: TDateTime): Int64;
function DateMath(const ADate: TDateTime;
                  const y, m, d, h, n, s, z: Integer): TDateTime;
function DateTrunc(const ADate: TDateTime;
                   const P: TDateTimePrecision): TDateTime;
function EndOfDay(const DT: TDateTime): TDateTime;
function ExtractTZ(var VDate: string): string;
function FirstOfMonth(const ADate: TDateTime): TDateTime;
function IntervalStr(const AInterval: TDateTime;
                     const LoP: TDateTimePrecision = dtpHour;
                     const HiP: TDateTimePrecision = dtpSecond;
                     const ADaySeparator: string = 'd'): string;
function ISO8601Date(const S: string; const LocalOffset: TDateTime = 0): TDateTime;
function ISO8601Str(const D: TDateTime; const P: TDateTimePrecision;
                    const UseT: Boolean = True;
                    const Basic: Boolean = False): string;
function IsValidTimeZone(const ATZ: string;
                         const AllowZulu: Boolean = True;
                         const AllowColon: Boolean = True;
                         const AllowHourOnly: Boolean = True): Boolean;
function LastDOM(const ADate: TDateTime): Integer; overload;
function LastDOM(const AMonth: Integer; const AYear: Integer = 0): Integer; overload;
function LastOfMonth(const ADate: TDateTime): TDateTime;
function LocalUTCOffset: TDateTime;
function MatchingMonthDay(const ADate: TDateTime; const ADay: Integer): Boolean;
function MinutesToTime(const AMinutes: Cardinal): TDateTime;
function MonthDiff(const AStart, AEnd: TDateTime): Currency;
function MonthStrToInt(const AMonth: string): Integer;
function NextMonthDay(const ADate: TDateTime; const ADay: Integer): TDateTime;
function PrevMonthDay(const ADate: TDateTime; const ADay: Integer): TDateTime;
function SameTime(const d1, d2: TDateTime): Boolean;
function MStoTime(const AMS: Cardinal): TDateTime;
function SecondsToTime(const ASeconds: Cardinal): TDateTime;
function StoDTP(const S: string;
                const ADefault: TDateTimePrecision): TDateTimePrecision;
function TimeToMinutes(const ATime: TDateTime): Cardinal;
function TimeToMS(const ATime: TDateTime): Cardinal;
function TimeToSeconds(const ATime: TDateTime): Cardinal;
function TimeZoneOffset(const ATimeZoneStr: string): TDateTime;
function TimeZoneStr(const ALocalOffset: TDateTime;
                     const AZulu: Boolean = True;
                     const AColon: Boolean = True;
                     const AHourOnly: Boolean = True): string;
function WeekDayDiff(const AStart, AEnd: TDateTime): Integer;
function WeekdayNum(const ADate: TDateTime): Integer;
function WeekdayOfMonth(const ADate: TDateTime; const Nth: Integer;
                        const DOW: TDayOfWeek): TDateTime;

function DtoS(const DT: TDateTime;
              const P: TDateTimePrecision = dtpDay): string;
function StoD(const DS: string): TDateTime;
function StoDDef(const DS: string; const ADefault: TDateTime): TDateTime;
function TryStoD(const DS: string; out ADate: TDateTime): Boolean;

function TtoS(const DT: TDateTIme;
              const P: TDateTimePrecision = dtpSecond): string;
function StoT(const TS: string): TDateTime;
function StoTDef(const TS: string; const ADefault: TDateTime): TDateTime;
function TryStoT(const TS: string; out ATime: TDateTime): Boolean;

function FormatDateTimeEx(const AFormat: string; ADateTime: TDateTime): string;
function ReplaceDateTokens(const APattern: string;
                           const ADate: TDateTime): string;
function TryAnyStrToDateTime(const DS: string; var DayPos: Integer;
                             out OutDate: TDateTime): Boolean;
function TryFormatToDateTime(const DS, AFormat: string;
                             out OutDate: TDateTime): Boolean;

implementation

uses Windows, llStrings;

const
  UPPER_DIGITS = ALPHA_UPPER + DIGIT_CHARS;

{+}
function AdjustDate(const AAging: string; var VDate: TDateTime): Boolean;
{
  Adjusts the given date based on the aging string. The aging string is a value
  or combination of values specifying what to add or subtract from the supplied
  date/time. The function returns a boolean indicating whether a valid aging
  string was supplied and VDate was updated.

  Aging syntax: NOW | #A | [+|-]#[X][#X...][#T|Z][F|L]

  If the aging string is the constant "NOW", the date is set to the current
  system date/time.

  Otherwise, # is a number and X represents one of the following letters:

  A - Absolute date/time in the form of YYYYMMDDHHNNSSZZZ (or a subset)
  S - Seconds
  N - miNutes
  H - Hours
  D - Days
  M - Months
  Y - Years
  T - Change the Time to the given HHNNSSZZZ value (or a subset).
  Z - Zero out the time/change to midnight (Equivalent to "00T").
  W - Change the Weekday of the date (1..7 = Sunday..Saturday).
  F - Change the date to the First of the month.
  L - Change the date to the Last of the month.

  If the aging string does not start with a plus or minus, minus is assumed.
  If only a number is supplied, it is assumed to be a number of days.

  The calculations are applied in the order listed above. If the resulting date
  has an invalid day of the month (e.g. Feb 29 in a non-leap year, Dec 31 minus
  one month, etc.) the day of the month is set to the last valid day.

  Milliseconds can only be adjusted by time replacement (T) or truncation (Z).

  When changing the Weekday, if the date lands on the specified day of the week
  after the other adjustments, it is not changed. Otherwise, the date will be
  adjusted to the previous/next matching weekday.

  Aging examples:
  1D7H        Subtracts 1 day and 7 hours (subtraction is assumed).
  +1DZ        Adjusts the date to tomorrow at midnight.
  -1D1430T    Adjusts the date to yesterday at 2:30PM.
  -1MF        Adjusts the date to the first of last month.
  20160802A   Changes the date to the absolute value 2016-08-02 00:00:00.000.
  +2W         If not already a Monday, adjust to the next Monday.
  +1D2W       Adjust to next Monday.
  7           Subtracts 7 days (subtraction and days are assumed).

  See also: DateMath, StoD, StoT
-}
var
  y, m, d, h, n, s, w, i, x: Integer;
  bSubtract, bAbsolute, bTrunc, bFirst, bLast, bDateMath: Boolean;
  sAging, sNum: string;
  dtTime: TDateTime;
begin
  Result := False;
  if AAging = '' then
    Exit;
  sAging := UpperCase(AAging);
  if sAging = 'NOW' then begin
    VDate := Now;
    Result := True;
    Exit;
  end;
  bSubtract := sAging[1] <> '+';
  if (sAging[1] = '-') or (sAging[1] = '+') then
    Delete(sAging, 1, 1);
  if sAging = '' then
    Exit;
  y := 0;
  m := 0;
  d := 0;
  h := 0;
  n := 0;
  s := 0;
  w := 0;
  dtTime := 0;
  bAbsolute := False;
  bTrunc := False;
  bFirst := False;
  bLast := False;
  bDateMath := False;
  sNum := '';
  i := 1;
  while i <= Length(sAging) do begin
    if IsDelimiter(DIGIT_CHARS, sAging, i) then
      sNum := sNum + sAging[i]
    else begin
      if sAging[i] = 'A' then begin
        if not TryStoD(sNum, VDate) then
          Exit;
        bAbsolute := True;
      end else if sAging[i] = 'T' then begin
        if not TryStoT(sNum, dtTime) then
          Exit
        else if dtTime = 0 then
          bTrunc := True;
      end else begin
        if sNum = '' then
          x := 0
        else if not TryStrToInt(sNum, x) then
          Exit;
        if IsDelimiter('ZFL', sAging, i) then begin
          if x > 0 then
            Exit; // Non-zero number where no number is expected
        end else if x = 0 then
          Exit; // Zero number where a number is required
        case sAging[i] of
          'Y': y := x;
          'M': m := x;
          'D': d := x;
          'H': h := x;
          'N': n := x;
          'S': s := x;
          'W': if x > 7 then Exit else w := x;
          'Z': bTrunc := True;
          'F': bFirst := True;
          'L': bLast := True;
          else Exit;
        end;
      end;
      bDateMath := True;
      sNum := '';
    end;
    Inc(i);
  end;
  if not bAbsolute then begin
    if bDateMath then begin
      if bSubtract then
        VDate := DateMath(VDate, -y, -m, -d, -h, -n, -s, 0)
      else
        VDate := DateMath(VDate, y, m, d, h, n, s, 0);
      if dtTime > 0 then
        ReplaceTime(VDate, dtTime);
      if bTrunc then
        VDate := Trunc(VDate);
      if w > 0 then begin
        if bSubtract then
          i := -1
        else
          i := 1;
        while DayOfWeek(VDate) <> w do
          VDate := VDate + i;
      end;
      if bFirst then
        VDate := FirstOfMonth(VDate);
      if bLast then
        VDate := LastOfMonth(VDate);
    end else if sNum <> '' then begin
      // If no letters were specified, just assume days.
      x := StrToIntDef(sNum, 0);
      if bSubtract then
        VDate := VDate - x
      else
        VDate := VDate + x;
    end else
      Exit;
  end;
  Result := True;
end;

{+}
function AgeOnDate(const ADOB, ADate: TDateTime): Integer;
{
  Returns a person's age on a given date as an integer number of years.

  See also: DateDiff
-}
var
  cy, cm, cd: Word; // Current year, month, day
  by, bm, bd: Word; // Birth year, month, day
begin
  DecodeDate(ADate, cy, cm, cd);
  DecodeDate(ADOB, by, bm, bd);
  Result := cy - by;
  { Subtract one if they haven't had their birthday yet this year: }
  if (Result > 0) and ((bm > cm) or ((bm = cm) and (bd > cd))) then
    Dec(Result);
end;

{+}
function AnsiDate(const DT: TDateTime): string;
{
  Just a quicker way to write FormatDateTime('yyyy-mm-dd', DT)

  Constants for use with FormatDateTime are also available: 
  ANSI_DATE = 'yyyy-mm-dd';
  ANSI_DATETIME = ANSI_DATE + ' hh:nn:ss';
  ANSI_DATETIMEZ = ANSI_DATETIME + '.zzz';
  
  See also: AnsiDateTime
-}
begin
  Result := FormatDateTime(ANSI_DATE, DT);
end;

{+}
function AnsiDateTime(const DT: TDateTime): string;
{
  Just a quicker way to write FormatDateTime('yyyy-mm-dd hh:nn:ss', DT)

  Constants for use with FormatDateTime are also available: 
  ANSI_DATE = 'yyyy-mm-dd';
  ANSI_DATETIME = ANSI_DATE + ' hh:nn:ss';
  ANSI_DATETIMEZ = ANSI_DATETIME + '.zzz';
  
  See also: AnsiDate
-}
begin
  Result := FormatDateTime(ANSI_DATETIME, DT);
end;

{+}
function AssumeCentury(const y: Integer): Integer;
{
  Given a two-digit year, use TwoDigitYearCenturyWindow to assume a century and
  return the adjusted year value using the logic as described in the help.
-}
// TODO: An overloaded version that takes a TFormatSettings parameter?
var
  c, m, d, w: Word;
begin
  Assert((y >= 0) and (y < 100), 'AssumeCentury requires input 0..99');
  DecodeDate(Date, c, m, d);
  Result := y + c - (c mod 100); // Assume the current century to start
  w := FormatSettings.TwoDigitYearCenturyWindow;
  if w > 0 then begin
    c := c - w;
    c := c mod 100;
    if y >= c then // >= pivot year means It goes in the previous century
      Result := Result - 100;
  end;
end;

{+}
function DateDiff(const AUnit: TDateTimePrecision;
                  const AStart, AEnd: TDateTime): Int64;
{
  Calculates the difference between two dates as an integer number of units,
  where the unit is determined by the AUnit TDateTimePrecision argument.

  Any fractional unit is discarded by truncation or rounding (ms). For example,
  if AUnit is dtpDay and the dates are 2 days and 20 hours apart, the result
  will be 2.

  Units larger than dtpYear (dtpDecade, etc.) are just treated as dtpYear.

  See also: AgeOnDate, MonthDiff
-}
var
  s, e, d: TDateTime; // start, end, diff
begin
  // Calculate as positive, but negate if start & end are reversed.
  s := Min(AStart, AEnd);
  e := Max(AStart, AEnd);
  d := e - s;
  case AUnit of
    dtpMillisecond:
      Result := Round(d * MSecsPerDay);
    dtpSecond:
      Result := Trunc(d * SecsPerDay);
    dtpMinute:
      Result := Trunc(d * MinsPerDay);
    dtpHour:
      Result := Trunc(d * HoursPerDay);
    dtpDay:
      Result := Trunc(d);
    dtpMonth:
      Result := Trunc(MonthDiff(s, e));
    else // Year
      Result := AgeOnDate(s, e);
  end;
  if AStart > AEnd then
    Result := -Result;
end;

{+}
function DateMath(const ADate: TDateTime;
                  const y, m, d, h, n, s, z: Integer): TDateTime;
{
  Allows you to manipulate a date by applying positive or negative values to
  any of the parts in any combination and returns the new TDateTime value.

  The calculations are applied smallest to largest (milliseconds up to years).
  If the finally calculated date has an invalid number of days (i.e. Feb 29 in
  a non-leap year, Dec 31 minus one month, etc.) the day of the month will be
  truncated to the last valid day of the month.

  Examples: DateMath(<2009-12-31 23:59:59.900>, 0, 0, 3, 0, 0, 0, 500) ->
              <2010-01-04 00:00:00.400>
            DateMath(<2009-12-31 00:00:00.000>, 0, -1, 0, 4, 0, 0, 0) ->
              <2009-11-30 04:00:00.000>

  See also: AdjustDate
-}
  procedure RollOver(const Lo, Hi: Integer; var Value, Larger: Integer);
  var
    i: Integer;
  begin
    i := (Hi - Lo) + 1;
    while Value > Hi do begin
      Inc(Larger);
      Dec(Value, i);
    end;
    while Value < Lo do begin
      Dec(Larger);
      Inc(Value, i);
    end;
  end;
var
  wy, wm, wd, wh, wn, ws, wz: Word;
  yy, mm, dd, hh, nn, ss, zz: Integer; // I need to allow for negatives
  x: Integer;
begin
  DecodeTime(ADate, wh, wn, ws, wz);
  zz := wz + z;
  ss := ws + s;
  nn := wn + n;
  hh := wh + h;
  x := 0;
  RollOver(0, 999, zz, ss);
  RollOver(0,  59, ss, nn);
  RollOver(0,  59, nn, hh);
  RollOver(0,  23, hh,  x);
  Result := Trunc(ADate) + x + d; // Adding days is easy integer math!
  DecodeDate(Result, wy, wm, wd);
  dd := wd;
  mm := wm + m;
  yy := wy + y;
  RollOver(1, 12, mm, yy);
  x := MonthDays[IsLeapYear(yy), mm];
  if dd > x then
    dd := x; // Truncate if the days of the month are no longer valid
  Result := EncodeDate(yy, mm, dd) + EncodeTime(hh, nn, ss, zz);
end;

{+}
function DateTrunc(const ADate: TDateTime;
                   const P: TDateTimePrecision): TDateTime;
{
  Truncates a date at a certain precision, similar to the PostgreSQL function
  called "date_trunc".

  Examples where ADate is 1995-10-15 09:45:30.500:
  DateTrunc(ADate, dtpMinute) -> 1995-10-15 09:45:00.000
  DateTrunc(ADate, dtpMonth)  -> 1995-10-01 00:00:00.000
  DateTrunc(ADate, dtpDecade) -> 1990-01-01 00:00:00.000
-}
var
  y, m, d, h, n, s, z: Word;
begin
  DecodeDate(ADate, y, m, d);
  DecodeTime(ADate, h, n, s, z);
  if P < dtpMillisecond then
    z := 0;
  if P < dtpSecond then
    s := 0;
  if P < dtpMinute then
    n := 0;
  if P < dtpHour then
    h := 0;
  if P < dtpDay then
    d := 1;
  if P < dtpMonth then
    m := 1;
  if P < dtpYear then begin
    case P of
      dtpDecade:      y := y - (y mod 10);
      dtpCentury:     y := y - (y mod 100);
      dtpMillennium:  y := y - (y mod 1000);
    end;
  end;
  Result := EncodeDate(y, m, d) + EncodeTime(h, n, s, z);
end;

{+}
function EndOfDay(const DT: TDateTime): TDateTime;
{
  Returns the very last possible instant of the given day.  This is useful for
  date range queries like "<timestamp> BETWEEN :Start AND :End", where :Start
  would be Trunc(DT) and :End would be EndOfDay(DT) (however, I find it's
  usually better to say "<timestamp> >= :Start and <timestamp> < :End" where
  End is midnight on the day AFTER the range end).
-}
begin
  Result := DT;
  ReplaceTime(Result, EncodeTime(23, 59, 59, 999));
end;

{+}
function ExtractTZ(var VDate: string): string;
{
  Extracts a time zone from a DtoS-type string and returns it. The original
  input is updated by having the time zone removed. If there is no time zone,
  the result is blank and the argument remains unchanged.
  
  See also: DtoS
-}
var
  i: Integer;
begin
  i := Pos('+', VDate);
  if i = 0 then
    i := Pos('-', VDate);
  if i > 0 then begin
    Result := Copy(VDate, i, MaxInt);
    SetLength(VDate, i - 1);
  end else if VDate[Length(VDate)] = 'Z' then begin  // Zulu?
    Result := 'Z';
    SetLength(VDate, Length(VDate) - 1);
  end else
    Result := '';
end;

{+}
function FirstOfMonth(const ADate: TDateTime): TDateTime;
{
  Returns the TDateTime argument adjusted to the first of the given month. The
  time remains unchanged.
-}
var
  y, m, d: Word;
begin
  DecodeDate(ADate, y, m, d);
  Result := ADate;
  ReplaceDate(Result, EncodeDate(y, m, 1));
end;

{+}
function IntervalStr(const AInterval: TDateTime;
                     const LoP: TDateTimePrecision = dtpHour;
                     const HiP: TDateTimePrecision = dtpSecond;
                     const ADaySeparator: string = 'd'): string;
{
  Converts a TDateTime into an interval string representing its duration since
  zero. AInterval would normally be the result of subtracting a start time from
  an end time.

  The format of the string depends on the LoP and HiP TDateTimePrecision values
  ("LowPrecision" and "HighPrecision", respectively). These default to dtpHours
  and dtpSeconds accordingly, which will result in a format of "hh:nn:ss".

  If HiP is dtpMillisecond, the format will have the milliseconds added as a
  decimal fraction of the seconds like "04:13:22.377"

  If LoP is dtpDay the format will have the number of days and the day
  separator string ("d" by default) prepended, e.g. "1d04:13:22". Otherwise, the
  hour value will be as large as it needs to be, e.g. "48:56:13". Another common
  day separator might be a period, "1.04:13:22".

  If LoP is dtpMonth, the format will be the same as if LoP were dtpHour, but
  with any insignificant leading zeros removed. For example, "00:00:00.123"
  would become "0.123", "00:05:13" -> "5:13", etc.

  If LoP is dtpYear, the format will be the same as setting LoP to dtpDay if the
  interval is greater than 24 hours, otherwise dtpHour.

  If LoP is dtpDecade or less, the format will be the smallest possible value of
  the format "0d0h0m0.000s" up to HiP. If LoP is dtpCentury or less, spaces
  will be added between each separator: "0d 0h 0m 0.000s". If ADaySeparator is
  a comma-separated list of four values, those will be used as the separators
  regardless of LoP.

  The following constants are available to make the options easier to remember:

  lopTrimZeros = dtpMonth;   // Synonyms for IntervalStr
  lopDayOrHour = dtpYear;
  lopSmallSeps = dtpDecade;
  lopSpaceSeps = dtpCentury;

-}
var
  d: Integer;
  h, m, s, z: Word;
  slSeps: TStringList;
begin
  d := Trunc(AInterval);
  DecodeTime(AInterval, h, m, s, z);
  if LoP > dtpDecade then begin
    if (LoP = dtpDay) or ((LoP = dtpYear) and (d > 0)) then begin
      Result := IntToStr(d) + ADaySeparator;
    end else if LoP >= dtpMonth then begin
      Result := '';
      h := h + (d * 24);
      if LoP > dtpHour then begin
        m := m + (h * 60);
        if LoP > dtpMinute then
          s := s + (m * 60);
      end;
    end;
    if (LoP <= dtpHour) and (HiP >= dtpHour) then
      Result := Result + Format('%.2d', [h]);
    if (LoP <= dtpMinute) and (HiP >= dtpMinute) then begin
      if LoP < dtpMinute then
        Result := Result + ':';
      Result := Result + Format('%.2d', [m]);
    end;
    if (LoP <= dtpSecond) and (HiP >= dtpSecond) then begin
      if LoP < dtpSecond then
        Result := Result + ':';
      Result := Result + Format('%.2d', [s])
    end;
    if HiP = dtpMillisecond then
      Result := Result + '.' + Format('%3.3d', [z]);
    if LoP = dtpMonth then begin // Trim any non-significant leading zeros:
      d := Length(Result);
      while (d > 1) and ((Result[1] = '0') or (Result[1] = ':')) do begin
        Delete(Result, 1, 1);
        Dec(d);
      end;
      if Result[1] = '.' then
        Insert('0', Result, 1);
    end;
  end else begin // 0d0h0m0.000s
    Result := '';
    slSeps := TStringList.Create;
    try
      slSeps.StrictDelimiter := True;
      slSeps.DelimitedText := ADaySeparator;
      if slSeps.Count <> 4 then begin
        if LoP < dtpDecade then
          slSeps.DelimitedText := 'd ,h ,m ,s'
        else
          slSeps.DelimitedText := 'd,h,m,s';
      end;
      if d > 0 then
        Result := IntToStr(d) + slSeps[0];
      if (HiP >= dtpHour) and ((h > 0) or (d > 0)) then
        Result := Result + IntToStr(h) + slSeps[1];
      if (HiP >= dtpMinute) and ((m > 0) or (h > 0) or (d > 0)) then
        Result := Result + IntToStr(m) + slSeps[2];
      if HiP >= dtpSecond then begin
        Result := Result + IntToStr(s);
        if HiP = dtpMSec then
          Result := Result + '.' + Format('%3.3d', [z]);
        Result := Result + slSeps[3];
      end else if Result = '' then begin
        case HiP of
          dtpDay:    Result := '0' + slSeps[0];
          dtpHour:   Result := '0' + slSeps[1];
          dtpMinute: Result := '0' + slSeps[2];
        end;
      end;
      Result := Trim(Result);
    finally
      slSeps.Free;
    end;
  end;
end;

{+}
function ISO8601Date(const S: string; const LocalOffset: TDateTime = 0): TDateTime;
{
  Converts a string in ISO 8601 format (yyyy-mm-ddThh:nn:ss.zzz) into a
  TDateTime. The time separator can be either a "T" or a space.  It handles both
  extended (with separators) and basic (no separators) formats. If a time zone
  specifier is present (Z, +, -) the time will be converted to UTC and then
  local time based on the given offset (or not, if zero). Time zones offsets can
  appear in any of the following formats: Z, +00, +05, -07:00, +0600

  See also: ISO8601Str
-}
var
  i: Integer;
  sDate, sTime, sTZ: string;
  d: TDateTime;
begin
  Result := 0;
  // Find the time separator (if present) & split the string:
  i := Pos('T', S);
  if i = 0 then
    i := Pos(' ', S);
  if i = 0 then
    i := Length(S) + 1;
  sDate := Copy(S, 1, i - 1);
  sTime := Copy(S, i + 1, MaxInt);
  try
    if sDate <> '' then begin
      sDate := StringReplace(sDate, '-', '', [rfReplaceAll]);
      Result := StoD(sDate);
    end;
    if sTime <> '' then begin
      // Find and set aside the time zone if present:
      i := Pos('+', sTime);
      if i = 0 then
        i := Pos('-', sTime);
      if i = 0 then
        i := Pos('Z', sTime); // Zulu (UTC)
      if i > 0 then begin
        sTZ := Copy(sTime, i, MaxInt);
        Delete(sTime, i, MaxInt);
      end else
        sTZ := '';
      // Convert to a TDateTime:
      sTime := StringReplace(sTime, ':', '', [rfReplaceAll]);
      sTime := StringReplace(sTime, '.', '', [rfReplaceAll]); // ms decimal
      Result := Result + StoT(sTime);
      // If given a timezone, adjust to UTC, then to local:
      if sTZ <> '' then begin
        if sTZ <> 'Z' then begin
          sTZ := StringReplace(sTZ, ':', '', [rfReplaceAll]);
          d := StoT(Copy(sTZ, 2, MaxInt)); // Skip leading +/-
          if sTZ[1] = '+' then // Subtract the offset to get UTC
            Result := Result - d
          else
            Result := Result + d;
        end;
        Result := Result + LocalOffset;
      end;
    end;
  except
    on E:EConvertError do
      raise EConvertError.CreateFmt('Invalid ISO8601 date/time value: "%s"', [S]);
  end;
end;

{+}
function ISO8601Str(const D: TDateTime; const P: TDateTimePrecision;
                    const UseT: Boolean = True;
                    const Basic: Boolean = False): string;
{
  Converts a TDateTime into an ISO 8601 format date string, carried out to the
  precision specified by P. If UseT is True, the string will contain "T" between
  the date and time elements (yyyy-mm-ddThh:nn:ss.zzz). If Basic is True, no
  separators will be used in the output (yyyymmddThhnnss). If the value holds
  only a time, the date portion will not be output. If you want a timezone
  offset, you need to add it yourself.

  See also: ISO8601Date, TimeZoneStr
-}
var
  sFmt: string;
  i: Integer;
begin
  if Basic then begin
    Result := DtoS(D, P);
    if P > dtpDay then begin
      if UseT then
        Insert('T', Result, 9)
      else
        Insert(' ', Result, 9);
      // If the value holds only a time, delete the date portion (18991230):
      if D < 1.0 then
        if UseT then
          Delete(Result, 1, 8)
        else
          Delete(Result, 1, 9);
    end;
  end else begin
    sFmt := ANSI_DATETIMEZ;
    // sFmt := 'yyyy-mm-dd hh:nn:ss.zzz';
    //    Cut sites ^  ^  ^  ^  ^  ^
    if P < dtpMillisecond then begin
      i := Ord(P) - Ord(dtpYear);
      if i < 0 then
        i := 0; // Precisions < dtpYear = dtpYear
      i := (i * 3) + 5; // Calc the cut site
      Delete(sFmt, i, MaxInt);
    end;
    Result := FormatDateTime(sFmt, D);
    if P > dtpDay then begin
      if UseT then
        Result[11] := 'T'; // Must be after FormatDateTime - it gets converted!
      if D < 1.0 then // If only a time, delete the date portion (1899-12-30)
        if UseT then
          Delete(Result, 1, 10)
        else
          Delete(Result, 1, 11);
    end;
  end;
end;

{+}
function IsValidTimeZone(const ATZ: string;
                         const AllowZulu: Boolean = True;
                         const AllowColon: Boolean = True;
                         const AllowHourOnly: Boolean = True): Boolean;
{
  Validates a time zone string in +/-HH[[:]MM] format.
  A blank is always valid. If AllowZulu is True, it may be "Z".
  If AllowColon is False, a colon separator is not allowed.
  If AllowHourOnly is False, minutes are required even if zero.
-}
var
  sTZ: string;
  i, iLen: Integer;
begin
  if ATZ = '' then begin
    Result := True;
    Exit;
  end;
  if SameText(ATZ, 'Z') then begin
    Result := AllowZulu;
    Exit;
  end;
  Result := (ATZ[1] = '+') or (ATZ[1] = '-'); // +/-HH[[:]MM]
  if Result then begin
    sTZ := Copy(ATZ, 2, MaxInt);
    i := Pos(':', sTZ);
    if i > 0 then begin
      Result := AllowColon;
      if Result and (i = 3) then begin
        Result := Length(sTZ) = 5; // Must have hour and minutes with a colon
        Delete(sTZ, i, 1);
      end else
        Result := False;
    end;
    if Result then begin
      iLen := Length(sTZ);
      Result := ((iLen = 4) or ((iLen = 2) and AllowHourOnly)) and IsNumeric(sTZ);
      if Result then begin
        i := StrToIntDef(Copy(sTZ, 1, 2), -1);
        Result := (i >= 0) and (i < 24);
        if Result and (iLen = 4) then begin
          i := StrToIntDef(Copy(sTZ, 3, 2), -1);
          Result := (i >= 0) and (i < 60);
        end;
      end;
    end;
  end;
end;

{+ LastDOM1}
function LastDOM(const ADate: TDateTime): Integer;
{
  Returns the integer last day of the month for the given date.
  Example: LastDOM(<2005-01-16>) -> 31

  Overloaded: LastDOM(const AMonth: Integer; const AYear: Integer)
-}
var
  y, m, d: Word;
begin
  DecodeDate(ADate, y, m, d);
  Result := MonthDays[IsLeapYear(y), m];
end;

{+ LastDOM2}
function LastDOM(const AMonth: Integer; const AYear: Integer = 0): Integer;
{
  Returns the integer last day of the given month and year. The current year is
  used for leap year logic if not provided.

  Examples: LastDOM(9)       -> 30
            LastDOM(2, 2004) -> 29

  Overloaded: LastDOM(const ADate: TDateTime)
-}
var
  y, m, d: Word;
begin
  if AYear = 0 then
    DecodeDate(Date, y, m, d)
  else
    y := AYear;
  Result := MonthDays[IsLeapYear(y), AMonth];
end;

{+}
function LastOfMonth(const ADate: TDateTime): TDateTime;
{
  Returns the TDateTime argument adjusted to the last day of the given month.
-}
var
  y, m, d: Word;
begin
  DecodeDate(ADate, y, m, d);
  Result := ADate;
  ReplaceDate(Result, EncodeDate(y, m, LastDOM(m, y)));
end;

{+}
function LocalUTCOffset: TDateTime;
{
  Get the local UTC offset from Windows using GetTimeZoneInformation. Add this
  value to a local time in order to convert it to UTC time.

  Raises an exception if GetTimeZoneInformation returns TIME_ZONE_ID_INVALID.
-}
var
  tzi: TTimeZoneInformation;
  r: DWORD;
  iBias: Integer;
begin
  r := GetTimeZoneInformation(tzi);
  case r of
    TIME_ZONE_ID_UNKNOWN:
      iBias := tzi.Bias;
    TIME_ZONE_ID_STANDARD:
      iBias := tzi.Bias + tzi.StandardBias;
    TIME_ZONE_ID_DAYLIGHT:
      iBias := tzi.Bias + tzi.DaylightBias;
    else // TIME_ZONE_ID_INVALID:
      raise Exception.Create('TIME_ZONE_ID_INVALID: ' +
                             SysErrorMessage(GetLastError));
  end;
  Result := iBias / MinsPerDay; // Bias is in minutes
end;

{+}
function MatchingMonthDay(const ADate: TDateTime; const ADay: Integer): Boolean;
{
  Returns True if the given date matches the specified day of the month.

  If ADay is greater than the number of days in the given month, the function
  will return True if the date is the last day of its month.

  Examples: MatchingMonthDay(<2008-09-24>, 15) -> False (no match)
            MatchingMonthDay(<2008-09-24>, 24) -> True (match)
            MatchingMonthDay(<2008-09-30>, 31) -> True (match, last DOM)

  See Also: NextMonthDay, PrevMonthDay

  Formerly known as "ValidMonthDay" which was too vague.
-}
var
  y, m, d: Word;
begin
  DecodeDate(ADate, y, m, d);
  Result := (d = ADay) or ( (ADay > d) and (d = LastDOM(m, y)) );
end;

{+}
function MinutesToTime(const AMinutes: Cardinal): TDateTime;
{
  Turns a positive integer number of minutes into the fractional part of a
  TDateTime. The integer portion of the result will always be zero, even if the
  input is greater than the number of minutes in a day (MinsPerDay=1440).

  See also: TimeToMinutes
-}
var
  m: Integer;
begin
  m := AMinutes mod MinsPerDay;
  Result := EncodeTime(m div 60, m mod 60, 0, 0);
end;

{+}
function MonthDiff(const AStart, AEnd: TDateTime): Currency;
{
  Calculates the difference between two dates in calendar months.

  A partial calendar month becomes a fraction in the form of # days divided by
  the total number of days in the month. Personally, I don't think there is a
  definitive way to calculate this.

  For example: 2005-01-01 to 2005-02-10 = 1.3571
    -> 1.0 for January plus 0.3571 for 10 out of 28 days in February.

  See also: DateDiff, AgeOnDate
-}
var
  dLo, dHi: TDateTime;
  ly, lm, ld, hy, hm, hd: Word;
  i, x: Integer;
begin
  dLo := Trunc(Min(AStart, AEnd));
  dHi := Trunc(Max(AStart, AEnd));
  DecodeDate(dLo, ly, lm, ld);
  DecodeDate(dHi, hy, hm, hd);
  // First determine the difference in whole years:
  i := hy - ly;
  if (i > 0) then begin
    if ((hm < lm) or ((hm = lm) and (hd < ld))) then begin
      i := i - 1; // That last one isn't a whole one!
      ly := hy - 1;
    end else begin
      ly := hy;
    end;
  end;
  Result := i * 12.0;
  // Are they equal or only the year is different?  If so, we're done.
  if (lm = hm) and (ld = hd) then
    Exit;
  // Is a single, partial calendar month all that remains?
  x := LastDOM(dHi);
  if (ly = hy) and (lm = hm) and ((ld > 1) or (hd < x)) then begin
    i := (hd - ld) + 1;
    Result := Result + (i / x);
    Exit;
  end;
  // Trim any beginning partial month:
  if ld > 1 then begin
    x := LastDOM(dLo);
    Result := Result + ( ((x - ld) + 1) / x );
    ld := 1;
    Inc(lm);
    if lm > 12 then begin
      lm := 1;
      Inc(ly);
    end
  end;
  dLo := EncodeDate(ly, lm, ld);
  // Trim any final partial month:
  x := LastDOM(dHi);
  if hd < x then begin
    Result := Result + (hd / x);
    Dec(hm);
    if hm < 1 then begin
      hm := 12;
      Dec(hy);
    end;
    dHi := EncodeDate(hy, hm, 28); // Day doesn't matter as long as it's > 1
  end;
  // Now count the remaining full calendar months:
  while dLo < dHi do begin
    Result := Result + 1.0;
    dLo := IncMonth(dLo, 1);
  end;
  if (Result > 0) and (AEnd < AStart) then
    Result := -Result;
end;

{+}
function MonthStrToInt(const AMonth: string): Integer;
{
  Returns the integer month number 1..12 of the given month string.

  It is case-insensitive and works with full month names or their abbreviations
  as defined by the ShortMonthNames variable. Returns zero if the string is not
  valid.

  Example: MonthStrToInt('January') -> 1
-}
{ Note: Other functions rely upon this function returning zero if the string is
  not a valid character month. DO NOT modify this function to also accept
  numeric input! }
var
  sTemp: string;
begin
  // Luckily, the short names are just the first 3 characters of the long names.
  sTemp := Copy(AMonth, 1,
    Length(FormatSettings.ShortMonthNames[1]));
  Result := 1;
  repeat
    if SameText(sTemp, FormatSettings.ShortMonthNames[Result]) then // Case-insensitive
      Exit;
    Inc(Result);
  until Result > 12;
  Result := 0;
end;

{+}
function MStoTime(const AMS: Cardinal): TDateTime;
{
  Turns a positive integer number of milliseconds into the fractional part of a
  TDateTime. The integer portion of the result will always be zero, even if the
  input is greater than the number of milliseconds in a day.

  See also: TimeToMS, TimeToSeconds, MinutesToTime, TimeToMinutes
-}
var
  h, m, s, z: Integer;
begin
  z := AMS mod MSecsPerDay;
  h := z div MSecsPerHour;
  z := z mod MSecsPerHour;
  m := z div MSecsPerMin;
  z := z mod MSecsPerMin;
  s := z div MSecsPerSec;
  z := z mod MSecsPerSec;
  Result := EncodeTime(h, m, s, z);
end;

{+}
function NextMonthDay(const ADate: TDateTime; const ADay: Integer): TDateTime;
{
  Returns a TDateTime representing the given date adjusted to the nearest future
  date matching the desired day of the month. The time, if any, is unchanged.

  Examples: NextMonthDay(<2004-05-15>, 20) -> <2004-05-20>
            NextMonthDay(<2004-05-15>, 15) -> <2004-06-15>

  If the desired day would result in an invalid date, the last valid day of the
  month will be used:

  Examples:  NextMonthDay(<2004-04-01>, 31) -> <2004-04-30>
             NextMonthDay(<2004-08-31>, 31) -> <2004-09-30>

  See also: PrevMonthDay, MatchingMonthDay
-}
var
  y, m, d, l: Word;
begin
  Result := ADate + 1; // At least one day past input
  DecodeDate(Result, y, m, d);
  if d = ADay then
    Exit; // Spot on!
  if d > ADay then begin
    // We've passed it for this month. Get it next month:
    Result := IncMonth(Result, 1);
    DecodeDate(Result, y, m, d);
  end;
  // Adjust to the lesser of the desired day or the last day of the month:
  d := ADay;
  l := LastDOM(Result);
  if d > l then
    d := l;
  ReplaceDate(Result, EncodeDate(y, m, d));
end;

{+}
function PrevMonthDay(const ADate: TDateTime; const ADay: Integer): TDateTime;
{
  Returns a TDateTime representing the given date adjusted to the nearest past
  date matching the desired day of the month - the opposite of NextMonthDay().

  Examples: PrevMonthDay(<2004-05-15>, 15) -> <2004-04-15>
            PrevMonthDay(<2004-05-31>, 31) -> <2004-04-30>

  See also: MatchingMonthDay, NextMonthDay
-}
var
  y, m, d, l: Word;
begin
  Result := ADate - 1; // At least one day prior to input
  DecodeDate(Result, y, m, d);
  if d = ADay then
    Exit; // Spot on!
  if d < ADay then begin
    // We're beyond it this month. Get it in the prior month:
    Result := IncMonth(Result, -1);
    DecodeDate(Result, y, m, d);
  end;
  // Adjust to the lesser of the desired day or the last day of the month:
  d := ADay;
  l := LastDOM(Result);
  if d > l then
    d := l;
  ReplaceDate(Result, EncodeDate(y, m, d));
end;

{+}
function SameTime(const d1, d2: TDateTime): Boolean;
{
  Compares two TDateTime values numerically, and returns True if the difference
  between them is less than one millisecond.

  Handy constant: OneMillisecond
-}
begin
  Result := Abs(d1 - d2) < OneMillisecond;
end;

{+}
function SecondsToTime(const ASeconds: Cardinal): TDateTime;
{
  Turns a positive integer number of seconds into the fractional part of a
  TDateTime. The integer portion of the result will always be zero, even if the
  input is greater than the number of seconds in a day (SecsPerDay=86400).

  See also: TimeToSeconds, MinutesToTime, TimeToMinutes
-}
var
  h, m, s: Integer;
begin
  s := ASeconds mod SecsPerDay;
  h := s div SecsPerHour;
  s := s mod SecsPerHour;
  m := s div SecsPerMin;
  s := s mod SecsPerMin;
  Result := EncodeTime(h, m, s, 0);
end;

{+}
function StoDTP(const S: string;
                const ADefault: TDateTimePrecision): TDateTimePrecision;
{
  Finds a string in the dtpWords array, or a single character in the dtpLetters
  array and returns the corresponding TDateTimePrecision. Matching is case
  insensitive. The default value is returned if no match is found.

  See also: TDateTimePrecision
-}
var
  c: Char;
begin
  Result := dtpMillennium;
  if Length(S) = 1 then begin
    c := S[1].ToUpper;
    while Result <= dtpMillisecond do begin
      if c = dtpLetters[Result] then
        Exit;
      if Result = dtpMillisecond then
        break;
      Inc(Result);
    end;
  end else begin
    while Result <= dtpMillisecond do begin
      if SameText(S, dtpWords[Result]) then
        Exit;
      if Result = dtpMillisecond then
        break;
      Inc(Result);
    end;
  end;
  Result := ADefault;
end;

{+}
function TimeToMinutes(const ATime: TDateTime): Cardinal;
{
  Takes the time portion of a TDateTime and turns it into a positive integer
  number of minutes past midnight 0..1439.

  See also: MinutesToTime, SecondsToTime, TimeToSeconds
-}
begin
  Result := Round(Frac(ATime) * MinsPerDay);
  if Result >= MinsPerDay then
    Result := MinsPerDay - 1; // Zero would be the next day, keep date the same.
end;

{+}
function TimeToMS(const ATime: TDateTime): Cardinal;
{
  Takes the time portion of a TDateTime and turns it into a positive integer
  number of milliseconds past midnight.

  See also: MStoTime, SecondsToTime, MinutesToTime, TimeToMinutes
-}
begin
  Result := Round(Frac(ATime) * MSecsPerDay);
  if Result >= MSecsPerDay then
    Result := MSecsPerDay - 1; // Zero would be the next day, keep date the same.
end;

{+}
function TimeToSeconds(const ATime: TDateTime): Cardinal;
{
  Takes the time portion of a TDateTime and turns it into a positive integer
  number of seconds past midnight 0..86399

  See also: SecondsToTime, MinutesToTime, TimeToMinutes
-}
begin
  Result := Round(Frac(ATime) * SecsPerDay);
  if Result >= SecsPerDay then
    Result := SecsPerDay - 1; // Zero would be the next day, keep date the same.
end;

{+}
function TimeZoneOffset(const ATimeZoneStr: string): TDateTime;
{
  Converts a time zone string in +/-hh[[:]mm] format into a TDateTime UTC offset
  value. If the input is blank or "Z", the offset is zero. This offset is added
  to local time to convert to UTC.

  See also: TimeZoneStr
-}
var
  z: string;
  h, m: Integer;
begin
  z := UpperCase(ATimeZoneStr);
  if (z = '') or (z = 'Z') then begin
    Result := 0;
    Exit;
  end;
  h := Pos(':', z);
  if h > 0 then
    Delete(z, h, 1);
  h := StrToIntDef(Copy(z, 2, 2), 0);
  m := StrToIntDef(Copy(z, 4, 2), 0);
  Result := ((h * 60) + m) / MinsPerDay;
  // An offset is the value you ADD to local time to get UTC.
  // A positive timezone is a negative adjustment:
  if z[1] = '+' then
    Result := -Result;
end;

{+}
function TimeZoneStr(const ALocalOffset: TDateTime;
                     const AZulu: Boolean = True;
                     const AColon: Boolean = True;
                     const AHourOnly: Boolean = True): string;
{
  Converts a local UTC offset TDateTime into a string in ISO 8601 format,
  suitable for concatenation to an ISO8601 date/time string.

  The default format is "+/-hh[:mm]", returning "Z" if the offset is zero. Set
  AZulu to False if you want "+00[:00]" instead of "Z".

  Setting AColon to False excludes the colon separator, like "+0700".

  If AHourOnly is set to False, the minutes will always be included, zero or
  otherwise. By default, minutes are only included when non-zero.

  See also: ISO8601Str, TimeZoneOffset
-}
begin
  if AZulu and (ALocalOffset = 0) then
    Result := 'Z'
  else begin
    // A negative offset is a positive adjustment to UTC, and vice versa.
    // Also, zero must always show positive (+00:00).
    if ALocalOffset <= 0 then
      Result := '+'
    else
      Result := '-';
    if AColon then
      Result := Result + FormatDateTime('hh:nn', ALocalOffset)
    else
      Result := Result + TtoS(ALocalOffset, dtpMinute);
    if AHourOnly and EndsWith('00', Result) then
      SetLength(Result, 3); // +HH or -HH
  end;
end;

{+}
function WeekDayDiff(const AStart, AEnd: TDateTime): Integer;
{
  Calculates the difference between two dates in weekdays.

  If the last day is not a weekday, it doesn't count. For example, Wednesday to
  Saturday = 2 (w-th, th-f).
-}
// TODO: AddWeekDays(TDateTime, Integer) - increment the date by # weekdays
var
  dLo, dHi: TDateTime;
  i: Integer;
begin
  Result := 0;
  dLo := Trunc(Min(AStart, AEnd));
  dHi := Trunc(Max(AStart, AEnd));
  while dLo < dHi do begin
    i := DayOfWeek(dLo);
    if (i > 1) and (i < 7) then
      Result := Result + 1;
    dLo := dLo + 1;
  end;
  if (Result > 0) then begin
    i := DayOfWeek(dHi);
    if (i = 1) or (i = 7) then
      Result := Result - 1;
    if AEnd < AStart then
      Result := -Result;
  end;
end;

{+}
function WeekdayNum(const ADate: TDateTime): Integer;
{
  Returns an integer indicating which weekday of the month the given date falls
  on. For example, 2017-01-22 is the 4th Sunday in January, so the function will
  return 4. Use DayOfWeek to find out which day of the week it is.

  See also: WeekdayOfMonth
-}
var
  y, m, d: Word;
begin
  DecodeDate(ADate, y, m, d);
  Result := 1;
  while d > 7 do begin
    Inc(Result);
    Dec(d, 7);
  end;
end;

{+}
function WeekdayOfMonth(const ADate: TDateTime; const Nth: Integer;
                        const DOW: TDayOfWeek): TDateTime;
{
  Returns the date that falls on the Nth weekday of the given month.

  <ADate> is the TDateTime that contains the month & year you wish the result
  to fall in.

  <Nth> specifies which instance of the day of the week you wish to find, such
  as first (1), second (2), etc. If this value is greater than the actual number
  of times the weekday occurs in the given month, the last will be returned.

  <DOW> is the day of the week you wish to find. If dowAnyDay is given, the day
  of the week will be taken from the current value of ADate.

  Example: To find the second Sunday in the month of February 2005:
           WeekdayOfMonth(<2005-02-25>, 2, dowSunday) -> <2005-02-13>

  See also: WeekdayNum
-}
var
  i, d, x: Integer;
begin
  // Start with the first of the month:
  Result := FirstOfMonth(ADate);
  d := 1;
  // Increment until we find the first correct day of the week:
  if DOW > dowAnyDay then
    x := Ord(DOW)
  else
    x := DayOfWeek(ADate);
  i := DayOfWeek(Result);
  while i <> x do begin
    Result := Result + 1;
    Inc(d);
    Inc(i);
    if i > 7 then
      i := 1;
  end;
  // Increment by 7 for the second & subsequent occurrences.
  // Just stop if the end of the month is reached.
  x := LastDOM(Result);
  for i := 2 to Nth do begin
    if (d + 7) > x then
      Break;
    Result := Result + 7;
    d := d + 7;
  end;
end;

{+}
function DtoS(const DT: TDateTime;
              const P: TDateTimePrecision = dtpDay): string;
{
  "Date to String" - Formats a TDateTime value as a string in a
  'yyyymmddhhnnsszzz' format. This locale-neutral format is handy for storage or
  transmission and sorts naturally. The inverse of DtoS is the StoD function.

  The optional TDateTimePrecision argument will determine the length of the
  returned string and defaults to dtpDay, or 'yyyymmdd'.

  Precisions lower than dtpYear are ignored. The smallest supported return value
  is 'yyyy'.
  
  See also: StoD, StoDDef, TryStoD, TtoS
-}
var
  sFmt: string;
  i: Integer;
begin
  sFmt := 'yyyymmddhhnnsszzz';
  // Chop     ^ ^ ^ ^ ^ ^
  if P < dtpMillisecond then begin
    i := Ord(P) - Ord(dtpYear);
    if i < 0 then
      i := 0; // Precisions < dtpYear = dtpYear
    i := (i * 2) + 4;
    SetLength(sFmt, i);
  end;
  Result := FormatDateTime(sFmt, DT);
end;

{+}
function StoD(const DS: string): TDateTime;
{
  "String to Date" - Converts a string in 'yyyymmddhhnnsszzz' format, or a
  subset thereof (see DtoS), to a TDateTime.

  Input strings shorter than the length of the full format will have their
  missing values defaulted appropriately: 1 for month or day, and zero for time
  values. The minimum accepted input is a 4-digit year.

  Invalid input will raise an EConvertError exception. See StoDDef and TryStoD
  for additional options when dealing with untrusted input.
  
  See also: DtoS, StoDDef, TryStoD
-}
var
  L: Integer;
  y, m, d, h, n, s, z: Word;
begin
  L := Length(DS);
  if L < 17 then
    if not (L in [4, 6, 8, 10, 12, 14]) then
      raise EConvertError.CreateFmt('Invalid StoD input length (%d): "%s"',
                                    [L, DS]);
  try
    y := StrToInt(Copy(DS, 1, 4));
    if L >= 6 then m := StrToInt(Copy(DS, 5, 2)) else m := 1;
    if L >= 8 then d := StrToInt(Copy(DS, 7, 2)) else d := 1;
    Result := EncodeDate(y, m, d);
    if L >= 10 then begin
      h := StrToInt(Copy(DS, 9, 2));
      if L >= 12 then n := StrToInt(Copy(DS, 11, 2)) else n := 0;
      if L >= 14 then s := StrToInt(Copy(DS, 13, 2)) else s := 0;
      if L >= 17 then z := StrToInt(Copy(DS, 15, 3)) else z := 0;
      Result := Result + EncodeTime(h, n, s, z);
    end;
  except
    on EConvertError do
      raise EConvertError.CreateFmt('Invalid StoD input: "%s"', [DS]);
  end;
end;

{+}
function StoDDef(const DS: string; const ADefault: TDateTime): TDateTime;
{
  StoDDef is to StoD as StrToIntDef is to StrToInt.

  See also: DtoS, StoD, TryStoD
-}
begin
  try
    Result := StoD(DS);
  except
    on EConvertError do
      Result := ADefault;
  end;
end;

{+}
function TryStoD(const DS: string; out ADate: TDateTime): Boolean;
{
  TryStoD is to StoD as TryStrToInt is to StrToInt.

  See also: DtoS, StoD, StoDDef
-}
begin
  try
    ADate := StoD(DS);
    Result := True;
  except
    on EConvertError do
      Result := False;
  end;
end;

{+}
function TtoS(const DT: TDateTIme;
              const P: TDateTimePrecision = dtpSecond): string;
{
  "Time to String" - Works much like DtoS, but on a time value only with a
  'hhnnsszzz' format. Precisions lower than dtpHour are ignored. The inverse
  function is StoT.

  See also: StoT, StoTDef, TryStoT
-}
var
  sFmt: string;
  i: Integer;
begin
  sFmt := 'hhnnsszzz';
  // Chop   ^ ^ ^
  if P < dtpMillisecond then begin
    i := Ord(P) - Ord(dtpHour);
    if i < 0 then
      i := 0;
    i := (i * 2) + 2;
    SetLength(sFmt, i);
  end;
  Result := FormatDateTime(sFmt, DT);
end;

{+}
function StoT(const TS: string): TDateTime;
{
  "String to Time" - Converts a time string in 'hhnnsszzz' format, or a subset
  thereof (see TtoS), to a TDateTime. It is the inverse of the TtoS function.

  Input strings shorter than the length of the full format will have their
  missing values defaulted to zero. Any extra digits beyond 9 are simply ignored
  (HL7 time values often have milliseconds carried out to four places).

  Invalid input will raise an EConvertError exception. See StoTDef and TryStoT
  for additional options when dealing with untrusted input.

  See also: TtoS, StoTDef, TryStoT
-}
var
  L: Integer;
  h, n, s, z: Word;
begin
  L := Length(TS);
  if L < 9 then
    if not (L in [2, 4, 6]) then
      raise EConvertError.CreateFmt('Invalid StoT input length (%d): "%s"',
                                    [L, TS]);
  try
    h := StrToInt(Copy(TS, 1, 2));
    if L >= 4 then n := StrToInt(Copy(TS, 3, 2)) else n := 0;
    if L >= 6 then s := StrToInt(Copy(TS, 5, 2)) else s := 0;
    if L >= 9 then z := StrToInt(Copy(TS, 7, 3)) else z := 0;
    Result := EncodeTime(h, n, s, z);
  except
    on EConvertError do
      raise EConvertError.CreateFmt('Invalid StoT input: "%s"', [TS]);
  end;
end;

{+}
function StoTDef(const TS: string; const ADefault: TDateTime): TDateTime;
{
  StoTDef is to StoT as StrToIntDef is to StrToInt.

  See also: TtoS, StoT, TryStoT
-}
begin
  try
    Result := StoT(TS);
  except
    on EConvertError do
      Result := ADefault;
  end;
end;

{+}
function TryStoT(const TS: string; out ATime: TDateTime): Boolean;
{
  TryStoT is to StoT as TryStrToInt is to StrToInt.

  See also: DtoT, StoT, StoTDef
-}
begin
  try
    ATime := StoT(TS);
    Result := True;
  except
    on EConvertError do
      Result := False;
  end;
end;

{+}
function FormatDateTimeEx(const AFormat: string; ADateTime: TDateTime): string;
{
  Adds a few new specifiers on top of those supported by FormatDateTime:

  Spec  Displays
  xd    Day of the year (001..366)
  xi    ISO8601 day of the week (1..7, starting Monday)
  xk    Day of the week (1..7, starting Sunday)
  xq    Quarter (1..4)
  xw    Week of the year (01..53)
  xy    Year that goes with xw (4-digit)
  xz    Year that goes with xw (2-digit)

  See also: ReplaceDateTokens
-}
var
  sFmt, sLF: string;
  Flags: TReplaceFlags;
  y, m, d: Word;
begin
  sFmt := AFormat;
  sLF := LowerCase(sFmt);
  if Pos('x', sLF) > 0 then begin
    Flags := [rfReplaceAll, rfIgnoreCase];
    if Pos('xd', sLF) > 0 then
      sFmt := StringReplace(sFmt, 'xd', Format('%.3d', [DayOfTheYear(ADateTime)]), Flags);
    if Pos('xi', sLF) > 0 then
      sFmt := StringReplace(sFmt, 'xi', Format('%d', [DayOfTheWeek(ADateTime)]), Flags);
    if Pos('xk', sLF) > 0 then
      sFmt := StringReplace(sFmt, 'xk', Format('%d', [DayOfWeek(ADateTime)]), Flags);
    if Pos('xq', sLF) > 0 then begin
      DecodeDate(ADateTime, y, m, d);
      d := m div 3;
      if m mod 3 <> 0 then
        Inc(d);
      sFmt := StringReplace(sFmt, 'xq', Format('%d', [d]), Flags);
    end;
    if (Pos('xw', sLF) > 0) or (Pos('xy', sLF) > 0) or (Pos('xz', sLF) > 0) then begin
      d := WeekOfTheYear(ADateTime, y);
      sFmt := StringReplace(sFmt, 'xw', Format('%.2d', [d]), Flags);
      sFmt := StringReplace(sFmt, 'xy', Format('%.4d', [y]), Flags);
      sFmt := StringReplace(sFmt, 'xz', Format('%.2d', [y mod 100]), Flags);
    end;
  end;
  Result := FormatDateTime(sFmt, ADateTime);
end;

{+}
function ReplaceDateTokens(const APattern: string;
                           const ADate: TDateTime): string;
{
  Replaces parameters in a string with values from the given date/time.
  Especially handy for date-based filenames.

  The tokens ARE case-sensitive:

  %y=Year, 2-digit
  %Y=Year, 4-digit
  %m=Month 01..12
  %M=Month abbreviation Jan..Dec
  %L=Month name January..December
  %d=Day of the month 01..31
  %D=Day of the year 001..366
  %h=Hour 00..23
  %H=Hour 01..12
  %n=Minute 00..59
  %s=Second 00..59
  %z=Millisecond 000..999
  %I=ISO8601 day of the week 1..7 (starting Monday)
  %J=Day of the week Sunday..Saturday
  %k=Day of the week 1..7 (starting Sunday)
  %K=Day of the week abbreviation Sun..Sat
  %p=a/p
  %P=AM/PM (based on locale)
  %q=Quarter 1..4
  %w=Week of the Year 01..53
  %x=Year that goes with week#, 2-digit
  %X=Year that goes with week#, 4-digit
  %C=Force any words returned to upper-case (e.g. Wed -> WED)
  %c=Force any words returned to lower-case (e.g. Wed -> wed)

  See also: FormatDateTimeEx
-}
var
  y, m, d, h, n, s, z, q, w, x, i, k: Word;
  sAP, sPM: string;
  iCase: Integer;
  procedure RepNum(const AToken, AFormat: string; const AValue: Word);
  begin
    if Pos(AToken, Result) > 0 then { Avoid unnecessary string reallocation }
      Result := StringReplace(Result, AToken, Format(AFormat, [AValue]),
                              [rfReplaceAll]);
  end;
  procedure RepStr(const AToken, AValue: string);
  begin
    if Pos(AToken, Result) > 0 then
      Result := StringReplace(Result, AToken, AValue, [rfReplaceAll]);
  end;
  function Caseify(const S: string): string;
  begin
    case iCase of
       0: Result := S;
       1: Result := AnsiUpperCase(S);
      -1: Result := AnsiLowerCase(S);
    end;
  end;
begin
  Result := APattern;
  DecodeDateTime(ADate, y, m, d, h, n, s, z);
  q := m div 3;
  if m mod 3 <> 0 then
    Inc(q);
  // Common/standard numeric replacements:
  RepNum('%Y', '%.4d', y);
  RepNum('%y', '%.2d', y mod 100);
  RepNum('%m', '%.2d', m);
  RepNum('%d', '%.2d', d);
  RepNum('%h', '%.2d', h);
  RepNum('%n', '%.2d', n);
  RepNum('%s', '%.2d', s);
  RepNum('%z', '%.3d', z);
  RepNum('%q', '%d',   q);
  // Skip the rest if there's nothing left to replace:
  if Pos('%', Result) = 0 then
    Exit;
  // Any case changers?
  iCase := 0;
  if Pos('%C', Result) > 0 then
    iCase := 1;
  if Pos('%c', Result) > 0 then
    iCase := -1;
  if iCase <> 0 then
    Result := StringReplace(Result, '%c', '', [rfIgnoreCase, rfReplaceAll]);
  // Prepare for AM/PM times:
  if h >= 12 then begin
    sAP := 'p';
    sPM := FormatSettings.TimePMString;
    if h > 12 then
      h := h - 12;
  end else begin
    sAP := 'a';
    sPM := FormatSettings.TimeAMString;
    if h = 0 then
      h := 12;
  end;
  i := DayOfTheWeek(ADate);
  k := DayOfWeek(ADate);
  w := WeekOfTheYear(ADate, x);
  RepNum('%D', '%.3d', DayOfTheYear(ADate));
  RepNum('%H', '%.2d', h);
  RepNum('%I', '%d', i);
  RepStr('%J', Caseify(FormatSettings.LongDayNames[k]));
  RepNum('%k', '%d', k);
  RepStr('%K', Caseify(FormatSettings.ShortDayNames[k]));
  RepStr('%L', Caseify(FormatSettings.LongMonthNames[m]));
  RepStr('%M', Caseify(FormatSettings.ShortMonthNames[m]));
  RepStr('%p', Caseify(sAP));
  RepStr('%P', Caseify(sPM));
  RepNum('%w', '%.2d', w);
  RepNum('%X', '%.4d', x);
  RepNum('%x', '%.2d', x mod 100);
end;

{+}
function TryAnyStrToDateTime(const DS: string; var DayPos: Integer;
                             out OutDate: TDateTime): Boolean;
{
  Works like TryStrToDateTime, but does its best to decipher formats other than
  the current locale formats (which it tries first). It also handles non-numeric
  months (see MonthStrToInt) and DtoS-type strings (StoD).

  The DayPos argument is used to help interpret ambiguous dates where both the
  month and day values are <= 12. DayPos can be set to -1..3 and behaves
  according to the following table:

  -1  Tries the current locale format first, then uses the default behavior of
      using the first value <= 12 as the month. DayPos will not be modified by
      the function.

  0   Like -1, but allows itself to be updated to 1-3 if a non-ambiguous date is
      encountered, so that the "learned" format can be used on future calls.

  1-3 Uses the number to choose the first, second or third part of the date as
      the day value when it otherwise can't tell the month and day apart. When
      DayPos is > 0, the function does not try the current locale format first.

  Requires DateUtils, so D6+ only.
-}
var
  bAmbiguous, bCharMonth, bKnownYear: Boolean;
  sTemp: string;
  p, i, m, d, y: Integer;
  sParts: array[0..3] of string;
  iParts: array[0..2] of Integer;
  OutTime: TDateTime;
begin
  OutDate := 0;
  bAmbiguous := True;
  bCharMonth := False;
  bKnownYear := False;

  // First, just try the default locale format & see if it works.
  // Treat DayPos > 0 like a user override and skip the default format attempt.
  sTemp := AnsiUpperCase(Trim(DS));
  if DayPos > 0 then
    Result := False
  else
    Result := TryStrToDateTime(sTemp, OutDate);
  if Result or (sTemp = '') then // Treat a blank the same as TryStrToDateTime
    Exit;

  // Parse the string into parts based on non-alphanumeric separators.
  // The first three parts are the date, and the rest is the time.
  p := 0;
  for i := 1 to Length(sTemp) do begin
    if (p = 3) or IsDelimiter(UPPER_DIGITS, sTemp, i) then
      sParts[p] := sParts[p] + sTemp[i]
    else
      Inc(p);
  end;

  // One part = DtoS format "yyyymmddhhnnss"
  if p = 0 then begin
    Result := TryStoD(sTemp, OutDate);
    Exit;
  end;

  // Two parts? WTF is that?
  if p = 1 then
    Exit;

  y := -1;
  m := -1;
  d := -1;

  try

    // Check for a character month value:
    for p := 0 to 2 do begin
      i := MonthStrToInt(sParts[p]);
      if i > 0 then begin
        iParts[p] := i;
        m := p;
        bCharMonth := True; // Combined with a 4-digit year = unambiguous date!
      end else begin
        iParts[p] := StrToInt(sParts[p]);
      end;
    end;

    // Standard ANSI/ISO yyyy-mm-dd format?
    if (m < 0) and (Length(sParts[0]) = 4) and
       (iParts[1] <= 12) and (iParts[2] <= 31) then begin
      y := 0;
      m := 1;
      d := 2;
      if iParts[2] > 12 then
        bAmbiguous := False;
    end else begin
      // Is there an obvious year value?
      for p := 0 to 2 do begin
        if (p <> m) and (iParts[p] > 31) then begin
          y := p;
          bKnownYear := True;
          if bCharMonth then
            bAmbiguous := False;
          Break;
        end;
      end;
    end;

    // If there's no month or year selected yet and DayPos was
    // given, use it as a clue:
    if (m < 0) and (y < 0) and (DayPos > 0) and
       (iParts[DayPos - 1] <= 31) then
      d := DayPos - 1;

    // How about a numeric month value?
    if (m < 0) then begin
      for p := 0 to 2 do begin
        if (p <> y) and (p <> d) and (iParts[p] <= 12) then begin
          m := p;
          Break;
        end;
      end;
    end;

    // The day is whatever is left over and <= 31:
    if (d < 0) then begin
      for p := 0 to 2 do begin
        if (p <> y) and (p <> m) and (iParts[p] <= 31) then begin
          d := p;
          if bAmbiguous and bKnownYear and (iParts[d] > 12) then
            bAmbiguous := False;
          Break;
        end;
      end;
    end;

    // Still no year?
    if y < 0 then
      for p := 0 to 2 do
        if (p <> m) and (p <> d) then
          y := p;

    // If we have all three candidates, try to make a date:
    if (y >= 0) and (m >= 0) and (d >= 0) then begin
      // If the month and day are both <= 12, which is which?
      // If DayPos was provided, use it to choose the best day:
      if bAmbiguous and (iParts[d] <= 12) and (m = (DayPos - 1)) then begin
        m := d;
        d := DayPos - 1;
      end;
      // If the year is only two digits, assume a century:
      if (iParts[y] < 100) and (Length(sParts[y]) < 4) then
        iParts[y] := AssumeCentury(iParts[y]);
      // Make the date:
      OutDate := EncodeDate(iParts[y], iParts[m], iParts[d]);
      Result := True;
      // If this date was unambiguous, set DayPos:
      if (DayPos = 0) and not bAmbiguous then
        DayPos := d + 1;
      // Was a time included?
      sTemp := Trim(sParts[3]);
      if sTemp <> '' then begin
        if TryStrToTime(sTemp, OutTime) then
          ReplaceTime(OutDate, OutTime)
        else
          Result := False; // Dang! We were so close...
      end;
    end;

  except
    on EConvertError do
      Result := False;
  end;

end; {TryAnyStrToDateTime}

{+}
function TryFormatToDateTime(const DS, AFormat: string;
                             out OutDate: TDateTime): Boolean;
{
  Works like TryStrToDateTime, but uses a date/time format string so it knows
  how to parse the string. Shorthand formats like "c" and "tt" are interpreted
  correctly, as are "m"s that follow "h"s.

  Note: The format MUST include separators or whitespace between date and time
  elements. Mashed-together formats like "yyyymmdd" will NOT work!

  Example: TryFormatToDateTime('06/30/2007 2:21 PM', 'mm/dd/yyyy h:nn ampm', D)

  Requires DateUtils, so D6+ only.

  Limitation: It doesn't currently handle single- or double-quoted literals.
-}
  procedure BreakString(const A: string; slList: TStringList);
  var
    sUpper, sWord: string;
    x: Integer;
  begin
    sUpper := AnsiUpperCase(A);
    sWord := '';
    for x := 1 to Length(sUpper) do begin
      if IsDelimiter(UPPER_DIGITS, sUpper, x) then begin
        sWord := sWord + sUpper[x];
      end else if sWord <> '' then begin
        slList.Add(sWord);
        sWord := '';
      end;
    end;
    if sWord <> '' then
      slList.Add(sWord);
  end; {BreakString}
var
  slValues, slFormats: TStringList;
  sTemp: string;
  i, j, p, iMax, iPrev: Integer;
  y, m, d, h, n, s, z: Word;
begin
  Result := False;
  OutDate := 0;
  if Trim(DS) = '' then
    Exit;
  slValues := TStringList.Create;
  slFormats := TStringList.Create;
  try
    // Replace the shorthand formats with their full-size counterparts:
    // TODO: Handle single- or double-quoted literals.
    sTemp := AnsiUpperCase(AFormat);
    if sTemp = '' then
      sTemp := 'C'; // FormatDateTime treats a blank like c
    sTemp := StringReplace(sTemp, 'C',
      FormatSettings.ShortDateFormat + ' ' + FormatSettings.LongTimeFormat, []);
    sTemp := StringReplace(sTemp, 'DDDDDD', FormatSettings.LongDateFormat, []);
    sTemp := StringReplace(sTemp, 'DDDDD', FormatSettings.ShortDateFormat, []);
    sTemp := StringReplace(sTemp, 'TT', FormatSettings.LongTimeFormat, []);
    sTemp := StringReplace(sTemp, 'T', FormatSettings.ShortTimeFormat, []);
    BreakString(sTemp, slFormats);
    BreakString(DS, slValues);
    y := 0;
    m := 0;
    d := 0;
    h := 0;
    n := 0;
    s := 0;
    z := 0;
    iPrev := 0;
    if slValues.Count <= slFormats.Count then
      iMax := slValues.Count
    else
      iMax := slFormats.Count;
    try
      for p := 0 to iMax - 1 do begin
        //                        '12345678'
        j := Pos(slFormats[p][1], 'YMDHNSZA');
        // If M[M] follows H, consider it minutes instead of months:
        if (j = 2) and (iPrev = 4) and (Length(slFormats[p]) <= 2) then
          j := 5;
        iPrev := j;
        case j of
          1: // y = year
            begin
              y := StrToInt(slValues[p]);
              // Assume a century if the year wasn't given with 4 digits:
              if Length(slFormats[p]) < 4 then
                y := AssumeCentury(y);
            end;
          2: // m = month
            begin
              i := MonthStrToInt(slValues[p]);
              if i > 0 then
                m := i
              else
                m := StrToInt(slValues[p]);
            end;
          3: // d = day
            // Ignore word days like "Saturday" from "dddd" and "ddd"
            if Length(slFormats[p]) <= 2 then
              d := StrToInt(slValues[p]);
          4: // h = hour
            h := StrToInt(slValues[p]);
          5: // n = minute
            n := StrToInt(slValues[p]);
          6: // s = second
            s := StrToInt(slValues[p]);
          7: // z = millisecond
            z := StrToInt(slValues[p]);
          8: // AM/PM
            if (h = 12) and (Pos('A', slValues[p]) > 0) then
              h := 0
            else if (h > 0) and (h < 12) and (Pos('P', slValues[p]) > 0) then
              h := h + 12;
        end; // Ignore everything else
      end;
      OutDate := EncodeDate(y, m, d) +
                 EncodeTime(h, n, s, z);
      Result := True;
    except
      on EConvertError do
        Result := False;
    end;
  finally
    slValues.Free;
    slFormats.Free;
  end;
end; {TryFormatToDateTime}

end.

