{*****************************************************************************}
{                                                                             }
{  llDice.pas                                                                 }
{                                                                             }
{  Dice rolling and random number routines                                    }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llDice;

interface

uses
  SysUtils, Math;

  function Dice(const ARolls: Integer;
                const ASides: Integer = 6;
                const AModify: Integer = 0): Integer; overload;

  function Dice(const ADiceStr: string): Integer; overload;

  function TryDice(const ADiceStr: string; out Value: Integer): Boolean;

  function RandRange(const ALo, AHi: Integer): Integer;

implementation

uses
  llIntegerList, llNumbers, llStrings;

{+ dice1}
function Dice(const ARolls: Integer;
              const ASides: Integer = 6;
              const AModify: Integer = 0): Integer; overload;
{
  Returns the sum of one or more virtual dice rolls (random numbers). A die of
  <ASides> sides is rolled <ARolls> times. The sum of these rolls is returned.
  Optionally, the sum can be modified by adding the value of <AModify>, which
  may be negative.

  If only the number of rolls is given, an unmodified 6-sided die is assumed.
-}
var
  i: Integer;
begin
  // Start with the modifier plus one per roll. Random is (0..n-1) not (1..n)
  Result := ARolls + AModify;
  for i := 1 to ARolls do
    Inc(Result, Random(ASides));
end;

{+ dice2}
function Dice(const ADiceStr: string): Integer; overload;
{
  Nerd alert! Accepts a string in "gaming notation", such as "3d6" or "2d4+1"
  and returns the integer sum of the dice roll(s). Returns zero on blank input
  and raises an EConvertError on bad input. Input without a "d" in it (case
  insensitive) is considered a constant and will simply be converted to an
  integer.

  Examples: Dice('3d6')   -> 3..18
            Dice('2d4+2') -> (2..8)+2 (modifier is added to the final sum)
            Dice('d%')    -> 1..100 (1 roll assumed, "%" is shorthand for 100)
            Dice('10')    -> 10
            Dice('foo')   -> EConvertError

  You may nest rolls by using parentheses. For example, if you wanted to roll
  3d6 and then roll that many 12-sided dice, you could use "(3d6)d12". Anywhere
  in the string where you would put a number can be replaced by another roll in
  parentheses. Nesting is effectively unlimited.

  The notation also allows for "rules" which are added after the die string and
  are separated with periods. Each rule name or its abbreviation is immediately
  followed by a number. Where it makes sense, this number may be negative.
  Multiple rules can be combined and are applied in the order shown in the
  following table:

  Rule (abbr)    Notes
  ----------------------------------------------------------------------------
  premod (b)*    Adds # to each die roll before any other rules.
  reroll (rr)    If a die roll is <= #, the die is rerolled.
  extra (x)      If a die roll is >= #, an extra die is rolled.
  postmod (a)*   Adds # to each die roll after any reroll/extra rules.
  max (hi)       Sets the maximum for any single die roll to # after mods.
  min (lo)       Sets the minimum for any single die roll to # after mods.
  retry (rm)     Retry misses up to # times. Requires the "hits" rule.
  drophigh (dh)  Drops (discards) the highest # rolls.
  droplow (dl)   Drops the lowest # rolls.
  keephigh (kh)  Keeps only the highest # rolls.
  keeplow (kl)   Keeps only the lowest # rolls.
  hits (h)       Returns the count of hits (rolls >= #) instead of the sum.
  fails (f)      Returns the count of fails (rolls <= #) instead of the sum.
  totalmax (th)  Sets the maximum total after all rules/modifiers to #.
  totalmin (tl)  Sets the minimum total after all rules/modifiers to #.
  ----------------------------------------------------------------------------
  *The premod and postmod abbreviations of "m" and "mm" have been deprecated in
  favor of "b" and "a" (before and after), respectively.
  
  The hits and fails rules are mutually exclusive.

  A modifier (the "+2" in "2d4+2") affects the final result, regardless of
  whether it is the sum of the rolls or a count of hits/fails. The modifier
  operator can be "+" (addition), "-" (subtraction), "*" (multiplication), or
  "/" (integer division). The totalmax and totalmin rules are applied AFTER the
  modifier.

  Examples:
    4d6.droplow1 (rolls 4d6, drops lowest roll)
    3d6.reroll2 (rolls 3d6, rerolling any ones or twos)
    8d6.hits5 (rolls 8d6, returns the count of "hit" rolls >= 5)
    4d8.premod-2.min1 (rolls 4d8, subtracts 2 from each roll, with a minimum
      individual roll of 1)

  Example: You are playing Dust Tactics and your laser squad is firing. You have
  8 total dice of lasers to fire. Hits are two of six sides on a Dust die, so
  hits5 is used to count a hit whenever a roll is 5 or 6. Lasers are special in
  that they "burn through", so each hit rolls an extra die that may result in
  another hit, ad infinitum. The extra5 rule is used so that any hit gets to
  roll an extra die. The final dice string is "8d6.hits5.extra5".

  Retrying misses allows additional rolls for any misses up to # of times. Using
  4d6.hits5.retry2 as an example, the 4d6 rolls are 1,5,6,4 - two hits, two
  misses. The first retry re-rolls the two misses (1,4) resulting in 6,3. The
  second retry allows the 3 from the first retry attempt to be re-rolled.

  See also: TryDice
-}
const
  MinInt = Low(Integer);
  ModOps = '+-*/';
var
  sDie, sTemp, sRule, sRuleName: string;
  Rolls: TIntegerList;
  i, j, k, iRolls, iSides, iModifier, mop, iMisses: Integer;
  iPreMod, iPostMod, iMax, iMin, iReroll, iExtra, iRetry: Integer;
  iHits, iFails, iDropHigh, iDropLow, iKeepHigh, iKeepLow: Integer;
  iTotalMax, iTotalMin: Integer;
  function RuleMatch(const ALong, AShort: string): Boolean;
  begin
    Result := (sRuleName = ALong) or (sRuleName = AShort);
  end;
  procedure EndlessLoop(const ARule: string; const AValue: Integer);
  begin
    raise EConvertError.CreateFmt(
      'The %s%d rule causes an endless loop condition', [ARule, AValue]);
  end;
begin
  Result := 0;
  if ADiceStr = '' then
    Exit;
  // First check for and replace any nested rolls with their results:
  sDie := LowerCase(ADiceStr);
  i := Pos('(', sDie);
  while i > 0 do begin
    j := FindPairPos(sDie, i);
    if j = 0 then
      raise EConvertError.Create('Mismatched parens in dice string');
    sTemp := Copy(sDie, i + 1, j - (i + 1));
    k := Dice(sTemp); // I <3 recursion!
    Delete(sDie, i, (j - i) + 1);
    Insert(IntToStr(k), sDie, i);
    i := Pos('(', sDie);
  end;
  // Now back to our regularly scheduled programming...
  iMin := MinInt;
  iMax := MaxInt;
  iPreMod := 0;
  iPostMod := 0;
  iReroll := MinInt;
  iExtra := MaxInt;
  iRetry := 0;
  iDropHigh := 0;
  iDropLow := 0;
  iKeepHigh := 0;
  iKeepLow := 0;
  iHits := MaxInt;
  iFails := MinInt;
  iTotalMin := MinInt;
  iTotalMax := MaxInt;
  sDie := StringReplace(sDie, '%', '100', [rfReplaceAll]);
  sRule := '';
  // Parse, preprocess and remove any rules:
  i := Pos('.', sDie);
  if i > 0 then begin
    sTemp := Copy(sDie, i + 1, MaxInt); // The main die string must always be first
    Delete(sDie, i, MaxInt);
    repeat
      i := Pos('.', sTemp);
      if i = 0 then
        i := Length(sTemp) + 1;
      sRule := Copy(sTemp, 1, i - 1);
      Delete(sTemp, 1, i);
      // First, bust out the numeric argument after the rule name, e.g. hits5
      j := Length(sRule);
      while IsDelimiter('0123456789+-', sRule, j) do
        Dec(j);
      if not TryStrToInt(Copy(sRule, j + 1, MaxInt), k) then
        raise EConvertError.Create('Invalid numeric in dice rule: ' + sRule);
      sRuleName := Copy(sRule, 1, j); // Used by RuleMatch
      // The "during rolling" rules:
      if RuleMatch('premod', 'b') or (sRuleName = 'm') then {'m' is deprecated}
        iPreMod := k
      else if RuleMatch('postmod', 'a') or (sRuleName = 'mm') then {'mm' is deprecated}
        iPostMod := k
      else if RuleMatch('reroll', 'rr') then
        iReroll := k
      else if RuleMatch('extra', 'x') then
        iExtra := k
      else if RuleMatch('max', 'hi') then
        iMax := k
      else if RuleMatch('min', 'lo') then
        iMin := k
      else if RuleMatch('retry', 'rm') then
        iRetry := k
      // The "after rolling" rules:
      else if RuleMatch('drophigh', 'dh') then
        iDropHigh := k
      else if RuleMatch('droplow', 'dl') then
        iDropLow := k
      else if RuleMatch('keephigh', 'kh') then
        iKeepHigh := k
      else if RuleMatch('keeplow', 'kl') then
        iKeepLow := k
      else if RuleMatch('hits', 'h') then
        iHits := k
      else if RuleMatch('fails', 'f') then
        iFails := k
      else if RuleMatch('totalmax', 'th') then
        iTotalMax := k
      else if RuleMatch('totalmin', 'tl') then
        iTotalMin := k
      else
        raise EConvertError.Create('Invalid dice rule: ' + sRule);
    until (sTemp = '');
  end;
  // Parse the main die string:
  // Number of rolls:
  i := Pos('d', sDie);
  if i = 0 then begin
    Result := StrToInt(sDie); // No "d"?  It's a constant.
    Exit;
  end;
  if i > 1 then
    iRolls := StrToInt(Copy(sDie, 1, i - 1))
  else
    iRolls := 1; // 1 roll if not specified, e.g. "d6" == "1d6"
  Delete(sDie, 1, i);
  if iRolls < 0 then // Negatives cause index out of bounds errors
    iRolls := 0;
  // Number of sides, modifier:
  iSides := 0; // sides (eliminate warning)
  iModifier := 0; // modifier
  mop := 0; // modifier operator
  for i := 2 to Length(sDie) do begin // It can't be the first character!
    if IsDelimiter(ModOps, sDie, i) then begin
      mop := Pos(sDie[i], ModOps);
      iSides := StrToInt(Copy(sDie, 1, i - 1));
      iModifier := StrToInt(Copy(sDie, i + 1, MaxInt));
      break;
    end;
  end;
  if mop = 0 then // no modifier
    iSides := StrToInt(sDie);
  if iSides < 0 then // Negatives cause overflow! (Zero will always roll a 1)
    iSides := 0;
  // If no rules and an easy modifier, just do it the fast/simple way.
  // sRule will be non-blank (the last rule processed) if we had any.
  if (sRule = '') and (mop <= 2) then begin
    if mop = 2 then
      iModifier := -iModifier;
    Result := Dice(iRolls, iSides, iModifier);
    Exit;
  end;
  // Check for endless loop conditions:
  if iReroll >= (iSides + iPreMod) then
    EndlessLoop('reroll', iReroll);
  if iExtra <= (1 + iPreMod) then
    EndlessLoop('extra', iExtra);
  // Roll the dice:
  Rolls := TIntegerList.Create;
  try
    Rolls.Sorted := True; // Sorted to make high/low stuff easy
    i := iRolls; // Set capacity to the # of rolls. Multiply if extra rolls possible.
    if (iExtra < MaxInt) or (iRetry > 0) then
      i := i * (Max(iRetry, iSides - iExtra + 1) + 1);
    Rolls.Capacity := i;
    repeat
      iMisses := 0; // Count misses for retry attempts
      while iRolls > 0 do begin
        repeat
          i := Random(iSides) + 1 + iPreMod;
        until i > iReroll;
        if i >= iExtra then
          Inc(iRolls); // extra roll!
        i := i + iPostMod;
        i := MinMax(iMin, i, iMax);
        Rolls.Add(i);
        if (iHits < MaxInt) and (iRetry > 0) and (i < iHits) then
          Inc(iMisses);
        Dec(iRolls);
      end;
      iRolls := iMisses;
      Dec(iRetry);
    until iRolls = 0;
    // Apply post-rolling rules:
    while (iDropHigh > 0) and (Rolls.Count > 0) do begin
      Rolls.Delete(Rolls.Count - 1);
      Dec(iDropHigh);
    end;
    while (iDropLow > 0) and (Rolls.Count > 0) do begin
      Rolls.Delete(0);
      Dec(iDropLow);
    end;
    while (iKeepHigh > 0) and (Rolls.Count > iKeepHigh) do
      Rolls.Delete(0);
    while (iKeepLow > 0) and (Rolls.Count > iKeepLow) do
      Rolls.Delete(Rolls.Count - 1);
    // Get the total:
    if iHits < MaxInt then begin
      // It's sorted so just find the first hit & count from there.
      i := Rolls.Count; // Start high as misses are most likely the majority
      while (i > 0) and (Rolls[i - 1] >= iHits) do
        Dec(i);
      Result := Rolls.Count - i;
    end else if iFails > MinInt then begin
      Result := Rolls.Count;
      while (Result > 0) and (Rolls[Result - 1] > iFails) do
        Dec(Result);
    end else
      Result := Rolls.Sum;
    // Apply the modifier to the final result, whether sum or hit/fail count:
    if iModifier <> 0 then begin
      case mop of
        1: Result := Result + iModifier;
        2: Result := Result - iModifier;
        3: Result := Result * iModifier;
        4: Result := Result div iModifier;
      end;
    end;
    // Cap the min/max:
    Result := MinMax(iTotalMin, Result, iTotalMax);
  finally
    Rolls.Free;
  end;
end;

{+}
function TryDice(const ADiceStr: string; out Value: Integer): Boolean;
{
  TryDice is to Dice as TryStrToInt is to StrToInt.
-}
begin
  try
    Value := Dice(ADiceStr);
    Result := True;
  except
    on EConvertError do
      Result := False;
  end;
end;

{+}
function RandRange(const ALo, AHi: Integer): Integer;
{
  Returns a random integer between <ALo> and <AHi>, inclusive.

  This is meant as an alternative to D5's System.RandomRange which is non-
  inclusive, and an equivalent to Math.RandomRange in D6+.
-}
var
  H, L: Integer;
begin
  if ALo <= AHi then begin
    L := ALo;
    H := AHi;
  end else begin
    L := AHi;
    H := ALo;
  end;
  Result := Random((H - L) + 1) + L;
end;

end.
