{*****************************************************************************}
{                                                                             }
{  llFileUtils.pas                                                            }
{                                                                             }
{  File reading/writing routines                                              }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llFileUtils;

interface

uses
  SysUtils, Classes, Windows;

const
  ordEOF = 26; // End Of File marker
  UTF8_BOM = #$EF#$BB#$BF; // Also defined in llStrings
  UTF8_BOM_LENGTH = 3;

  function LoadStream(const FS: TFileStream;
                      const ATrimEOF: Boolean = True;
                      const ATrimBOM: Boolean = True): string;

  function CreateEmptyFile(const AFile: string): Boolean;
  function CreateSimpleFile(const AFile, AContent: string): Boolean;
  function ReadSimpleFile(const AFile: string; out OContent: string): Boolean;

  {+ /trim=2}
  function ReadStreamLine(const FS: TFileStream;
                          var VEncoding: TEncoding;
                          out OLine, OEOL: string): Boolean; overload;

  function ReadStreamLine(const FS: TFileStream;
                          var VEncoding: TEncoding;
                          out OLine: string): Boolean; overload;
  {
  Reads the next line of text from an open TFileStream and places it into the
  OLine argument. The function returns True if a line was read, False if at EOF.
  The file stream's position is left at the start of the next line of text.

  The function properly handles non-ASCII (UTF-8, etc.) files, and accepts any
  combination of end-of-line characters: CRLF, LF, or CR. The EOL character(s)
  found at the end of the current line are placed into the OEOL argument (if
  using that version of the overloaded function). If the line was ended by the
  end of the file, OEOL will be blank.

  VEncoding should be set to nil on the first call for a given file. The first
  read will look for a BOM and set VEncoding so it can be used on all subsequent
  calls.

  See also: ReadStreamLineEOL
  -}

  function ReadStreamLineEOL(const FS: TFileStream;
                             const EOL: string;
                             var VEncoding: TEncoding;
                             out OLine: string): Boolean;

  function TruncateLog(const AFile: string;
                       const AMaxBytes: Int64;
                       const APercentKept: Cardinal = 75;
                       const ATextMarker: string = ''): Boolean;

implementation

uses llFiles, llStrings;

{+}
function LoadStream(const FS: TFileStream;
                    const ATrimEOF: Boolean = True;
                    const ATrimBOM: Boolean = True): string;
{
  Reads an entire file stream into a string and returns the result. Clearly,
  this is meant for small files. Handles non-ASCII files (UTF-8, etc.).

  The file stream is assumed to be open and positioned at zero. Otherwise you
  will just get the rest of the file starting at the current position without
  benefit of any BOM if available.

  If ATrimEOF is True (the default), trailing EOF characters (0x1A, ASCII 26)
  will be removed from the result if present.

  If ATrimBOM is True (the default), a byte order marker will be removed from
  the start of the result if present.
-}
var
  iSize, iOffset: Integer;
  Buffer: TBytes;
  Enc: TEncoding;
begin
  iSize := FS.Size - FS.Position; // Hopefully < 2GB...
  SetLength(Buffer, iSize);
  FS.ReadBuffer(Buffer, iSize);
  Enc := nil;
  iOffset := TEncoding.GetBufferEncoding(Buffer, Enc);
  if ATrimEOF then
    while (iSize > 0) and (Buffer[iSize] = ordEOF) do
      Dec(iSize);
  if not ATrimBOM then
    iOffset := 0;
  if not Assigned(Enc) then
    Enc := TEncoding.Default;
  Result := Enc.GetString(Buffer, iOffset, iSize - iOffset);
end;

{+}
function CreateEmptyFile(const AFile: string): Boolean;
{
  Creates a zero byte file, returning True if successful. If the file already
  exists, it will be overwritten/truncated. Actually just calls CreateSimpleFile
  with no content.

  See also: CreateSimpleFile, ReadSimpleFile
-}
begin
  Result := CreateSimpleFile(AFile, '');
end;

{+}
function CreateSimpleFile(const AFile, AContent: string): Boolean;
{
  Creates a file with a simple string for content, returning a boolean success.
  If the file already exists, it will be overwritten/truncated. If AContent is
  blank, the file will have zero bytes. The file is written using UTF-8 encoding
  without a BOM.

  See also: CreateEmptyFile, ReadSimpleFile
-}
var
  OutStream: TFileStream;
  Buffer: TBytes;
begin
  try
    OutStream := TFileStream.Create(AFile, fmCreate or fmShareExclusive);
    try
      if AContent <> '' then begin
        Buffer := TEncoding.UTF8.GetBytes(AContent);
        OutStream.Write(Buffer[0], Length(Buffer));
      end;
    finally
      OutStream.Free;
    end;
    Result := True;
  except
    // Eat it
    Result := False;
  end;
end;

{+}
function ReadSimpleFile(const AFile: string; out OContent: string): Boolean;
{
  The counterpart to CreateSimpleFile. If the file exists and can be read, the
  function returns True and OContent will contain the file contents. If the file
  does not exist or cannot be opened, the function returns False.

  See also: CreateEmptyFile, CreateSimpleFile, LoadStream
-}
var
  InStream: TFileStream;
begin
  Result := False;
  if FileExists(AFile) then begin
    try
      InStream := TFileStream.Create(AFile, fmOpenRead or fmShareDenyNone);
      try
        OContent := LoadStream(InStream);
        Result := True;
      finally
        InStream.Free;
      end;
    except
      // Eat it
    end;
  end;
end;

function ReadStreamLine(const FS: TFileStream;
                        var VEncoding: TEncoding;
                        out OLine, OEOL: string): Boolean;
const
  BufSize = 255;
  ordCR = 13;
  ordLF = 10;
  cCRLF = #13#10;
var
  bMore: Boolean;
  iNewPos: Int64;
  i, o, iRead, iOffset: Integer;
  Buffer: TBytes;
begin
  Result := False;
  SetLength(Buffer, BufSize + 1);
  bMore := True;
  OLine := '';
  OEOL := '';
  iNewPos := FS.Position;
  iRead := FS.Read(Buffer, BufSize);
  while (iRead > 0) and bMore do begin
    // If this is the first read, see if there's a BOM:
    if iNewPos > 0 then
      iOffset := 0
    else
      iOffset := TEncoding.GetBufferEncoding(Buffer, VEncoding);
    Result := True;
    for i := iOffset to (iRead - 1) do begin
      o := Buffer[i];
      if (o = ordCR) or (o = ordLF) then begin
        bMore := False;
        OLine := OLine + VEncoding.GetString(Buffer, iOffset, i - iOffset);
        iNewPos := iNewPos + i + 1;
        OEOL := Char(Buffer[i]);
        // If CR, look for a following LF
        if o = ordCR then begin
          if i < (iRead - 1) then begin
            if Buffer[i + 1] = ordLF then begin
              Inc(iNewPos);
              OEOL := cCRLF;
            end;
          end else begin
            // Oh snap! It's a CR at the end of the buffer - read one more char
            iRead := FS.Read(Buffer, 1);
            if (iRead = 1) and (Buffer[0] = ordLF) then begin
              Inc(iNewPos);
              OEOL := cCRLF;
            end;
          end;
        end;
        FS.Seek(iNewPos, soBeginning);
        break;
      end;
    end;
    if bMore then begin // No EOL found
      OLine := OLine + VEncoding.GetString(Buffer, iOffset, iRead - iOffset);
      iNewPos := iNewPos + iRead;
      iRead := FS.Read(Buffer, BufSize);
    end;
  end;
end;

function ReadStreamLine(const FS: TFileStream;
                        var VEncoding: TEncoding;
                        out OLine: string): Boolean; overload;
var
  sEOL: string;
begin
  // Just call the other one and discard the EOL value:
  Result := ReadStreamLine(FS, VEncoding, OLine, sEOL);
end;

{+}
function ReadStreamLineEOL(const FS: TFileStream;
                           const EOL: string;
                           var VEncoding: TEncoding;
                           out OLine: string): Boolean;
{
  Just like ReadStreamLine, but you specify the exact end-of-line character(s)
  to look for. EOL can be virtually any string of single-byte characters, but
  must match exactly. The function returns True if a line was read, False if at
  EOF.

  VEncoding should be set to nil on the first call for a given file. The first
  read will look for a BOM and set VEncoding so it can be used on all subsequent
  calls.

  See also: ReadStreamLine
-}
const
  BufSize = 255;
var
  bMore, bAll: Boolean;
  iNewPos: Int64;
  iLenEOL, iRead, iOffset, i, j, r, iEOL: Integer;
  Buffer, EolBuf: TBytes;
  sTest: string;
begin
  iLenEOL := Length(EOL);
  Assert(iLenEOL > 0, 'Non-blank EOL value required.');
  SetLength(Buffer, BufSize);
  SetLength(EolBuf, iLenEOL);
  iEOL := Ord(EOL[1]);
  Result := False;
  bMore := True;
  OLine := '';
  iNewPos := FS.Position;
  iRead := FS.Read(Buffer, BufSize);
  while (iRead > 0) and bMore do begin
    Result := True;
    if iNewPos > 0 then
      iOffset := 0
    else
      iOffset := TEncoding.GetBufferEncoding(Buffer, VEncoding);
    for i := 0 to (iRead - 1) do begin
      if Buffer[i] = iEOL then begin
        if iLenEOL > 1 then begin
          bAll := True; // Flags if we matched the whole EOL
          if (i + iLenEOL) <= iRead then begin
            for j := 2 to iLenEOL do
              if Buffer[i + (j - 1)] <> Ord(EOL[j]) then
                bAll := False;
          end else begin
            // Take what's left in our buffer,
            sTest := '';
            for j := i to (iRead - 1) do
              sTest := sTest + Char(Buffer[j]);
            // read in the remaining bytes required for a whole EOL,
            j := iLenEOL - Length(sTest);
            r := FS.Read(EolBuf, j);
            // and if able to read it, see if it matches:
            if r = j then begin
              for r := 0 to (j - 1) do
                sTest := sTest + Char(EolBuf[r]);
              if sTest = EOL then
                iNewPos := iNewPos + j
              else
                bAll := False;
            end else
              bAll := False;
            // Put our position back where it was if it was a false alarm
            if not bAll then
              FS.Seek(iNewPos + iRead, soBeginning);
          end;
          if not bAll then
            Continue;
        end;
        bMore := False;
        OLine := OLine + VEncoding.GetString(Buffer, iOffset, i - iOffset);
        iNewPos := iNewPos + i + iLenEOL;
        FS.Seek(iNewPos, soBeginning);
        break;
      end;
    end;
    if bMore then begin // No EOL found, go back for more
      OLine := OLine + VEncoding.GetString(Buffer, iOffset, iRead - iOffset);
      iNewPos := iNewPos + iRead;
      iRead := FS.Read(Buffer, BufSize);
    end;
  end;
end;

{+}
function TruncateLog(const AFile: string;
                     const AMaxBytes: Int64;
                     const APercentKept: Cardinal = 75;
                     const ATextMarker: string = ''): Boolean;
{
  If the given text file is greater than AMaxBytes in size, the file is opened
  and the oldest lines are removed to make room for newer log entries. The final
  size will be the requested percentage of AMaxBytes or a bit less (up to the
  next end-of-line) to ensure that the log does not start with a partial line.

  Selecting a APercentKept to a value less than 100% is advised to reduce how
  often the file needs to be shortened.

  If ATextMarker is supplied, additional lines will be deleted until one that
  starts with ATextMarker is found (case-sensitive). If the text marker is not
  found at all, the log file will just be shortened as if it were not supplied.

  Returns a boolean indicating success. Not needing to shorten the file also
  counts as success.
-}
var
  sTemp, sLine, sEOL: string;
  InStream, OutStream: TFileStream;
  iSize, iPos: Int64;
  bFound: Boolean;
  Buffer: TBytes;
  enc: TEncoding;
begin
  Assert((APercentKept <= 100) and (APercentKept > 0),
         'TruncateLog APercentKept must be 1..100');
  Result := False;
  if FFSize(AFile, faAnyFile) <= AMaxBytes then begin
    Result := True;
    Exit; // Nothing to do!
  end;
  try
    InStream := TFileStream.Create(AFile, fmOpenRead or fmShareDenyWrite);
    try
      sTemp := UniqueFileName(AFile + '.tmp');
      OutStream := TFileStream.Create(sTemp, fmCreate or fmShareExclusive);
      try
        // Find out the encoding and write a BOM on output if needed:
        SetLength(Buffer, 10);
        InStream.Read(Buffer, 10);
        TEncoding.GetBufferEncoding(Buffer, enc);
        SetLength(Buffer, 0);
        Buffer := enc.GetPreamble;
        if Length(Buffer) > 0 then
          OutStream.Write(Buffer, Length(Buffer));
        // Our desired final size (plus one in case we land on LF exactly):
        iSize := Round(AMaxBytes * (APercentKept / 100)) + 1;
        // Seek into InStream past the part being discarded:
        InStream.Seek(InStream.Size - iSize, soBeginning);
        // Position to the start of the next line:
        ReadStreamLine(InStream, enc, sLine, sEOL);
        // Do we need to look for a text marker?
        if ATextMarker <> '' then begin
          iPos := InStream.Position;
          repeat
            bFound := StartsWith(ATextMarker, sLine);
          until bFound or not ReadStreamLine(InStream, enc, sLine, sEOL);
          if bFound then begin
            sLine := sLine + sEOL;
            Buffer := enc.GetBytes(sLine);
            OutStream.Write(Buffer, Length(Buffer));
          end else
            InStream.Position := iPos; // Put it back where we found it
        end;
        // Copy the remainder to the output:
        OutStream.CopyFrom(InStream, InStream.Size - InStream.Position);
      finally
        OutStream.Free;
      end;
    finally
      InStream.Free;
    end;
    // Now replace the original with the temp file:
    if SysUtils.DeleteFile(AFile) then
      Result := SysUtils.RenameFile(sTemp, AFile)
    else // Should this raise an exception? No, just return False.
      SysUtils.DeleteFile(sTemp);
  except
    // Eat it, return False
    Result := False;
  end;
end;

end.
