{*****************************************************************************}
{                                                                             }
{  llFiles.pas                                                                }
{                                                                             }
{  File/Directory routines                                                    }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llFiles;

interface

uses
  Windows, SysUtils, Classes, Math; {Max, RoundTo}

const
  DefaultPrettySizeB  = ' B';
  DefaultPrettySizeKB = ' KB';
  DefaultPrettySizeMB = ' MB';
  DefaultPrettySizeGB = ' GB';
  DefaultPrettySizeKilo = 1024;
  DefaultPrettySizeDecimals = 1;

type
  {+ /trim=2 TFileInfo}
  TFileInfo = class(TObject)
  {-}
  private
    FPath: string;
    FName: string;
    FTime: Integer;
    FSize: Int64;
    FAttr: Integer;
    FModified: TDateTime;
    FCreated: TDateTime;
    FAccessed: TDateTime;
    procedure SetPath(const Value: string);
    function GetFilename: string;
    procedure SetFilename(const Value: string);
  {+ /trim=2 /addline=OFF TFileInfo}
  public
    constructor Create;
    constructor CreateSR(const APath: string; const SR: TSearchRec);
    property Filename: string read GetFilename write SetFilename; // with path
    property Path: string read FPath write SetPath;
    property Name: string read FName write FName;
    property Time: Integer read FTime write FTime;
    property Size: Int64 read FSize write FSize;
    property Attr: Integer read FAttr write FAttr;
    property Timestamp: TDateTime read FModified write FModified; // Synonym
    property Modified: TDateTime read FModified write FModified;
    property Created: TDateTime read FCreated write FCreated;
    property Accessed: TDateTime read FAccessed write FAccessed;
    procedure Clone(const ASource: TFileInfo);
    procedure Update(const APath: string; const SR: TSearchRec);
    function UpdateFromFile(const AFile: string): Boolean;
  end;
  {+ /trim=OFF /addline=OFF

  A TObject-based version of TSearchRec that includes a path and pre-converted
  file times.

  When storing these in TStringList.Objects, you can use the following functions
  with TStringList.CustomSort to sort them by time or size:

  SortFileInfoByTime
  SortFileInfoByTimeDesc
  SortFileInfoBySize
  SortFileInfoBySizeDesc

  See also: TRecurser
  -}

  TFileSorting = (sortNatural, sortName, sortTime, sortSize, sortCustom);

  {+ /trim=2 TRecurser}
  TRecurser = class(TObject)
  {-}
  private
    FStartPath: string;
    FWildcard: string;
    FAttr: Integer;
    FAnyAttr: Boolean;
    FMinDepth: Integer;
    FMaxDepth: Integer;
    FSorting: TFileSorting;
    FSortDesc: Boolean;
    FCustomSort: TStringListSortCompare;
    FOnFile: TNotifyEvent;
    FOnBeforeSub: TNotifyEvent;
    FOnAfterSub: TNotifyEvent;
    FFileInfo: TFileInfo;
    FEvent: Char;
    FFullPath: string;
    FError: string;
    FCancel: Boolean;
    FSkip: Boolean;
    FDepth: Integer;
    FBottom: Integer;
    function GetAttr: Integer;
    procedure SetAttr(const Value: Integer);
    function GetWildcard: string;
    procedure SetWildcard(const Value: string);
  {+ /trim=2 /addline=OFF TRecurser}
  public
    constructor Create;
    property Wildcard: string read GetWildcard write SetWildcard;
    property Attr: Integer read GetAttr write SetAttr;
    property MinDepth: Integer read FMinDepth write FMinDepth;
    property MaxDepth: Integer read FMaxDepth write FMaxDepth;
    property Sorting: TFileSorting read FSorting write FSorting;
    property SortDescending: Boolean read FSortDesc write FSortDesc;
    property CustomSort: TStringListSortCompare read FCustomSort write FCustomSort;
    property OnFile: TNotifyEvent read FOnFile write FOnFile;
    property OnBeforeSub: TNotifyEvent read FOnBeforeSub write FOnBeforeSub;
    property OnAfterSub: TNotifyEvent read FOnAfterSub write FOnAfterSub;
    property Cancelled: Boolean read FCancel write FCancel;
    property Skip: Boolean read FSkip write FSkip;
    property FullPath: string read FFullPath;
    property FileInfo: TFileInfo read FFileInfo;
    property Event: Char read FEvent;
    property Depth: Integer read FDepth;
    property Bottom: Integer read FBottom;
    property ErrorStr: string read FError;
    function Process: Boolean;
  end;
  {+ /trim=OFF /addline=OFF

  Use TRecurser to perform actions on files and/or subdirectories while
  recursing a directory tree.

  Set Wildcard to the starting directory and file wildcard you want to process,
  then call Process. All subdirectories (*.*) are recursed; the wildcard only
  controls the filenames that are processed. Subdirectories are processed in
  alphabetical order.

  The OnFile action is called for every file that matches the wildcard, in the
  order determined by the Sorting option.

  The OnBeforeSub action is called before each subdirectory is entered. The
  OnAfterSub action is called after processing a subdirectory, and could be used
  if you wanted to do something like remove empty subdirectories after deleting
  files.

  ----- Read/Write Properties (default)

  Wildcard
    Set this to the initial directory and file wildcard, e.g. 'D\Path\*.*'. If a
    path is not included, the current directory (at the time the property is
    set) is assumed.

  Attr (0)
    This value is passed as the Attr argument to FindFirst to control what types
    of files you want to find based on file attributes. It defaults to zero,
    which is all normal files. Hidden subdirectories will be recursed if Attr
    includes faHidden.

    If Attr is negative, only files that match the attribute mask exactly will
    be added to the list. For example, to *only* include files with the archive
    attribute set, use -faArchive.

  MinDepth (-1)
    If you don't want to start processing until you get to a certain level of
    subdirectory, set this. The initial directory has a depth of zero. For
    example, if you wanted to skip files in the starting directory and process
    them in subdirectories only, set MinDepth := 1. Defaults to -1 (no minimum).

  MaxDepth (MaxInt)
    If you want to recurse only to a certain subdirectory depth, set this. The
    initial directory has a depth of zero. Defaults to MaxInt (no maximum).

  Sorting (sortNatural)
    Choose the order that files from a directory are sorted in before calling
    the OnFile events. The TFileSorting options are sortNatural (the order they
    are returned by FindFirst/FindNext), sortName, sortTime, sortSize, or
    sortCustom. If custom, the CustomSort procedure must also be assigned.

  SortDescending (False)
    Set to True if you want the files sorted in descending order based on the
    Sorting option. This is done by simply reversing the stringlist's order
    after sorting in ascending order. SortDescending is ignored when Sorting is
    sortNatural.

  CustomSort (nil)
    A TStringListSortCompare function to use when Sorting is set to sortCustom.
    The list's Objects are populated with TFileInfo objects. If SortDescending
    is True, the list will be reversed after sorting with this function.

  OnFile
    Set this to a standard TNotifyEvent procedure to call for each file found.
    It is passed a reference to the TRecurser as Sender. FullPath will be set to
    the fully qualified filename. If you need additional information about the
    file (date, size, etc.), reference the FileInfo property.

  OnBeforeSub
  OnAfterSub
    Set these to a TNotifyEvent procedure to call before and/or after processing
    a subdirectory, respectively. These procs are not called on the starting
    directory, only subdirectories. The proc is passed a reference to the
    TRecurser object as Sender. FullPath will be set to the fully qualified
    subdirectory name including a trailing backslash.

  Cancelled (False)
    Cancelled can be set in any of the event procedures to cancel the process
    and return. If cancelled, the Process function returns False and ErrorStr
    will be set to 'Cancelled'.

  Skip
    Skip can be set in the OnFile or OnBeforeSub events to skip processing of
    the current directory and its children. Sibling directories and their
    children will continue to be processed. Setting it during OnFile will stop
    processing any remaining files from the current directory, and won't enter
    any of its subdirectories. Setting it from OnBeforeSub will prevent entering
    that particular subdirectory and will skip the call to OnAfterSub.

  ----- Readonly Properties

  FullPath
    Reference this during an event procedure to get the name of the file or
    subdirectory being processed. Directory names will always have a trailing
    backslash. If Process fails, FullPath can tell you where it stopped.

  FileInfo
    This TFileInfo object contains information about the current file being
    processed. It is available only during the OnFile event, and is otherwise
    unassigned.

  Event
    If you wanted to use a single procedure for all three events, this character
    can be used to tell you which one is currently being called:
    F=OnFile, B=OnBeforeSub, A=OnAfterSub.

  Depth
    The current depth of the directory being processed. The starting directory
    has a depth of zero. In the OnBeforeSub and OnAfterSub events, Depth is the
    depth of the subdirectory's parent, not the depth inside the subdirectory.
    Example: Wildcard='D:\Foo\*.*', OnAfterSub for D:\Foo\Bar\ reports a Depth
    of 0, the depth of D:\Foo where Bar lives. In the OnFile event for
    D:\Foo\Bar\test.txt, Depth would be 1.

  Bottom
    The highest depth reached during the process. If read during an event before
    Process has finished, it is the highest depth reached *so far*.

  ErrorStr
    If Process returns False, this will be set to the error that caused it to
    fail. If Cancelled was set to True, ErrorStr will be 'Cancelled'.

  ----- Methods

  Process
    After setting the other properties, call Process to begin the recursion. If
    successful, the function will return True. If it fails or is cancelled, it
    will return False and ErrorStr will contain the reason why.

  See also: TFileInfo
  -}

  function AddDot(const AExt: string): string;
  function StripDot(const AExt: string): string;

  function AddSlash(const APath: string): string;
  function StripSlash(const APath: string): string;
  function SlashPath(const APath: string; const ASlash: Boolean): string;

  function StripExt(const AFile: string): string;

  function AttributeStr(const Attr: Integer;
                        const AFormat: string = 'ADHRS'): string;
  function StrToAttribute(const S: string): Integer;

  function DirectoryIsEmpty(const ADir: string): Boolean;

  function DiskFreeByFile(const AFile: string): Int64;

  function EllipsisPath(const APath: string; const ALength: Integer): string;

  function ExtractFileExtNoDot(const AFile: string): string;
  function ExtractFileNameNoExt(const AFile: string): string;

  function FFDate(const AFile: string; const Attr: Integer = 0): TDateTime;
  function FFSize(const AFile: string; const Attr: Integer = 0): Integer;
  function FFSize64(const AFile: string; const Attr: Integer = 0): Int64;

  function FileCopy(const ASource, ATarget: string;
                    const APreserveAttr: Boolean = False;
                    const APreserveTimes: Boolean = False;
                    const AMove: Boolean = False): Boolean;

  function FileTimeToDateTime(const AFileTime: TFileTime): TDateTime;

  function FindInPath(const AFile: string; out OResult: string): Boolean;

  function FixFilePath(const AFile, APath: string): string;

  function IsDotDir(const SR: TSearchRec): Boolean;
  function IsSubDir(const SR: TSearchRec): Boolean;

  function IsLegalFilename(const S: string): Boolean;

  function ListFiles(const AList: TStrings;
                     const Path: string;
                     const Attr: Integer = 0;
                     const IncludePath: Boolean = False;
                     const Recurse: Boolean = False;
                     const UseFileInfo: Boolean = False): Integer;
  function ListSubdirs(const AList: TStrings;
                       const ADir: string;
                       const IncludeHidden: Boolean = False;
                       const IncludePath: Boolean = False): Integer;

  function LoadStrings(const AFile: string; const AList: TStrings): Boolean;

  function PrettySize(const ASize: Int64): string;

  function SameFile(const F1, F2: string; const APath: string = ''): Boolean;

  function SecureDelete(const AFile: string; const APasses: Integer = 1): Boolean;

  function UniqueFileName(const AFile: string): string;
  function UnusedFileName(const AFile: string): string; deprecated;

  function WildCopy(const ASource, ATarget: string;
                    const Attr: Integer = 0;
                    const ARecurse: Boolean = False;
                    const APreserveAttr: Boolean = False;
                    const APreserveTimes: Boolean = False;
                    const ACopyEmptySubDirs: Boolean = False): Boolean;

  function WildDelete(const AFileSpec: string;
                      const Attr: Integer = 0;
                      const ARecurse: Boolean = False): Integer;

  function WildExists(const AFileSpec: string;
                      const Attr: Integer = 0;
                      const ARecurse: Boolean = False): Boolean;

  {+ /trim=2 SortFileInfoBy}
  function SortFileInfoByTime(List: TStringList; Index1, Index2: Integer): Integer;
  function SortFileInfoByTimeDesc(List: TStringList; Index1, Index2: Integer): Integer;
  function SortFileInfoBySize(List: TStringList; Index1, Index2: Integer): Integer;
  function SortFileInfoBySizeDesc(List: TStringList; Index1, Index2: Integer): Integer;
  {+ /trim=2
    Use with TStringList.CustomSort after calling ListFiles with FileInfo to
    sort the list by the files' date or size. Items that compare equally on the
    main comparison will be compared by string value and will honor the list's
    CaseSensitive setting.

    See also: ListFiles, ReverseStringList, TFileInfo
  -}

  // TFileStream-related - see llFileUtils for Unicode compatible replacements

var
  PrettySizeKilo: Integer;
  PrettySizeDecimals: Integer;
  PrettySizeB: string;
  PrettySizeKB: string;
  PrettySizeMB: string;
  PrettySizeGB: string;

implementation

uses llStrings, ShellAPI;

{ TFileInfo }

constructor TFileInfo.Create;
begin
  inherited;
end;

constructor TFileInfo.CreateSR(const APath: string; const SR: TSearchRec);
begin
  Create;
  Update(APath, SR);
end;

procedure TFileInfo.Clone(const ASource: TFileInfo);
begin
  Path := ASource.Path;
  Name := ASource.Name;
  Time := ASource.Time;
  Size := ASource.Size;
  Attr := ASource.Attr;
  Modified := ASource.Modified;
  Created := ASource.Created;
  Accessed := ASource.Accessed;
end;

function TFileInfo.GetFilename: string;
begin
  Result := FPath + FName;
end;

procedure TFileInfo.SetFilename(const Value: string);
begin
  Path := ExtractFilePath(Value);
  Name := ExtractFileName(Value);
end;

procedure TFileInfo.SetPath(const Value: string);
begin
  FPath := AddSlash(Value);
end;

procedure TFileInfo.Update(const APath: string; const SR: TSearchRec);
begin
  Path := APath;
  Name := SR.Name;
  {$WARN SYMBOL_DEPRECATED OFF}
  Time := SR.Time; // TSearchRec.Time is deprecated since XE
  {$WARN SYMBOL_DEPRECATED ON}
  Size := SR.Size;
  Attr := SR.Attr;
  Modified := SR.TimeStamp;
  Created := FileTimeToDateTime(SR.FindData.ftCreationTime);
  Accessed := FileTimeToDateTime(SR.FindData.ftLastAccessTime);
end;

function TFileInfo.UpdateFromFile(const AFile: string): Boolean;
var
  SR: TSearchRec;
  sPath: string;
begin
  if FindFirst(AFile, 0, SR) = 0 then begin
    sPath := ExtractFilePath(AFile);
    if sPath = '' then
      sPath := GetCurrentDir;
    Update(sPath, SR);
    FindClose(SR);
    Result := True;
  end else
    Result := False;
end;

function SortFileInfoByTime(List: TStringList; Index1, Index2: Integer): Integer;
var
  t1, t2: TDateTime;
begin
  t1 := TFileInfo(List.Objects[Index1]).Modified;
  t2 := TFileInfo(List.Objects[Index2]).Modified;
  if t1 > t2 then
    Result := 1
  else if t1 < t2 then
    Result := -1
  else if List.CaseSensitive then // Sort by name if times are equal
    Result := AnsiCompareStr(List[Index1], List[Index2])
  else
    Result := AnsiCompareText(List[Index1], List[Index2]);
end;

function SortFileInfoByTimeDesc(List: TStringList; Index1, Index2: Integer): Integer;
begin
  Result := -SortFileInfoByTime(List, Index1, Index2);
end;

function SortFileInfoBySize(List: TStringList; Index1, Index2: Integer): Integer;
var
  s1, s2: Int64;
begin
  s1 := TFileInfo(List.Objects[Index1]).Size;
  s2 := TFileInfo(List.Objects[Index2]).Size;
  if s1 > s2 then
    Result := 1
  else if s1 < s2 then
    Result := -1
  else if List.CaseSensitive then // Sort by name if sizes are equal
    Result := AnsiCompareStr(List[Index1], List[Index2])
  else
    Result := AnsiCompareText(List[Index1], List[Index2]);
end;

function SortFileInfoBySizeDesc(List: TStringList; Index1, Index2: Integer): Integer;
begin
  Result := -SortFileInfoBySize(List, Index1, Index2);
end;

{ TRecurser }

constructor TRecurser.Create;
begin
  inherited;
  FMinDepth := -1;
  FMaxDepth := MaxInt;
  FSorting := sortNatural;
  FAnyAttr := True;
end;

function TRecurser.GetAttr: Integer;
begin
  if FAnyAttr then
    Result := FAttr
  else
    Result := -FAttr;
end;

function TRecurser.GetWildcard: string;
begin
  Result := FStartPath + FWildcard;
end;

procedure TRecurser.SetAttr(const Value: Integer);
begin
  FAnyAttr := Value >= 0;
  FAttr := Abs(Value);
end;

procedure TRecurser.SetWildcard(const Value: string);
begin
  FStartPath := ExtractFilePath(Value);
  FWildcard := ExtractFilename(Value);
  if FStartPath = '' then
    FStartPath := AddSlash(GetCurrentDir);
end;

function TRecurser.Process: Boolean;
  procedure ProcessDir(const APath: string; const ADepth: Integer);
  var
    SR: TSearchRec;
    slFiles, slSubs: TStringList;
    i: Integer;
  begin
    // APath is assumed to include a trailing backslash
    if ADepth > FBottom then
      FBottom := ADepth;
    if Assigned(FOnFile) and (ADepth >= FMinDepth) then begin
      if FindFirst(APath + FWildcard, FAttr, SR) = 0 then begin
        FSkip := False;
        FDepth := ADepth;
        FEvent := 'F';
        if FSorting = sortNatural then begin
          // Don't bother building a list when sorting is sortNatural
          FFileInfo := TFileInfo.Create;
          try
            repeat
              if ((SR.Attr and faDirectory) = 0) and
                 (FAnyAttr or ((SR.Attr and FAttr) <> 0)) then begin
                FFullPath := APath + SR.Name;
                FFileInfo.Update(APath, SR);
                FOnFile(Self);
                if FCancel or FSkip then
                  Exit;
              end;
            until FindNext(SR) <> 0;
          finally
            FindClose(SR);
            FreeAndNil(FFileInfo);
          end;
        end else begin
          slFiles := TStringList.Create;
          try
            repeat
              if ((SR.Attr and faDirectory) = 0) and
                 (FAnyAttr or ((SR.Attr and FAttr) <> 0)) then
                slFiles.AddObject(APath + SR.Name, TFileInfo.CreateSR(APath, SR));
            until FindNext(SR) <> 0;
            FindClose(SR);
            if slFiles.Count > 1 then begin
              case FSorting of
                sortName: slFiles.Sort;
                sortTime: slFiles.CustomSort(SortFileInfoByTime);
                sortSize: slFiles.CustomSort(SortFileInfoBySize);
                sortCustom: slFiles.CustomSort(FCustomSort);
              end;
              if FSortDesc then
                ReverseStringList(slFiles);
            end;
            for i := 0 to slFiles.Count - 1 do begin
              FFullPath := slFiles[i];
              FFileInfo := TFileInfo(slFiles.Objects[i]);
              FOnFile(Self);
              if FCancel or FSkip then
                Exit;
            end;
          finally
            FFileInfo := nil;
            for i := 0 to slFiles.Count - 1 do
              TFileInfo(slFiles.Objects[i]).Free;
            slFiles.Free;
          end;
        end;
      end;
    end;
    if ADepth < FMaxDepth then begin
      slSubs := TStringList.Create;
      try
        ListSubdirs(slSubs, APath, (FAttr and faHidden) <> 0, False);
        slSubs.Sort;
        for i := 0 to slSubs.Count - 1 do begin
          FSkip := False;
          FFullPath := AddSlash(APath + slSubs[i]);
          if Assigned(FOnBeforeSub) and (ADepth >= FMinDepth) then begin
            FDepth := ADepth;
            FEvent := 'B';
            FOnBeforeSub(Self);
            if FCancel then
              Break;
            if FSkip then
              Continue;
          end;
          ProcessDir(FFullPath, ADepth + 1);
          if FCancel then
            Break;
          if Assigned(FOnAfterSub) and (ADepth >= FMinDepth) then begin
            FFullPath := AddSlash(APath + slSubs[i]);
            FDepth := ADepth;
            FEvent := 'A';
            FOnAfterSub(Self);
            if FCancel then
              Break;
          end;
        end;
      finally
        slSubs.Free;
      end;
    end;
  end;
begin
  Result := False;
  FCancel := False;
  if (FStartPath = '') or (FWildcard = '') then begin
    FError := 'Wildcard required';
    Exit;
  end;
  if not (Assigned(FOnFile) or
          Assigned(FOnBeforeSub) or
          Assigned(FOnAfterSub)) then begin
    FError := 'No events assigned';
    Exit;
  end;
  if (FSorting = sortCustom) and not Assigned(FCustomSort) then begin
    FError := 'CustomSort function not assigned';
    Exit;
  end;
  FError := '';
  FBottom := 0;
  FFullPath := FStartPath;
  try
    ProcessDir(FStartPath, 0);
    if FCancel then
      FError := 'Cancelled'
    else
      Result := True;
  except
    on e:Exception do
      FError := e.Message;
  end;
end;

{*****************************************************************************}

{+}
function AddDot(const AExt: string): string;
{
  Adds a dot to the beginning of an extension if one is not already present.
  A blank string will not be modified.
-}
begin
  Result := AExt;
  if (Result <> '') and (Result[1] <> '.') then
    Result := '.' + Result;
end;

{+}
function StripDot(const AExt: string): string;
{
  Removes a dot from the beginning of an extension if present.
-}
begin
  Result := AExt;
  while (Result <> '') and (Result[1] = '.') do
    Delete(Result, 1, 1);
end;

{+}
function AddSlash(const APath: string): string;
{
  Just a shorter way to write IncludeTrailingPathDelimiter.
-}
begin
  Result := IncludeTrailingPathDelimiter(APath);
end;

{+}
function StripSlash(const APath: string): string;
{
  Just a shorter way to write ExcludeTrailingPathDelimiter.
-}
begin
  Result := ExcludeTrailingPathDelimiter(APath);
end;

{+}
function SlashPath(const APath: string; const ASlash: Boolean): string;
{
  Calls AddSlash/StripSlash based on a True/False value, respectively.
-}
begin
  if ASlash then
    Result := AddSlash(APath)
  else
    Result := StripSlash(APath);
end;

{+}
function StripExt(const AFile: string): string;
{
  Just an easy-to-type wrapper for ChangeFileExt(AFilename, '');
-}
begin
  Result := ChangeFileExt(AFile, '');
end;

{+}
function AttributeStr(const Attr: Integer;
                      const AFormat: string = 'ADHRS'): string;
{
  Turns a file attribute integer (TSearchRec.Attr) into a string of letters
  indicating the attributes present.

  The order, case and positioning of the letters are determined by the AFormat
  argument. If AFormat has a 6th character, it is used as a placeholder in the
  output for missing attributes, resulting in a fixed-length return value. The
  first five characters of AFormat MUST be the letters 'ADHRS' in any order.

  Examples (where Attr is faArchive or faReadOnly):
  AttributeStr(Attr)           -> 'AR'
  AttributeStr(Attr, 'adhrs ') -> 'a  r '
  AttributeStr(Attr, 'DHRSA-') -> '--R-A'

  See also: StrToAttribute
-}
var
  bFixed: Boolean;
  sUpper: string;
  procedure AddLtr(const fa: Integer; const Ltr: Char);
  var
    p: Integer;
  begin
    if (Attr and fa) <> 0 then begin
      p := Pos(Ltr, sUpper);
      if p > 0 then
        Result[p] := AFormat[p];
    end;
  end;
begin
  bFixed := Length(AFormat) > 5;
  if bFixed then
    Result := StringOfChar(AFormat[6], 5)
  else
    Result := StringOfChar(' ', 5);
  sUpper := UpperCase(Copy(AFormat, 1, 5));
  AddLtr(faArchive,   'A');
  AddLtr(faDirectory, 'D');
  AddLtr(faHidden,    'H');
  AddLtr(faReadOnly,  'R');
  AddLtr(faSysFile,   'S');
  if not bFixed then
    Result := StringReplace(Result, ' ', '', [rfReplaceAll]); // Strip spaces
end;

{+}
function StrToAttribute(const S: string): Integer;
{
  Converts a string containing letters (ADHRS) into a numeric file attribute
  value. If the string contains an asterisk (*), faAnyFile is returned.
  Characters other than ADHRS and * are ignored. A blank or invalid string will
  return a result of zero.

  See also: AttributeStr
-}
var
  sUpper: string;
  i: Integer;
begin
  Result := 0;
  sUpper := UpperCase(S);
  for i := 1 to Length(sUpper) do
    case sUpper[i] of
      'A': Result := Result or faArchive;
      'D': Result := Result or faDirectory;
      'H': Result := Result or faHidden;
      'R': Result := Result or faReadonly;
      'S': Result := Result or faSysFile;
      '*': Result := faAnyFile;
    end;
end;

{+}
function DirectoryIsEmpty(const ADir: string): Boolean;
{
  Returns True if the directory in question contains no files or subdirectories.
  It will also return True if the directory doesn't exist, so watch for that.
-}
var
  SR: TSearchRec;
begin
  Result := True;
  if FindFirst(AddSlash(ADir) + '*.*', faAnyFile, SR) = 0 then begin
    repeat
      Result := IsDotDir(SR); // Anything but . and .. means it's not empty!
    until (FindNext(SR) <> 0) or not Result;
    FindClose(SR);
  end;
end;

{+}
function DiskFreeByFile(const AFile: string): Int64;
{
  Calls DiskFree(n) for the drive letter specified in the path of the argument.
  If the filename does not include a path, the current directory is assumed.
  If the call fails on a UNC path, the function returns -1.
-}
var
  iDrive: Integer;
  sDrive: string;
  iFree64, iTotal64: Int64;
begin
  sDrive := ExtractFileDrive(AFile);
  if sDrive = '' then
    iDrive := 0 // 0 = current dir
  else if sDrive[1] = PathDelim then begin // UNC path
    sDrive := ExtractFileDir(AFile);
    if GetDiskFreeSpaceEx(PChar(sDrive), iFree64, iTotal64, nil) then
      Result := iFree64
    else
      Result := -1;
    Exit;
  end else begin
    sDrive := UpperCase(sDrive);
    iDrive := (Ord(sDrive[1]) - Ord('A')) + 1;
  end;
  Result := DiskFree(iDrive);
end;

{+}
function EllipsisPath(const APath: string; const ALength: Integer): string;
{
  If a path string is longer than the desired length, middle subdirectories are
  replaced with an ellipsis until the total length is under the limit.

  Example: EllipsisPath('C:\Foo\BarBaz\Ray01.txt', 20) -> "C:\Foo\...\Ray01.txt"

  If there isn't enough room for even the final directory or file name, the left
  side will be truncated and replaced with an ellipsis, e.g. "...SuperLongDir\".

  The smallest possible return value is "..." for lengths <= 3.
-}
var
  slDirs: TStringList;
  iLast, iLeft, iRight, iLenL, iLenR: Integer;
  procedure MakeResult;
  var
    r: Integer;
  begin
    Result := '';
    for r := 0 to iLeft - 1 do
      Result := Result + slDirs[r] + PathDelim;
    Result := Result + '...';
    for r := iRight + 1 to iLast do
      Result := Result + PathDelim + slDirs[r];
  end;
begin
  Result := APath; // C:\Foo\Bar\Baz\
  if Length(Result) > ALength then begin
    slDirs := TStringList.Create;
    try
      TokenList(Result, PathDelim, slDirs); // C: Foo Bar Baz ""
      iLast := slDirs.Count - 1;
      if slDirs[iLast] = '' then begin // Did it end with a backslash?
        slDirs.Delete(iLast);
        Dec(iLast);
        slDirs[iLast] := slDirs[iLast] + PathDelim;
      end;
      // Start as close to the middle as possible.
      iLeft := slDirs.Count div 2;
      iRight := iLeft;
      MakeResult;
      while (Length(Result) > ALength) and (iLeft > 0) do begin
        // Work from the middle out, going for the longest parts first to remove
        // the fewest directories as possible to reach the desired length.
        iLenL := Length(slDirs[iLeft - 1]); // iLeft is known to be > 0
        if iRight < (iLast - 1) then
          iLenR := Length(slDirs[iRight + 1])
        else
          iLenR := 0;
        if iLenL > iLenR then
          Dec(iLeft)
        else if iLenR > iLenL then
          Inc(iRight)
        else if iRight < iLast - 2 then
          Inc(iRight)
        else if (iLeft > 1) or (iRight >= iLast - 1) then
          Dec(iLeft)
        else
          Inc(iRight);
        MakeResult;
      end;
      if Length(Result) > ALength then
        Result := '...' + Rightmost(slDirs[iLast], Max(ALength - 3, 0));
    finally
      slDirs.Free;
    end;
  end;
end;

{+}
function ExtractFileExtNoDot(const AFile: string): string;
{
  Just like ExtractFileExt, but removes the leading dot if present.
-}
begin
  Result := StripDot(ExtractFileExt(AFile));
end;

{+}
function ExtractFileNameNoExt(const AFile: string): string;
{
  Like ExtractFileName, but also removes any file extension.

  Example: ExtractFileNameNoExt('d:\path\file.ext') -> 'file'
-}
begin
  Result := StripExt(ExtractFileName(AFile));
end;

{+}
function FFDate(const AFile: string; const Attr: Integer = 0): TDateTime;
{
  Retrieves a file's modified date/time from FindFirst, or zero if not found.
  Supply Attr if you need to find hidden files, etc. Directories are ignored.

  If given a wildcard instead of a single filename, the date of the newest file
  will be returned.

  See also: FileTimeToDateTime, FFSize, FFSize64
-}
var
  SR: TSearchRec;
  d: TDateTime;
begin
  Result := 0;
  if FindFirst(AFile, Attr, SR) = 0 then begin
    repeat
      if (SR.Attr and faDirectory) = 0 then begin
        d := SR.TimeStamp;
        if d > Result then
          Result := d;
      end;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
end;

{+}
function FFSize(const AFile: string; const Attr: Integer = 0): Integer;
{
  This actually just calls FFSize64 and caps the result at MaxInt.

  See also: FFSize64, FFDate
-}
var
  iSize: Int64;
begin
  iSize := FFSize64(AFile, Attr);
  if iSize <= MaxInt then
    Result := iSize
  else
    Result := MaxInt;
end;

{+}
function FFSize64(const AFile: string; const Attr: Integer = 0): Int64;
{
  Unlike the FileSize function that requires a handle to an open file, this one
  just gets it from FindFirst. Supply Attr if you need to find hidden files, 
  etc. Directories are ignored.
  
  Returns the size of the file(s) in bytes or -1 if not found. If given a
  wildcard instead of a filename, the sum of the sizes is returned.

  See also: FFSize, FFDate
-}
var
  SR: TSearchRec;
begin
  if FindFirst(AFile, Attr, SR) = 0 then begin
    Result := 0;
    repeat
      if (SR.Attr and faDirectory) = 0 then
        Result := Result + SR.Size;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end else
    Result := -1;
end;

{+}
function FileCopy(const ASource, ATarget: string;
                  const APreserveAttr: Boolean = False;
                  const APreserveTimes: Boolean = False;
                  const AMove: Boolean = False): Boolean;
{
  Copies a file and returns True if successful. If the target file exists, it
  will be overwritten. If AMove is set to True, the original will be deleted
  after a successful copy.

  The Windows.CopyFile function copies a file along with its existing file
  attributes, last modified date, and encryption (if any, and assuming the
  target drive supports it). By default, this copies a file without all that
  extra baggage and sets the creation and last modified times to current.

  The attributes and/or file times CAN be preserved on the copy if either of the
  two optional arguments are set to True.
-}
var
  wShare: Word;
  fsSource, fsTarget: TFileStream;
  ftCreated, ftAccessed, ftModified: TFileTime;
  pCreated, pAccessed, pModified: PFileTime;
begin
  Result := False;
  if not FileExists(ASource) then
    Exit;
  pCreated := @ftCreated; // The pointers must have somewhere to point
  pAccessed := @ftAccessed;
  pModified := @ftModified;
  try
    if AMove then
      wShare := fmShareExclusive // If this fails, so will delete - find out early
    else
      wShare := fmShareDenyNone;
    fsSource := TFileStream.Create(ASource, fmOpenRead or wShare);
    try
      if APreserveTimes then
        if not GetFileTime(fsSource.Handle, pCreated, pAccessed, pModified) then
          Exit;
      if FileExists(ATarget) then
        if not DeleteFile(ATarget) then
          Exit;
      fsTarget := TFileStream.Create(ATarget, fmCreate or fmShareExclusive);
      try
        fsTarget.CopyFrom(fsSource, 0);
        if APreserveTimes then
          Result := SetFileTime(fsTarget.Handle, pCreated, pAccessed, pModified)
        else
          Result := True;
      finally
        fsTarget.Free;
      end;
    finally
      fsSource.Free;
    end;
    if Result and APreserveAttr then
      Result := SetFileAttributes(PChar(ATarget),
                                  GetFileAttributes(PChar(ASource)));
    if Result and AMove then
      Result := DeleteFile(ASource);
  except
    // Eat it and return False
  end;
end;

{+}
function FileTimeToDateTime(const AFileTime: TFileTime): TDateTime;
{
  Converts a TSearchRec.FindData file time to a TDateTime, adjusting for the
  local time zone.

  See also: TFileInfo, FFDate
-}
var
  SystemTime, LocalTime: TSystemTime;
begin
  FileTimeToSystemTime(AFileTime, SystemTime);
  SystemTimeToTzSpecificLocalTime(nil, SystemTime, LocalTime);
  Result := SystemTimeToDateTime(LocalTime);
end;

{+}
function FindInPath(const AFile: string; out OResult: string): Boolean;
{
  Checks for the existence of a file in the current directory and PATH
  environment variable, returning True if found. The fully qualified name of
  the found file is placed into OResult. If found in the current directory,
  OResult will not contain a path, just the filename.

  If the input filename contains a path, it will be ignored/removed.
-}
var
  sName: string;
  slPath: TStringList;
  i: Integer;
begin
  Result := True;
  sName := ExtractFileName(AFile);
  // Check current dir first:
  OResult := sName;
  if FileExists(OResult) then
    Exit;
  // Then PATH:
  slPath := TStringList.Create;
  try
    slPath.Delimiter := ';';
    slPath.StrictDelimiter := True;
    slPath.DelimitedText := GetEnvironmentVariable('Path');
    for i := 0 to slPath.Count - 1 do begin
      OResult := AddSlash(slPath[i]) + sName;
      if FileExists(OResult) then
        Exit;
    end;
  finally
    slPath.Free;
  end;
  // Would have exited early if found
  OResult := '';
  Result := False;
end;

{+}
function FixFilePath(const AFile, APath: string): string;
{
  Expands a filename based on the fully qualified path provided.

  Essentially, it's just like Expand[UNC]FileName, but you don't need to worry
  about whether it's UNC or not, and it accepts a specific path rather than
  expanding based on the current directory.

  Examples: FixFilePath('..\foo.pas', 'c:\dev\project\') -> 'c:\dev\foo.pas'
            FixFilePath('foo.pas', '\\ray\shared') -> \\ray\shared\foo.pas'

  A blank filename will result in blank output.

  See also: SameFile
-}
var
  bUNC: Boolean;
  p: string;
begin
  Result := AFile;
  if Result <> '' then begin // Blank is blank.
    // If the filename is already fully qualified, don't mess with the path:
    p := ExtractFileDrive(Result);
    if p = '' then begin
      p := ExtractFileDrive(APath);
      bUNC := Length(p) > 1;
      if Result[1] = PathDelim then begin // Root - prepend drive or unc share name
        if bUNC then
          p := StripSlash(p)
        else
          p := p + DriveDelim;
        Result := p + Result;
      end else // It's relative, e.g. '..\foo.pas'
        Result := AddSlash(APath) + Result;
    end else
      bUNC := Length(p) > 1;
    // Expand any . or .. notations away:
    if bUNC then
      Result := ExpandUNCFileName(Result)
    else
      Result := ExpandFileName(Result);
  end;
end;

{+}
function IsDotDir(const SR: TSearchRec): Boolean;
{
  Returns True if the TSearchRec is the . or .. directory.

  See also: IsSubDir
-}
begin
  Result := ((SR.Attr and faDirectory) <> 0) and
            ( (SR.Name = '.') or (SR.Name = '..') );
end;

{+}
function IsSubDir(const SR: TSearchRec): Boolean;
{
  Returns True if the TSearchRec contains a directory that is neither . nor ..

  See also: IsDotDir
-}
begin
  Result := ((SR.Attr and faDirectory) <> 0) and
            (SR.Name <> '.') and (SR.Name <> '..');
end;

{+}
function IsLegalFilename(const S: string): Boolean;
{
  Returns True if the filename does not contain any of the reserved characters
  or names that are not valid for a Windows filename. The function expects the
  filename ONLY -- do not include the path because slashes are reserved
  characters. This can also be used for a directory name as long as it is passed
  by itself. Blank input returns False.

  The list of reserved characters, names and other rules according to MSDN:

  < (less than)
  > (greater than)
  : (colon)
  " (double quote)
  / (forward slash)
  \ (backslash)
  | (vertical bar or pipe)
  ? (question mark)
  * (asterisk)

  Any character with an ASCII value 0..31.

  Reserved names: CON, PRN, AUX, NUL, COM1..COM9, and LPT1..LPT9

  Do not end a file or directory name with a space or a period.

  2017-05-26: Added reserved name $MFT after reading about a crash bug when used
  as a directory name.
-}
const
  ReservedChars = '<>:"/\|?*';
  ReservedNames = ':CON:PRN:AUX:NUL:COM1:COM2:COM3:COM4:COM5:COM6:COM7:COM8:COM9:LPT1:LPT2:LPT3:LPT4:LPT5:LPT6:LPT7:LPT8:LPT9:$MFT:';
var
  i, j: Integer;
begin
  Result := False;
  j := Length(S);
  if j = 0 then
    Exit;
  if (S[j] = ' ') or (S[j] = '.') then
    Exit;
  for i := 1 to j do
    if IsDelimiter(ReservedChars, S, i) or (Ord(S[i]) < 32) then
      Exit;
  if Pos(':' + UpperCase(S) + ':', ReservedNames) > 0 then
    Exit;
  // All tests passed:
  Result := True;
end;

{+}
function ListFiles(const AList: TStrings;
                   const Path: string;
                   const Attr: Integer = 0;
                   const IncludePath: Boolean = False;
                   const Recurse: Boolean = False;
                   const UseFileInfo: Boolean = False): Integer;
{
  Adds filenames to a TStrings/TStringList list and returns the count of files
  added. Directories are excluded from the list. The TStringList.Object of each
  file is set to the TSearchrec.Time value to allow you to filter or sort on the
  time after retrieving the list.

  If UseFileInfo is set to True, Objects will instead be filled with TFileInfo
  objects. IMPORTANT: The caller is responsible for freeing these objects!

  The <Path> and <Attr> arguments determine what files are added to the list
  and equate to the FindFirst arguments of the same name.

  If <Attr> is negative, only files that match the attribute mask will be
  added to the list. For example, to include both normal and hidden files, use
  faHidden. To list *only* files with the archive attribute, use -faArchive.

  By default the filenames are added to the list without modification. If
  IncludePath is True, the filenames will be fully qualified based on the Path.

  If you want the function to recurse subdirectories, set Recurse to True. All
  subdirectories are followed, not just those matching the wildcard. Recursion
  forces IncludePath to True. Hidden subdirectories will be recursed if the main
  Attr argument includes faHidden.

  See also: StringListIntObjectSortCompare, SortFileInfoByDate/Size,
            TFileInfo, TRecurser

-}
var
  SR: TSearchRec;
  a: Integer;
  bAny: Boolean;
  sPath, sPre, sWild: string;
  slSubdirs: TStringList;
begin
  Result := 0;
  sPath := ExtractFilePath(Path);
  if IncludePath or Recurse then
    sPre := sPath
  else
    sPre := '';
  a := Abs(Attr);
  if FindFirst(Path, a, SR) = 0 then begin
    bAny := Attr >= 0;
    repeat
      if ((SR.Attr and faDirectory) = 0) and
         (bAny or ((SR.Attr and a) <> 0)) then begin
        if UseFileInfo then
          AList.AddObject(sPre + SR.Name, TFileInfo.CreateSR(sPath, SR))
        else
          {$WARN SYMBOL_DEPRECATED OFF}
          AList.AddObject(sPre + SR.Name, TObject(SR.Time));
          {$WARN SYMBOL_DEPRECATED ON}
        Inc(Result);
      end;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
  if Recurse then begin
    slSubdirs := TStringList.Create;
    try
      // Find hidden subdirs if the main Attr argument includes faHidden:
      if ListSubdirs(slSubdirs, sPre, (a and faHidden) <> 0, True) > 0 then begin
        slSubdirs.Sort;
        sWild := ExtractFileName(Path);
        for a := 0 to slSubdirs.Count - 1 do
          Result := Result +
            ListFiles(AList, AddSlash(slSubdirs[a]) + sWild,
                      Attr, IncludePath, Recurse, UseFileInfo);
      end;
    finally
      slSubdirs.Free;
    end;
  end;
end;

{+}
function ListSubdirs(const AList: TStrings;
                     const ADir: string;
                     const IncludeHidden: Boolean = False;
                     const IncludePath: Boolean = False): Integer;
{
  Adds the subdirectories of ADir to a TStrings/TStringList list and returns the
  count of subdirectories added.

  Set IncludeHidden to True if you want hidden directories included in the list.

  Set IncludePath to True if you want the fully qualified pathname added to the
  list. If False (the default), just the subdirectory names will be added.
-}
var
  sPre: string;
  a: Integer;
  SR: TSearchRec;
begin
  Result := 0;
  if IncludePath then
    sPre := AddSlash(ADir)
  else
    sPre := '';
  if IncludeHidden then
    a := faDirectory or faHidden
  else
    a := faDirectory;
  if FindFirst(AddSlash(ADir) + '*.*', a, SR) = 0 then begin
    repeat
      if IsSubDir(SR) then begin
        AList.Add(sPre + SR.Name);
        Inc(Result);
      end;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
end;

{+}
function LoadStrings(const AFile: string; const AList: TStrings): Boolean;
{
  This function is a substitute for TStrings.LoadFromFile when the file may be
  in use by a program that is writing to it, such as a busy log file.

  TStrings.LoadFromFile opens the file with fmShareDenyWrite and raises an
  exception if unable to open the file with those rights. This function uses
  fmShareDenyNone and then calls TStrings.LoadFromStream, returning True if
  successful.

  See also: FSReadLn, FSReadLnGetEOL
-}
var
  fs: TFileStream;
begin
  Result := False;
  try
    fs := TFileStream.Create(AFile, fmOpenRead or fmShareDenyNone);
    try
      AList.LoadFromStream(fs);
      Result := True;
    finally
      fs.Free;
    end;
  except
    // Eat it and return False
  end;
end;

{+}
function PrettySize(const ASize: Int64): string;
{
  Turns a file size into a "pretty" string in one of these formats:
  999 B
  999 KB
  999[.9] MB
  999[.9] GB

  The default size strings are B, KB, MB and GB, all preceded by a space. You
  can change these defaults by assigning new values to the PrettySizeB|KB|MB|GB
  variables. The default values are defined as the constants
  DefaultPrettySizeB|KB|MB|GB.

  The number of decimal places on MB & GB values is determined by the
  PrettySizeDecimals variable, which defaults to 1 (DefaultPrettySizeDecimals).
  Setting it to zero will just round to the nearest whole number.

  By default, a kilobyte is 1024 bytes, a MB is 1024 KB, etc. You can change
  this to 1000 by assigning a new value to PrettySizeKilo. Its default value is
  the constant DefaultPrettySizeKilo.
-}
var
  z: Double;
  sFmt: string;
begin
  if ASize < 1000 then
    Result := IntToStr(ASize) + PrettySizeB
  else begin
    z := ASize / PrettySizeKilo;
    if Ceil(z) < 1000 then
      Result := IntToStr(Ceil(z)) + PrettySizeKB
    else begin
      sFmt := '##0';
      if PrettySizeDecimals > 0 then
        sFmt := sFmt + FormatSettings.DecimalSeparator + StringOfChar('0', PrettySizeDecimals);
      z := z / PrettySizeKilo;
      if RoundTo(z, -PrettySizeDecimals) < 1000.0 then
        Result := FormatFloat(sFmt, z) + PrettySizeMB
      else begin
        z := z / PrettySizeKilo;
        Result := FormatFloat(sFmt, z) + PrettySizeGB;
      end;
    end;
  end;
end;

{+}
function SameFile(const F1, F2: string; const APath: string = ''): Boolean;
{
  Takes two filenames and returns a boolean indicating whether the two strings
  actually refer to the same file. Internally, it uses FixFilePath to do so.
  If you do not supply the APath argument, it defaults to the current directory.

  Example: SameFile('foo.txt', '.\Foo.txt') -> True

  The function will return True if both filenames are blank.

  See also: FixFilePath
-}
var
  sPath, sFile1, sFile2: string;
begin
  sPath := APath;
  if sPath = '' then
    sPath := GetCurrentDir;
  sFile1 := FixFilePath(F1, sPath);
  sFile2 := FixFilePath(F2, sPath);
  Result := SameText(sFile1, sFile2);
end;

{+}
function SecureDelete(const AFile: string; const APasses: Integer = 1): Boolean;
{
  Deletes a file after overwriting it. Returns a boolean indicating success.
  The file will be overwritten APasses times, alternating between zeros and
  ones.
-}
const
  BufSize = 4096;
var
  Buffer0: array[0..BufSize] of Byte;
  Buffer1: array[0..BufSize] of Byte;
  i: Integer;
  bZero: Boolean;
  fs: TFileStream;
  iSize: Int64;
begin
  // It may end up a little longer since we're writing in 4K blocks, but it will
  // also serve to obfuscate the original size.
  FillMemory(@Buffer0, BufSize, 0);
  if APasses > 1 then
    FillMemory(@Buffer1, BufSize, $FF);
  try
    for i := 1 to APasses do begin
      bZero := i mod 2 = 1;
      fs := TFileStream.Create(AFile, fmOpenReadWrite or fmShareExclusive);
      try
        iSize := fs.Size;
        while iSize > 0 do begin
          if bZero then
            fs.Write(Buffer0, BufSize)
          else
            fs.Write(Buffer1, BufSize);
          iSize := iSize - BufSize;
        end;
      finally
        fs.Free;
      end;
    end;
    Result := DeleteFile(AFile);
  except
    Result := False; // Eat it and just return False
  end;
end;

{+}
function UniqueFileName(const AFile: string): string;
{
  Takes a fully qualified filename and if it exists, adds a number to the end
  of it until it no longer exists and returns the new filename. If the file does
  not exist, the filename is returned unchanged.

  Useful for exports and backups where you don't want to overwrite any earlier
  versions of a file. The number has at least two digits (file.txt, file00.txt,
  file01.txt, etc.) to improve name-based sorting should there be more than
  nine files.
-}
var
  sName, sExt: string;
  i: Integer;
begin
  Result := AFile;
  i := 0;
  while FileExists(Result) do begin
    if i = 0 then begin // One time only
      sName := StripExt(AFile);
      sExt := ExtractFileExt(AFile);
    end;
    Result := sName + Format('%.2d', [i]) + sExt;
    Inc(i);
  end;
end;

{+}
function UnusedFileName(const AFile: string): string; deprecated;
{
  Deprecated. Use UniqueFileName instead.
-}
begin
  Result := UniqueFileName(AFile);
end;

{+}
function WildCopy(const ASource, ATarget: string;
                  const Attr: Integer = 0;
                  const ARecurse: Boolean = False;
                  const APreserveAttr: Boolean = False;
                  const APreserveTimes: Boolean = False;
                  const ACopyEmptySubDirs: Boolean = False): Boolean;
{
  A wildcard copy function that returns a boolean indicating success.

  It uses the FileCopy function and its options to copy the files.

  It is always best to use fully qualified path names for both ASource and
  ATarget.

  ATarget can be just a directory where you want the files copied to, or it may
  include a filename wildcard. The target wildcard can be used to rename the
  files as they are copied (e.g. copy *.txt to *.bak). The root target directory
  must already exist, but any recursed subdirectories will be created if needed.

  The Attr argument is used when calling FindFirst, and will recurse into hidden
  subdirectories if it contains faHidden.

  See also: FileCopy, ListSubDirs, WildExists
-}
var
  SR: TSearchRec;
  sSrcPath, sSrcWild, sTgtPath, sTgtWild, sFormat, sName, sExt: string;
  bSameName: Boolean;
  i: Integer;
  slSubs: TStringList;
begin
  Result := False;
  sSrcPath := ExtractFilePath(ASource);
  sSrcWild := ExtractFileName(ASource);
  // Is ATarget just a path, or a path and wildcard?
  if Pos('*', ATarget) > 0 then begin
    sTgtPath := ExtractFilePath(ATarget);
    sTgtWild := ExtractFileName(ATarget);
  end else begin
    sTgtPath := AddSlash(ATarget);
    sTgtWild := '';
  end;
  // Make sure the target path exists
  if not DirectoryExists(sTgtPath) then
    Exit;
  // Prepare our target filename format string
  sFormat := sTgtWild;
  if (sFormat = '*.*') or (sFormat = '*') then
    sFormat := '';
  bSameName := IsBlank(sFormat);
  if not bSameName then begin
    i := Pos('.', sFormat);
    if i = 0 then begin
      // If there is no dot and only one asterisk, replace it with the whole filename
      if CountIn('*', sFormat) = 1 then
        sFormat := StringReplace(sFormat, '*', '%2:s', []);
      i := Length(sFormat) + 1;
    end;
    sName := Copy(sFormat, 1, i - 1);
    sExt := Copy(sFormat, i, MaxInt);
    sName := StringReplace(sName, '*', '%0:s', [rfReplaceAll]);
    sExt := StringReplace(sExt, '*', '%1:s', [rfReplaceAll]);
    sFormat := sName + sExt;
  end;
  // Copy files
  if FindFirst(ASource, Attr, SR) = 0 then begin
    repeat
      if bSameName then
        sName := SR.Name
      else begin
        sName := Format(sFormat, [ExtractFileNameNoExt(SR.Name),
                                  ExtractFileExtNoDot(SR.Name),
                                  SR.Name]);
        // Discard a trailing dot if present:
        i := Length(sName);
        while (i > 0) and (sName[i] = '.') do begin
          Dec(i);
          SetLength(sName, i);
        end;
      end;
      if not FileCopy(sSrcPath + SR.Name, sTgtPath + sName,
                      APreserveAttr, APreserveTimes) then
        Exit;
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
  // Recurse subdirs
  if ARecurse then begin
    slSubs := TStringList.Create;
    try
      if ListSubDirs(slSubs, sSrcPath, (Attr and faHidden) <> 0, False) > 0 then begin
        for i := 0 to slSubs.Count - 1 do begin
          sName := sSrcPath + AddSlash(slSubs[i]) + sSrcWild;
          if ACopyEmptySubDirs or WildExists(sName, Attr, ARecurse) then begin
            sExt := sTgtPath + slSubs[i];
            if not DirectoryExists(sExt) then
              if not CreateDir(sExt) then
                Exit;
            if not WildCopy(sName, AddSlash(sExt) + sTgtWild, Attr, ARecurse,
                            APreserveAttr, APreserveTimes, ACopyEmptySubDirs) then
              Exit;
          end;
        end;
      end;
    finally
      slSubs.Free;
    end;
  end;
  // If we got to the end, we have succeeded!
  Result := True;
end;

function WildDelete(const AFileSpec: string;
                    const Attr: Integer = 0;
                    const ARecurse: Boolean = False): Integer;
{
  Given a wildcard filespec (and optional attribute), it will return the number
  of matching files the function was able to find and delete.
  If ARecurse is True, that includes any matching files in subdirectories.
-}
var
  SR: TSearchRec;
  sPath, sWild: string;
  slSubs: TStringList;
  i: Integer;
begin
  Result := 0;
  sPath := ExtractFilePath(AFileSpec);
  sWild := ExtractFileName(AFileSpec);
  if FindFirst(AFileSpec, Attr, SR) <> 0 then begin
    try
      repeat
        if DeleteFile(sPath + SR.Name) then
          Inc(Result);
      until FindNext(SR) <> 0;
    finally
      FindClose(SR);
    end;
  end;
  if ARecurse then begin
    slSubs := TStringList.Create;
    try
      if ListSubDirs(slSubs, sPath, (Attr and faHidden) <> 0) > 0 then
        for i := 0 to slSubs.Count - 1 do
          Result := Result +
            WildDelete(AddSlash(sPath + slSubs[i]) + sWild, Attr, ARecurse);
    finally
      slSubs.Free;
    end;
  end;
end;

{+}
function WildExists(const AFileSpec: string;
                    const Attr: Integer = 0;
                    const ARecurse: Boolean = False): Boolean;
{
  Given a wildcard filespec (and optional attribute), it will return true if
  any file specified by the wildcard exists, including looking in subdirectories
  if ARecurse is True. Also works as a substitute for FileExists.
-}
var
  SR: TSearchRec;
  sPath, sWild: string;
  slSubs: TStringList;
  i: Integer;
begin
  Result := FindFirst(AFileSpec, Attr, SR) = 0;
  if Result then begin
    FindClose(SR); // The pre-recursion version ended here
    Exit;
  end;
  if ARecurse then begin
    sPath := ExtractFilePath(AFileSpec);
    sWild := ExtractFileName(AFileSpec);
    slSubs := TStringList.Create;
    try
      if ListSubDirs(slSubs, sPath, (Attr and faHidden) <> 0) > 0 then begin
        for i := 0 to slSubs.Count - 1 do begin
          Result := WildExists(AddSlash(sPath + slSubs[i]) + sWild, Attr, ARecurse);
          if Result then
            break;
        end;
      end;
    finally
      slSubs.Free;
    end;
  end;
end;

initialization

  PrettySizeB  := DefaultPrettySizeB;
  PrettySizeKB := DefaultPrettySizeKB;
  PrettySizeMB := DefaultPrettySizeMB;
  PrettySizeGB := DefaultPrettySizeGB;
  PrettySizeKilo := DefaultPrettySizeKilo;
  PrettySizeDecimals := DefaultPrettySizeDecimals;

end.

