{*****************************************************************************}
{                                                                             }
{  llIntegerList.pas                                                          }
{                                                                             }
{  A class for keeping a list of integers, very similar to TStringList.       }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llIntegerList;

interface

uses SysUtils, Classes, Contnrs;

type
  EIntegerListError = class(Exception);

  {+ /trim=2 TIntegerList}
  TIntegerList = class(TObject)
  {-}
  private
    FDelimiter: Char;
    FDescending: Boolean;
    FDuplicates: TDuplicates;
    FList: TObjectList;
    FNumberFormat: string;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    FOwnsObjects: Boolean;
    FQuoteChar: Char;
    FSorted: Boolean;
    FSum64: Int64;
    FUpdateCount: Integer;
    function GetCapacity: Integer;
    function GetCommaText: string;
    function GetCount: Integer;
    function GetObject(Index: Integer): TObject;
    function GetSum: Integer;
    function GetValue(Index: Integer): Integer;
    procedure SetCapacity(const Value: Integer);
    procedure SetCommaText(const Value: string);
    procedure SetDescending(const Value: Boolean);
    procedure SetObject(Index: Integer; const Value: TObject);
    procedure SetSorted(const Value: Boolean);
    procedure SetValue(Index: Integer; const Value: Integer);
    function GetDelimitedText: string;
    procedure SetDelimitedText(const Value: string);
  protected
    function AddItem(const Value: Integer; const AObject: TObject): Integer;
    procedure Changed; virtual;
    procedure Changing; virtual;
    function CompareItems(const Index1, Index2: Integer;
                          const Index2IsValue: Boolean): Integer;
    procedure DeleteItem(const Index: Integer);
    procedure ExchangeItems(const Index1, Index2: Integer);
    procedure InsertItem(const Index, Value: Integer; const AObject: TObject);
    procedure QuickSort(L, R: Integer);
    procedure RangeCheck(const Index: Integer);
    procedure SortCheck;
  {+ /trim=2 /addline=OFF TIntegerList}
  public
    constructor Create; overload;
    constructor Create(const AOwnsObjects: Boolean); overload;
    destructor Destroy; override;
    property Capacity: Integer read GetCapacity write SetCapacity;
    property CommaText: string read GetCommaText write SetCommaText;
    property Count: Integer read GetCount;
    property DelimitedText: string read GetDelimitedText write SetDelimitedText;
    property Delimiter: Char read FDelimiter write FDelimiter;
    property Descending: Boolean read FDescending write SetDescending;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property NumberFormat: string read FNumberFormat write FNumberFormat;
    property Objects[Index: Integer]: TObject read GetObject write SetObject;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
    property OwnsObjects: Boolean read FOwnsObjects write FOwnsObjects;
    property QuoteChar: Char read FQuoteChar write FQuoteChar;
    property Sorted: Boolean read FSorted write SetSorted;
    property Sum: Integer read GetSum;
    property Sum64: Int64 read FSum64;
    property Values[Index: Integer]: Integer read GetValue write SetValue; default;
    function Add(const Value: Integer): Integer;
    function AddObject(const Value: Integer; const AObject: TObject): Integer;
    procedure BeginUpdate;
    procedure Clear;
    procedure Clone(const Source: TIntegerList);
    procedure Delete(const Index: Integer);
    procedure EliminateDupes;
    procedure EndUpdate;
    procedure Exchange(const Index1, Index2: Integer);
    function Find(const Value: Integer; var Index: Integer): Boolean;
    function IndexOf(const Value: Integer): Integer;
    function IndexOfEx(const Value, Offset: Integer): Integer;
    function IndexOfObject(const AObject: TObject): Integer;
    procedure Insert(const Index, Value: Integer);
    procedure InsertObject(const Index, Value: Integer; const AObject: TObject);
    procedure Sort;
  end;
  {+ /trim=OFF /addline=OFF

  TIntegerList works just like TStringList, but carries Integers instead of
  strings.  There are only a few significant differences:

  The Sum and Sum64 properties are the sum of all the numbers in the list as an
  Integer and Int64, respectively.  If the actual sum overflows the range of an
  Integer, Sum will be set to MaxInt and Sum64 will contain the correct value.
  Should the sum exceed the range of an Int64, the class will fall over and die.

  The Descending property can be used to reverse the sorting order. Custom
  sorting is not supported.

  The OwnsObjects property allows the list's Objects to be freed automatically
  when deleted, just like TObjectList.OwnsObjects. When False, object lifetime
  is the responsibility of the programmer, as it is when using TStringList.

  The CommaText property allows you to get or set the list as a simple string of
  comma-delimited integers, such as '3,21,19,11'.  Assigning a string that
  contains non-integer values will result in an EConvertError exception.

  The DelimitedText property works in a similar, but more flexible manner. When
  output, the NumberFormat property determines the format of each value written
  to the string. Set it to any valid numeric format string, e.g. '%.3d' or '%m',
  including non-decimal format strings. If the format string does not contain
  one of the decimal-format letters (d,u,x) it will be considered a float-type
  format. Quoting of the output is done only where required.

  On input, DelimitedText safely handles things like quoted values and non-
  numeric characters like thousands separators. Any fractional digits in the
  input are truncated. If the NumberFormat property contains the letter "x", the
  values will be interpreted as hex. Scientific notation is not supported as an
  input format.
  -}

implementation

uses llStrings;

{ TIntegerList }

type
  TIntListItem = class(TObject)
  public
    FValue: Integer;
    FObject: TObject;
  end;

{ Constructors/Destructors }

constructor TIntegerList.Create;
begin
  inherited;
  FDuplicates := dupAccept;
  FList := TObjectList.Create(True); // Automatic Free of our TIntListItems
  FDelimiter := ',';
  FQuoteChar := '"';
end;

constructor TIntegerList.Create(const AOwnsObjects: Boolean);
begin
  inherited Create;
  FOwnsObjects := AOwnsObjects;
end;

destructor TIntegerList.Destroy;
begin
  if FOwnsObjects then
    Clear;
  FList.Free;
  inherited;
end;

{ Private }

function TIntegerList.GetCapacity: Integer;
begin
  Result := FList.Capacity;
end;

function TIntegerList.GetCommaText: string;
var
  i: Integer;
begin
  if Count > 0 then begin
    Result := IntToStr(TIntListItem(FList[0]).FValue);
    for i := 1 to Count - 1 do
      Result := Result + ',' + IntToStr(TIntListItem(FList[i]).FValue);
  end else
    Result := '';
end;

function TIntegerList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TIntegerList.GetDelimitedText: string;
var
  i: Integer;
  bPlain, bDecimal: Boolean;
  sDelims: string;
  function ItemStr(const Index: Integer): string;
  var
    j: Integer;
  begin
    if bPlain then
      Result := IntToStr(TIntListItem(FList[Index]).FValue)
    else begin
      if bDecimal then
        Result := Format(FNumberFormat, [TIntListItem(FList[Index]).FValue])
      else // Must be a float-type (%n, %m, etc.)
        Result := Format(FNumberFormat, [TIntListItem(FList[Index]).FValue * 1.0]);
      for j := 1 to Length(Result) do
        if IsDelimiter(sDelims, Result, j) then begin
          Result := Enquote(Result, FQuoteChar);
          break;
        end;
    end;
  end;
begin
  if Count > 0 then begin
    bPlain := FNumberFormat = '';
    bDecimal := False;
    for i := 1 to Length(FNumberFormat) do
      if IsDelimiter('dux', FNumberFormat, i) then
        bDecimal := True;
    sDelims := FDelimiter + ' ' + FQuoteChar;
    Result := ItemStr(0);
    for i := 1 to Count - 1 do
      Result := Result + FDelimiter + ItemStr(i);
  end else
    Result := '';
end;

function TIntegerList.GetObject(Index: Integer): TObject;
begin
  RangeCheck(Index);
  Result := TIntListItem(FList[Index]).FObject;
end;

function TIntegerList.GetSum: Integer;
begin
  if (FSum64 < MaxInt) and (FSum64 > Low(Integer)) then
    Result := Integer(FSum64)
  else
    Result := MaxInt; // MaxInt shall serve as our overflow indicator.
end;

function TIntegerList.GetValue(Index: Integer): Integer;
begin
  RangeCheck(Index);
  Result := TIntListItem(FList[Index]).FValue;
end;

procedure TIntegerList.SetCapacity(const Value: Integer);
var
  bShrinking: Boolean;
begin
  bShrinking := Value < Capacity;
  if bShrinking then
    BeginUpdate;
  try
    if FOwnsObjects then // Free any objects we're about to lose by shrinking
      while FList.Count > Value do
        DeleteItem(FList.Count - 1);
    FList.Capacity := Value;
  finally
    if bShrinking then
      EndUpdate;
  end;
end;

procedure TIntegerList.SetCommaText(const Value: string);
var
  i, j: Integer;
begin
  BeginUpdate;
  try
    Clear;
    if Value <> '' then begin
      j := 1;
      for i := 1 to Length(Value) do begin
        if Value[i] = ',' then begin
          Add(StrToInt(Copy(Value, j, i - j)));
          j := i + 1;
        end;
      end;
      Add(StrToInt(Copy(Value, j, Length(Value) + 1 - j)));
    end;
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.SetDelimitedText(const Value: string);
var
  iLen, i, j: Integer;
  sNum: string;
  bHex: Boolean;
  procedure AddNum;
  var
    x: Integer;
  begin
    if bHex then
      Add(StrToInt('$' + sNum))
    else begin
      x := Pos(FormatSettings.DecimalSeparator, sNum); // Truncate any fractional part
      if x > 0 then
        System.Delete(sNum, x, MaxInt);
      for x := Length(sNum) downto 1 do // Remove any non-numeric parts ("$,)
        if not IsDelimiter('0123456789-', sNum, x) then
          System.Delete(sNum, x, 1);
      Add(StrToInt(sNum));
    end;
  end;
begin
  BeginUpdate;
  try
    Clear;
    iLen := Length(Value);
    if iLen > 0 then begin
      bHex := Pos('x', FNumberFormat) > 0;
      i := 1;
      j := 1;
      while i <= ilen do begin
        if Value[i] = FDelimiter then begin
          sNum := Copy(Value, j, i - j);
          AddNum;
          j := i + 1;
        end else if Value[i] = FQuoteChar then
          i := CloseQuotePos(Value, i); // llStrings
        Inc(i);
      end;
      sNum := Copy(Value, j, iLen + 1 - j);
      AddNum;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.SetDescending(const Value: Boolean);
begin
  if FDescending <> Value then begin
    FDescending := Value;
    if Sorted then begin
      Sorted := False;
      Sorted := True; // Re-sort
    end;
  end;
end;

procedure TIntegerList.SetObject(Index: Integer; const Value: TObject);
begin
  RangeCheck(Index);
  TIntListItem(FList[Index]).FObject := Value;
end;

procedure TIntegerList.SetSorted(const Value: Boolean);
begin
  if FSorted <> Value then begin
    if Value then
      Sort;
    FSorted := Value;
  end;
end;

procedure TIntegerList.SetValue(Index: Integer; const Value: Integer);
begin
  SortCheck;
  Dec(FSum64, GetValue(Index)); // GetValue does our RangeCheck for us!
  Inc(FSum64, Value);
  TIntListItem(FList[Index]).FValue := Value;
end;

{ Protected }

function TIntegerList.AddItem(const Value: Integer;
                              const AObject: TObject): Integer;
begin
  if not Sorted then
    Result := Count
  else
    if Find(Value, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: raise EIntegerListError.Create('Duplicates not allowed');
      end;
  InsertItem(Result, Value, AObject);
end;

procedure TIntegerList.Changed;
begin
  if (FUpdateCount = 0) and Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TIntegerList.Changing;
begin
  if (FUpdateCount = 0) and Assigned(FOnChanging) then
    FOnChanging(Self);
end;

function TIntegerList.CompareItems(const Index1, Index2: Integer;
                                   const Index2IsValue: Boolean): Integer;
var
  v1, v2: Integer;
begin
  v1 := TIntListItem(FList[Index1]).FValue;
  if Index2IsValue then
    v2 := Index2
  else
    v2 := TIntListItem(FList[Index2]).FValue;
  if v1 < v2 then
    Result := -1
  else if v1 > v2 then
    Result := 1
  else // equal
    Result := 0;
  if FDescending then
    Result := -Result;
end;

procedure TIntegerList.DeleteItem(const Index: Integer);
begin
  BeginUpdate;
  try
    Dec(FSum64, TIntListItem(FList[Index]).FValue); // Delete() calls RangeCheck
    if FOwnsObjects then
      TIntListItem(FList[Index]).FObject.Free;
    FList.Delete(Index);
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.ExchangeItems(const Index1, Index2: Integer);
begin
  BeginUpdate;
  try
    FList.Exchange(Index1, Index2);
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.InsertItem(const Index, Value: Integer;
                                  const AObject: TObject);
var
  Item: TIntListItem;
begin
  BeginUpdate;
  try
    Item := TIntListItem.Create;
    Item.FValue := Value;
    Item.FObject := AObject;
    FList.Insert(Index, Item);
    Inc(FSum64, Value);
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.QuickSort(L, R: Integer);
var
  I, J, P: Integer;
begin
  BeginUpdate;
  try
    // Cribbed from TStringList.QuickSort
    repeat
      I := L;
      J := R;
      P := (L + R) shr 1;
      repeat
        while CompareItems(I, P, False) < 0 do Inc(I); // while SCompare(Self, I, P) < 0 do Inc(I);
        while CompareItems(J, P, False) > 0 do Dec(J); // while SCompare(Self, J, P) > 0 do Dec(J);
        if I <= J then
        begin
          ExchangeItems(I, J);
          if P = I then
            P := J
          else if P = J then
            P := I;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then QuickSort(L, J);
      L := I;
    until I >= R;
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.RangeCheck(const Index: Integer);
begin
  if (Index < 0) or (Index >= Count) then
    raise EIntegerListError.Create('Index out of range: ' + IntToStr(Index));
end;

procedure TIntegerList.SortCheck;
begin
  if FSorted then
    raise EIntegerListError.Create('Operation not allowed on sorted list');
end;

{ Public }

function TIntegerList.Add(const Value: Integer): Integer;
begin
  Result := AddItem(Value, nil);
end;

function TIntegerList.AddObject(const Value: Integer;
                                const AObject: TObject): Integer;
begin
  Result := AddItem(Value, AObject);
end;

procedure TIntegerList.BeginUpdate;
begin
  if FUpdateCount = 0 then
    Changing;
  Inc(FUpdateCount);
end;

procedure TIntegerList.Clear;
var
  i: Integer;
begin
  BeginUpdate;
  try
    if FOwnsObjects then
      for i := FList.Count - 1 downto 0 do
        TIntListItem(FList[i]).FObject.Free;
    FList.Clear;
    FSum64 := 0;
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.Clone(const Source: TIntegerList);
var
  i: Integer;
begin
  BeginUpdate;
  try
    Clear;
    { Set Sorted to False before copying stuff over. If the Source IS sorted, we
      are getting it in the right order anyway and turning Sorted on afterwards
      will be essentially cost-free. Adding things to an unsorted list is much
      easier since we don't have to worry about *where* to stick each item. }
    Sorted := False;
    Duplicates := dupAccept; // In case Source.Duplicates was changed AFTER adding stuff
    Capacity := Source.Capacity;
    OwnsObjects := Source.OwnsObjects;
    Descending := Source.Descending;
    NumberFormat := Source.NumberFormat;
    for i := 0 to Source.Count - 1 do
      InsertItem(i, TIntListItem(Source.FList[i]).FValue,
                    TIntListItem(Source.FList[i]).FObject);
    Sorted := Source.Sorted;
    Duplicates := Source.Duplicates;
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.Delete(const Index: Integer);
begin
  RangeCheck(Index);
  DeleteItem(Index);
end;

procedure TIntegerList.EndUpdate;
begin
  Dec(FUpdateCount);
  if FUpdateCount = 0 then
    Changed;
end;

procedure TIntegerList.EliminateDupes;
var
  i: Integer;
begin
  BeginUpdate;
  try
    Sort;
    for i := Count - 1 downto 1 do
      if CompareItems(i, i - 1, False) = 0 then
        DeleteItem(i);
  finally
    EndUpdate;
  end;
end;

procedure TIntegerList.Exchange(const Index1, Index2: Integer);
begin
  SortCheck;
  RangeCheck(Index1);
  RangeCheck(Index2);
  ExchangeItems(Index1, Index2);
end;

function TIntegerList.Find(const Value: Integer; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := CompareItems(I, Value, True); // CompareStrings(FList^[I].FString, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if Duplicates <> dupAccept then L := I;
      end;
    end;
  end;
  Index := L;
end;

function TIntegerList.IndexOf(const Value: Integer): Integer;
begin
  if Sorted then begin
    if not Find(Value, Result) then
      Result := -1;
  end else begin
    for Result := 0 to Count - 1 do
      if TIntListItem(FList[Result]).FValue = Value then
        Exit;
    Result := -1;
  end;
end;

function TIntegerList.IndexOfEx(const Value, Offset: Integer): Integer;
begin
  for Result := Offset to Count - 1 do
    if TIntListItem(FList[Result]).FValue = Value then
      Exit;
  Result := -1;
end;

function TIntegerList.IndexOfObject(const AObject: TObject): Integer;
begin
  for Result := 0 to Count - 1 do
    if TIntListItem(FList[Result]).FObject = AObject then
      Exit;
  Result := -1;
end;

procedure TIntegerList.Insert(const Index, Value: Integer);
begin
  SortCheck;
  if Index <> Count then
    RangeCheck(Index); // We must allow inserting items at the end
  InsertItem(Index, Value, nil);
end;

procedure TIntegerList.InsertObject(const Index, Value: Integer;
                                    const AObject: TObject);
begin
  SortCheck;
  if Index <> Count then
    RangeCheck(Index); // We must allow inserting items at the end
  InsertItem(Index, Value, AObject);
end;

procedure TIntegerList.Sort;
begin
  if not Sorted and (Count > 1) then
    QuickSort(0, Count - 1);
end;

end.

