{*****************************************************************************}
{                                                                             }
{  llLike.pas                                                                 }
{                                                                             }
{  A class to do database-style LIKE pattern matching of strings.             }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llLike;

interface

uses
  SysUtils, Classes, StrUtils;

type
  ELikePatternError = class(Exception);

  {+ /trim=2 TLikePattern}
  TLikePattern = class(TObject)
  {-}
  private
    FOneChar: Char;
    FAnyChar: Char;
    FEscChar: Char;
    FCaseSen: Boolean;
    FPattern: string;
    FAnchorStart: Boolean;
    FAnchorEnd: Boolean;
    FParts: TStringList;
    FBlankError: Boolean;
    procedure SetPattern(const Value: string);
    procedure SetAnyChar(const Value: Char);
    procedure SetEscChar(const Value: Char);
    procedure SetOneChar(const Value: Char);
    procedure SetCaseSen(const Value: Boolean);
    function ValidWildCard(const Value: Char): Boolean;
  {+ /trim=2 /addline=OFF TLikePattern}
  public
    constructor Create;
    destructor Destroy; override;                          // defaults:
    property AnyChar: Char read FAnyChar write SetAnyChar; // %
    property OneChar: Char read FOneChar write SetOneChar; // _
    property EscChar: Char read FEscChar write SetEscChar; // \
    property CaseSensitive: Boolean read FCaseSen write SetCaseSen; // True
    property Pattern: string read FPattern write SetPattern;
    property ErrorOnBlankPattern: Boolean read FBlankError write FBlankError;
    function Matches(const S: string): Boolean;
    procedure Reset; // Resets to default state, blank pattern
    function TryPattern(const APattern: string; out OError: string): Boolean;
  end;
  {+ /trim=OFF /addline=OFF

  TLikePattern is used to do database-style LIKE pattern matching of strings.

  Assign a LIKE pattern using the Pattern property, then check one or more
  strings to see if they match the pattern using the Matches method. The method
  returns True when the string matches the pattern.

  The LIKE pattern follows the same rules as described in the PostgreSQL
  database docs, by default using the underscore for a one-character match
  (OneChar), a percent sign for a zero-or-more-character match (AnyChar) and
  the backslash to escape literal underscores and percent signs (EscChar).

  Setting an invalid pattern (e.g. conflicting wildcards, no text, etc.) will
  raise an ELikePatternError exception. The TryPattern function allows you to
  attempt setting the pattern from untrusted input to see if it is valid without
  having to set up your own try/except. If the function returns False, the error
  message will be in the OError argument.

  Changing the wildcards or escape character will clear the current pattern, so
  set the pattern last after changing the other settings. Spaces and letters are
  not allowed as wildcard/escape characters. Attempting to set an invalid
  wildcard character will silently leave the character unchanged.

  The CaseSensitive property is True by default. This setting can be changed
  without losing the current pattern, but it does cause it to be re-parsed.

  Calling Matches against a blank/unassigned pattern will raise an exception by
  default. The ErrorOnBlankPattern property controls this. If the property is
  set to False, calling Matches against a blank or failed pattern will just
  return False instead of raising an exception.

  Usage example:

  LP := TLikePattern.Create;
  try
    LP.Pattern := '%ray%';
    if LP.Matches(sSomeString) then ...
  finally
    LP.Free;
  end;
  -}

implementation

uses
  llStrings;

const
  OrMore = 1000000000; // One beelion!

{ TLikePattern }

constructor TLikePattern.Create;
begin
  inherited;
  FParts := TStringList.Create;
  Reset;
end;

destructor TLikePattern.Destroy;
begin
  FParts.Free;
  inherited;
end;

function TLikePattern.ValidWildCard(const Value: Char): Boolean;
begin
  Result := (Ord(Value) > Ord(' ')) and (Pos(Value, ALPHA_BOTH) = 0);
end;

procedure TLikePattern.SetAnyChar(const Value: Char);
begin
  if (Value <> FAnyChar) and ValidWildCard(Value) then begin
    FAnyChar := Value;
    SetPattern('');
  end;
end;

procedure TLikePattern.SetOneChar(const Value: Char);
begin
  if (Value <> FOneChar) and ValidWildCard(Value) then begin
    FOneChar := Value;
    SetPattern('');
  end;
end;

procedure TLikePattern.SetEscChar(const Value: Char);
begin
  if (Value <> FEscChar) and ValidWildCard(Value) then begin
    FEscChar := Value;
    SetPattern('');
  end;
end;

procedure TLikePattern.SetCaseSen(const Value: Boolean);
var
  i: Integer;
begin
  if Value <> FCaseSen then begin
    FCaseSen := Value;
	  // If going from True to False, we can just uppercase the existing parts.
	  // Otherwise, it needs to be re-parsed.
    if FParts.Count > 0 then begin
      if FCaseSen then
		    SetPattern(FPattern)
	    else
        for i := 0 to FParts.Count - 1 do
          FParts[i] := AnsiUpperCase(FParts[i]);
    end;
  end;
end;

procedure TLikePattern.Reset;
begin
  FAnyChar := '%'; // Other common values: *
  FOneChar := '_'; // Other common values: ?
  FEscChar := '\';
  FCaseSen := True;
  FBlankError := True;
  SetPattern('');
end;

procedure TLikePattern.SetPattern(const Value: string);
var
  sPtn, sPart: string;
  i, p, m: Integer;
  c: Char;
  procedure BadWilds(const APos: Integer);
  var
    iPos: Integer;
  begin
    iPos := APos;
    if not FAnchorStart then
      Inc(iPos); // Adjust for the deleted AnyChar
    FParts.Clear;
    raise ELikePatternError.Create(
      'Invalid adjacent wildcards at position ' + IntToStr(iPos));
  end;
  procedure NoText;
  begin
    FParts.Clear;
    raise ELikePatternError.Create(
      'Pattern must contain at least one non-wildcard character.');
  end;
begin
  FParts.Clear;
  FPattern := Value;
  if FPattern = '' then
    Exit;
  // Make sure our wildcard/escape characters are all different:
  if (AnyChar = OneChar) or (AnyChar = EscChar) or (OneChar = EscChar) then
    raise ELikePatternError.CreateFmt(
      'Conflicting wildcard/escape characters: %s%s%s',
      [AnyChar, OneChar, EscChar]);
  // If case insensitive, parse the pattern as upper case.
  if CaseSensitive then
    sPtn := FPattern
  else
    sPtn := AnsiUpperCase(FPattern);
  // Check for start anchoring:
  FAnchorStart := sPtn[1] <> AnyChar;
  if not FAnchorStart then begin
    if (Length(sPtn) > 1) and (sPtn[2] = AnyChar) then
      BadWilds(1); // Double AnyChars
    Delete(sPtn, 1, 1);
  end else if sPtn[1] = OneChar then begin
    // Starting with OneChar[+OneChar...]+AnyChar falsely anchors to the start.
    p := Length(sPtn);
    m := 2;
    while (m <= p) and (sPtn[m] = OneChar) do
      Inc(m);
    if (m <= p) and (sPtn[m] = AnyChar) then
      BadWilds(m);
  end;
  // Check for end anchoring:
  p := Length(sPtn); // Watch out for an escaped trailing AnyChar
  if p = 0 then
    NoText;
  FAnchorEnd := (sPtn[p] <> AnyChar) or ((p > 1) and (sPtn[p - 1] = EscChar));
  if not FAnchorEnd then begin
    Dec(p);
    if (p > 0) and (sPtn[p] = AnyChar) then // Adjacent AnyChars?
      if (p = 1) or (sPtn[p - 1] <> EscChar) then // An escaped literal is OK
        BadWilds(p);
    Delete(sPtn, Length(sPtn), 1);
  end else if sPtn[p] = OneChar then begin
    // Ending with AnyChar+OneChar[+OneChar...] falsely anchors to the end
    Dec(p);
    while (p >= 1) and (sPtn[p] = OneChar) do
      Dec(p);
    if (p > 0) and (sPtn[p] = AnyChar) then begin
      Dec(p);
      if (p > 0) and (sPtn[p] <> EscChar) then
        BadWilds(p);
    end;
  end;
  { Go through the pattern, check for escaped wildcards and build our list of
    parts for matching. The FParts stringlist contains an entry for each bit of
    text between the wildcards.

    Those separated by an AnyChar match have an Object value of zero, meaning
    that any number of characters can separate the parts, including zero.

    Those separated by a OneChar match have an Object value > 0, meaning that
    exactly that number of characters must separate the parts.

    When a OneChar and AnyChar appear together, it means that *at least* that
    specific number of characters must separate the parts, but could be more.
    In those cases, the Object value will be the number of OneChar wildcards
    plus the OrMore constant, which is a very large number.
  }
  sPart := '';
  i := 1;
  p := Length(sPtn);
  m := 0; // eliminate warning
  while i <= p do begin
    c := sPtn[i];
    if c = EscChar then begin
      if i < p then begin
        Inc(i);
        sPart := sPart + sPtn[i]; // A literal EscChar must be doubled!
        sPtn[i] := 'X'; // Change it to make looking backwards for doubled AnyChars simpler
      end else begin
        FParts.Clear;
        raise ELikePatternError.Create('Trailing Escape character');
      end;
    end else if (c = AnyChar) or (c = OneChar) then begin
      // If this is a OneChar, how many are in a row?
      if c = OneChar then begin
        m := 1;
        while (i < p) and (sPtn[i + 1] = OneChar) do begin
          Inc(m);
          Inc(i);
        end;
      end;
      // If this part is blank, we have leading/adjacent wildcards.
      // Two AnyChars in a row are not valid, but other combos are.
      if sPart = '' then begin
        if c = AnyChar then begin
          // Double AnyChars?
          if (i = 1) or (sPtn[i - 1] = AnyChar) then
            BadWilds(i);
          // Otherwise, add the OrMore constant to the previous part's modifier
          m := Integer(FParts.Objects[FParts.Count - 1]);
          if m < OrMore then
            FParts.Objects[FParts.Count - 1] := TObject(m + OrMore);
          Inc(i);
          Continue;
        end else if i > 1 then begin
          // If this isn't a leading OneChar, make the previous part's modifier
          // m + OrMore.
          m := m + Integer(FParts.Objects[FParts.Count - 1]);
          if m < OrMore then
            m := m + OrMore;
          FParts.Objects[FParts.Count - 1] := TObject(m);
          Inc(i);
          Continue;
        end;
      end;
      FParts.Add(sPart);
      sPart := '';
      if c = OneChar then
        FParts.Objects[FParts.Count - 1] := TObject(m);
    end else
      sPart := sPart + c;
    Inc(i);
  end;
  if sPart <> '' then
    FParts.Add(sPart);
  if (FParts.Count = 0) or ((FParts.Count = 1) and (FParts[0] = '')) then
    NoText;
end;

function TLikePattern.TryPattern(const APattern: string;
                                 out OError: string): Boolean;
begin
  Result := False;
  try
    SetPattern(APattern);
    Result := True;
    OError := '';
  except
    on e:ELikePatternError do
      OError := e.Message;
    else
      raise;
  end;
end;

function TLikePattern.Matches(const S: string): Boolean;
var
  sText: string;
  i, j, k, p, m, iFirstPart, iLastPart, iEndPos: Integer;
  bOrMore: Boolean;
begin
  Result := False;
  // A blank/invalid pattern always fails a match
  if FParts.Count = 0 then begin
    if FBlankError then
      raise ELikePatternError.Create('Pattern not defined');
    Exit;
  end;
  if CaseSensitive then
    sText := S
  else
    sText := AnsiUpperCase(S);
  // First, determine where we start looking. If the first part is blank, it
  // means we have one or more OneChar matches at the start.
  if FParts[0] = '' then begin
    m := Integer(FParts.Objects[0]);
    if m > OrMore then
      m := m - OrMore;
    i := 1 + m;
    iFirstPart := 1;
  end else begin
    i := 1; // First character
    iFirstPart := 0; // Part zero
  end;
  iLastPart := FParts.Count - 1;
  // If anchored, make sure those parts of the pattern appear where expected:
  if FAnchorStart and
     (Copy(sText, i, Length(FParts[iFirstPart])) <> FParts[iFirstPart]) then
    Exit;
  if FAnchorEnd then begin
    m := Integer(FParts.Objects[iLastPart]); // End with any OneChar matches?
    j := (Length(sText) - m) + 1;
    k := Length(FParts[iLastPart]);
    j := j - k;
    iEndPos := j;
    if Copy(sText, j, k) <> FParts[iLastPart] then
      Exit;
    // Is this a simple match of the whole string?
    if FAnchorStart and FAnchorEnd and (FParts.Count = 1) and
       (Copy(sText, 1, Length(sText) - m) <> FParts[0]) then
      Exit;
  end else
    iEndPos := 0; // Eliminate warning
  // Find the first instance of the first part of the pattern:
  i := PosEx(FParts[iFirstPart], sText, i);
  while i > 0 do begin
    // See if we have matches for the rest of the pattern parts:
    Result := True;
    j := i;
    m := Integer(FParts.Objects[iFirstPart]);
    bOrMore := m > OrMore;
    if bOrMore then
      m := m - OrMore;
    for p := (iFirstPart + 1) to iLastPart do begin
      k := j + Length(FParts[p - 1]) + m;
      j := PosEx(FParts[p], sText, k);
      // If m > 0, there is a OneChar match between parts and the next part must
      // be found exactly where expected (j must equal k).
      if (m > 0) and (j <> k) and not bOrMore then
        j := 0;
      if j = 0 then begin
        Result := False;
        break;
      end;
      m := Integer(FParts.Objects[p]);
      bOrMore := m > OrMore;
      if bOrMore then
        m := m - OrMore;
    end;
    {
    If anchored to the end and the last part is separated by a OneChar match,
    the last part must have been found exactly where expected. If the last part
    is separated by an AnyChar match, we already know it has been found at the
    end and can assume a "greedy" match.
    }
    if Result and FAnchorEnd and (j <> iEndPos) then begin
      if iLastPart > 0 then
        m := Integer(FParts.Objects[iLastPart - 1])
      else
        m := 0;
      if (m > 0) and (m < OrMore) then
        Result := False;
    end;
    // Did we match everything? Success! If anchored to start, no more tries.
    if Result or FAnchorStart then
      break;
    // Next attempt:
    Inc(i);
    i := PosEx(FParts[iFirstPart], sText, i);
  end;
end;

end.

