{*****************************************************************************}
{                                                                             }
{  llLogicIni.pas                                                             }
{                                                                             }
{  A comment-preserving ini file class based on TMemIniFile.                  }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llLogicIni;

interface

uses
  SysUtils, Classes, llBool, llDates, llStrings;

const
  IniCommentChar = ';';
  IniBlankValue = '""';

type
  {+ /trim=2 TLogicIni}
  TLogicIni = class(TObject)
  {-}
  private
    FData: TStringList;
    FFileName: string;
    FWantComments: Boolean; // Default=True
    FDestroying: Boolean;
    FDirty: Boolean;
    FOldDates: Boolean;
    FPreserveWS: Boolean; // Preserves whitespace before/after values
    FEncoding: TEncoding;
    function AddSection(const Section: string): TStringList;
    function GetCaseSensitive: Boolean;
    function GetSection(const Section: string): TStringList;
    function GetSectionCount: Integer;
    function GetSectionName(Index: Integer): string;
    function GetSorted: Boolean;
    function GetValue(const Section, Ident: string; out Value: string): Boolean;
    function RemoveComments(const AStrings: TStrings): Boolean;
    procedure SetCaseSensitive(const Value: Boolean);
    procedure SetSorted(const Value: Boolean);
    procedure SetValue(const Section, Ident, Value: string);
    procedure SetWantComments(const Value: Boolean);
  {+ /trim=2 /addline=OFF TLogicIni}
  public
    constructor Create; overload;
    constructor Create(const AFileName: string;
                       const AWantComments: Boolean = True;
                       const APreserveWhitespace: Boolean = False); overload;
    destructor Destroy; override;
    property CaseSensitive: Boolean read GetCaseSensitive write SetCaseSensitive;
    property Dirty: Boolean read FDirty;
    property FileName: string read FFileName write FFileName;
    property OldDateTimeFormat: Boolean read FOldDates write FOldDates;
    property PreserveWhitespace: Boolean read FPreserveWS write FPreserveWS;
    property SectionCount: Integer read GetSectionCount;
    property SectionName[Index: Integer]: string read GetSectionName;
    property Sorted: Boolean read GetSorted write SetSorted;
    property WantComments: Boolean read FWantComments write SetWantComments;
    procedure Clear;
    procedure ClearSection(const Section: string); // Leaves empty section
    function CopySection(const Source, Target: string): Boolean;
    procedure DeleteKey(const Section, Ident: string);
    procedure DeleteValue(const Section, Ident: string);
    procedure EraseSection(const Section: string); // Removes completely
    procedure GetStrings(Strings: TStrings; const AWantComments: Boolean);
    procedure LoadFromFile(const AFileName: string);
    function ReadBool(const Section, Ident: string; Default: Boolean): Boolean;
    function ReadDate(const Section, Ident: string; Default: TDateTime): TDateTime;
    function ReadDateTime(const Section, Ident: string; Default: TDateTime): TDateTime;
    function ReadFloat(const Section, Ident: string; Default: Double): Double;
    function ReadInteger(const Section, Ident: string; Default: Longint): Longint;
    function ReadInt64(const Section, Ident: string; Default: Int64): Int64;
    procedure ReadSections(Strings: TStrings);
    procedure ReadSectionValues(const Section: string; Strings: TStrings;
                                const AWantComments: Boolean = True);
    function ReadString(const Section, Ident, Default: string): string;
    function ReadTime(const Section, Ident: string; Default: TDateTime): TDateTime;
    function RenameSection(const AOld, ANew: string): Boolean;
    procedure SaveToFile(const AFileName: string);
    function SectionExists(const Section: string): Boolean;
    procedure SetStrings(Strings: TStrings);
    procedure SortSections;
    procedure SortValues;
    procedure UpdateFile;
    function ValueExists(const Section, Ident: string): Boolean;
    procedure WriteBool(const Section, Ident: string; Value: Boolean);
    procedure WriteDate(const Section, Ident: string; Value: TDateTime);
    procedure WriteDateTime(const Section, Ident: string; Value: TDateTime;
                            const P: TDateTimePrecision = dtpSecond);
    procedure WriteFloat(const Section, Ident: string; Value: Double);
    procedure WriteInteger(const Section, Ident: string; Value: Longint);
    procedure WriteInt64(const Section, Ident: string; Value: Int64);
    procedure WriteSectionValues(const Section: string; Strings: TStrings);
    procedure WriteString(const Section, Ident, Value: string);
    procedure WriteTime(const Section, Ident: string; Value: TDateTime;
                        const P: TDateTimePrecision = dtpSecond);
  end;
  {+ /trim=OFF /addline=OFF

  TLogicIni is my own take on TMemIniFile that preserves comments (if desired)
  and stores date and time values in a non-regional manner using the DtoS/TtoS
  and related functions from the llDates unit. It doesn't use the fancy hashing
  stringlist that TMemIniFile does, but slow ini lookups have never been a
  bottleneck for me.

  You can still read and write dates and times in the classic format by setting
  the OldDateTimeFormat property to True. The class will always *read* a date or
  time in the old format (by detecting the presence of separators), but it won't
  write them unless you tell it to. Changing this setting WILL NOT change any
  existing values - they will remain as-is until touched by a WriteDate/Time
  call. The precision arguments are ignored when using the old formats.

  It also includes (always) a "no-name" section that can hold comments or name-
  value pairs at the beginning of the file before any section headers. You can
  read/write values in this section using a blank section name ('').

  When reading a value, the default is only returned if the value is not present
  in the ini file at all. If present and set to blank, the blank (or data type
  equivalent) is returned. This allows you to distinguish between a missing
  value and one deliberately set to blank.

  Because TStringLists remove a name when the value is set to blank, values that
  are set to blank are stored as empty double quotes (""), defined in the handy
  constant IniBlankValue. When you read and write individual values, blanks are
  automatically translated from/to this value. To remove a value completely, use
  the DeleteValue method. When you read a whole section, blanks will still be
  listed with the empty quotes.

  Setting the FileName property DOES NOT automatically load a new file! It just
  changes the filename used by UpdateFile. To load an ini file from disk after
  the constructor has been called you must use the LoadFromFile method. If you
  want the name remembered for later UpdateFile calls, set FileName as well.

  Hint: If you set Sorted to True, it is best not to refer to sections or values
  by index while you're changing stuff!
  -}

implementation

{ TLogicIni }

constructor TLogicIni.Create;
begin
  inherited;
  FDestroying := False;
  FWantComments := True;
  FPreserveWS := False;
  FOldDates := False;
  FData := TStringList.Create;
  FData.CaseSensitive := False;
  FData.Sorted := False;
  AddSection(''); // The no-name section is always present and first
  FDirty := False;
end;

constructor TLogicIni.Create(const AFileName: string;
                             const AWantComments: Boolean;
                             const APreserveWhitespace: Boolean);
begin
  Create;
  WantComments := AWantComments;
  PreserveWhitespace := APreserveWhitespace;
  FFileName := AFileName;
  LoadFromFile(AFileName);
end;

destructor TLogicIni.Destroy;
begin
  if Assigned(FData) then begin
    FDestroying := True;
    Clear;
    FData.Free;
  end;
  inherited;
end;

{----- Private/Protected methods ----------------------------------------------}

function TLogicIni.AddSection(const Section: string): TStringList;
begin
  // If the section already exists, it is returned.
  Result := GetSection(Section);
  if not Assigned(Result) then begin
    Result := TStringList.Create;
    Result.CaseSensitive := FData.CaseSensitive;
    Result.Sorted := FData.Sorted;
    FData.AddObject(Section, Result);
    FDirty := True;
  end;
end;

function TLogicIni.GetCaseSensitive: Boolean;
begin
  Result := FData.CaseSensitive;
end;

function TLogicIni.GetSection(const Section: string): TStringList;
var
  i: Integer;
begin
  i := FData.IndexOf(Section);
  if i >= 0 then
    Result := TStringList(FData.Objects[i])
  else
    Result := nil;
end;

function TLogicIni.GetSectionCount: Integer;
begin
  Result := FData.Count;
end;

function TLogicIni.GetSectionName(Index: Integer): string;
begin
  Result := FData[Index];
end;

function TLogicIni.GetSorted: Boolean;
begin
  Result := FData.Sorted;
end;

function TLogicIni.GetValue(const Section, Ident: string; out Value: string): Boolean;
var
  Sec: TStringList;
  i: Integer;
begin
  // Returns True if the value exists.
  Result := False;
  Sec := GetSection(Section);
  if Assigned(Sec) then begin
    i := Sec.IndexOfName(Ident);
    if i >= 0 then begin
      Result := True;
      Value := Sec.ValueFromIndex[i];
      if Value = IniBlankValue then
        Value := '';
    end;
  end;
end;

function TLogicIni.RemoveComments(const AStrings: TStrings): Boolean;
var
  i: Integer;
begin
  Result := False; // Returns True if any were removed so you can set FDirty
  for i := AStrings.Count - 1 downto 0 do
    if IsBlank(AStrings[i]) or (AStrings[i][1] = IniCommentChar) then begin
      AStrings.Delete(i);
      Result := True;
    end;
end;

procedure TLogicIni.SetCaseSensitive(const Value: Boolean);
var
  i: Integer;
begin
  if FData.CaseSensitive <> Value then begin
    FData.CaseSensitive := Value;
    for i := 0 to FData.Count - 1 do
      TStringList(FData.Objects[i]).CaseSensitive := Value;
  end;
end;

procedure TLogicIni.SetSorted(const Value: Boolean);
var
  i: Integer;
begin
  if FData.Sorted <> Value then begin
    FData.Sorted := Value;
    for i := 0 to FData.Count - 1 do
      TStringList(FData.Objects[i]).Sorted := Value;
    FDirty := True;
  end;
end;

procedure TLogicIni.SetValue(const Section, Ident, Value: string);
var
  Sec: TStringList;
begin
  Assert(Ident <> '', 'Blank ini value names are not supported.');
  {
    Setting a TStringList.Values[] property to blank removes the entry
    completely. We don't want that - it makes it impossible to distinguish
    between a value deliberately set to blank and one that is not present.
    When reading, we would always get the default value instead of the blank!
  }
  { Used to do it this way:
  if Value = '' then
    Sec.Add(Ident + '=')
  }
  Sec := AddSection(Section); // Returns existing one if already present
  if Value = '' then
    Sec.Values[Ident] := IniBlankValue
  else
    Sec.Values[Ident] := Value;
  FDirty := True;
end;

procedure TLogicIni.SetWantComments(const Value: Boolean);
var
  s: Integer;
  Sec: TStringList;
begin
  if FWantComments <> Value then begin
    FWantComments := Value;
    if not FWantComments then begin
      for s := 0 to FData.Count - 1 do begin
        Sec := TStringList(FData.Objects[s]);
        if RemoveComments(Sec) then
          FDirty := True;
      end;
    end;
  end;
end;

{----- Public/Published methods -----------------------------------------------}

procedure TLogicIni.Clear;
var
  i: Integer;
begin
  for i := 0 to FData.Count - 1 do
    FData.Objects[i].Free;
  FData.Clear;
  if not FDestroying then
    AddSection(''); // The no-name section is always present and first
  FDirty := False;
  FEncoding := nil;
end;

procedure TLogicIni.ClearSection(const Section: string);
var
  Sec: TStringList;
begin
  Sec := GetSection(Section);
  if Assigned(Sec) then begin
    FDirty := FDirty or (Sec.Count > 0);
    Sec.Clear;
  end;
end;

function TLogicIni.CopySection(const Source, Target: string): Boolean;
var
  Src, Tgt: TStringList;
begin
  // Copies an entire section to a new name, overwriting the target section if
  // it exists.
  Result := False;
  if Source <> Target then begin
    Src := GetSection(Source);
    if Assigned(Src) then begin
      Tgt := AddSection(Target); // Returns existing section if present
      Tgt.Assign(Src);
      Result := True;
      FDirty := True;
    end;
  end;
end;

procedure TLogicIni.DeleteKey(const Section, Ident: string);
begin
  DeleteValue(Section, Ident); // This is the proc name in TMemIniFile
end;

procedure TLogicIni.DeleteValue(const Section, Ident: String);
var
  Sec: TStringList;
begin
  Sec := GetSection(Section);
  if Assigned(Sec) then begin
    Sec.Values[Ident] := '';
    FDirty := True;
  end;
end;

procedure TLogicIni.EraseSection(const Section: string);
var
  i: Integer;
begin
  if Section = '' then
    TStringList(FData.Objects[0]).Clear // The no-name section can't be removed, only emptied
  else begin
    i := FData.IndexOf(Section);
    if i > 0 then begin
      FData.Objects[i].Free;
      FData.Delete(i);
    end;
  end;
  FDirty := True;
end;

procedure TLogicIni.GetStrings(Strings: TStrings; const AWantComments: Boolean);
var
  s, i, iMax: Integer;
  Section: TStringList;
begin
  Strings.Clear;
  iMax := FData.Count - 1;
  for s := 0 to iMax do begin
    // The no-name section is special - no section header is output & no
    // trailing blank line is added if it's empty.
    if s > 0 then
      Strings.Add('[' + FData[s] + ']');
    Section := TStringList(FData.Objects[s]);
    for i := 0 to Section.Count - 1 do
      if AWantComments or (Section[i] = '') or (Section[i][1] <> IniCommentChar) then
        Strings.Add(Section[i]);
    if (s < iMax) and ((s > 0) or (Section.Count > 0)) then
      Strings.Add('');
  end;
end;

procedure TLogicIni.LoadFromFile(const AFileName: string);
var
  slTemp: TStringList;
begin
  if FileExists(AFileName) then begin
    slTemp := TStringList.Create;
    try
      slTemp.LoadFromFile(AFileName);
      SetStrings(slTemp);
    finally
      slTemp.Free;
    end;
  end else
    Clear;
  FDirty := False;
end;

function TLogicIni.ReadBool(const Section, Ident: string; Default: Boolean): Boolean;
var
  sValue: string;
begin
  if GetValue(Section, Ident, sValue) then
    Result := StoB(sValue, Default)
  else
    Result := Default;
end;

function TLogicIni.ReadDate(const Section, Ident: string; Default: TDateTime): TDateTime;
begin
  Result := ReadDateTime(Section, Ident, Default);
end;

function TLogicIni.ReadDateTime(const Section, Ident: string; Default: TDateTime): TDateTime;
var
  sValue: string;
begin
  if GetValue(Section, Ident, sValue) then begin
    if Pos(FormatSettings.DateSeparator, sValue) = 0 then
      Result := StoDDef(sValue, Default)
    else
      Result := StrToDateTimeDef(sValue, Default); // Old format still supported
  end else
    Result := Default;
end;

function TLogicIni.ReadFloat(const Section, Ident: string; Default: Double): Double;
var
  sValue: string;
begin
  if GetValue(Section, Ident, sValue) then
    Result := StrToFloatDef(sValue, Default)
  else
    Result := Default;
end;

function TLogicIni.ReadInt64(const Section, Ident: string; Default: Int64): Int64;
var
  sValue: string;
begin
  if GetValue(Section, Ident, sValue) then
    Result := StrToInt64Def(sValue, Default)
  else
    Result := Default;
end;

function TLogicIni.ReadInteger(const Section, Ident: string; Default: Integer): Longint;
var
  sValue: string;
begin
  if GetValue(Section, Ident, sValue) then
    Result := StrToIntDef(sValue, Default)
  else
    Result := Default;
end;

procedure TLogicIni.ReadSections(Strings: TStrings);
var
  i: Integer;
begin
  Strings.Clear;
  for i := 1 to FData.Count - 1 do // Skip the no-name section (0)
    Strings.Add(FData[i]);
end;

procedure TLogicIni.ReadSectionValues(const Section: string; Strings: TStrings;
                                      const AWantComments: Boolean);
var
  i: Integer;
begin
  i := FData.IndexOf(Section);
  if i >= 0 then begin
    Strings.Assign(TStrings(FData.Objects[i]));
    // Do we need to strip comments out of these?
    if FWantComments and not AWantComments then
      RemoveComments(Strings);
  end else
    Strings.Clear;
end;

function TLogicIni.ReadString(const Section, Ident, Default: string): string;
begin
  if not GetValue(Section, Ident, Result) then
    Result := Default;
end;

function TLogicIni.ReadTime(const Section, Ident: string; Default: TDateTime): TDateTime;
var
  sValue: string;
begin
  if GetValue(Section, Ident, sValue) then begin
    if Pos(FormatSettings.TimeSeparator, sValue) = 0 then
      Result := StoTDef(sValue, Default)
    else
      Result := StrToTimeDef(sValue, Default); // Old format still supported
  end else
    Result := Default;
end;

function TLogicIni.RenameSection(const AOld, ANew: string): Boolean;
var
  i: Integer;
begin
  Result := False;
  if (AOld <> ANew) and not SectionExists(ANew) then begin
    i := FData.IndexOf(AOld);
    // You can't rename the no-name section...
    if i > 0 then begin
      FData[i] := ANew;
      Result := True;
      FDirty := True;
    // ...but you can copy it to the new name & clear it!
    end else if CopySection(AOld, ANew) then begin
      ClearSection(AOld);
      Result := True;
      FDirty := True;
    end;
  end;
end;

procedure TLogicIni.SaveToFile(const AFileName: string);
var
  slTemp: TStringList;
begin
  slTemp := TStringList.Create;
  try
    GetStrings(slTemp, FWantComments);
    slTemp.SaveToFile(AFileName, FEncoding);
  finally
    slTemp.Free;
  end;
  FDirty := False;
end;

function TLogicIni.SectionExists(const Section: string): Boolean;
begin
  Result := FData.IndexOf(Section) >= 0;
end;

procedure TLogicIni.SetStrings(Strings: TStrings);
var
  i, j: Integer;
  Sec: TStringList;
  sLine, sValue: string;
begin
  // Read in the strings, assigning each section to its own stringlist:
  Clear;
  Sec := GetSection(''); // The no-name section for loose comments/values
  for i := 0 to Strings.Count - 1 do begin
    sLine := Strings[i];
    if IsBlank(sLine) then
      Continue;
    if not FPreserveWS then
      sLine := Trim(sLine);
    j := Length(sLine);
    if (sLine[1] = '[') and (sLine[j] = ']') then begin // New Section?
      Delete(sLine, j, 1);
      Delete(sLine, 1, 1);
      Sec := AddSection(Trim(sLine));
    end else begin
      // Clean up the data by removing spaces around the equal signs:
      if sLine[1] <> IniCommentChar then begin
        if not FPreserveWS then begin
          j := Pos('=', sLine);
          if j > 0 then begin
            sValue := Trim(Copy(sLine, j + 1, MaxInt));
            if sValue = '' then
              sValue := IniBlankValue;
            sLine := Trim(Copy(sLine, 1, j - 1)) + '=' + sValue;
          end;
        end;
      end else if not FWantComments then
        Continue;
      Sec.Add(sLine);
    end;
  end;
  FEncoding := Strings.Encoding;
  FDirty := True;
end;

procedure TLogicIni.SortSections;
begin
  FData.Sort;
  FDirty := True;
end;

procedure TLogicIni.SortValues;
var
  s: Integer;
begin
  for s := 0 to FData.Count - 1 do
    TStringList(FData.Objects[s]).Sort;
  FDirty := True;
end;

procedure TLogicIni.UpdateFile;
begin
  if FFileName <> '' then
    SaveToFile(FFileName);
end;

function TLogicIni.ValueExists(const Section, Ident: string): Boolean;
var
  sValue: string;
begin
  Result := GetValue(Section, Ident, sValue);
end;

procedure TLogicIni.WriteBool(const Section, Ident: string; Value: Boolean);
begin
  SetValue(Section, Ident, BtoS(Value));
end;

procedure TLogicIni.WriteDate(const Section, Ident: string; Value: TDateTime);
begin
  if FOldDates then
    SetValue(Section, Ident, DateToStr(Value))
  else
    SetValue(Section, Ident, DtoS(Value, dtpDay));
end;

procedure TLogicIni.WriteDateTime(const Section, Ident: string; Value: TDateTime;
                                  const P: TDateTimePrecision);
begin
  if FOldDates then
    SetValue(Section, Ident, DateTimeToStr(Value))
  else
    SetValue(Section, Ident, DtoS(Value, P));
end;

procedure TLogicIni.WriteFloat(const Section, Ident: string; Value: Double);
begin
  SetValue(Section, Ident, FloatToStr(Value));
end;

procedure TLogicIni.WriteInt64(const Section, Ident: string; Value: Int64);
begin
  SetValue(Section, Ident, IntToStr(Value));
end;

procedure TLogicIni.WriteInteger(const Section, Ident: string; Value: Integer);
begin
  SetValue(Section, Ident, IntToStr(Value));
end;

procedure TLogicIni.WriteSectionValues(const Section: string; Strings: TStrings);
var
  Sec: TStringList;
begin
  Sec := GetSection(Section);
  if not Assigned(Sec) then
    Sec := AddSection(Section);
  Sec.Assign(Strings);
end;

procedure TLogicIni.WriteString(const Section, Ident, Value: string);
begin
  SetValue(Section, Ident, Value);
end;

procedure TLogicIni.WriteTime(const Section, Ident: string; Value: TDateTime;
                              const P: TDateTimePrecision);
begin
  if FOldDates then
    SetValue(Section, Ident, TimeToStr(Value))
  else
    SetValue(Section, Ident, TtoS(Value, P));
end;

end.
