{*****************************************************************************}
{                                                                             }
{  llMRU.pas                                                                  }
{                                                                             }
{  TMRU implements a Most-Recently-Used list as a submenu.                    }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llMRU;

interface

uses
  Windows, SysUtils, Classes, Menus, Registry;

type
  {$IFOPT M-}
    {$DEFINE NO_TYPEINFO}
    {$M+}
  {$ELSE}
    {$UNDEF NO_TYPEINFO}
  {$ENDIF}
  {+ /trim=2 TMRU}
  TMRU = class(TObject)
  {-}
  private
    FMaxItems: Integer;
    FMenu: TMenuItem;
    FOnChange: TNotifyEvent;
    FOnClick: TNotifyEvent;
    FRegKey: string;
    FAutoSave: Boolean;
    function GetItemByIndex(Index: Integer): string;
    function GetCount: Integer;
    procedure SetMaxItems(const Value: Integer);
    procedure DoOnChange;
    procedure DoOnClick(Sender: TObject);
    procedure SetMenu(const Value: TMenuItem);
    procedure SetRegKey(const Value: string);
  {+ /trim=2 /addline=OFF TMRU}
  public
    constructor Create(AParentMenu: TMenuItem; const ARegKey: string);
    destructor Destroy; override;
    property Items[Index: Integer]: string read GetItemByIndex; default;
    property Count: Integer read GetCount;
    procedure Clear;
    procedure Erase;
    procedure Load;
    procedure Save;
    procedure Update(const S: string);
    procedure Delete(const Index: Integer); overload;
    procedure Delete(const S: string); overload;
    procedure RemoveDeletedFiles;
  published
    property ParentMenu: TMenuItem read FMenu write SetMenu;
    property RegistryKey: string read FRegKey write SetRegKey;
    property MaxItems: Integer read FMaxItems write SetMaxItems;
    property AutoSave: Boolean read FAutoSave write FAutoSave;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
  end;
  {+ /trim=OFF /addline=OFF

  TMRU implements a Most-Recently-Used list as a submenu.

  Start by creating your main menu and adding a menu item that will contain the
  MRU items as a submenu. Caption this with text like "Recent Files..." or
  something. Create an instance of TMRU and attach it to this menu item using
  the first Create argument. You can also use a TPopupMenu.Items property in
  place of a regular menu item. Changing the ParentMenu property after creation
  will re-load the MRU from the current registry key.

  It also needs a registry key where it can store its data (HKCU). By default,
  the MRU stores 10 items. Set MaxItems to change this (the value is recalled
  when loading a previously saved MRU). Setting the RegistryKey property after
  creation will re-load the MRU from the new location.

  Set the OnClick event to be notified when the user clicks on one of the items
  in the MRU. The Sender will be the menuitem that was clicked, and the Caption
  of that menu item contains the selected value. It is not necessary (but
  harmless) to call Update when responding to this event - the MRU already did
  so prior to firing it.

  By default, AutoSave is False, and you must call the Save method to save the
  state of the MRU (usually in OnCloseQuery) to the registry. If set to True,
  the MRU will automatically save its data to the registry whenever it is
  updated. The Erase method deletes any previously stored registry values and
  sets AutoSave to False. If you want to be notified when the MRU list changes,
  assign the OnChange event.

  In your program, whenever you open a file (the MRU can be used to keep track
  of anything, not just files), just call the Update method with the text of the
  item you just selected/opened. The item will be added to the MRU or moved to
  the top if already present. The items in the MRU are maintained in a case-
  insensitive manner.

  If you need to manually remove items from the list (for example, files that no
  longer exist), use one of the overloaded Delete methods. You can also call the
  RemoveDeletedFiles procedure to easily check for and get rid of deleted files.
  -}
  {$IFDEF NO_TYPEINFO}
    {$M-}
  {$ENDIF}
  
  procedure CleanMRUList(const AList: TStrings);

  procedure UpdateMRUList(const AList: TStrings; const AValue: string;
                          const AMax: Integer = 0;
                          const ACaseSensitive: Boolean = False);

implementation

{ Stand-alone functions }

{+}
procedure CleanMRUList(const AList: TStrings);
{
  Removes items from an MRU list of filenames where the file no longer exists.
-}
var
  i: Integer;
begin
  for i := AList.Count - 1 downto 0 do begin
    try
      if (AList[i] = '') or not FileExists(AList[i]) then
        AList.Delete(i);
    except
      // Eat it - might be a disconnected drive or something
    end;
  end;
end;

{+}
procedure UpdateMRUList(const AList: TStrings; const AValue: string;
                        const AMax: Integer = 0;
                        const ACaseSensitive: Boolean = False);
{
  Updates a list of strings like a Most-Recently-Used list by inserting or
  moving the given value to the top of the list, and removing the oldest (last)
  entry if the list is full (AMax > 0).
-}
var
  i: Integer;
begin
  // Is this value already in the list?
  if ACaseSensitive then begin
    i := AList.Count - 1;
    while i >= 0 do begin
      if AList[i] = AValue then
        break;
      Dec(i);
    end;
  end else begin
    i := AList.IndexOf(AValue);
    if i = 0 then // Update it to the current value - case may have changed
      AList[i] := AValue;
  end;
  if i <> 0 then begin // If already first, we're done.
    // Delete the existing value or the oldest value(s) if full:
    if i > 0 then
      AList.Delete(i)
    else if AMax > 0 then
      while AList.Count >= AMax do
        AList.Delete(AList.Count - 1);
    // Insert the value at the top of the list:
    AList.Insert(0, AValue);
  end;
end;

{ TMRU }

const
  BlankItemTag = 999;

constructor TMRU.Create(AParentMenu: TMenuItem; const ARegKey: string);
begin
  inherited Create;
  FMaxItems := 10;
  FAutoSave := False;
  ParentMenu := AParentMenu;
  RegistryKey := ARegKey; // Loads the MRU
end;

destructor TMRU.Destroy;
begin
  if Assigned(FMenu) then
    FMenu.Clear;
  inherited;
end;

procedure TMRU.Clear;
var
  BlankItem: TMenuItem;
begin
  FMenu.Clear;
  BlankItem := TMenuItem.Create(FMenu);
  BlankItem.AutoHotkeys := maManual;
  BlankItem.Caption := '(no items)';
  BlankItem.Tag := BlankItemTag;
  BlankItem.Enabled := False;
  FMenu.Add(BlankItem);
  DoOnChange;
end;

procedure TMRU.Delete(const Index: Integer);
begin
  FMenu.Delete(Index);
  if FMenu.Count > 0 then
    DoOnChange
  else
    Clear;
end;

procedure TMRU.Delete(const S: string);
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if SameText(Items[i], S) then
      Delete(i);
end;

procedure TMRU.DoOnChange;
begin
  if AutoSave then
    Save;
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TMRU.DoOnClick(Sender: TObject);
begin
  if TMenuItem(Sender).Tag = BlankItemTag then
    Exit;
  Update(TMenuItem(Sender).Caption);
  if Assigned(FOnClick) then
    FOnClick(Sender);
end;

procedure TMRU.Erase;
var
  Reg: TRegistry;
  i, iCount: Integer;
begin
  // Removes all trace of the MRU list from the registry & turns off AutoSave.
  AutoSave := False; // It is assumed you don't want it saved again
  if FRegKey = '' then
    Exit;
  Reg := TRegistry.Create(KEY_ALL_ACCESS);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(FRegKey, False) then begin
      if Reg.ValueExists('MRUCount') then begin
        iCount := Reg.ReadInteger('MRUCount');
        Reg.DeleteValue('MRUCount');
        if Reg.ValueExists('MRUMax') then
          Reg.DeleteValue('MRUMax');
        for i := 0 to iCount - 1 do
          if Reg.ValueExists('MRU' + IntToStr(i)) then
            Reg.DeleteValue('MRU' + IntToStr(i));
      end;
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

function TMRU.GetCount: Integer;
begin
  Result := FMenu.Count;
  if (Result = 1) and (FMenu.Items[0].Tag = BlankItemTag) then
    Result := 0;
end;

function TMRU.GetItemByIndex(Index: Integer): string;
begin
  Result := FMenu.Items[Index].Caption;
end;

procedure TMRU.Load;
var
  SavedEvent: TNotifyEvent;
  Reg: TRegistry;
  i, iCount: Integer;
  sName: string;
  bSaveSave: Boolean;
begin
  if (FRegKey = '') or not Assigned(FMenu) then
    Exit;
  SavedEvent := FOnChange;
  FOnChange := nil; // Do just a single OnChange after loading
  bSaveSave := AutoSave;
  AutoSave := False; // Do not AutoSave while loading!
  Reg := TRegistry.Create(KEY_READ);
  try
    Clear;
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKeyReadOnly(FRegKey) then begin
      if Reg.ValueExists('MRUCount') then begin
        iCount := Reg.ReadInteger('MRUCount');
        FMaxItems := Reg.ReadInteger('MRUMax');
        for i := iCount - 1 downto 0 do begin
          sName := 'MRU' + IntToStr(i);
          if Reg.ValueExists(sName) then
            Update(Reg.ReadString(sName));
        end;
      end;
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
    FOnChange := SavedEvent;
    DoOnChange;
    AutoSave := bSaveSave;
  end;
end;

procedure TMRU.Save;
var
  Reg: TRegistry;
  i, iOldCount: Integer;
begin
  if Assigned(FMenu) and (FRegKey <> '') then begin
    Reg := TRegistry.Create(KEY_ALL_ACCESS);
    try
      Reg.RootKey := HKEY_CURRENT_USER;
      if Reg.OpenKey(FRegKey, True) then begin
        if Reg.ValueExists('MRUCount') then
          iOldCount := Reg.ReadInteger('MRUCount') - 1
        else
          iOldCount := -1;
        Reg.WriteInteger('MRUCount', FMenu.Count);
        Reg.WriteInteger('MRUMax', MaxItems);
        for i := 0 to FMenu.Count - 1 do
          Reg.WriteString('MRU' + IntToStr(i), FMenu.Items[i].Caption);
        for i := FMenu.Count to iOldCount do
          if Reg.ValueExists('MRU' + IntToStr(i)) then
            Reg.DeleteValue('MRU' + IntToStr(i));
        Reg.CloseKey;
      end;
    finally
      Reg.Free;
    end;
  end;
end;

procedure TMRU.SetMaxItems(const Value: Integer);
begin
  if (Value >= 0) and (Value <> FMaxItems) then begin
    FMaxItems := Value;
    if FMaxItems = 0 then
      Clear
    else if FMenu.Count > FMaxItems then begin
      while FMenu.Count > FMaxItems do
        FMenu.Delete(FMenu.Count - 1);
      DoOnChange;
    end;
  end;
end;

procedure TMRU.RemoveDeletedFiles;
var
  i: Integer;
  bChanged, bSaveSave: Boolean;
begin
  bChanged := False;
  bSaveSave := AutoSave;
  AutoSave := False; // Save only once when finished.
  try
    for i := Count - 1 downto 0 do begin
      try
        if not FileExists(FMenu.Items[i].Caption) then begin
          Delete(i);
          bChanged := True;
        end;
      except
        // Bury it - might be a disconnected drive or something
      end;
    end;
  finally
    AutoSave := bSaveSave;
    if bChanged then
      DoOnChange;
  end;
end;

procedure TMRU.SetMenu(const Value: TMenuItem);
begin
  FMenu := Value;
  FMenu.AutoHotkeys := maManual;
  if FRegKey <> '' then
    Load
  else
    Clear;
end;

procedure TMRU.SetRegKey(const Value: string);
begin
  FRegKey := Value;
  Load;
end;

procedure TMRU.Update(const S: string);
var
  i: Integer;
  NewItem: TMenuItem;
begin
  if FMaxItems = 0 then
    Exit;
  i := FMenu.Count - 1;
  while i >= 0 do begin
    if SameText(S, FMenu.Items[i].Caption) then
      break
    else if FMenu.Items[i].Tag = BlankItemTag then
      FMenu.Delete(i);
    Dec(i);
  end;
  if i = 0 then
    Exit; // It's already the first item in the list!
  if (i < 0) and (FMenu.Count >= FMaxItems) then
    i := FMenu.Count - 1;
  if i >= 0 then
    FMenu.Delete(i);
  NewItem := TMenuItem.Create(FMenu);
  NewItem.AutoHotkeys := maManual;
  NewItem.Caption := S;
  NewItem.OnClick := DoOnClick;
  FMenu.Insert(0, NewItem);
  DoOnChange;
end;

end.
