{*****************************************************************************}
{                                                                             }
{  llMsgDlg.pas                                                               }
{                                                                             }
{  Convenient wrappers for MessageBox-type functions.                         }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llMsgDlg;

interface

uses
  Dialogs, Controls, Forms, Windows;

{+}
function CenteredMsgDlg(const AMsg, ACaption: string;
                        DlgType: TMsgDlgType;
                        Buttons: TMsgDlgButtons;
                        AForm: TForm = nil;
                        const HelpCtx: Integer = 0;
                        const HelpFileName: string = ''): Integer;
{
  An adaptation of Dialogs.MessageDlgPosHelp that accepts a caption, centers the
  dialog over the calling form, and plays a sound like Application.MessageBox.
  If the caption is blank, the default MessageDlg caption is used. If the form
  is not provided, the dialog appears in its default position of screen center.
-}

{+}
procedure MsgAlert(const Text: string; const Caption: string = '';
                   AForm: TForm = nil);
{
  Shows a CenteredMsgDlg with a warning icon and an OK button.
-}

{+}
procedure MsgError(const Text: string; const Caption: string = '';
                   AForm: TForm = nil);
{
  Shows a CenteredMsgDlg with an error icon and an OK button.
-}

{+}
procedure MsgInfo(const Text: string; const Caption: string = '';
                  AForm: TForm = nil);
{
  Shows a CenteredMsgDlg with an information icon and an OK button.
-}

{+}
function MsgOKCancel(const Text: string; const Caption: string = '';
                     AForm: TForm = nil): Boolean;
{
  Returns a True/False answer to an OK/Cancel CenteredMsgDlg.
-}

{+}
function MsgRetryCancel(const Text: string; const Caption: string = '';
                        AForm: TForm = nil): Boolean;
{
  Returns a True/False answer to a Retry/Cancel CenteredMsgDlg.
-}

{+}
function MsgYesNo(const Text: string; const Caption: string = '';
                  AForm: TForm = nil): Boolean;
{
  Returns a True/False answer to a Yes/No CenteredMsgDlg.
-}

{+}
function MsgYesNoCancel(const Text: string; const Caption: string = '';
                        AForm: TForm = nil): Integer;
{
  Returns mrYes, mrNo or mrCancel in response to a CenteredMsgDlg.
-}

implementation

function CenteredMsgDlg(const AMsg, ACaption: string;
                        DlgType: TMsgDlgType;
                        Buttons: TMsgDlgButtons;
                        AForm: TForm = nil;
                        const HelpCtx: Integer = 0;
                        const HelpFileName: string = ''): Integer;
var
  x, y: Integer;
  pTop: TPoint;
begin
  // Play a sound like Application.MessageBox:
  case DlgType of
    mtWarning:
      x := MB_ICONEXCLAMATION;
    mtError:
      x := MB_ICONHAND;
    mtInformation:
      x := MB_ICONASTERISK;
    mtConfirmation:
      x := MB_ICONQUESTION;
    else
      x := MB_OK;
  end;
  MessageBeep(x);
  // Adapted from Dialogs.MessageDlgPosHelp
  with CreateMessageDialog(AMsg, DlgType, Buttons) do
    try
      if ACaption <> '' then
        Caption := ACaption;
      HelpContext := HelpCtx;
      HelpFile := HelpFileName;
      if Assigned(AForm) then begin
        // If the top left shows up as 0,0 this form is most likely parented
        // inside another. Convert to screen coordinates.
        pTop.X := AForm.Left;
        pTop.Y := AForm.Top;
        if (pTop.X = 0) and (pTop.Y = 0) then
          pTop := AForm.ClientToScreen(pTop);
        x := pTop.X + (AForm.Width div 2);
        y := pTop.Y + (AForm.Height div 2);
        x := x - (Width div 2);
        y := y - (Height div 2);
        if x < 0 then
          x := 0;
        if y < 0 then
          y := 0;
        Left := x;
        Top := y;
        MakeFullyVisible; // Fixes partially offscreen or split between monitors
      end else
        Position := poScreenCenter;
      Result := ShowModal;
    finally
      Free;
    end;
end;

procedure MsgAlert(const Text, Caption: string; AForm: TForm);
begin
  CenteredMsgDlg(Text, Caption, mtWarning, [mbOK], AForm);
end;

procedure MsgError(const Text, Caption: string; AForm: TForm);
begin
  CenteredMsgDlg(Text, Caption, mtError, [mbOK], AForm);
end;

procedure MsgInfo(const Text, Caption: string; AForm: TForm);
begin
  CenteredMsgDlg(Text, Caption, mtInformation, [mbOK], AForm);
end;

function MsgOKCancel(const Text, Caption: string; AForm: TForm): Boolean;
begin
  Result :=
    CenteredMsgDlg(Text, Caption, mtConfirmation, mbOKCancel, AForm) = mrOk;
end;

function MsgRetryCancel(const Text, Caption: string; AForm: TForm): Boolean;
begin
  Result :=
    CenteredMsgDlg(Text, Caption, mtConfirmation,
                   [mbRetry, mbCancel], AForm) = mrRetry;
end;

function MsgYesNo(const Text, Caption: string; AForm: TForm): Boolean;
begin
  Result :=
    CenteredMsgDlg(Text, Caption, mtConfirmation, [mbYes, mbNo], AForm) = mrYes;
end;

function MsgYesNoCancel(const Text, Caption: string; AForm: TForm): Integer;
begin
  Result := CenteredMsgDlg(Text, Caption, mtConfirmation, mbYesNoCancel, AForm);
end;

end.
