{*****************************************************************************}
{                                                                             }
{  llNetwork.pas                                                              }
{                                                                             }
{  Networking utility routines                                                }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llNetwork;

interface

uses
  Windows, Classes, SysUtils, WinSock, Winapi.IpHlpApi, Winapi.IpTypes;

const
  MaxIPv4 = High(Cardinal);
  { Constants for GetExtendedTCPTable function calls }
  IpHlpApiDLL = 'iphlpapi.dll';
  MIB_TCP_STATE: array[0..12] of string =
    ('UNKNOWN', 'CLOSED', 'LISTENING', 'SYN_SENT ','SYN_RECEIVED',
     'ESTABLISHED', 'FIN_WAIT_1', 'FIN_WAIT_2', 'CLOSE_WAIT', 'CLOSING',
     'LAST_ACK', 'TIME_WAIT', 'DELETE_TCB');
  ANY_SIZE = 1;

  AF_UNSPEC = 0; // GetAdapterAddresses Address Family constants
  AF_INET = 2;
  AF_INET6 = 23;

  IF_TYPE_OTHER = 1; // PIP_ADAPTER_ADDRESSES.IfType constants
  IF_TYPE_ETHERNET = 6;
  IF_TYPE_PPP = 23;
  IF_TYPE_SOFTWARE_LOOPBACK = 24;
  IF_TYPE_ATM = 37;
  IF_TYPE_IEEE80211 = 71; // Wireless
  IF_TYPE_TUNNEL = 131;
  IF_TYPE_IEEE1394 = 144; // Firewire

type
  { Types for GetExtendedTCPTable function calls }
  TCP_TABLE_CLASS = Integer;

  TGetExtendedTcpTable =
    function (pTcpTable: Pointer; dwSize: PDWORD; bOrder: BOOL; lAf: ULONG;
              TableClass: TCP_TABLE_CLASS; Reserved: ULONG): DWORD; stdcall;

  TTCPTableOption = (ttoListener, ttoConnections, ttoAll); // GetTCPTableData AOption

  TTCPTableClass = (TCP_TABLE_BASIC_LISTENER,
                    TCP_TABLE_BASIC_CONNECTIONS,
                    TCP_TABLE_BASIC_ALL,
                    TCP_TABLE_OWNER_PID_LISTENER,
                    TCP_TABLE_OWNER_PID_CONNECTIONS,
                    TCP_TABLE_OWNER_PID_ALL,
                    TCP_TABLE_OWNER_MODULE_LISTENER,
                    TCP_TABLE_OWNER_MODULE_CONNECTIONS,
                    TCP_TABLE_OWNER_MODULE_ALL); // Ord() == TableClass

  TTCPState = (tcpUnknown, tcpClosed, tcpListening, tcpSynSent, tcpSynReceived,
               tcpEstablished, tcpFinWait1, tcpFinWait2, tcpCloseWait, tcpClosing,
               tcpLastAck, tcpTimeWait, tcpDeleteTCB); // Ord() == dwState

  PMibTcpRowBasic = ^TMibTcpRowBasic;
  TMibTCPRowBasic = packed record
    dwState     : DWORD;
    dwLocalAddr : DWORD;
    dwLocalPort : DWORD;
    dwRemoteAddr: DWORD;
    dwRemotePort: DWORD;
  end;

  PMibTcpRowOwnerPid = ^TMibTcpRowOwnerPid;
  TMibTcpRowOwnerPid = packed record
    dwState     : DWORD;
    dwLocalAddr : DWORD;
    dwLocalPort : DWORD;
    dwRemoteAddr: DWORD;
    dwRemotePort: DWORD;
    dwOwningPid : DWORD;
  end;

  PMIB_TCPTABLE_BASIC = ^MIB_TCPTABLE_BASIC;
  MIB_TCPTABLE_BASIC = packed record
   dwNumEntries: DWORD;
   table: Array [0..ANY_SIZE - 1] of TMibTcpRowBasic;
  end;

  PMIB_TCPTABLE_OWNER_PID = ^MIB_TCPTABLE_OWNER_PID;
  MIB_TCPTABLE_OWNER_PID = packed record
   dwNumEntries: DWORD;
   table: Array [0..ANY_SIZE - 1] of TMibTcpRowOwnerPid;
  end;
  { End GetExtendedTCPTable types }

  function GetTCPTableData(const ACSVData: TStringList;
                           const AOption: TTCPTableOption = ttoAll): Cardinal;
  function GetTCPTableDataF(const AFunction: TGetExtendedTcpTable;
                            const ACSVData: TStringList;
                            const AOption: TTCPTableOption = ttoAll): Cardinal;

  function CardinalToIPv4(const AIP: Cardinal;
                          const ABigEndian: Boolean = False): string;
  function Int64toIPv4(const AIP: Int64): string;
  function IPv4toCardinal(const AIP: string;
                          const ABigEndian: Boolean = False): Cardinal;
  function IPv4toInt64(const AIP: string): Int64;

  {+ /trim=2}
  function IsPrivateIPv4(const AIP: Int64): Boolean; overload;
  function IsPrivateIPv4(const AIP: string): Boolean; overload;
  {
    Returns True if the IP argument is within one of the private IP ranges:
    10.0.0.0/8
    172.16.0.0/12
    192.168.0.0/16

    See also: MatchesCIDRv4
  -}

  {+ /trim=2}
  function IsValidIPv4(const AIP: Int64): Boolean; overload;
  function IsValidIPv4(const AIP: string): Boolean; overload;
  {
    Returns True if the IP argument is valid.
    "0.0.0.0" to "255.255.255.255" for strings, or 0..MaxIPv4 for Int64.

    See also: IPv4toCardinal, IPv4toInt64
  -}

  {+ /trim=2}
  function MatchesCIDRv4(const AAddr, AMask, ABits: Int64): Boolean; overload;
  function MatchesCIDRv4(const AAddr, AMask: string): Boolean; overload;
  {
    Returns a boolean indicating if the IPv4 dotted-decimal IP Address (or Int64
    equivalent) matches the CIDR-format network mask.

    Example: MatchesCIDRv4('192.168.1.1', '192.168.0.0/16') --> True

    If the string mask does not include a number of bits after a slash, it is
    assumed to be "/0". A zero bit mask simply checks for an exact match. A 32 bit
    mask matches *anything* (always returns True).

    If either value is not a valid IP or the number of bits is not 0..32, the
    function will return False.

    See also: IPv4toCardinal, IPv4toInt64, IsValidIPv4
  -}

  function CIDRv4Range(const AMask: string; out OMin, OMax: Cardinal): Integer;

  function CommonCIDRv4(const AIPList: TStrings;
                        out OMask: Int64; out OBits: Integer): string;

implementation

uses llNumbers;

{+}
function GetTCPTableData(const ACSVData: TStringList;
                         const AOption: TTCPTableOption = ttoAll): Cardinal;
{
  Gets Netstat-like info on TCP connections in a simple CSV format:

  LocalIp,LocalPort,RemoteIp,RemotePort,StateString,State,PID

  Returns an error code, or zero if successful.

  The AOption argument specifies whether you want only listening, only
  established connections, or all TCP connections (the default).

  TTCPTableOption = (ttoListener, ttoConnections, ttoAll);

  This unit contains all the types needed to work with the GetExtendedTcpTable
  API function. This function is primarily here as an example but can be useful
  as a "quickie".

  This function loads the dll, gets the proc address, and then calls the
  GetTCPTableDataF function that takes the function pointer as the first
  argument. If you have loaded the dll and proc address on your own (if making
  repeated calls), you can just call that function directly.

  See also: GetTCPTableDataF
-}
var
  hDLL: THandle;
  pFunc: TGetExtendedTcpTable;
begin
  hDLL := LoadLibrary(IpHlpApiDLL);
  if hDLL = 0 then begin
    Result := GetLastError;
    Exit;
  end;
  try
    pFunc := GetProcAddress(hDLL, 'GetExtendedTcpTable');
    if Assigned(pFunc) then
      Result := GetTCPTableDataF(pFunc, ACSVData, AOption)
    else
      Result := GetLastError;
  finally
    FreeLibrary(hDLL);
  end;
end;

{+}
function GetTCPTableDataF(const AFunction: TGetExtendedTcpTable;
                          const ACSVData: TStringList;
                          const AOption: TTCPTableOption = ttoAll): Cardinal;
{
  The workhorse behind GetTCPTableData, used when you already have the dll
  loaded and the GetExtendedTcpTable function pointer available.

  See also: GetTCPTableData
-}
var
  iClass, i: Integer;
  TableSize: DWORD;
  pTcpTable: PMIB_TCPTABLE_OWNER_PID;
  sData: string;
  function FixPort(const APort: DWORD): string;
  var
    iPort: Cardinal;
  begin
    // The port numbers returned in the table use only the lower 16 bits, and
    // are big-endian.
    iPort := ((APort and $FF) shl 8) + ((APort and $FF00) shr 8);
    Result := IntToStr(iPort);
  end;
begin
  Assert(Assigned(AFunction), 'GetTCPTableDataF: Function pointer not initialized');
  // Turn off range checking - accessing the table rows beyond zero raises an
  // error because they're defined as [0..ANY_SIZE - 1]
  {$IFOPT R+}
    {$DEFINE RANGEON}
    {$R-}
  {$ELSE}
    {$UNDEF RANGEON}
  {$ENDIF}
  // Set the class based on what we want to get back:
  case AOption of
    ttoListener:    iClass := Ord(TCP_TABLE_OWNER_PID_LISTENER);
    ttoConnections: iClass := Ord(TCP_TABLE_OWNER_PID_CONNECTIONS);
    else {ttoAll}   iClass := Ord(TCP_TABLE_OWNER_PID_ALL);
  end;
  // Find out how much memory we need to allocate and get it:
  TableSize := 0;
  Result := AFunction(nil, @TableSize, False, AF_INET, iClass, 0);
  if Result <> ERROR_INSUFFICIENT_BUFFER then
    Exit;
  GetMem(pTcpTable, TableSize);
  try
    Result := AFunction(pTcpTable, @TableSize, TRUE, AF_INET, iClass, 0);
    if Result = NO_ERROR then begin
      for i := 0 to pTcpTable.dwNumEntries - 1 do begin
        // Format: LocalIp,LocalPort,RemoteIp,RemotePort,StateString,State,PID
        sData := CardinalToIPv4(pTcpTable^.Table[i].dwLocalAddr, True) + ',';
        sData := sData + FixPort(pTcpTable^.Table[i].dwLocalPort) + ',';
        sData := sData + CardinalToIPv4(pTcpTable^.Table[i].dwRemoteAddr, True) + ',';
        sData := sData + FixPort(pTcpTable^.Table[i].dwRemotePort) + ',';
        sData := sData + MIB_TCP_STATE[pTcpTable^.Table[i].dwState] + ',';
        sData := sData + IntToStr(pTcpTable^.Table[i].dwState) + ',';
        sData := sData + IntToStr(pTcpTable^.Table[i].dwOwningPid);
        ACSVData.Add(sData);
      end;
    end;
  finally
    FreeMem(pTcpTable);
  end;
  // Turn range checking back on if it was on before
  {$IFDEF RANGEON}
    {$R+}
    {$UNDEF RANGEON}
  {$ENDIF}
end;

{+}
function CardinalToIPv4(const AIP: Cardinal;
                        const ABigEndian: Boolean = False): string;
{
  Converts a Cardinal into a dotted-decimal IPv4 address string.
  If ABigEndian is True, the bytes will be reversed first.

  See also: Int64toIPv4, IPv4toCardinal, IPv4toInt64, IsValidIPv4,
    SwapEndianCardinal
-}
var
  i64: Int64;
begin
  if ABigEndian then
    i64 := SwapEndianCardinal(AIP)
  else
    i64 := AIP;
  Result := Int64toIPv4(i64);
end;

{+}
function Int64toIPv4(const AIP: Int64): string;
{
  Converts an Int64 into a dotted-decimal IPv4 address string.

  If the numeric IP is out of range (0..MaxIPv4) the function returns an
  empty string. MaxIPv4 is a constant representing High(Cardinal).

  See also: CardinaltoIPv4, IPv4toCardinal, IPv4toInt64, IsValidIPv4
-}
var
  i, s, x: Integer;
begin
  Result := '';
  if (AIP < 0) or (AIP > MaxIPv4) then
    Exit;
  for i := 0 to 3 do begin
    s := 24 - (i * 8);
    x := (AIP shr s) and $FF;
    Result := Result + IntToStr(x);
    if i < 3 then
      Result := Result + '.';
  end;
end;

{+}
function IPv4toCardinal(const AIP: string;
                        const ABigEndian: Boolean = False): Cardinal;
{
  Converts a dotted-decimal IPv4 address string into a Cardinal. If the string
  is not a valid IP, the function returns zero.

  Note that 0.0.0.0 is a valid IP address, but will also return zero. If you are
  expecting that, IPv4toInt64 may be a better choice as it will return -1 for an
  invalid address.

  If ABigEndian is True, the bytes will be reversed in the result.

  See also: CardinaltoIPv4, Int64toIPv4, IPv4toInt64, IsValidIPv4,
    SwapEndianCardinal
-}
var
  i: Int64;
begin
  i := IPv4toInt64(AIP);
  if i >= 0 then
    Result := Int64Rec(i).Lo
  else
    Result := 0;
  if ABigEndian then
    Result := SwapEndianCardinal(Result);
end;

{+}
function IPv4toInt64(const AIP: string): Int64;
{
  Converts a dotted-decimal IPv4 address string into an Int64.
  If the string is not a valid IP, the function returns -1.

  See also: CardinaltoIPv4, Int64toIPv4, IPv4toCardinal, IsValidIPv4
-}
var
  slOctets: TStringList;
  i, s: Integer;
  iIP, x: Int64;
begin
  Result := -1;
  i := Length(AIP);
  if (i < 7) or (i > 15) then // 0.0.0.0 to 255.255.255.255
    Exit;
  iIP := 0;
  slOctets := TStringList.Create;
  try
    slOctets.Delimiter := '.';
    slOctets.StrictDelimiter := True;
    slOctets.DelimitedText := AIP;
    if slOctets.Count <> 4 then
      Exit;
    for i := 0 to 3 do begin
      s := Length(slOctets[i]);
      if (s < 1) or (s > 3) then
        Exit;
      x := StrToIntDef(slOctets[i], -1);
      if (x < 0) or (x > 255) then
        Exit;
      s := 24 - (i * 8);
      iIP := iIP + (x shl s);
    end;
    Result := iIP;
  finally
    slOctets.Free;
  end;
end;

function IsPrivateIPv4(const AIP: Int64): Boolean; overload;
begin
  // 10.0.0.0/8     =  167772160
  // 172.16.0.0/12  = 2886729728
  // 192.168.0.0/16 = 3232235520
  Result := MatchesCIDRv4(AIP,  167772160,  8) or
            MatchesCIDRv4(AIP, 2886729728, 12) or
            MatchesCIDRv4(AIP, 3232235520, 16);
end;

function IsPrivateIPv4(const AIP: string): Boolean; overload;
begin
  Result := IsPrivateIPv4(IPv4toInt64(AIP));
end;

function IsValidIPv4(const AIP: Int64): Boolean; overload;
begin
  Result := (AIP >= 0) and (AIP <= MaxIPv4);
end;

function IsValidIPv4(const AIP: string): Boolean;
begin
  Result := IsValidIPv4(IPv4toInt64(AIP));
end;

function MatchesCIDRv4(const AAddr, AMask, ABits: Int64): Boolean; overload;
var
  iBits: Integer;
begin
  Result := False;
  if IsValidIPv4(AAddr) and IsValidIPv4(AMask) then begin
    iBits := 32 - ABits;
    if (iBits > 0) and (iBits < 32) then
      Result := (AAddr shr iBits) = (AMask shr iBits)
    else if ABits = 0 then
      Result := AAddr = AMask
    else if ABits = 32 then
      Result := True;
  end;
end;

function MatchesCIDRv4(const AAddr, AMask: string): Boolean;
var
  sTemp: string;
  i, iBits: Integer;
  iAddr, iMask: Int64;
begin
  Result := False;
  sTemp := AMask;
  i := Pos('/', sTemp);
  if i > 0 then begin
    if not TryStrToInt(Copy(sTemp, i + 1, MaxInt), iBits) then
      Exit; // Not a valid mask
    SetLength(sTemp, i - 1);
  end else // Without a number of bits, assume 0 (exact match)
    iBits := 0;
  iMask := IPv4toInt64(sTemp);
  iAddr := IPv4toInt64(AAddr);
  Result := MatchesCIDRv4(iAddr, iMask, iBits);
end;

{+}
function CIDRv4Range(const AMask: string; out OMin, OMax: Cardinal): Integer;
{
  Takes a string IPv4 CIDR mask and calculates the lowest and highest Cardinal
  IP addresses it contains and places them into the variable arguments.
  The function returns the number of bits the mask covers, or -1 if the mask is
  not valid.
-}
var
  sTemp: string;
  i, iBits: Integer;
  iMask: Int64;
begin
  Result := -1;
  sTemp := AMask;
  i := Pos('/', sTemp);
  if i > 0 then begin
    if not TryStrToInt(Copy(sTemp, i + 1, MaxInt), iBits) then
      Exit; // Not a valid mask
    SetLength(sTemp, i - 1);
  end else // Without a number of bits, assume 0 (exact match)
    iBits := 0;
  iMask := IPv4toInt64(sTemp);
  if iMask >= 0 then begin
    OMin := Cardinal(iMask);
    if (iBits > 0) and (iBits < 32) then begin
      OMax := MaxIPv4 shl iBits;
      OMax := OMax shr iBits;
      OMax := OMax + OMin;
    end else if iBits = 32 then
      OMax := OMin
    else begin
      OMin := 0;
      OMax := MaxIPv4;
    end;
    Result := iBits;
  end;
end;

{+}
function CommonCIDRv4(const AIPList: TStrings;
                      out OMask: Int64; out OBits: Integer): string;
{
  Takes a list of string IPv4 addresses as input and calculates their shared
  network mask using the highest number of bits all the addresses have in
  common. The Mask and Bits are updated in the numeric output parameters, and
  the CIDR string (e.g. "172.16.0.0/24") is the return value.

  If the list is empty or any address in the list is invalid, the function will
  return an empty string. The content of the output parameters is undefined.

  If the addresses have absolutely no bits in common (for example, 10.x.x.x and
  192.168.x.x), the output parameters will be zero and the return string will be
  "0.0.0.0/0".
-}
var
  iDiff, i, b: Integer;
  iIP: Int64;
begin
  Result := '';
  if AIPList.Count = 0 then
    Exit;
  // Grab the first address:
  OMask := IPv4ToInt64(AIPList[0]);
  if OMask < 0 then
    Exit;
  iDiff := 0;
  for i := 1 to AIPList.Count - 1 do begin
    iIP := IPv4ToInt64(AIPList[i]);
    if iIP < 0 then
      Exit;
    // How many bits does this have in common with what we have so far?
    b := 31;
    while (b >= 0) and ((iIP shr b) = (OMask shr b)) do
      Dec(b);
    if b >= 0 then
      Inc(b);
    // Keep track of the highest difference:
    if b > iDiff then
      iDiff := b;
    if iDiff = 32 then
      break;
    OMask := OMask and iIP;
  end;
  iIP := MaxIPv4; // An all-ones network mask
  iIP := iIP shr iDiff;
  iIP := iIP shl iDiff;
  OMask := OMask and iIP;
  OBits := 32 - iDiff;
  Result := Int64ToIPv4(OMask) + '/' + IntToStr(OBits);
end;

end.


