{*****************************************************************************}
{                                                                             }
{  llNumbers.pas                                                              }
{                                                                             }
{  Number and math routines                                                   }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llNumbers;

interface

uses
  SysUtils, Math;

  {+ /trim=2}
  function BinStrToCardinal(const S: string): Cardinal;
  function BinStrToInt(const S: string): Integer;
  function BinStrToInt64(const S: string): Int64;
  {
    Converts a binary string (e.g. '0101') into a number. If the string is
    smaller than the number of bits in the result type, the string is right-
    justified into the least significant bits of the number. If the string is
    longer than the size of the result, only the rightmost characters are used.

    See also: IntToBinStr
  -}

  {+ /trim=2}
  function IntToBinStr(const Value: Integer; const Bits: Integer = 0): string; overload;
  function IntToBinStr(const Value: Cardinal; const Bits: Integer = 0): string; overload;
  function IntToBinStr(const Value: Int64; const Bits: Integer = 0): string; overload;
  {
    Converts an integer type number into a binary string (e.g. '0101'). If the
    number of Bits is not supplied or is less than 1, the default size of the
    input numeric is used.

    See also: BinStrToCardinal, BinStrToInt, BinStrToInt64
  -}

  function GCD(const a, b: Integer): Integer;

  {+ /trim=2}
  function MinMax(const AMin, AValue, AMax: Integer): Integer; overload;
  function MinMax(const AMin, AValue, AMax: Cardinal): Cardinal; overload;
  function MinMax(const AMin, AValue, AMax: Int64): Int64; overload;
  function MinMax(const AMin, AValue, AMax: Single): Single; overload;
  function MinMax(const AMin, AValue, AMax: Double): Double; overload;
  function MinMax(const AMin, AValue, AMax: Extended): Extended; overload;
  function MinMax(const AMin, AValue, AMax: Currency): Currency; overload;
  {
    Translates directly to Max(AMin, Min(AValue, AMax)), but with less chance
    of the programmer messing it up. :)

    Overloaded just like the Min and Max functions in the Math unit.
  -}

  function OrdinalStr(const AInt: Integer): string;

  {+ /trim=2}
  function PrettyInt(const Value: Integer): string; overload;
  function PrettyInt(const Value: Int64): string; overload;
  function PrettyInt(const Value: Cardinal): string; overload;
  {
    Returns the integer as a string with thousands separators.
  -}

  {+ /trim=2}
  function RotL(const AValue: Word; const ABits: Integer): Word; overload;
  function RotL(const AValue, ABits: Integer): Integer; overload;
  function RotL(const AValue: Cardinal; const ABits: Integer): Cardinal; overload;
  function RotL(const AValue: Int64; const ABits: Integer): Int64; overload;

  function RotR(const AValue: Word; const ABits: Integer): Word; overload;
  function RotR(const AValue, ABits: Integer): Integer; overload;
  function RotR(const AValue: Cardinal; const ABits: Integer): Cardinal; overload;
  function RotR(const AValue: Int64; const ABits: Integer): Int64; overload;
  {
    Bitwise left/right rotation function for various integer types.
  -}

  function SwapEndianCardinal(const AValue: Cardinal): Cardinal; register;
  function SwapEndianInt(const AValue: Integer): Integer; register;

implementation

function BinStrToCardinal(const S: string): Cardinal;
var
  i, iLo: Integer;
  j: Cardinal;
begin
  Result := 0;
  j := 1;
  iLo := Length(S) - (SizeOf(Result) * 8) - 1;
  for i := Length(S) downto Max(iLo, 1) do begin
    if S[i] = '1' then
      Result := Result or j;
    j := j shl 1;
  end;
end;

function BinStrToInt(const S: string): Integer;
var
  i, j, iLo: Integer;
begin
  Result := 0;
  j := 1;
  iLo := Length(S) - (SizeOf(Result) * 8) - 1;
  for i := Length(S) downto Max(iLo, 1) do begin
    if S[i] = '1' then
      Result := Result or j;
    j := j shl 1;
  end;
end;

function BinStrToInt64(const S: string): Int64;
var
  i, iLo: Integer;
  j: Int64;
begin
  Result := 0;
  j := 1;
  iLo := Length(S) - (SizeOf(Result) * 8) - 1;
  for i := Length(S) downto Max(iLo, 1) do begin
    if S[i] = '1' then
      Result := Result or j;
    j := j shl 1;
  end;
end;

function IntToBinStr(const Value: Integer; const Bits: Integer = 0): string;
var
  b, i, s, z: Integer;
begin
  z := SizeOf(Value) * 8;
  b := Bits;
  if (b < 1) or (b > z) then
    b := z;
  Result := StringOfChar('0', b);
  z := b - 1;
  s := 1;
  for i := 0 to z do begin
    if (Value and s) <> 0 then
      Result[b - i] := '1';
    if i < z then
      s := s shl 1;
  end;
end;

function IntToBinStr(const Value: Cardinal; const Bits: Integer = 0): string;
var
  b, i, z: Integer;
  s: Cardinal;
begin
  z := SizeOf(Value) * 8;
  b := Bits;
  if (b < 1) or (b > z) then
    b := z;
  Result := StringOfChar('0', b);
  z := b - 1;
  s := 1;
  for i := 0 to z do begin
    if (Value and s) <> 0 then
      Result[b - i] := '1';
    if i < z then
      s := s shl 1;
  end;
end;

function IntToBinStr(const Value: Int64; const Bits: Integer = 0): string;
var
  b, i, z: Integer;
  s: Int64;
begin
  z := SizeOf(Value) * 8;
  b := Bits;
  if (b < 1) or (b > z) then
    b := z;
  Result := StringOfChar('0', b);
  z := b - 1;
  s := 1;
  for i := 0 to z do begin
    if (Value and s) <> 0 then
      Result[b - i] := '1';
    if i < z then
      s := s shl 1;
  end;
end;

{+}
function GCD(const a, b: Integer): Integer;
{
  Calculates the greatest common divisor of two integers (Euclidean algorithm).
-}
begin
  if b = 0 then
    Result := a
  else
    Result := GCD(b, a mod b);
end;

{------------------------------------------------------------------------------}

function MinMax(const AMin, AValue, AMax: Integer): Integer;
begin
  Result := Max(AMin, Min(AValue, AMax));
end;

function MinMax(const AMin, AValue, AMax: Cardinal): Cardinal;
begin
  Result := Max(AMin, Min(AValue, AMax));
end;

function MinMax(const AMin, AValue, AMax: Int64): Int64;
begin
  Result := Max(AMin, Min(AValue, AMax));
end;

function MinMax(const AMin, AValue, AMax: Single): Single;
begin
  Result := Max(AMin, Min(AValue, AMax));
end;

function MinMax(const AMin, AValue, AMax: Double): Double;
begin
  Result := Max(AMin, Min(AValue, AMax));
end;

function MinMax(const AMin, AValue, AMax: Extended): Extended;
begin
  Result := Max(AMin, Min(AValue, AMax));
end;

function MinMax(const AMin, AValue, AMax: Currency): Currency;
begin
  Result := Max(AMin, Min(AValue, AMax));
end;

{------------------------------------------------------------------------------}

{+}
function OrdinalStr(const AInt: Integer): string;
{
  Returns the number as an ordinal, like '1st', '2nd', '3rd', '4th', etc.
  Negative values are converted to positive.
-}
var
  i: Integer;
begin
  i := Abs(AInt);
  Result := IntToStr(i);
  if (i >= 4) and (i <= 20) then
    i := 0 // All teens are "th"
  else
    i := i mod 10;
  case i of
    1:   Result := Result + 'st';
    2:   Result := Result + 'nd';
    3:   Result := Result + 'rd';
    else Result := Result + 'th';
  end;
end;

{------------------------------------------------------------------------------}

function PrettyInt(const Value: Integer): string; overload;
begin
  Result := Format('%.0n', [Value * 1.0]);
end;

function PrettyInt(const Value: Int64): string; overload;
begin
  Result := Format('%.0n', [Value * 1.0]);
end;

function PrettyInt(const Value: Cardinal): string; overload;
begin
  Result := Format('%.0n', [Value * 1.0]);
end;

{------------------------------------------------------------------------------}

function RotL(const AValue: Word; const ABits: Integer): Word; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shl b) or (AValue shr (s - b));
end;

function RotL(const AValue, ABits: Integer): Integer; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shl b) or (AValue shr (s - b));
end;

function RotL(const AValue: Cardinal; const ABits: Integer): Cardinal; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shl b) or (AValue shr (s - b));
end;

function RotL(const AValue: Int64; const ABits: Integer): Int64; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shl b) or (AValue shr (s - b));
end;

function RotR(const AValue: Word; const ABits: Integer): Word; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shr b) or (AValue shl (s - b));
end;

function RotR(const AValue, ABits: Integer): Integer; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shr b) or (AValue shl (s - b));
end;

function RotR(const AValue: Cardinal; const ABits: Integer): Cardinal; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shr b) or (AValue shl (s - b));
end;

function RotR(const AValue: Int64; const ABits: Integer): Int64; overload;
var
  s, b: Integer;
begin
  s := SizeOf(AValue) * 8;
  b := ABits mod s;
  Result := (AValue shr b) or (AValue shl (s - b));
end;

{------------------------------------------------------------------------------}

{+}
function SwapEndianCardinal(const AValue: Cardinal): Cardinal; register;
{
  Reverses the bytes in a 32-bit Cardinal.

  See also: SwapEndianInt
-}
asm
  bswap eax
end;

{+}
function SwapEndianInt(const AValue: Integer): Integer; register;
{
  Reverses the bytes in a 32-bit Integer.

  See also: SwapEndianCardinal
-}
asm
  bswap eax
end;

end.
