{*****************************************************************************}
{                                                                             }
{  llPrinting.pas                                                             }
{                                                                             }
{  Printing routines                                                          }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llPrinting;

interface

  uses Windows, Classes, SysUtils, Printers, WinSpool, Graphics;

  procedure GetBinNames(BinStrings: TStrings);
  function ChangePrinterBin(const ABin: Integer): Integer;

  procedure SizeMemoLines(Canvas: TCanvas; Lines: TStrings;
                          const MMLength: Integer;
                          const MMIndent: Integer = 0);

implementation

{$HINTS OFF}
{+}
procedure GetBinNames(BinStrings: TStrings);
{
  Fills a TStrings variable with the names of the available paper sources (bins)
  for the currently selected printer.  The bin numbers that you can pass to the
  ChangePrinterBin() function are stored in the Objects part of the list.

  Retrieved from the Delphi newsgroup archives, originally by Peter Below.
-}
type
  TBinName = Array [0..23] of Char;
  TBinNameArray = Array [1..255] of TBinName; // Array [1..High(Integer) div Sizeof( TBinName )] of TBinName;
  PBinnameArray = ^TBinNameArray;
  TBinArray = Array [1..255] of Word; // Array [1..High(Integer) div Sizeof(Word)] of Word;
  PBinArray = ^TBinArray;
Var
  Device, Driver, Port: Array [0..255] of Char;
  hDevMode: THandle;
  i, numBinNames, numBins, temp: Integer;
  pBinNames: PBinnameArray;
  pBins: PBinArray;
begin
  Printer.GetPrinter(Device, Driver, Port, hDevMode);
  Printer.SetPrinter(Device, Driver, Port, 0);
  Printer.GetPrinter(Device, Driver, Port, hDevMode);
  numBinNames :=
    WinSpool.DeviceCapabilities( Device, Port, DC_BINNAMES, Nil, Nil );
  numBins :=
    WinSpool.DeviceCapabilities( Device, Port, DC_BINS, Nil, Nil );
  If numBins <> numBinNames Then Begin
    raise Exception.Create(
      'DeviceCapabilities reports different number of bins and bin names!');
  End; { If }
  If numBinNames > 0 Then Begin
    pBins := Nil; // This causes a "value assigned to...not used" hint
    GetMem( pBinNames, numBinNames * Sizeof( TBinname ));
    GetMem( pBins, numBins * Sizeof(Word));
    try
      WinSpool.DeviceCapabilities( Device, Port, DC_BINNAMES,
                                   Pchar( pBinNames ), Nil);
      WinSpool.DeviceCapabilities( Device, Port, DC_BINS,
                                   Pchar( pBins ), Nil );
      BinStrings.Clear;
      For i:= 1 To numBinNames Do Begin
        temp := pBins^[i];
        BinStrings.AddObject(pBinNames^[i], TObject( temp ));
      End;
    finally
      FreeMem( pBinNames );
      If pBins <> Nil Then
        FreeMem( pBins );
    end;
  End;
end;
{$HINTS ON}

{+}
function ChangePrinterBin(const ABin: Integer): Integer;
{
  Changes the paper source bin of the currently selected printer.  See the
  GetBinNames procedure for info on retrieving the valid bin numbers.  This
  procedure must be called *before* Printer.BeginDoc to take effect.

  I changed it to a function to return the current bin value before the change.

  Retrieved from the Delphi newsgroup archives, originally by Peter Below and
  originally called "SelectPrinterBin".
-}
var
  Device, Driver, Port: array[0..80] of Char;
  DevMode: THandle;
  pDevmode: PDeviceMode;
begin
  Result := 0;
  // Get printer device name etc.
  Printer.GetPrinter(Device, Driver, Port, DevMode);
  // force reload of DEVMODE
  Printer.SetPrinter(Device, Driver, Port, 0) ;
  // get DEVMODE handle
  Printer.GetPrinter(Device, Driver, Port, DevMode);
  If Devmode <> 0 Then Begin
    // lock it to get pointer to DEVMODE record
    pDevMode := GlobalLock( Devmode );
    If pDevmode <> Nil Then
    try
      With pDevmode^ Do Begin
        Result := dmDefaultsource; // Return current value
        // modify paper size
        dmDefaultsource := ABin;
        // tell printer driver that dmDefaultsource field contains
        // data it needs to inspect.
        dmFields := dmFields or DM_DEFAULTSOURCE;
      End;
    finally
      // unlock devmode handle.
      GlobalUnlock( Devmode );
    end;
  End; { If }
end;

{+}
procedure SizeMemoLines(Canvas: TCanvas; Lines: TStrings;
                        const MMLength, MMIndent: Integer);
{
  Given a stringlist full of text, the function will adjust the length of each
  line to best fit/wrap on a line of a given length in millimeters.  When
  finished, this same stringlist will contain the adjusted lines.

  If desired, the first line can be indented (shortened) by a certain length in
  millimeters by supplying a MMIndent value > 0.  To indent all lines except
  the first, pass a negative MMIndent value.
-}
var
  LogX, iPixels, iIndent, iTextWidth, i, n, t, iAdjust: Integer;
  slTemp: TStringList;
  sTemp: String;
begin
  LogX := GetDeviceCaps(Canvas.Handle, LOGPIXELSX);
  iPixels := Round((MMLength / 25.4) * LogX);
  iIndent := Round((MMIndent / 25.4) * LogX);
  if iIndent > 0 then
    iPixels := iPixels - iIndent;
  iAdjust := Canvas.TextWidth('A');
  slTemp := TStringList.Create;
  try
    { Instead of doing slTemp.Assign(Lines), copy the Text property in order
      to remove any "soft" line breaks that may have been introduced by a
      containing control (e.g. TMemo) and only consider the "hard" line breaks
      entered by the user. }
    slTemp.Text := Lines.Text;
    Lines.Clear;
    for i := 0 to slTemp.Count - 1 do begin
      sTemp := TrimRight(slTemp[i]);
      iTextWidth := Canvas.TextWidth(sTemp);
      while iTextWidth > iPixels do begin
        t := Length(sTemp);
        n := Trunc((iPixels / iTextWidth) * t);
        while (Canvas.TextWidth(Copy(sTemp, 1, n)) > iPixels) and (n > 2) do
          n := n - 3;
        while (Canvas.TextWidth(Copy(sTemp, 1, n)) < (iPixels - iAdjust)) and
              (n < t) do
          n := n + 1;
        t := n;
        while (n > 0) and (sTemp[n] <> ' ') and (sTemp[n] <> '-') do
          n := n - 1; // Trim to the last space or dash prior to max length
        if n = 0 then
          n := t; // If there are no breaks, we've got to print something!
        Lines.Add(Copy(sTemp, 1, n));
        iPixels := iPixels + iIndent;
        iIndent := 0;
        Delete(sTemp, 1, n);
        sTemp := Trim(sTemp);
        iTextWidth := Canvas.TextWidth(sTemp);
      end;
      Lines.Add(sTemp);
      iPixels := iPixels + iIndent;
      iIndent := 0;
    end;
  finally
    slTemp.Free;
  end;
end;

end.
