{*****************************************************************************}
{                                                                             }
{  llSettings.pas                                                             }
{                                                                             }
{  Procedures and functions dealing with saving & loading program settings.   }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llSettings;

interface

uses SysUtils, Classes, Contnrs, Windows, Forms, Registry;

const
  sfpTrue    =  1; // SaveFormPos constants
  sfpDefault =  0;
  sfpFalse   = -1;
  sfpDelete  = -2;
  sfpFlagName = 'SaveWindowPos';
  sfpOtherInfoDelete = '$$sfpOtherInfoDelete$$';

type
  {+ /trim=2}
  TllRegistry = class(TRegistry)
  public
    function DefBool(const AName: string; const ADefault: Boolean): Boolean;
    function DefCurr(const AName: string; const ADefault: Currency): Currency;
    function DefDate(const AName: string; const ADefault: TDateTime): TDateTime;
    function DefFloat(const AName: string; const ADefault: Double): Double;
    function DefInt(const AName: string; const ADefault: Integer): Integer;
    function DefStr(const AName, ADefault: string): string;
  end;
  {+ /trim=OFF
  A subclass of TRegistry that adds some default read methods with nice, short
  names. If the value does not exist, the default value is returned.

  Internally, they all work like this (DefStr example):

  if ValueExists(AName) then
    Result := ReadString(AName)
  else
    Result := ADefault;
  -}

  TllPrefType = (ptBoolean, ptCurrency, ptDateTime, ptFloat, ptInteger,
                 ptString, ptEncodedString);

  TllPrefItem = class(TObject)
  private
    FName: string;
    FType: TllPrefType;
    FValue: string;
    FMemoryOnly: Boolean;
    function GetBool: Boolean;
    function GetCurr: Currency;
    function GetDate: TDateTime;
    function GetFloat: Double;
    function GetInt: Integer;
    procedure SetBool(const Value: Boolean);
    procedure SetCurr(const Value: Currency);
    procedure SetDate(const Value: TDateTime);
    procedure SetFloat(const Value: Double);
    procedure SetInt(const Value: Integer);
  public
    property Name: string read FName;
    property PrefType: TllPrefType read FType write FType;
    property MemoryOnly: Boolean read FMemoryOnly write FMemoryOnly;
    property AsBool: Boolean read GetBool write SetBool;
    property AsCurr: Currency read GetCurr write SetCurr;
    property AsDate: TDateTime read GetDate write SetDate;
    property AsFloat: Double read GetFloat write SetFloat;
    property AsInt: Integer read GetInt write SetInt;
    property AsStr: string read FValue write FValue;
  end;

  {+ /trim=2 TllPrefs}
  TllPrefs = class(TObject)
  {-}
  private
    FRegKey: string;
    FItems: TObjectList;
    FLastName: string;
    FLastIndex: Integer;
    function GetCount: Integer;
    function GetItemByName(AName: string): TllPrefItem;
  {+ /trim=2 /addline=OFF TllPrefs}
  public
    constructor Create;
    destructor Destroy; override;
    property Count: Integer read GetCount;
    property Items[AName: string]: TllPrefItem read GetItemByName; default;
    property RegKey: string read FRegKey write FRegKey;
    function Add(const AName: string; const AType: TllPrefType): TllPrefItem;
    procedure Clear;
    procedure Delete(const AName: string; const ADeleteFromRegistry: Boolean);
    function ItemByIndex(const AIndex: Integer): TllPrefItem;
    function IndexOfName(const AName: string): Integer; // -1 if not found
    function Load: Boolean;
    function LoadPrefsFromKey: Boolean;
    function Save: Boolean;
  end;
  {

    Maintains a list of values stored in the registry. Example usage:

    Prefs := TllPrefs.Create;
    Prefs.RegKey := 'Software\RayMarron\LogicLib'; // Required
    Prefs.Add('Max', ptInteger).AsInt := 10; // Set a default value when added
    Prefs.Add('Title', ptString).AsStr := 'foo';
    Prefs.Load;
    ...
    Label.Caption := Prefs['Title'].AsStr;
    ...
    Prefs.Save;
    Prefs.Free;

    Each preference's Name must be unique (case-insensitive). An assertion will
    be raised if a duplicate name is added.

    Each preference's PrefType property is one of the following values:

      TllPrefType = (ptBoolean, ptCurrency, ptDateTime, ptFloat, ptInteger,
                     ptString, ptEncodedString);

    An encoded string is stored/retrieved using SimpleEncode/SimpleDecode.

    Get/Set the preference's value using one of the As... properties: AsBool,
    AsCurr, AsDate, AsFloat, AsInt, and AsStr.

    The Load/Save functions return True if successful. Only defined values are
    loaded and saved. If a value does not exist in the registry, it defaults to
    its current value; Assign each value a sensible default when Added.

    Set the MemoryOnly property if a value should not be saved to the registry.

    See also: TllRegistry, SimpleDecode, SimpleEncode
  -}

{+}
function SaveFormPos(const ARegKey: string;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const ASaveFlag: Integer = sfpDefault): Boolean; overload;

function SaveFormPos(const AReg: TRegistry;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const ASaveFlag: Integer = sfpDefault): Boolean; overload;
{
  Saves a form's position and size to HKCU. Values named Top, Left, Height,
  Width, and Maximized are saved. Use LoadFormPos to read these settings back.

  If you plan on storing the position of more than one form, give each a unique
  registry key or use APrefix to differentiate the values. A common value for
  APrefix would be the form's ClassName.

  A boolean value in this key called "SaveWindowPos" (sfpFlagName) controls
  whether these settings get saved or not. It defaults to True. You can change
  this value (e.g. make it a user preference) with the ASaveFlag argument and
  handy constants:

  sfpTrue    - Set it to True
  sfpDefault - Use the currently stored value or create it as True
  sfpFalse   - Set it to False (LoadFormPos still reads previously saved values)
  sfpDelete  - Set it to False and delete any saved values including AOtherInfo

  The function returns True if the save flag ends up True and the settings were
  saved.

  The optional AOtherInfo argument allows you to provide a tag-along list of
  name-value pairs that will be stored as strings to the same registry key and
  can later be retrieved using LoadFormPos. This is handy for other window state
  information like splitter positions, etc. Note that APrefix does NOT apply to
  the value names in AOtherInfo. If an AOtherInfo value is set to the special
  constant sfpOtherInfoDelete, its registry entry will be removed if present.

  There are two overloaded versions of the function - One that takes a registry
  key name, and one that takes an already open TRegistry object. The former will
  open the key, call the latter, then close the key.

  See also: LoadFormPos
-}

{+}
function LoadFormPos(const ARegKey: string;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const APosOnly: Boolean = False): Boolean; overload;

function LoadFormPos(const AReg: TRegistry;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const APosOnly: Boolean = False): Boolean; overload;
{
  Loads a form's position and size from registry settings previously saved to
  HKCU by the SaveFormPos function. Returns True if the settings were present
  and loaded. If there are no saved settings and the form has an owner, it
  centers the form over its owner.

  If AOtherInfo is provided as a list of name-value pairs, matching items in the
  registry key will be retrieved and written to the list. Missing AOtherInfo
  values do not affect the function's boolean return value, so it is suggested
  that you preload the list with sensible default values.

  APrefix is a string tacked on to the beginning of the registry entries so you
  can store more than one form's position in the same key. A common usage would
  be the form's ClassName. APrefix does not apply to AOtherInfo value names.

  If APosOnly is True, only the form's position will be retrieved, not its size.

  There are two overloaded versions of the function - One that takes a registry
  key name, and one that takes an already open TRegistry object. The former will
  open the key, call the latter, then close the key.

  See also: SaveFormPos
-}

function EraseRegStrings(const ARegKey: string; const AList: TStrings): Boolean;
function LoadRegStrings(const ARegKey: string; const AList: TStrings): Boolean;
function SaveRegStrings(const ARegKey: string; const AList: TStrings): Boolean;

implementation

uses llBool, llDates, llStrings;

const
  PrefRegKeyError = 'TllPrefs.RegKey not assigned';

{ TllPrefItem }

function TllPrefItem.GetBool: Boolean;
begin
  Result := StoB(FValue);
end;

function TllPrefItem.GetCurr: Currency;
begin
  Result := StrToCurrDef(FValue, 0);
end;

function TllPrefItem.GetDate: TDateTime;
begin
  Result := StoDDef(FValue, 0);
end;

function TllPrefItem.GetFloat: Double;
begin
  Result := StrToFloatDef(FValue, 0);
end;

function TllPrefItem.GetInt: Integer;
begin
  Result := StrToIntDef(FValue, 0);
end;

procedure TllPrefItem.SetBool(const Value: Boolean);
begin
  FValue := BtoS(Value);
end;

procedure TllPrefItem.SetCurr(const Value: Currency);
begin
  FValue := CurrToStr(Value);
end;

procedure TllPrefItem.SetDate(const Value: TDateTime);
begin
  FValue := DtoS(Value, dtpMillisecond);
end;

procedure TllPrefItem.SetFloat(const Value: Double);
begin
  FValue := FloatToStr(Value);
end;

procedure TllPrefItem.SetInt(const Value: Integer);
begin
  FValue := IntToStr(Value);
end;

{ TllPrefs }

function TllPrefs.Add(const AName: string;
                      const AType: TllPrefType): TllPrefItem;
begin
  // Usage: Prefs.Add('Foo', ptString).AsStr := 'foo';
  Assert(IndexOfName(AName) < 0,
         'TllPrefs duplicate preference name ' + Enquote(AName));
  Result := TllPrefItem.Create;
  Result.FName := AName;
  Result.FType := AType;
  FItems.Add(Result);
end;

procedure TllPrefs.Clear;
begin
  FItems.Clear;
end;

constructor TllPrefs.Create;
begin
  FItems := TObjectList.Create(True);
end;

procedure TllPrefs.Delete(const AName: string;
                          const ADeleteFromRegistry: Boolean);
var
  i: Integer;
  Reg: TllRegistry;
begin
  i := IndexOfName(AName);
  if i >= 0 then begin
    FItems.Delete(i);
    FLastName := '';
    if ADeleteFromRegistry then begin
      Assert(FRegKey <> '', PrefRegKeyError);
      Reg := TllRegistry.Create(KEY_ALL_ACCESS);
      try
        Reg.RootKey := HKEY_CURRENT_USER;
        if Reg.OpenKey(FRegKey, False) then begin
          if Reg.ValueExists(AName) then
            Reg.DeleteValue(AName);
          Reg.CloseKey;
        end;
      finally
        Reg.Free;
      end;
    end;
  end;
end;

destructor TllPrefs.Destroy;
begin
  FItems.Free;
  inherited;
end;

function TllPrefs.GetCount: Integer;
begin
  Result := FItems.Count;
end;

function TllPrefs.GetItemByName(AName: string): TllPrefItem;
var
  i: Integer;
begin
  i := IndexOfName(AName);
  if i >= 0 then
    Result := ItemByIndex(i)
  else
    Result := nil;
end;

function TllPrefs.IndexOfName(const AName: string): Integer;
begin
  if SameText(FLastName, AName) then
    Result := FLastIndex
  else begin
    Result := FItems.Count - 1;
    while Result >= 0 do begin
      if SameText(ItemByIndex(Result).Name, AName) then
        break;
      Dec(Result);
    end;
    if Result >= 0 then begin
      FLastName := AName;
      FLastIndex := Result;
    end;
  end;
end;

function TllPrefs.ItemByIndex(const AIndex: Integer): TllPrefItem;
begin
  Result := TllPrefItem(FItems[AIndex]);
end;

function TllPrefs.Load: Boolean;
var
  Reg: TllRegistry;
  i: Integer;
  Pref: TllPrefItem;
begin
  Result := False;
  Assert(FRegKey <> '', PrefRegKeyError);
  Reg := TllRegistry.Create(KEY_READ);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKeyReadOnly(FRegKey) then begin
      for i := 0 to FItems.Count - 1 do begin
        Pref := ItemByIndex(i);
        if Pref.MemoryOnly then
          Continue;
        case Pref.PrefType of
          ptBoolean:
            Pref.AsBool := Reg.DefBool(Pref.Name, Pref.AsBool);
          ptCurrency:
            Pref.AsCurr := Reg.DefCurr(Pref.Name, Pref.AsCurr);
          ptDateTime:
            Pref.AsDate := Reg.DefDate(Pref.Name, Pref.AsDate);
          ptFloat:
            Pref.AsFloat := Reg.DefFloat(Pref.Name, Pref.AsFloat);
          ptInteger:
            Pref.AsInt := Reg.DefInt(Pref.Name, Pref.AsInt);
          ptString:
            Pref.AsStr := Reg.DefStr(Pref.Name, Pref.AsStr);
          ptEncodedString:
            Pref.AsStr := SimpleDecode(Reg.DefStr(Pref.Name, Pref.AsStr));
        end;
      end;
      Reg.CloseKey;
      Result := True;
    end;
  finally
    Reg.Free;
  end;
end;

function TllPrefs.LoadPrefsFromKey: Boolean;
var
  Reg: TllRegistry;
  i: Integer;
  slKeys: TStringList;
  pt: TllPrefType;
begin
  { This works like Load (and actually calls it) but will first add PrefItems
    for any registry entries that exist in the key and are not yet in PrefItems.
    It is of limited use because it can only tell whether a value is a string,
    integer, or datetime (binary).
  }
  pt := ptString; // Eliminate warning
  Assert(FRegKey <> '', PrefRegKeyError);
  Reg := TllRegistry.Create(KEY_READ);
  slKeys := TStringList.Create;
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKeyReadOnly(FRegKey) then begin
      Reg.GetKeyNames(slKeys);
      for i := 0 to slKeys.Count - 1 do begin
        if IndexOfName(slKeys[i]) < 0 then begin
          case Reg.GetDataType(slKeys[i]) of
            rdString, rdExpandString: pt := ptString;
            rdInteger: pt := ptInteger;
            rdBinary: pt := ptDateTime; // Hope for the best!
            rdUnknown: Continue; // Unsupported
          end;
          Add(slKeys[i], pt);
        end;
      end;
      Reg.CloseKey;
    end;
  finally
    slKeys.Free;
    Reg.Free;
  end;
  Result := Load;
end;

function TllPrefs.Save: Boolean;
var
  Reg: TllRegistry;
  i: Integer;
  Pref: TllPrefItem;
begin
  Result := False;
  Assert(FRegKey <> '', PrefRegKeyError);
  Reg := TllRegistry.Create(KEY_ALL_ACCESS);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(FRegKey, True) then begin
      for i := 0 to FItems.Count - 1 do begin
        Pref := ItemByIndex(i);
        if Pref.MemoryOnly then
          Continue;
        case Pref.PrefType of
          ptBoolean:
            Reg.WriteBool(Pref.Name, Pref.AsBool);
          ptCurrency:
            Reg.WriteCurrency(Pref.Name, Pref.AsCurr);
          ptDateTime:
            Reg.WriteDateTime(Pref.Name, Pref.AsDate);
          ptFloat:
            Reg.WriteFloat(Pref.Name, Pref.AsFloat);
          ptInteger:
            Reg.WriteInteger(Pref.Name, Pref.AsInt);
          ptString:
            Reg.WriteString(Pref.Name, Pref.AsStr);
          ptEncodedString:
            Reg.WriteString(Pref.Name, SimpleEncode(Pref.AsStr));
        end;
      end;
      Reg.CloseKey;
      Result := True;
    end;
  finally
    Reg.Free;
  end;
end;

{ TllRegistry }

function TllRegistry.DefBool(const AName: string;
                             const ADefault: Boolean): Boolean;
begin
  if ValueExists(AName) then
    Result := ReadBool(AName)
  else
    Result := ADefault;
end;

function TllRegistry.DefCurr(const AName: string;
                             const ADefault: Currency): Currency;
begin
  if ValueExists(AName) then
    Result := ReadCurrency(AName)
  else
    Result := ADefault;
end;

function TllRegistry.DefDate(const AName: string;
                             const ADefault: TDateTime): TDateTime;
begin
  if ValueExists(AName) then
    Result := ReadDateTime(AName)
  else
    Result := ADefault;
end;

function TllRegistry.DefFloat(const AName: string;
                              const ADefault: Double): Double;
begin
  if ValueExists(AName) then
    Result := ReadFloat(AName)
  else
    Result := ADefault;
end;

function TllRegistry.DefInt(const AName: string;
                            const ADefault: Integer): Integer;
begin
  if ValueExists(AName) then
    Result := ReadInteger(AName)
  else
    Result := ADefault;
end;

function TllRegistry.DefStr(const AName, ADefault: string): string;
begin
  if ValueExists(AName) then
    Result := ReadString(AName)
  else
    Result := ADefault;
end;

{ Other Stuff }

function SaveFormPos(const ARegKey: string;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const ASaveFlag: Integer = sfpDefault): Boolean; overload;
var
  Reg: TRegistry;
begin
  Result := False;
  Reg := TRegistry.Create(KEY_ALL_ACCESS);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(ARegKey, True) then begin
      Result := SaveFormPos(Reg, AForm, AOtherInfo, APrefix, ASaveFlag);
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

function SaveFormPos(const AReg: TRegistry;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const ASaveFlag: Integer = sfpDefault): Boolean; overload;
var
  i: Integer;
  sName: string;
begin
  if ASaveFlag <= sfpFalse then
    Result := False
  else if (ASaveFlag = sfpDefault) and
          AReg.ValueExists(APrefix + sfpFlagName) then
    Result := AReg.ReadBool(APrefix + sfpFlagName)
  else
    Result := True; // Default is True
  AReg.WriteBool(APrefix + sfpFlagName, Result);
  if Result then begin
    AReg.WriteBool(APrefix + 'Maximized', AForm.WindowState = wsMaximized);
    if AForm.WindowState = wsNormal then begin
      AReg.WriteInteger(APrefix + 'Top', AForm.Top);
      AReg.WriteInteger(APrefix + 'Left', AForm.Left);
      AReg.WriteInteger(APrefix + 'Height', AForm.Height);
      AReg.WriteInteger(APrefix + 'Width', AForm.Width);
    end;
    if Assigned(AOtherInfo) then begin
      for i := 0 to AOtherInfo.Count - 1 do begin
        if AOtherInfo.ValueFromIndex[i] = sfpOtherInfoDelete then begin
          if AReg.ValueExists(AOtherInfo.Names[i]) then
            AReg.DeleteValue(AOtherInfo.Names[i]);
        end else if AOtherInfo.Names[i] <> '' then
          AReg.WriteString(AOtherInfo.Names[i], AOtherInfo.ValueFromIndex[i])
        else if AOtherInfo[i] <> '' then
          AReg.WriteString(AOtherInfo[i], '');
      end;
    end;
    Result := True;
  end else if ASaveFlag = sfpDelete then begin
    AReg.DeleteValue(APrefix + 'Top');
    AReg.DeleteValue(APrefix + 'Left');
    AReg.DeleteValue(APrefix + 'Height');
    AReg.DeleteValue(APrefix + 'Width');
    AReg.DeleteValue(APrefix + 'Maximized');
    if Assigned(AOtherInfo) then
      for i := 0 to AOtherInfo.Count - 1 do begin
        sName := Coalesce([AOtherInfo.Names[i], AOtherInfo[i]]);
        if (sName <> '') and AReg.ValueExists(sName) then
          AReg.DeleteValue(sName);
      end;
  end;
end;

function LoadFormPos(const ARegKey: string;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const APosOnly: Boolean = False): Boolean; overload;
var
  Reg: TRegistry;
begin
  Result := False;
  Reg := TRegistry.Create(KEY_READ);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKeyReadOnly(ARegKey) then begin
      Result := LoadFormPos(Reg, AForm, AOtherInfo, APrefix, APosOnly);
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

function LoadFormPos(const AReg: TRegistry;
                     const AForm: TForm;
                     const AOtherInfo: TStrings = nil;
                     const APrefix: string = '';
                     const APosOnly: Boolean = False): Boolean; overload;
var
  i: Integer;
  sName: string;
  F: TForm;
begin
  Result := False;
  if AReg.ValueExists(APrefix + 'Top') then begin
    AForm.Top := AReg.ReadInteger(APrefix + 'Top');
    AForm.Left := AReg.ReadInteger(APrefix + 'Left');
    if not APosOnly then begin
      AForm.Height := AReg.ReadInteger(APrefix + 'Height');
      AForm.Width := AReg.ReadInteger(APrefix + 'Width');
    end;
    AForm.MakeFullyVisible; // Fixes partially offscreen or split between monitors
    Result := True;
  end;
  if AReg.ValueExists(APrefix + 'Maximized') then begin
    if AReg.ReadBool(APrefix + 'Maximized') then
      AForm.WindowState := wsMaximized;
    Result := True;
  end;
  if Assigned(AOtherInfo) then
    for i := 0 to AOtherInfo.Count - 1 do begin
      sName := Coalesce([AOtherInfo.Names[i], AOtherInfo[i]]);
      if (sName <> '') and AReg.ValueExists(sName) then
        AOtherInfo[i] := sName + '=' + AReg.ReadString(sName);
    end;
  // If there are no saved settings, try to center it over the owner form:
  if (not Result) and Assigned(AForm.Owner) and (AForm.Owner is TForm) then begin
    F := TForm(AForm.Owner);
    AForm.Top := F.Top + (F.Height div 2) - (AForm.Height div 2);
    AForm.Left := F.Left + (F.Width div 2) - (AForm.Width div 2);
    AForm.MakeFullyVisible;
  end;
end;

{+}
function EraseRegStrings(const ARegKey: string; const AList: TStrings): Boolean;
{
  Takes a list of name-value pairs and deletes the values with the same names
  from an HKCU registry key. Returns a boolean indicating the successful
  deletion of all the existing named settings. If the registry key does not
  exist, it will also return True.

  TStrings.Values doesn't work on blanks -- it removes the name from the list
  when the value is set to an empty string. Any values you want to erase that
  don't currently have a value can just be added to AList as a name without an
  equal sign.

  See Also: SaveRegStrings, LoadRegStrings
-}
var
  Reg: TRegistry;
  i: Integer;
  sName: string;
begin
  Reg := TRegistry.Create(KEY_ALL_ACCESS);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(ARegKey, False) then begin
      Result := True;
      for i := 0 to AList.Count - 1 do begin
        sName := Coalesce([AList.Names[i], AList[i]]);
        if (sName <> '') and Reg.ValueExists(sName) then
          if not Reg.DeleteValue(sName) then
            Result := False;
      end;
      Reg.CloseKey;
    end else
      Result := not Reg.KeyExists(ARegKey);
  finally
    Reg.Free;
  end;
end;

{+}
function LoadRegStrings(const ARegKey: string; const AList: TStrings): Boolean;
{
  Takes a list of name-value pairs and loads the values with the same names from
  an HKCU registry key. Returns a boolean indicating success in opening the key.
  Missing registry values simply leave the default value alone and do not affect
  the return value.

  TStrings.Values doesn't work on blanks -- it removes the name from the list
  when the value is set to an empty string. Any values you want to read that
  don't currently have a value can just be added to AList as a name followed by
  an equal sign. Likewise, any blank values read in are stored in the list this
  way.

  See Also: SaveRegStrings, EraseRegStrings
-}
var
  Reg: TRegistry;
  i: Integer;
  sName: string;
begin
  Result := False;
  Reg := TRegistry.Create(KEY_READ);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKeyReadOnly(ARegKey) then begin
      for i := 0 to AList.Count - 1 do begin
        sName := Coalesce([AList.Names[i], AList[i]]);
        if (sName <> '') and Reg.ValueExists(sName) then
          AList[i] := sName + '=' + Reg.ReadString(sName);
      end;
      Reg.CloseKey;
      Result := True;
    end;
  finally
    Reg.Free;
  end;
end;

{+}
function SaveRegStrings(const ARegKey: string; const AList: TStrings): Boolean;
{
  Takes a list of name-value pairs and saves them to an HKCU registry key as
  strings. Returns a boolean indicating the success of opening the registry key.

  TStrings.Values doesn't work on blanks -- it removes the name from the list
  when the value is set to an empty string. If you want to save a blank value,
  just add the item name to the list without an equal sign/value.

  See Also: LoadRegStrings, EraseRegStrings
-}
var
  Reg: TRegistry;
  i: Integer;
begin
  Result := False;
  Reg := TRegistry.Create(KEY_ALL_ACCESS);
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(ARegKey, True) then begin
      for i := 0 to AList.Count - 1 do
        if AList.Names[i] <> '' then
          Reg.WriteString(AList.Names[i], AList.ValueFromIndex[i])
        else if AList[i] <> '' then
          Reg.WriteString(AList[i], '');
      Result := True;
      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

end.

