unit llSimpleSched;

interface

uses
  SysUtils, Classes, llDates;

type
  {+ /trim=2 TSimpleSchedule}
  TSimpleSchedule = class(TObject)
  {-}
  private
    FStartMin: Cardinal; // Minutes past midnight
    FOnMin: Cardinal;
    FOffMin: Cardinal;
    function GetAlwaysOn: Boolean;
    function GetCurrentState: Boolean;
    function GetScheduleStr: string;
  {+ /trim=2 /addline=OFF TSimpleSchedule}
  public
    constructor Create;
    property AlwaysOn: Boolean read GetAlwaysOn;
    property CurrentState: Boolean read GetCurrentState;
    property ScheduleStr: string read GetScheduleStr;
    property StartMinute: Cardinal read FStartMin;
    property OnMinutes: Cardinal read FOnMin;
    property OffMinutes: Cardinal read FOffMin;
    function SetSchedule(const AStartMin, AOnMin, AOffMin: Cardinal): Boolean;
    function StateAt(const AWhen: TDateTime; out OUntil: TDateTime): Boolean;
  end;
  {+ /trim=OFF /addline=OFF

  TSimpleSchedule is a class to define and monitor a simple schedule with a
  resolution of a minute. The primary application is inside a loop such as a
  service's Execute loop.

  There are three types of schedules: Always on, Start/Stop and Cycle. Always on
  is the default state of the schedule. A Start/Stop schedule is when the
  schedule starts at a given time and stops at another, once per day. A cycle
  schedule is where it is ON for a number of minutes and then OFF for a number
  of minutes, over and over.

  StartMinute is the number of minutes past midnight that the schedule starts.
  The schedule is always ON at the start time.

  OnMinutes is how long the schedule is ON, starting at the start time. When
  Always on, this value would be 1440 - the number of minutes in a day.

  OffMinutes is how long the schedule is OFF after being on. When Always on,
  this value is zero.

  On a Start/Stop schedule, the On+Off minutes must add up to 1440. On a cycle
  schedule, On+Off must be an even divisor of 1440 so the schedule starts at the
  same time each day.

  Use SetSchedule to assign the schedule values. They must all be set at once to
  properly validate them. The function will return True if the values are valid
  and the schedule was updated. If False, the schedule is not changed.
  
  Examples          Start  On    Off   ScheduleStr
  ---------------------------------------------------------------
  Always on         0      1440  0     Always on
  09:00-17:00       540    480   960   Start 09:00 Stop 17:00
  15 minute cycle   0      15    15    Start 00:00 On 15 Off 15
  ---------------------------------------------------------------

  The StateAt function will return the state of the schedule at a given time
  (usually Now) where True is on and False is off. The output parameter OUntil
  will be set to the next time the schedule state will change. If the schedule
  is Always on, OUntil will be set to zero.

  If you just want the state of the schedule at the current time and don't care
  about when it will change next, read the CurrentState property. However, using
  StateAt and checking OUntil against the current time is more efficient in a
  tight loop.

  Usage example:

  var
    MySched: TSimpleSchedule;
    bState: Boolean;
    dtUntil: TDateTime;
  begin
    MySched := TSimpleSchedule.Create;
    try
      if not MySched.SetSchedule(FSchedStart, FSchedOn, FSchedOff) then
        raise Exception.Create('Invalid schedule!');
      bState = MySched.StateAt(Now, dtUntil); // dtUntil = 0 when Always On
      repeat
        if bState then
          DoSomething;
        if (dtUntil > 0) and (Now >= dtUntil) then
          bState := MySched.StateAt(Now, dtUntil);
        Sleep(500);
      until Terminated;
    finally
      MySched.Free;
    end;
  end;

  The llSimpleSchedEdit unit contains a form that can be used to edit a
  schedule. Here is a usage example:

  var
    EditForm: TSchedEditForm;
  begin
    EditForm := TSchedEditForm.Create(Self);
    try
      EditForm.SetSchedule(FSchedStart, FSchedOn, FSchedOff);
      if EditForm.ShowModal = mrOk then begin
        EditForm.GetSchedule(FSchedStart, FSchedOn, FSchedOff);
        ScheduleLabel.Caption := EditForm.ScheduleStr;
      end;
    finally
      EditForm.Free;
    end;
  end;
  -}

implementation

{ TSimpleSchedule }

constructor TSimpleSchedule.Create;
begin
  inherited;
  // Initialize to always on:
  FStartMin := 0;
  FOnMin := MinsPerDay;
  FOffMin := 0;
end;

function TSimpleSchedule.GetAlwaysOn: Boolean;
begin
  Result := FOffMin = 0;
end;

function TSimpleSchedule.GetCurrentState: Boolean;
var
  dtUntil: TDateTime;
begin
  // Use this if you don't care when it will change next
  Result := StateAt(Now, dtUntil);
end;

function TSimpleSchedule.GetScheduleStr: string;
const
  cFmt = 'hh:nn';
begin
  if AlwaysOn then
    Result := 'Always on'
  else if (FOnMin + FOffMin) < MinsPerDay then
    Result := 'Start ' + FormatDateTime(cFmt, MinutesToTime(FStartMin)) +
              ' On ' + IntToStr(FOnMin) +
              ' Off ' + IntToStr(FOffMin)
  else
    Result := 'Start ' + FormatDateTime(cFmt, MinutesToTime(FStartMin)) +
              ' Stop ' + FormatDateTime(cFmt, MinutesToTime(FStartMin + FOnMin));
end;

function TSimpleSchedule.SetSchedule(const AStartMin, AOnMin, AOffMin: Cardinal): Boolean;
var
  m: Cardinal;
begin
  // Returns True if the schedule is valid and was applied.
  // If Off minutes = 0, schedule is Always On. Automatic pass.
  // StartMinute must be < MinsPerDay.
  // On minutes must be > 0.
  // On+Off must be <= MinsPerDay
  // On+Off must be an even divisor of MinsPerDay.
  if AOffMin = 0 then begin // Always on
    Result := True;
    FStartMin := 0;
    FOnMin := MinsPerDay;
    FOffMin := 0;
  end else begin
    m := AOnMin + AOffMin;
    Result := (AStartMin < MinsPerDay) and
              (AOnMin > 0) and
              (m <= MinsPerDay) and
              (MinsPerDay mod m = 0);
    if Result then begin
      FStartMin := AStartMin;
      FOnMin := AOnMin;
      FOffMin := AOffMin;
    end;
  end;
end;

function TSimpleSchedule.StateAt(const AWhen: TDateTime;
                                 out OUntil: TDateTime): Boolean;
var
  dtOn, dtOff: TDateTime;
begin
  if AlwaysOn then begin
    OUntil := 0;
    Result := True;
  end else begin
    // Start at the latest start time <= AWhen
    OUntil := Trunc(AWhen);
    if FStartMin > 0 then
      OUntil := OUntil + MinutesToTime(FStartMin);
    if OUntil > AWhen then
      OUntil := OUntil - 1;
    dtOn := MinutesToTime(FOnMin);
    dtOff := MinutesToTime(FOffMin);
    Result := False; // Turns on at start time
    while OUntil <= AWhen do begin
      Result := not Result;
      if Result then
        OUntil := OUntil + dtOn
      else
        OUntil := OUntil + dtOff;
    end;
  end;
end;

end.
