{*****************************************************************************}
{                                                                             }
{  llStrings.pas                                                              }
{                                                                             }
{  String handling routines                                                   }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llStrings;

interface

uses
  Windows, Classes, SysUtils, StrUtils, Character;

{+ /pad=2 ALPHA_LOWER}
const
  DefaultQuoteChar = '"';
  ALPHA_LOWER = 'abcdefghijklmnopqrstuvwxyz';
  ALPHA_UPPER = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  ALPHA_BOTH  = ALPHA_UPPER + ALPHA_LOWER;
  DIGIT_CHARS = '0123456789';
  ALPHA_NUMERIC = ALPHA_BOTH + DIGIT_CHARS;
  UTF8_BOM = #$EF#$BB#$BF; // Also defined in llFileUtils
  UTF8_BOM_LENGTH = 3;
  PairOpen  = '({[<';
  PairClose = ')}]>';
{+ /pad=0
  Handy constants including digits, the lower & upper case alphabets, and the
  UTF8 Byte-order-marker, which I have to skip over when working with some text
  files (especially csv files created by SQL Server).
-}

type
  {+ /trim=2 TStringObject}
  TStringObject = class(TObject)
  public
    Value: string;
  end;
  {+ /trim=OFF
  A simple object that contains only a string. Handy for storing a string in an
  Object property.

  See also: StringListStringObjectSortCompare
  -}

type
  {+ /trim=2 TllCharCase}
  TllCharCase = (ecNormal, ecUpperCase, ecLowerCase, ecStartCase);
  {+ /trim=OFF
  Based on StdCtrls.TEditCharCase for use with the ChangeCase function.
  This variation adds ecStartCase to capitalize the first letter of each word.
  It is not called ecTitleCase since that technically doesn't capitalize some
  words (see Wikipedia article "Letter case").
  -}

  function ChangeCase(const S: string; const ACase: TllCharCase): string;

  function CloseQuotePos(const S: string; const AIndex: Integer;
                         const Backwards: Boolean = False): Integer;
  function FindPairPos(const S: string; const AIndex: Integer): Integer;
  function FindUnquotedPairPos(const S: string; const AIndex: Integer;
                               const AQuote: Char = DefaultQuoteChar): Integer;

  function CharsOnly(const AInput, AValid: string;
                     const MostlyGood: Boolean = True): string;
  function DigitsOnly(const AInput: string): string;

  function Coalesce(const S: array of string): string;

  function CountIn(const APart, AWhole: string): Integer;
  procedure CountUniqueStrings(const AList: TStrings; const AValue: string;
                               const AModifier: Cardinal = 0);

  function Dequote(const S: string; const AQuote: Char = DefaultQuoteChar): string;
  function DequoteStr(const S: string): string;
  function Enquote(const S: string; const AQuote: Char = DefaultQuoteChar): string;

  function Ellipsis(const S: string; const ALength: Integer;
                    const ARight: Boolean = True): string;

  function EscapedReplace(const ASource, AOld, ANew, AEsc: string;
                          const AFlags: TReplaceFlags;
                          const ARemoveEsc: Boolean): string;

  function FormatDigits(const ADigits, AFormat: string;
                        const AEscape: Char = '\'): string;

  function HasChars(const S, AChars: string): Boolean;

  function IndexOfPartial(const AItems: TStrings; const AValue: String;
                          const ADefault: Integer = -1;
                          const ACaseSensitive: Boolean = True): Integer;
  function IndexOfPrefix(const AItems: TStrings; const AValue: string;
                         const ADefault: Integer = -1;
                         const ACaseSensitive: Boolean = True): Integer;

  function InterleaveStrings(const S1, S2: string): string;

  function IsAlpha(const S: string; const Index: Integer = 0;
                   const AllowSpace: Boolean = False): Boolean;
  function IsBlank(const S: string): Boolean;
  function IsCharsOnly(const S, AValid: string;
                       const AIndex: Integer = 0): Boolean;
  function IsDigit(const S: string; const Index: Integer): Boolean;
  function IsNumeric(const S: string; const Index: Integer = 0;
                     const AllowSymbols: Boolean = False): Boolean;

  function LastPos(const Substr, S: string; const Offset: Integer = 0): Integer;
  function UnquotedLastPos(const Substr, S: string; const Offset: Integer = 0;
                           const AQuote: Char = DefaultQuoteChar): Integer;

  function Leftmost(const S: string; const ACount: Integer): string;
  function Rightmost(const S: string; const ACount: Integer): string;

  function PadC(const AInput: string; const ALen: Integer;
                const AChar: Char = ' '): string;
  function PadL(const AInput: string; const ALen: Integer;
                const AChar: Char = ' '): string;
  function PadR(const AInput: string; const ALen: Integer;
                const AChar: Char = ' '): string;

  procedure ReverseStringList(const AList: TStringList);

  function SimpleDecode(const AValue: string): string;
  function SimpleEncode(const AValue: string): string;

  function Spaces(const ALen: Integer): string;

  function StartsWith(const APart, AWhole: string;
                      const CaseSensitive: Boolean = True): Boolean;
  function EndsWith(const APart, AWhole: string;
                    const CaseSensitive: Boolean = True): Boolean;

  function StartCase(const S: string): string;

  procedure StringListDeepFree(const AList: TStringList; const AMaxDepth: Integer);

  function StringListIntObjectSortCompare(List: TStringList; Index1, Index2: Integer): Integer;
  function StringListStringObjectSortCompare(List: TStringList; Index1, Index2: Integer): Integer;

  function StripParens(const S: string): string;

  {+ /trim=2 TokenList}
  procedure TokenList(const AText, AToken: string; const AList: TStrings); overload;
  procedure TokenList(const AText, AToken: string; const AList: TStrings;
                      const AQuote: Char); overload;
  {
    Breaks up a string at each occurrence of <AToken> and adds these substrings to
    a TStrings descendant, <AList>. The list is not cleared first, so do this
    manually if desired. If the input is blank, nothing is added to the list.

    Tokens longer than a single character are fully supported.

    The two overloaded versions control how quoting is handled. In the version
    without a quote character, quoting is totally ignored and the string is split
    at every instance of the token.

    The version with a quote character will ignore quoted instances of the token
    and preserves all quoting and spaces when adding items to the list. Contrast
    with ParseCSVLine, which unquotes and optionally trims the separated values.

    See also: ParseCSVLine, SimpleCSV
  -}

  function UnquotedPos(const APart, AWhole: string;
                       const AQuote: Char = DefaultQuoteChar;
                       const AOffset: Integer = 1): Integer;
  function UnquotedWordPos(const AWord, AText: string;
                           const AOffset: Integer = 1;
                           const AQuote: Char = DefaultQuoteChar;
                           const AWordChars: string = ''): Integer;

  function WordPos(const AWord, AText: string; const AOffset: Integer = 1;
                   const AWordChars: string = ''): Integer;

implementation

const
  ordSpace = Ord(' '); // #32
  SimpleEncodeStart = '\0x';
  SimpleEncodeStartLen = 3;
  SimpleEncodeEnd = '\';
  SimpleEncodeEndLen = 1;

{+}
function ChangeCase(const S: string; const ACase: TllCharCase): string;
{
  Changes the case of a string based on the value of the second argument. If
  ACase is ecNormal, the string is unchanged. Using ecStartCase will change the
  first letter of each word to uppercase and the rest to lower.

    TllCharCase = (ecNormal, ecUpperCase, ecLowerCase, ecStartCase);

  See also: TllCharCase, StartCase
-}
begin
  if ACase = ecNormal then
    Result := S
  else if ACase = ecUpperCase then
    Result := AnsiUpperCase(S)
  else // ecLowerCase or ecStartCase
    Result := AnsiLowerCase(S);
  if (ACase = ecStartCase) and (Length(S) > 0) then
    Result := StartCase(Result);
end;

{+}
function CloseQuotePos(const S: string; const AIndex: Integer;
                       const Backwards: Boolean = False): Integer;
{
  Given a string and an index to the opening quote mark, the function returns
  the index of the final closing quote character, properly skipping any doubled
  literals. If a closing quote can't be located, the function returns zero.

  If Backwards is True, the index points to the closing quote character and the
  function looks backwards for the opening quote.

  Whatever character is present at the given index is used as the quote
  character. It could be a single quote, double quote, asterisk, or anything.

  Example: CloseQuotePos('cow="moo" dog="bark"', 5) -> 9
                              ^   ^
                     given this   it finds this

  See also: FindPairPos
-}
var
  Quote: Char;
  iLen: Integer;
begin
  Quote := S[AIndex];
  if Backwards then begin
    Result := AIndex - 1;
    while Result > 0 do begin
      if S[Result] = Quote then begin
        if (Result > 1) and (S[Result - 1] = Quote) then begin // a literal?
          Dec(Result);
        end else
          Exit;
      end;
      Dec(Result);
    end;
  end else begin
    iLen := Length(S);
    Result := AIndex + 1;
    while Result <= iLen do begin
      if S[Result] = Quote then begin
        if (Result < iLen) and (S[Result + 1] = Quote) then begin // a literal?
          Inc(Result);
        end else
          Exit;
      end;
      Inc(Result);
    end;
  end;
  Result := 0; // If we haven't exited early, we have failed.
end;

{+}
function FindPairPos(const S: string; const AIndex: Integer): Integer;
{
  Given a string and the index of a character of a matched pair (i.e.
  parentheses, square brackets, curly braces, or angle brackets) the function
  returns the index of its matching opposite, fully allowing for nested pairs.
  The function returns zero if the matching opposite is not found.

  If the character at the index is an opening pair character, the function
  seeks forward in the string for the closing character. If it is a closing pair
  character, it seeks backwards for the opening character.

  If the character at the index position is not one of the valid open/close
  characters, an assertion, range check, or access violation exception will be
  raised depending on compiler options.

  Example: FindPairPos('r = ((x + 5) * (y - z)) / 2', 5) -> 23
                            ^                 ^
                   given this                 it finds this

  Formerly known as CloseBracketPos and ClosePairPos.

  See also: FindUnquotedPairPos, CloseQuotePos
-}
var
  iLen, iStep, iNesting: Integer;
  OpenChar, CloseChar: Char;
begin
  OpenChar := S[AIndex];
  iLen := Pos(OpenChar, PairOpen);
  if iLen > 0 then begin
    CloseChar := PairClose[iLen];
    iStep := 1;
  end else begin
    iLen := Pos(OpenChar, PairClose);
    Assert(iLen > 0, 'Pair character expected at index.');
    CloseChar := PairOpen[iLen];
    iStep := -1;
  end;
  iLen := Length(S);
  iNesting := 1;
  Result := AIndex + iStep;
  while (Result > 0) and (Result <= iLen) do begin
    if S[Result] = OpenChar then
      Inc(iNesting)
    else if S[Result] = CloseChar then begin
      Dec(iNesting);
      if iNesting = 0 then
        Exit;
    end;
    Inc(Result, iStep);
  end;
  Result := 0; // If we haven't exited early, we have failed.
end;

{+}
function FindUnquotedPairPos(const S: string; const AIndex: Integer;
                             const AQuote: Char = DefaultQuoteChar): Integer;
{
  Works just like FindPairPos, but makes sure the matching pair character is not
  inside a quoted section of the string. The character at the starting index is
  assumed to not be quoted.

  See also: FindPairPos, UnquotedPos, CloseQuotePos
-}
const
  cPairOpen  = '({[<';
  cPairClose = ')}]>';
var
  iLen, iStep, iNesting: Integer;
  OpenChar, CloseChar: Char;
  bInQuote: Boolean;
begin
  OpenChar := S[AIndex];
  iLen := Pos(OpenChar, cPairOpen);
  if iLen > 0 then begin
    CloseChar := cPairClose[iLen];
    iStep := 1;
  end else begin
    iLen := Pos(OpenChar, cPairClose);
    Assert(iLen > 0, 'Pair character expected at index.');
    CloseChar := cPairOpen[iLen];
    iStep := -1;
  end;
  bInQuote := False; // Assume our starting index is unquoted
  iLen := Length(S);
  iNesting := 1;
  Result := AIndex + iStep;
  while (Result > 0) and (Result <= iLen) do begin
    if (S[Result] = OpenChar) and not bInQuote then
      Inc(iNesting)
    else if (S[Result] = CloseChar) and not bInQuote then begin
      Dec(iNesting);
      if iNesting = 0 then
        Exit;
    end else if S[Result] = AQuote then begin
      bInQuote := not bInQuote;
    end;
    Inc(Result, iStep);
  end;
  Result := 0; // If we haven't exited early, we have failed.
end;

{+}
function CharsOnly(const AInput, AValid: string;
                   const MostlyGood: Boolean = True): string;
{
  Strips any characters from AInput that don't appear in AValid. The comparison
  is case-sensitive.

  There are two shorthand values available for the valid characters string:
  @@ = All alpha characters, upper and lower
  ## = All numeric digits 0..9

  The optional MostlyGood argument is used to select the algorithm used to do
  the work. By default, an algorithm optimized for mostly good input is used -
  the fewer bad characters, the less work it does. When False, an algorithm that
  does less work when more bad characters are present is used.

  Example:  CharsOnly('FooBar1', '123abcABC') -> 'Ba1'

  See also: DigitsOnly, IsCharsOnly
-}
var
  sValid: string;
  i, j, iLen: Integer;
  function GoodChar: Boolean;
  begin
    Result := Pos(AInput[i], sValid) > 0;
  end;
begin
  sValid := StringReplace(AValid, '##', DIGIT_CHARS, []);
  sValid := StringReplace(sValid, '@@', ALPHA_BOTH, []);
  iLen := Length(AInput);
  // Note: The D5 IsDelimiter help is wrong: the index is 1-based, not 0-based!
  if MostlyGood then begin
    // This algorithm is optimized for mostly good input; less bad = less work
    Result := AInput;
    for i := iLen downto 1 do
      if not GoodChar then
        Delete(Result, i, 1);
  end else begin
    // This algorithm is optimized for more bad input; less good = less work
    SetLength(Result, iLen);
    j := 0;
    for i := 1 to iLen do
      if GoodChar then begin
        Inc(j);
        Result[j] := AInput[i];
      end;
    if j < iLen then
      SetLength(Result, j);
  end;
end;

{+}
function DigitsOnly(const AInput: string): string;
{
  Strips everything except numeric digits from the input string. Handy for
  stripping punctuation from phone number strings. Since this is needed so
  frequently, it's more optimized than calling CharsOnly(AInput, '##').

  Example: DigitsOnly('1(800)555-1212') -> '18005551212'

  See also: CharsOnly, IsNumeric
-}
var
  i: Integer;
begin
  Result := AInput;
  for i := Length(AInput) downto 1 do
    if not IsDelimiter(DIGIT_CHARS, AInput, i) then
      Delete(Result, i, 1);
end;

{+}
function Coalesce(const S: array of string): string;
{
  Much like the SQL function of the same name, it returns the first non-blank
  value from a list of strings. If none qualify, an empty string is returned.

  See also: IsBlank
-}
var
  i: Integer;
begin
  Result := '';
  for i := 0 to High(S) do
    if not IsBlank(S[i]) then begin
      Result := S[i];
      break;
    end;
end;

{+}
function CountIn(const APart, AWhole: string): Integer;
{
  Counts the number of times a substring appears in a string.
  If the substring is a repeating value like 'aa', every possible instance of
  the value will be counted, for example CountIn('aa', 'aaaa') == 3.
-}
var
  i: Integer;
begin
  Result := 0;
  i := PosEx(APart, AWhole);
  while i > 0 do begin
    Inc(Result);
    i := PosEx(APart, aWhole, i + 1);
  end;
end;

{+}
procedure CountUniqueStrings(const AList: TStrings; const AValue: string;
                             const AModifier: Cardinal = 0);
{
  Keep a count of unique strings in a TStrings.Objects properties. Pass the list
  and a string value. If the value does not yet appear in the list, it will be
  added with a count of 1 in the Object value. Otherwise, the count in the
  Object value will be incremented.

  If AModifier is > 0, the count will always be at least AModifier + 1. This is
  typically a very high number, so you can distinguish the modifier from the
  actual count.
-}
var
  i: Integer;
  j: Cardinal;
begin
  i := AList.IndexOf(AValue);
  if i < 0 then
    AList.AddObject(AValue, TObject(AModifier + 1))
  else begin
    j := Cardinal(AList.Objects[i]) + 1;
    if j < AModifier then
      j := j + AModifier;
    AList.Objects[i] := TObject(j);
  end;
end;

{+}
function Dequote(const S: string; const AQuote: Char = DefaultQuoteChar): string;
{
  Works just like AnsiDequotedStr, but that doesn't exist in Delphi 5.
  In addition, D6 & D7 have a bug where AnsiDequotedStr will return a pair of
  empty quotes ("" or '') unchanged instead of returning an empty string!

  See also: Enquote
-}
var
  iLen: Integer;
begin
  Result := S;
  iLen := Length(S);
  if (iLen > 1) and (S[1] = AQuote) and (S[iLen] = AQuote) then begin
    if iLen > 2 then begin
      Delete(Result, iLen, 1);
      Delete(Result, 1, 1);
      Result := StringReplace(Result, AQuote + AQuote, AQuote, [rfReplaceAll]);
    end else
      Result := '';
  end;
end;

{+}
function DequoteStr(const S: string): string;
{
  Works just like Dequote, but uses the first character of the string as the
  quote character. If the string does not start AND end with the same non-
  alphanumeric character, it will be returned unchanged.

  See also: StripParens
-}
var
  iLen: Integer;
begin
  iLen := Length(S);
  if (iLen >= 2) and (S[1] = S[iLen]) and not IsDelimiter(ALPHA_NUMERIC, S, 1) then
    Result := Dequote(S, S[1])
  else
    Result := S;
end;

{+}
function Enquote(const S: string; const AQuote: Char = DefaultQuoteChar): string;
{
  Works much like QuotedStr & AnsiQuotedStr, but in one convenient function
  with a nice, short name.

  See also: Dequote
-}
begin
  Result := AnsiQuotedStr(S, AQuote);
  { AnsiQuotedStr is more efficient than StringReplace
  Result := AQuote +
            StringReplace(S, AQuote, AQuote + AQuote, [rfReplaceAll]) +
            AQuote;
  }
end;

{+}
function Ellipsis(const S: string; const ALength: Integer;
                  const ARight: Boolean = True): string;
{
  If a string is longer than the desired length, it is shortened and has an
  ellipsis (...) placed at the end. If ARight is False, the ellipsis will be
  placed on the left (beginning) of the string.
-}
begin
  if Length(S) <= ALength then
    Result := S
  else if ARight then
    Result := Copy(S, 1, ALength - 3) + '...'
  else
    Result := '...' + Rightmost(S, ALength - 3);
end;

{+}
function EscapedReplace(const ASource, AOld, ANew, AEsc: string;
                        const AFlags: TReplaceFlags;
                        const ARemoveEsc: Boolean): string;
{
  Works just like StringReplace, but allows for escaped values to be skipped in
  the replacement.

  The escape value can be a single character or a string. Any AOld value that is
  preceeded by the AEsc string is not replaced with ANew.

  After completing your replacements, you will want to remove your escape
  sequence from the string as well. Set ARemoveEsc to True to have it removed.
  It is not done automatically in case you have multiple escaped values to
  replace. It only replaces AEsc+AOld with AOld, not every instance of AEsc.

  Known limitation: Escaped escape characters are not supported.
-}
var
  sEscOld, sEscTemp: string;
begin
  // First, see if any escaped replacement values are even present.
  // If not, we can skip a whole bunch of replacement work.
  sEscOld := AEsc + AOld; // Avoid repeated concatenation
  if Pos(AnsiUpperCase(sEscOld), AnsiUpperCase(ASource)) = 0 then begin
    Result := StringReplace(ASource, AOld, ANew, AFlags);
    Exit;
  end;
  // Replace our escaped values with a temporary placeholder unlikely to exist.
  // Make sure our placeholder value doesn't also contain AOld.
  sEscTemp := 'E.S,C;T^E*M-P';
  sEscTemp := StringReplace(sEscTemp, AOld, '', [rfIgnoreCase, rfReplaceAll]);
  Result := StringReplace(ASource, sEscOld, sEscTemp, AFlags);
  // Now do the regular replace of remaining unescaped values:
  Result := StringReplace(Result, AOld, ANew, AFlags);
  // Restore the placeholders to their original value:
  Result := StringReplace(Result, sEscTemp, sEscOld, AFlags);
  // Remove the escape strings if requested:
  if ARemoveEsc then
    Result := StringReplace(Result, sEscOld, AOld, AFlags);
end;

{+}
function FormatDigits(const ADigits, AFormat: string;
                      const AEscape: Char = '\'): string;
{
  Right-justifies/overlays a string of characters (usually digits) into a format
  string. Especially handy for phone number/SSN formatting, but it could
  conceivably be used on any type of input.

  Ex: FormatDigits('6025551212', '(099)999-9999')  ->  '(602)555-1212'
      FormatDigits('5551212',    '(099)999-9999')  ->  '555-1212'
      FormatDigits('6025551212', '999.999.9999')   ->  '602.555.1212'
      FormatDigits('bar',        'foo')            ->  'foobar'

  All digit characters are always output even if the format string is shorter
  or blank. Output stops when you run out of digit characters, even though there
  may be more format string remaining.

  Format string rules:
  9 = Replace this character with a character from the digit string.
  0 = Same as 9 but always includes the next format character to the left, even
      if you have run out of digit characters.
  * = Any other character is copied to the output as a literal.

  To output a literal 0 or 9, precede it with the escape character. The default
  escape character is a backslash, but it can be changed if you need backslashes
  in your output.

  Example: FormatDigits('123456', '999\0999') -> '1230456'
-}
var
  d, f, r: Integer;
begin
  Assert((AEscape <> '0') and (AEscape <> '9'), 'Escape character cannot be 0 or 9.');
  d := Length(ADigits);
  f := Length(AFormat);
  r := d + f;
  SetLength(Result, r);
  while d > 0 do begin
    if f > 0 then begin
      if (AFormat[f] = '0') or (AFormat[f] = '9') then begin
        if (f > 1) and (AFormat[f - 1] = AEscape) then begin // literal 0 or 9?
          Result[r] := AFormat[f];
          Dec(f);
        end else begin
          Result[r] := ADigits[d];
          Dec(d);
          if AFormat[f] = '0' then begin
            Dec(r);
            Dec(f);
            Result[r] := AFormat[f];
          end;
        end;
      end else
        Result[r] := AFormat[f];
      Dec(r);
      Dec(f);
    end else begin
      Result[r] := ADigits[d];
      Dec(r);
      Dec(d);
    end;
  end;
  Delete(Result, 1, r);
end;

{+}
function HasChars(const S, AChars: string): Boolean;
{
  Returns True if the string contains at least one of the characters present in
  AChars. It can contain other characters as well. Returns False if AChars is
  blank.
-}
var
  i: Integer;
begin
  if AChars <> '' then begin
    Result := True;
    for i := 1 to Length(S) do
      if Pos(S[i], AChars) > 0 then
        Exit;
  end;
  Result := False;
end;

{+}
function IndexOfPartial(const AItems: TStrings; const AValue: string;
                        const ADefault: Integer = -1;
                        const ACaseSensitive: Boolean = True): Integer;
{
  Returns the index of the first value in AItems that starts with AValue.
  If there is no match, the function returns ADefault (-1).

  See also: IndexOfPrefix, StartsWith
-}
var
  i: Integer;
begin
  Result := ADefault;
  for i := 0 to AItems.Count - 1 do
    if StartsWith(AValue, AItems[i], ACaseSensitive) then begin
      Result := i;
      Break;
    end;
end;

{+}
function IndexOfPrefix(const AItems: TStrings; const AValue: string;
                       const ADefault: Integer = -1;
                       const ACaseSensitive: Boolean = True): Integer;
{
  Returns the index of the AItems value that is the prefix/start of AValue.
  If AValue does not start with any of the AItems strings, the function returns
  ADefault (-1).

  See also: IndexOfPartial, StartsWith
-}
var
  i: Integer;
begin
  Result := ADefault;
  for i := 0 to AItems.Count - 1 do
    if StartsWith(AItems[i], AValue, ACaseSensitive) then begin
      Result := i;
      Break;
    end;
end;

{+}
function InterleaveStrings(const S1, S2: string): string;
{
  Combines two strings by alternating characters.

  Example: InterleaveStrings('abc', '1234') -> 'a1b2c34'
-}
var
  i, r, L1, L2, MaxLen: Integer;
begin
  L1 := Length(S1);
  L2 := Length(S2);
  SetLength(Result, L1 + L2);
  if L1 >= L2 then
    MaxLen := L1
  else
    MaxLen := L2;
  r := 1;
  for i := 1 to MaxLen do begin
    if i <= L1 then begin
      Result[r] := S1[i];
      Inc(r);
    end;
    if i <= L2 then begin
      Result[r] := S2[i];
      Inc(r);
    end;
  end;
end;

{+}
function IsAlpha(const S: string; const Index: Integer = 0;
                 const AllowSpace: Boolean = False): Boolean;
{
  Returns True if the character at the requested Index is an ASCII letter, or a
  space if AllowSpace is True. If Index is zero, the entire string is checked.
  A blank string always returns False.

  See also: IsCharsOnly, IsDigit, IsNumeric
-}
var
  sValid: string;
begin
  sValid := ALPHA_BOTH;
  if AllowSpace then
    sValid := sValid + ' ';
  Result := IsCharsOnly(S, sValid, Index);
end;

{+}
function IsBlank(const S: string): Boolean;
{
  Returns True if the string is empty or contains only "trimmable" characters.
  If you're just going to discard it if blank, why waste time creating another
  copy by actually trimming it?
-}
var
  i: Integer;
begin
  Result := False;
  for i := 1 to Length(S) do
    if Ord(S[i]) > ordSpace then
      Exit;
  Result := True;
end;

{+}
function IsCharsOnly(const S, AValid: string;
                     const AIndex: Integer = 0): Boolean;
{
  Returns a boolean indicating whether the string would be returned unchanged
  from a call to CharsOnly. In other words, it contains only the characters in
  AValid. If AIndex is non-zero, only the character at the given index is
  checked. An empty string returns False.

  See also: CharsOnly, IsAlpha, IsDigit, IsNumeric
-}
var
  sValid: string;
  i: Integer;
  function BadChar(const X: Integer): Boolean;
  begin
    Result := Pos(S[X], sValid) = 0;
  end;
begin
  Result := False;
  if S = '' then
    Exit;
  sValid := StringReplace(AValid, '##', DIGIT_CHARS, []);
  sValid := StringReplace(sValid, '@@', ALPHA_BOTH, []);
  if AIndex = 0 then begin
    for i := 1 to Length(S) do
      if BadChar(i) then
        Exit;
  end else
    if BadChar(AIndex) then
      Exit;
  Result := True;
end;

{+}
function IsDigit(const S: string; const Index: Integer): Boolean;
{
  Returns true if the character at the given index is a number 0-9.
  An empty string or index out of range returns False.

  If you want to check the whole string, use IsNumeric.
-}
begin
  Result := IsDelimiter(DIGIT_CHARS, S, Index);
end;

{+}
function IsNumeric(const S: string; const Index: Integer = 0;
                   const AllowSymbols: Boolean = False): Boolean;
{
  Returns True if the character at the given Index is a numeric digit. If the
  AllowSymbols argument is True, it also allows a decimal and leading +/-. If
  Index is zero, the entire string is checked. An empty string returns False.

  See also: IsCharsOnly, IsDigit, DigitsOnly
-}
var
  sValid: string;
begin
  sValid := DIGIT_CHARS;
  if AllowSymbols then
    sValid := sValid + '.+-';
  Result := IsCharsOnly(S, sValid, Index);
  if Result and AllowSymbols and (Index = 0) then begin
    // Make sure the symbols are valid
    if LastPos('+', S) > 1 then
      Result := False;
    if LastPos('-', S) > 1 then
      Result := False;
    if CountIn('.', S) > 1 then
      Result := False;
  end;
end;

{+}
function LastPos(const Substr, S: string; const Offset: Integer = 0): Integer;
{
  Essentially a backwards-looking PosEx.

  Returns the position of the last occurrence of the substring, or zero if not
  found. Accepts a starting offset like PosEx, the default of zero being the
  equivalent of Length(S). Actually, any value outside 1..Length(S) will become
  Length(S).

  Examples: LastPos('foo', 'foobarfoobar')    -> 7
            LastPos('foo', 'foobarfoobar', 6) -> 1

  See also: UnquotedLastPos
-}
var
  iSub, i, j: Integer;
begin
  Result := 0;
  iSub := Length(Substr);
  if iSub = 0 then
    Exit; // Can't look for a null string!
  i := Length(S);
  // Offset is given as the start of the substr, so adjust accordingly since
  // we need to start looking from the end:
  if (Offset > 0) and (Offset < i - iSub + 1) then
    i := Offset + iSub - 1;
  while i >= iSub do begin
    if S[i] = Substr[iSub] then begin
      j := 1;
      while (j < iSub) and (S[i - j] = Substr[iSub - j]) do
        Inc(j);
      if j = iSub then begin
        Result := i - iSub + 1;
        Exit;
      end;
    end;
    Dec(i);
  end;
end;

{+}
function UnquotedLastPos(const Substr, S: string; const Offset: Integer = 0;
                         const AQuote: Char = DefaultQuoteChar): Integer;
{
  Works just like LastPos, but finds only substrings that are not contained in
  quotes. The quote character defaults to ". The substring at the starting
  offset is assumed to not be quoted.

  See also: LastPos
-}
var
  iSub, i, j: Integer;
  bInQuote: Boolean;
begin
  Result := 0;
  iSub := Length(Substr);
  if iSub = 0 then
    Exit; // Can't look for a null string!
  i := Length(S);
  // Offset is given as the start of the substr, so adjust accordingly since
  // we need to start looking from the end:
  bInQuote := False;
  if (Offset > 0) and (Offset < i - iSub + 1) then
    i := Offset + iSub - 1;
  while i >= iSub do begin
    if (S[i] = Substr[iSub]) and not bInQuote then begin
      j := 1;
      while (j < iSub) and (S[i - j] = Substr[iSub - j]) do
        Inc(j);
      if j = iSub then begin
        Result := i - iSub + 1;
        Exit;
      end;
    end else if S[i] = AQuote then
      bInQuote := not bInQuote;
    Dec(i);
  end;
end;

{+}
function Leftmost(const S: string; const ACount: Integer): string;
{
  Returns the leftmost <count> characters of the input string. If the input is
  shorter than the count, all of the input will be returned. Sounds just like
  the Copy function, right? Not quite! A negative count will return all but the
  right # characters.

  Examples: Leftmost('foobar', 3)  = 'foo'
            Leftmost('foobar', 10) = 'foobar'
            Leftmost('foobar', -1) = 'fooba'

  See also: Rightmost
-}
begin
  if ACount >= 0 then
    Result := Copy(S, 1, ACount)
  else
    Result := Copy(S, 1, Length(S) + ACount);
end;

{+}
function Rightmost(const S: string; const ACount: Integer): string;
{
  Similar to the D7+ StrUtils.RightStr function, this returns the rightmost
  <count> characters of the input string. If the input is shorter than the
  count, all of the input will be returned. This variation also allows for a
  negative count that returns all but the left # characters.

  Examples: Rightmost('foobar', 3)  = 'bar'
            Rightmost('foobar', 10) = 'foobar'
            Rightmost('foobar', -1) = 'oobar'

  See also: Leftmost
-}
var
  iLen: Integer;
begin
  if ACount >= 0 then begin
    iLen := Length(S);
    if ACount >= iLen then
      Result := S
    else
      Result := Copy(S, (iLen - ACount) + 1, ACount);
  end else
    Result := Copy(S, 1 + Abs(ACount), MaxInt);
end;

{+}
function PadC(const AInput: string; const ALen: Integer;
              const AChar: Char = ' '): string;
{
  Pads a string equally on both sides with <AChar> to a length of <ALen>. If
  <AInput> is longer than <ALen>, the middle <ALen> characters will be returned.

  If <ALen> is negative, the string will only be padded if shorter than the
  desired length. Longer strings will be left alone.

  See also: PadL, PadR
-}
var
  iDiff, iHalf: Integer;
begin
  iDiff := Abs(ALen) - Length(AInput);
  if (ALen >= 0) or (iDiff > 0) then begin
    iHalf := Abs(iDiff) div 2;
    if iDiff > 0 then
      Result := StringOfChar(AChar, (iDiff - iHalf)) +
                AInput +
                StringOfChar(AChar, iHalf)
    else
      Result := Copy(AInput, iHalf + 1, ALen);
  end else
    Result := AInput;
end;

{+}
function PadL(const AInput: string; const ALen: Integer;
              const AChar: Char = ' '): string;
{
  Pads a string on the left with <AChar> to a length of <ALen>. If <AInput> is
  longer than <ALen>, the rightmost <ALen> characters of <AInput> will be
  returned.

  If <ALen> is negative, the string will only be padded if shorter than the
  desired length. Longer strings will be left alone.

  See also: PadC, PadR
-}
var
  iDiff: Integer;
begin
  iDiff := Abs(ALen) - Length(AInput);
  if (ALen >= 0) or (iDiff > 0) then begin
    if iDiff > 0 then
      Result := StringOfChar(AChar, iDiff) + AInput
    else
      Result := Copy(AInput, Abs(iDiff) + 1,  ALen);
  end else
    Result := AInput;
end;

{+}
function PadR(const AInput: string; const ALen: Integer;
              const AChar: Char = ' '): string;
{
  Pads a string on the right with <AChar> to a length of <ALen>. If <AInput> is
  longer than <ALen>, the leftmost <ALen> characters of <AInput> will be
  returned.

  If <ALen> is negative, the string will only be padded if shorter than the
  desired length. Longer strings will be left alone.

  See also: PadC, PadL
-}
var
  iDiff: Integer;
begin
  iDiff := Abs(ALen) - Length(AInput);
  if (ALen >= 0) or (iDiff > 0) then begin
    if iDiff > 0 then
      Result := AInput + StringOfChar(AChar, iDiff)
    else
      Result := Copy(AInput, 1, ALen);
  end else
    Result := AInput;
end;

{+}
procedure ReverseStringList(const AList: TStringList);
{
  Simply reverses the order of a TStringList. Handy if you need a list sorted in
  descending order, but don't want to write a custom sort function.
  If AList.Sorted is True when called, it will be set to False.

  See also: StringListIntObjectSortCompare, StringListStringObjectSortCompare
-}
var
  i, j: Integer;
begin
  AList.Sorted := False;
  AList.BeginUpdate;
  i := 0;
  j := AList.Count - 1;
  while i < j do begin
    AList.Exchange(i, j);
    Inc(i);
    Dec(j);
  end;
  AList.EndUpdate;
end;

{+}
function SimpleDecode(const AValue: string): string;
{
  Reverses the encoding done by SimpleEncode.

  Example: SimpleDecode('foo\0x0D0A\bar') -> 'foo'#13#10'bar'

  See also: SimpleEncode
-}
const
  ordJunk = Ord('?');
var
  iStart, iEnd, i, j, k: Integer;
  sOld, sNew: string;
begin
  Result := AValue;
  iStart := Pos(SimpleEncodeStart, Result);
  while iStart > 0 do begin
    iEnd := Pos(SimpleEncodeEnd, Result, iStart + 1);
    if iEnd = 0 then
      raise EConvertError.Create('Unterminated SimpleEncode sequence');
    sOld := Copy(Result, iStart + SimpleEncodeStartLen,
                         (iEnd - iStart) - SimpleEncodeStartLen);
    j := Length(sOld);
    if j > 0 then begin
      // If odd in length (sketchy!), add a leading zero
      if j mod 2 <> 0 then begin
        sOld := '0' + sOld;
        Inc(j);
      end;
      SetLength(sNew, j div 2);
      i := 1;
      k := 1;
      while i < j do begin
        sNew[k] := Char( StrToIntDef('$' + Copy(sOld, i, 2), ordJunk) );
        Inc(k);
        Inc(i, 2);
      end;
    end else // Zero length means it was an escaped SimpleEncodeStart!
      sNew := SimpleEncodeStart;
    Delete(Result, iStart, (iEnd - iStart) + SimpleEncodeEndLen);
    Insert(sNew, Result, iStart);
    iStart := PosEx(SimpleEncodeStart, Result, iStart + 1);
  end;
end;

{+}
function SimpleEncode(const AValue: string): string;
{
  Changes any low binary characters (under ASCII 32/Hex 0x20) to encoded values,
  leaving the string as human-readable as possible.

  The low binary characters are encoded as hex and surrounded by "\0x" and "\".
  Multiple adjacent characters will create a run of encoding between a single
  set of start/end markers.

  Example: SimpleEncode('foo'#13#10'bar') -> 'foo\0x0D0A\bar'

  See also: SimpleDecode
-}
var
  i, j, k: Integer;
  sHex: string;
begin
  // If the value actually contains SimpleEncodeStart, escape it by following it
  // immediately with SimpleEncodeEnd, creating an "empty" encoded section.
  Result := StringReplace(AValue, SimpleEncodeStart,
    SimpleEncodeStart + SimpleEncodeEnd, [rfReplaceAll]);
  i := Length(Result);
  while i > 0 do begin
    k := Ord(Result[i]);
    if k < 32 then begin
      // If it's a run of binary, it can all go in one block \0xdddddd...\
      j := i + 1;
      sHex := SimpleEncodeEnd;
      repeat
        sHex := IntToHex(k, 2) + sHex;
        Dec(i);
        if i > 0 then
          k := Ord(Result[i]);
      until (i = 0) or (k >= 32);
      sHex := SimpleEncodeStart + sHex;
      Inc(i);
      Delete(Result, i, j - i);
      Insert(sHex, Result, i);
    end;
    Dec(i);
  end;
end;

{+}
function Spaces(const ALen: Integer): string;
{
  Returns a string of <ALen> spaces.
-}
begin
  Result := StringOfChar(' ', ALen);
end;

{+}
function StartCase(const S: string): string;
{
  Capitalizes the first letter of each word in a string. Words are assumed to be
  separated by whitespace. Characters other than the first of each word are not
  modified. If you want the other characters in each word set to lower case,
  provide the function with all lower case input, i.e. StartCase(LowerCase(S)).

  Note: It is not called TitleCase since that technically doesn't capitalize
  some words (see Wikipedia article "Letter case").

  See also: TllCharCase, ChangeCase
-}
var
  i: Integer;
begin
  Result := S;
  if Result <> '' then begin
    Result[1] := UpCase(Result[1]);
    for i := 2 to Length(Result) do
      if Ord(Result[i - 1]) <= ordSpace then
        Result[i] := UpCase(Result[i]);
  end;
end;

{+}
function StartsWith(const APart, AWhole: string;
                    const CaseSensitive: Boolean = True): Boolean;
{
  Returns True if <APart> is the beginning of <AWhole>. It will also return
  True if <APart> is null (''). CaseSensitive is self-explanatory.
-}
var
  p, w: Integer;
begin
  p := Length(APart);
  if p = 0 then
    Result := True
  else if Length(AWhole) >= p then begin
    w := 1;
    if CaseSensitive then begin
      while (w <= p) and (APart[w] = AWhole[w]) do
        Inc(w);
    end else begin
      while (w <= p) and (APart[w].ToUpper = AWhole[w].ToUpper) do
        Inc(w);
    end;
    Result := w > p;
  end else
    Result := False;
end;

function EndsWith(const APart, AWhole: string;
                  const CaseSensitive: Boolean = True): Boolean;
{
  Returns True if <APart> is the end of <AWhole>. It will also return True if
  <APart> is null (''). CaseSensitive is self-explanatory.
-}
var
  p, w: Integer;
begin
  p := Length(APart);
  if p = 0 then
    Result := True
  else begin
    w := Length(AWhole);
    if w >= p then begin
      if CaseSensitive then begin
        while (p > 0) and (APart[p] = AWhole[w]) do begin
          Dec(p);
          Dec(w);
        end;
      end else begin
        while (p > 0) and (APart[p].ToUpper = AWhole[w].ToUpper) do begin
          Dec(p);
          Dec(w);
        end;
      end;
      Result := p = 0;
    end else
      Result := False;
  end;
end;

{+}
procedure StringListDeepFree(const AList: TStringList; const AMaxDepth: Integer);
{
  Traverses a TStringList, freeing any assigned objects and emptying the list.
  If AMaxDepth is > 0, it also recurses into those objects that are TStringLists
  and does the same until AMaxDepth is reached. If AMaxDepth is negative,
  recursion continues until the bottom is reached.
-}
var
  i: Integer;
begin
  for i := AList.Count - 1 downto 0 do begin
    if Assigned(AList.Objects[i]) then begin
      if (AMaxDepth <> 0) and (AList.Objects[i] is TStringList) then
        StringListDeepFree(TStringList(AList.Objects[i]), AMaxDepth - 1);
      AList.Objects[i].Free;
    end;
    AList.Delete(i);
  end;
end;

{+}
function StringListIntObjectSortCompare(List: TStringList;
                                        Index1, Index2: Integer): Integer;
{
  Use with TStringList.CustomSort to sort a stringlist by the integer values of
  its Objects properties. Items that compare equally on the object value will be
  compared by string value and will honor the list's CaseSensitive setting.

  To sort descending, just follow the sort with a call to ReverseStringList.

  See also: ReverseStringList, StringListStringObjectSortCompare, ListFiles
-}
var
  i1, i2: Integer;
begin
  i1 := Integer(List.Objects[Index1]);
  i2 := Integer(List.Objects[Index2]);
  if i1 > i2 then
    Result := 1
  else if i1 < i2 then
    Result := -1
  else if List.CaseSensitive then // Sort by string value if integers are equal
    Result := AnsiCompareStr(List[Index1], List[Index2])
  else
    Result := AnsiCompareText(List[Index1], List[Index2]);
end;

{+}
function StringListStringObjectSortCompare(List: TStringList;
                                           Index1, Index2: Integer): Integer;
{
  Use with TStringList.CustomSort to sort a stringlist by the TStringObject
  values of its Objects properties. Items that compare equally on the object
  value will be compared by string value. Both comparisons will honor the list's
  CaseSensitive setting. Unassigned objects will be compared as if they contain
  an empty string.

  See also: TStringObject, ReverseStringList, StringListIntObjectSortCompare
-}
var
  s1, s2: string;
  function GetVal(const AIndex: Integer): string;
  begin
    if Assigned(List.Objects[AIndex]) and
       (List.Objects[AIndex] is TStringObject) then
      Result := TStringObject(List.Objects[AIndex]).Value
    else
      Result := '';
  end;
begin
  s1 := GetVal(Index1);
  s2 := GetVal(Index2);
  if List.CaseSensitive then begin
    Result := AnsiCompareStr(s1, s2);
    if Result = 0 then
      Result := AnsiCompareStr(List[Index1], List[Index2]);
  end else begin
    Result := AnsiCompareText(s1, s2);
    if Result = 0 then
      Result := AnsiCompareText(List[Index1], List[Index2]);
  end;
end;

{+}
function StripParens(const S: string): string;
{
  Strips parentheses, square brackets, curly braces, angle brackets or quote
  characters from the input string if the first character in the string is one
  of those characters and the last character is the matching pair.

  See also: DequoteStr
-}
var
  i, iLen: Integer;
begin
  Result := S;
  iLen := Length(S);
  if iLen < 2 then
    Exit;
  i := Pos(S[1], PairOpen);
  if (i > 0) and (S[iLen] = PairClose[i]) then
    Result := Copy(S, 2, iLen - 2)
  else if (S[1] = '"') or (S[1] = '''') then
    Result := DequoteStr(S);
end;

procedure TokenList(const AText, AToken: string; const AList: TStrings); overload;
var
  i, j, t: Integer;
begin
  // This version ignores any quoting
  if AText = '' then
    Exit;
  t := Length(AToken);
  j := 1;
  i := Pos(AToken, AText);
  while i > 0 do begin
    AList.Add(Copy(AText, j, i - j));
    j := i + t;
    i := PosEx(AToken, AText, j);
  end;
  AList.Add(Copy(AText, j, MaxInt));
end;

procedure TokenList(const AText, AToken: string; const AList: TStrings;
                    const AQuote: Char); overload;
var
  i, j, t: Integer;
begin
  // This version honors and preserves quoting.
  if AText = '' then
    Exit;
  t := Length(AToken);
  j := 1;
  i := UnquotedPos(AToken, AText, AQuote);
  while i > 0 do begin
    AList.Add(Copy(AText, j, i - j));
    j := i + t;
    i := UnquotedPos(AToken, AText, AQuote, j);
  end;
  AList.Add(Copy(AText, j, MaxInt));
end;

{+}
function UnquotedPos(const APart, AWhole: string;
                     const AQuote: Char = DefaultQuoteChar;
                     const AOffset: Integer = 1): Integer;
{
  Returns the position of the substring within the larger string, but only if
  the substring is not enclosed in quotes. The function returns zero if not
  found. The quote character defaults to ".

  Example: UnquotedPos('foo', '"foobar"=foo') -> 10
-}
var
  i: Integer;
  bInQuote: Boolean;
begin
  i := 1;
  bInQuote := False;
  Result := PosEx(APart, AWhole, AOffset);
  while Result > 0 do begin
    while i < Result do begin
      if AWhole[i] = AQuote then
        bInQuote := not bInQuote;
      Inc(i);
    end;
    if not bInQuote then
      break;
    Result := PosEx(APart, AWhole, Result + 1);
  end;
end;

{+}
function UnquotedWordPos(const AWord, AText: string;
                         const AOffset: Integer = 1;
                         const AQuote: Char = DefaultQuoteChar;
                         const AWordChars: string = ''): Integer;
{
  A combination of UnquotedPos and WordPos to find the position of AWord within
  AText as an unquoted, separate word. See the other functions for argument
  information. Returns 0 if not found.

  See also: UnquotedPos, WordPos
-}
var
  i, j: Integer;
begin
  // If WordPos and UnquotedPos agree, that's our answer!
  Result := 0;
  i := AOffset - 1;
  if i < 0 then
    i := 0;
  repeat
    i := WordPos(AWord, AText, i + 1, AWordChars);
    if i > 0 then begin
      j := UnquotedPos(AWord, AText, AQuote, i);
      if i = j then
        Result := i;
    end else
      j := i;
  until i = j;
end;

{+}
function WordPos(const AWord, AText: string; const AOffset: Integer = 1;
                 const AWordChars: string = ''): Integer;
{
  Returns the position of the first "separate word" substring of AWord found in
  AText. The word must be surrounded by non-word characters to be considered
  separate, otherwise it is considered to be part of a larger word.

  By default, letters, numbers and underscores make up the characters that are
  considered part of "words". If you need different ones, use the optional
  WordChars argument to supply the entire list, including both upper and lower
  case letters if needed.

  Offset indicates where to start searching within AText, just like the Offset
  argument to PosEx.

  Examples:  WordPos('delete', 'Do not delete it!') -> 8
             WordPos('delete', 'I deleted it!')     -> 0 (part of a larger word)
-}
var
  sWC: String;
  i, j, k, t, w, tw: Integer;
  bMatch: Boolean;
  function NotWordChar(const AIndex: Integer): Boolean;
  begin
    Result := Pos(AText[AIndex], sWC) = 0;
  end;
begin
  Result := 0;
  w := Length(AWord);
  if w = 0 then
    Exit; // Can't search for a null word!
  if AWordChars = '' then
    sWC := ALPHA_BOTH + '_' + DIGIT_CHARS
  else
    sWC := AWordChars;
  i := AOffset;
  if i < 1 then
    i := 1;
  t := Length(AText);
  tw := t - (w - 1); // No sense checking past where AWord will fit!
  while i <= tw do begin
    // First character of our word?
    if AText[i] = AWord[1] then begin
      // Check the character to the left:
      if i > 1 then
        bMatch := NotWordChar(i - 1)
      else
        bMatch := True; // (start of line)
      if bMatch then begin
        // See that the rest of the word matches:
        j := i + 1;
        k := 2;
        while bMatch and (k <= w) do begin
          bMatch := AText[j] = AWord[k];
          Inc(j);
          Inc(k);
        end;
        if bMatch then begin
          // Check the character to the right:
          if j <= t then
            bMatch := NotWordChar(j);
          // Did all tests pass?
          if bMatch then begin
            Result := i;
            Exit;
          end;
        end;
      end;
    end;
    Inc(i);
  end;
end;

end.

