{*****************************************************************************}
{                                                                             }
{  llWinApi.pas                                                               }
{                                                                             }
{  Convenient wrappers for WinAPI and other windowing functions.              }
{                                                                             }
{  See LogicLib.dpk for copyright/licensing information.                      }
{                                                                             }
{*****************************************************************************}

unit llWinApi;

interface

uses
  Windows, SysUtils, Messages, Registry, SHFolder, ShlObj;

  function BrowseForFolder(const AHandle: HWND; const ACaption: string;
                           var AFolder: string): Boolean;

  {+ /trim=2}
  function ExecuteCommand(const ACommand: string; const AWaitMS: Integer;
                          const AShowWindow: Word = SW_SHOWNORMAL;
                          const ACreateFlags: Cardinal =
               CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS): Cardinal; overload;

  function ExecuteCommand(const ACommand: string; const AWait: Boolean;
                          const AShowWindow: Word = SW_SHOWNORMAL;
                          const ACreateFlags: Cardinal =
               CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS): Cardinal; overload;
  {
    Executes a command and optionally waits for it to finish.

    If waiting, the function returns the exit code of the process or WAIT_TIMEOUT
    if it times out before finishing. Otherwise, it returns zero (NO_ERROR,
    ERROR_SUCCESS) or GetLastError based on the success of CreateProcess.

    There are two overloads: An integer AWaitMS argument, and a boolean AWait.
    The integer variation takes the number of milliseconds to wait, 0 to not wait,
    or -1 to wait forever. The boolean overload calls the integer one using -1 for
    True and 0 for False.

    Adapted from Peter Below's WinExecAndWait32V2.

    See also: GetLastErrorStr
  -}

  function ExpandEnvars(const S: string): string;

  function GetAssociatedProgram(const AFile: string): string;
  function GetAssociatedProgramByExt(const AExt: string): string;

  function GetLastErrorStr(const AErrNo: Cardinal = 0;
                           const AIncludeErrNo: Boolean = False): string;

  function GetShellFolder(const CSIDL: Integer;
                          const Slash: Boolean = False): string;

  function GetTempDir(const ASlash: Boolean = False): string;

  function GetTempFile(const APath: string;
                       const APrefix: string = 'TMP'): string;

  function GetWinCompName: string;
  function GetWinDir(const Slash: Boolean = False): string;
  function GetWinUserName: string;

  function KillDirectory(const ADir: string): Boolean;

  function OpenAs_RunDLL(Wnd: HWnd; Instance: HInst; CmdLine: PChar;
    nCmdShow: Integer): Bool; stdcall; external 'shell32.dll' name 'OpenAs_RunDLLA';

  procedure OpenWith(const AFileName: string; const hWindow: HWND = 0;
                     const ShowCmd: Integer = SW_SHOWNORMAL);

  procedure PostKey(hWindow: HWND; Key: Word);

  function ReplaceFile(const ATarget, ASource: string): Boolean;
  function RecycleFile(const AFile: string): Boolean;

implementation

uses ShellAPI, llDates, llFiles, llFileUtils, llStrings;

// Used only by BrowseForFolder
function BrowseCallbackProc(hwnd: HWND; uMsg: UINT; lParam: LPARAM; lpData: LPARAM): Integer; stdcall;
begin
  if (uMsg = BFFM_INITIALIZED) then
    SendMessage(hwnd, BFFM_SETSELECTION, 1, lpData);
  BrowseCallbackProc := 0;
end;

{+}
function BrowseForFolder(const AHandle: HWND; const ACaption: string;
                         var AFolder: string): Boolean;
{
  A handy wrapper around SHBrowseForFolder.  Makes a nice, modern-looking
  replacement for SelectDirectory.

  Returns True if the user selects a directory.  The initial directory is
  passed as AFolder, which also holds the new value when successful.
-}
const
  BIF_STATUSTEXT           = $0004;
  BIF_NEWDIALOGSTYLE       = $0040;
  BIF_RETURNONLYFSDIRS     = $0080;
  BIF_SHAREABLE            = $0100;
  BIF_USENEWUI             = BIF_EDITBOX or BIF_NEWDIALOGSTYLE;
var
  BrowseInfo: TBrowseInfo;
  ItemIDList: PItemIDList;
  TempIDList: PItemIDList;
  PathBuf: array[0..MAX_PATH] of Char;
  DisplayBuf: array[0..MAX_PATH] of Char;
begin
  Result := False;
  SHGetSpecialFolderLocation(AHandle, CSIDL_DRIVES, TempIDList);
  BrowseInfo.hwndOwner := AHandle;
  BrowseInfo.pidlRoot := TempIDList;
  BrowseInfo.pszDisplayName := DisplayBuf;
  BrowseInfo.lpszTitle := PChar(ACaption);
  BrowseInfo.lpfn := @BrowseCallbackProc;
  BrowseInfo.lParam := LongInt(PChar(AFolder));
  BrowseInfo.ulFlags := BIF_RETURNONLYFSDIRS or BIF_USENEWUI;

  ItemIDList := SHBrowseForFolder(BrowseInfo);

  if ItemIDList <> nil then begin
    if SHGetPathFromIDList(ItemIDList, PathBuf) then
    begin
      AFolder := PathBuf;
      Result := True
    end;
    GlobalFreePtr(ItemIDList);
  end;
end;

function ExecuteCommand(const ACommand: string; const AWaitMS: Integer;
                        const AShowWindow: Word = SW_SHOWNORMAL;
                        const ACreateFlags: Cardinal =
               CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS): Cardinal; overload;
  function WaitForIt(hProcess: THandle): Boolean;
  var
    ret, ms: Cardinal;
    msg: TMsg;
    dtStop: TDateTime;
  begin
    if AWaitMS < 0 then begin
      ms := INFINITE;
      dtStop := Now + 999;
    end else begin
      ms := AWaitMS;
      dtStop := Now + MStoTime(ms);
    end;
    repeat
      { If this used WaitForSingleObject, the calling thread (this program)
        couldn't process paint messages and could get stuck in a deadlock if
        unable to process some other process' important message. }
      ret := MsgWaitForMultipleObjects(
               1, { 1 handle to wait on }
               hProcess, { the handle }
               False, { wake on any event }
               ms, { wait timeout }
               QS_PAINT or { wake on paint messages }
               QS_SENDMESSAGE { or messages from other threads }
               );
      if ret = WAIT_FAILED then
        break; { can do little here }
      if ret = (WAIT_OBJECT_0 + 1) then begin
        { Woke on a message, process paint messages only. Calling
          PeekMessage gets messages sent from other threads processed. }
        while PeekMessage(msg, 0, WM_PAINT, WM_PAINT, PM_REMOVE) do
          DispatchMessage(msg);
      end;
    until (ret = WAIT_OBJECT_0) or (Now >= dtStop);
    Result := (ret = WAIT_OBJECT_0);
  end;
var
  sCmd: string;
  si: TStartupInfo;
  pi: TProcessInformation;
begin
  FillChar(si, SizeOf(si), 0);
  si.cb := SizeOf(si);
  si.dwFlags := STARTF_USESHOWWINDOW;
  si.wShowWindow := AShowWindow;
  sCmd := Trim(ACommand); // A const should not be passed to CreateProcessW
  if CreateProcess(nil, PChar(sCmd), nil, nil, False,
                   ACreateFlags, nil, nil, si, pi) then begin
    try
      if AWaitMS <> 0 then begin
        if WaitForIt(pi.hProcess) then
          GetExitCodeProcess(pi.hProcess, Result)
        else
          Result := WAIT_TIMEOUT;
      end else
        Result := NO_ERROR;
    finally
      CloseHandle(pi.hProcess);
      CloseHandle(pi.hThread);
    end;
  end else
    Result := GetLastError;
end;

function ExecuteCommand(const ACommand: string; const AWait: Boolean;
                        const AShowWindow: Word = SW_SHOWNORMAL;
                        const ACreateFlags: Cardinal =
               CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS): Cardinal; overload;
var
  ms: Integer;
begin
  if AWait then
    ms := -1
  else
    ms := 0;
  Result := ExecuteCommand(ACommand, ms, AShowWindow, ACreateFlags);
end;

{+}
function ExpandEnvars(const S: string): string;
{
  A handy string-based wrapper around ExpandEnvironmentStrings.
  Turns things like "%SystemRoot%" into "C:\WINDOWS", etc.
-}
var
  iLen, i: Integer;
  sTemp: string;
begin
  iLen := Length(S) + MAX_PATH;
  repeat
    SetLength(sTemp, iLen);
    i := ExpandEnvironmentStrings(PChar(S), PChar(sTemp), iLen);
    if i = 0 then
      raise Exception.Create(SysErrorMessage(GetLastError))
    else if i > iLen then // needs more room!
      iLen := i
    else
      break;
  until False;
  Result := Copy(sTemp, 1, i - 1); // PChar(sTemp);
end;

{+}
function GetAssociatedProgram(const AFile: string): string;
{
  Uses the FindExecutable API function to get the program associated with the
  given file's extension.  The file must exist and should be a fully qualified
  filename.  Returns an empty string if unsuccessful.

  Note: If the file itself is executable (.exe, .bat, etc.) the return value
  is the filename itself (usually as a short filename using ~1 notation).
-}
var
  sFile, sDir: string;
  Buffer: array[0..MAX_PATH] of Char;
begin
  Result := '';
  sFile := AFile;
  sDir := ExtractFilePath(sFile);
  if FindExecutable(PChar(sFile), PChar(sDir), Buffer) > 32 then
    Result := Buffer;
end;

{+}
function GetAssociatedProgramByExt(const AExt: string): string;
{
  Get the program associated with a given extension.
  The extension can be provided with or without a leading period.
  This actually just creates a temp file with the given extension and calls
  GetAssociatedProgram.
-}
var
  sExt, sPath, sFile: string;
begin
  Result := '';
  sExt := AExt;
  if sExt = '' then
    Exit;
  if sExt[1] <> '.' then
    Insert('.', sExt, 1);
  sPath := GetShellFolder(CSIDL_LOCAL_APPDATA, True);
  sFile := UniqueFileName(sPath + 'temp' + IntToStr(GetTickCount) + sExt);
  if CreateSimpleFile(sFile, 'Delete me!') then begin
    try
      Result := GetAssociatedProgram(sFile);
    finally
      DeleteFile(sFile);
    end;
  end;
end;

{+}
function GetLastErrorStr(const AErrNo: Cardinal = 0;
                         const AIncludeErrNo: Boolean = False): string;
{
  Returns the text for a system error number.  If unable to get the text, the
  function just returns the number as a string.

  The function will return the text for the given error number argument, or the
  results of the GetLastError function if passed a zero.

  If AIncludeErrNo is True, the error number will be appended in parentheses.
-}
var
  dwErr, dwFlags, dwChars: Cardinal;
  Buffer: array[0..MAX_PATH] of Char;
begin
  if AErrNo = 0 then
    dwErr := GetLastError
  else
    dwErr := AErrNo;
  dwFlags := FORMAT_MESSAGE_FROM_SYSTEM;
  dwChars := FormatMessage(dwFlags, nil, dwErr, 0, Buffer, MAX_PATH, nil);
  if dwChars > 0 then begin
    Result := Trim(Copy(Buffer, 1, dwChars)); // Trim the trailing CRLF
    if AIncludeErrNo then
      Result := Result + ' (' + IntToStr(dwErr) + ')';
  end else
    Result := IntToStr(dwErr);
end;

{+}
function GetShellFolder(const CSIDL: Integer;
                        const Slash: Boolean = False): string;
{
  Handy wrapper around the ShGetFolderPath function that returns a string.
  Adds/Strips a trailing slash based on the boolean argument.
  Delphi6+: The CSIDL constants are declared in the SHFolder unit.
  Delphi5: The CSIDL constants are declared here and in ShlObj.
-}
var
  szPath: array[0..MAX_PATH] of Char;
begin
  if SHGetFolderPath(0, CSIDL, 0, SHGFP_TYPE_CURRENT, szPath) = 0 then
    Result := SlashPath(szPath, Slash)
  else
    Result := '';
end;

{+}
function GetTempDir(const ASlash: Boolean = False): string;
{
  Just a wrapper around the Windows.GetTempPath function that returns a string
  so you don't have to declare a buffer. ASlash determines if the result ends
  with a trailing backslash.
-}
var
  lpTemp: array[0..MAX_PATH] of Char;
  iLen: DWORD;
begin
  iLen := GetTempPath(MAX_PATH, lpTemp);
  if iLen > (MAX_PATH + 1) then
    raise Exception.CreateFmt('GetTempPath requires %d characters', [iLen]);
  Result := lpTemp; // Always has a slash
  if not ASlash then
    Result := StripSlash(Result);
end;

{+}
function GetTempFile(const APath: string;
                     const APrefix: string = 'TMP'): string;
{
  Wrapper around GetTempFileName() that conveniently returns a string.
-}
var
  lpTempFile: array[0..MAX_PATH] of Char;
begin
  if GetTempFileName(PChar(APath), PChar(APrefix), 0, lpTempFile) = 0 then
    raise Exception.Create('GetTempFileName API failed. ' +
                           SysErrorMessage(GetLastError));
  Result := lpTempFile;
end;

{+}
function GetWinCompName: string;
{
  Returns the current Windows computer name or a null string if unsuccessful.
-}
var
  Buffer: array [0..MAX_PATH] of Char;
  nSize: DWORD;
begin
  nSize := MAX_PATH - 1;
  if GetComputerName(Buffer, nSize) then
    Result := Buffer
  else
    Result := '';
end;

{+}
function GetWinDir(const Slash: Boolean = False): string;
{
  Wrapper around GetWindowsDirectory() that conveniently returns a string.
  The Slash argument will add a trailing backslash if True, and remove one
  if False (the default).
-}
var
  Buffer: array[0..MAX_PATH] of Char;
begin
  if GetWindowsDirectory(Buffer, MAX_PATH) <> 0 then
    Result := SlashPath(Buffer, Slash)
  else
    Result := '';
end;

{+}
function GetWinUserName: string;
{
  Returns the current Windows username or a null string if unsuccessful.
-}
var
  Buffer: array [0..MAX_PATH] of Char;
  nSize: DWORD;
begin
  nSize := MAX_PATH - 1;
  if GetUserName(Buffer, nSize) then
    Result := Buffer
  else
    Result := '';
end;

{+}
function KillDirectory(const ADir: string): Boolean;
{
  Removes a directory and anything it may contain, including subdirectories.
  Returns a boolean indicating success.
-}
var
  fos: TSHFileOpStruct;
  sDir: string;
begin
  FillChar(fos, SizeOf(fos), 0);
  fos.wFunc := FO_DELETE;
  sDir := ADir + #0; // Double-null terminated lists in pFrom and pTo!
  fos.pFrom := PChar(sDir);
  fos.fFlags := FOF_NOCONFIRMATION or FOF_SILENT or FOF_NOERRORUI;
  Result := (ShFileOperation(fos) = 0) and not fos.fAnyOperationsAborted;
end;

{+}
procedure OpenWith(const AFileName: string; const hWindow: HWND = 0;
                   const ShowCmd: Integer = SW_SHOWNORMAL);
{
  This is a ShellExecute equivalent but shows the "Open with..." dialog on file
  types without an existing association.
-}
begin
  if ShellExecute(hWindow, 'open', PChar(AFileName), nil, nil, ShowCmd) =
     SE_ERR_NOASSOC then
    OpenAs_RunDLL(hWindow, 0, PChar(AFileName), ShowCmd);
end;

{+}
procedure PostKey(hWindow: HWND; Key: Word);
{
  Sends a WM_KEYDOWN & WM_KEYUP to a window/control/component.
-}
begin
  if IsWindow(hWindow) then
  begin
    PostMessage(hWindow, WM_KEYDOWN, Key,
                MakeLong(0, MapVirtualKey(Key, 0)));
    PostMessage(hWindow, WM_KEYUP, Key,
                MakeLong(0, MapVirtualKey(Key, 0) or $C000));
  end;
end;

{+}
function ReplaceFile(const ATarget, ASource: string): Boolean;
{
  Replaces the target file with the source file and returns a boolean indicating
  success.

  The following sequence of operations is used to prevent loss of the original
  in case anything goes wrong:
    1. Rename Target to Temp filename (if it exists)
    2. Move Source to Target (works even if across volumes)
    3. Delete Temp file (if needed)
-}
  function DoMove: Boolean;
  var
    fos: TSHFileOpStruct;
    sFrom, sTo: string;
  begin
    FillChar(fos, SizeOf(fos), 0);
    fos.wFunc := FO_MOVE;
    sFrom := ASource + #0; // double-null terminated
    fos.pFrom := PChar(sFrom);
    sTo := ATarget + #0; // double-null terminated
    fos.pTo := PChar(sTo);
    fos.fFlags := FOF_NOCONFIRMATION or FOF_SILENT or FOF_NOERRORUI;
    Result := (ShFileOperation(fos) = 0) and not fos.fAnyOperationsAborted;
  end;
var
  sTempFile: string;
begin
  Result := False;
  if FileExists(ASource) then begin
    if FileExists(ATarget) then begin
      sTempFile := UniqueFileName(ATarget);
      if RenameFile(ATarget, sTempFile) then
        if DoMove then
          Result := DeleteFile(sTempFile);
    end else
      Result := DoMove;
  end;
end;

{+}
function RecycleFile(const AFile: string): Boolean;
{
  Silently deletes a file and allows it to be recovered from the Recycle Bin.
  Returns a boolean indicating success.

  Note that this doesn't work from a service. It only works when called from an
  interactive user session.
-}
var
  fos: TSHFileOpStruct;
  sFrom: string;
begin
  FillChar(fos, SizeOf(fos), 0);
  fos.wFunc := FO_DELETE;
  sFrom := AFile + #0; // double-null terminated
  fos.pFrom := PChar(sFrom);
  fos.fFlags := FOF_ALLOWUNDO or FOF_NOCONFIRMATION or FOF_SILENT;
  Result := (ShFileOperation(fos) = 0);
end;

end.
